%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Some basic data and predicates that are used by diffent modules.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- dynamic plprofiling/1, quietmode/1,
	   compileWithCompact/0, compileWithDebug/0, compileWithFailPrint/0,
	   standardLoadPath/1.


quietmode(no).	% yes if environment should work in quiet mode (option -q)
plprofiling(no). % perform profiling with Prolog profiler
%compileWithCompact.  % add if FlatCurry should be compactified
                      % before compiling into Prolog
%compileWithDebug.    % include if it should be compiled with debugging code
%compileWithFailPrint.% include this if it should be compiled without code for
                     % printing failure of reductions (predicate "failprint")
standardLoadPath([]). % standard load path (might be set like CURRYPATH)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% writing outputs:

% write on standard out if not in quiet mode:
writeNQ(T) :- quietmode(no) -> write(T); true.
nlNQ :- quietmode(no) -> nl; true.

% write on user error if not in quiet mode:
writeErrNQ(T) :- quietmode(no) -> write(user_error,T); true.
nlErrNQ :- quietmode(no) -> nl(user_error); true.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get name of temporary Prolog file:
getNewPrologFileName(PrologFile) :-
	pid(PID),
	number_chars(PID,PIDS),
	append("/tmp/pakcsprog",PIDS,P1), append(P1,".pl",ProgS),
	atom_chars(PrologFile,ProgS),
	appendAtom('rm -f ',PrologFile,RmCmd), system(RmCmd).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Conversion functions for characters:
% During (Prolog) runtime, characters are represented as atoms
% with a leading quote.

% is the argument internal representation of a character constructor?
isCharCons(S) :- atom(S), atom_chars(S,[39|_]).

% a list is considered as a string if all elements are
% character constants:
isString(L) :- var(L), !, fail.
isString([]).
isString([X|Xs]) :- isCharCons(X), isString(Xs).


% relate a Curry character representation with an ASCII value of a character:
char_int(C,N) :- var(N), !,
	(atom_chars(C,[39,N]) -> true
	  ; (atom_chars(C,[39,78,85,76]) % 'NUL
	      -> N=0
	       ; write(user_error,'INTERNAL ERROR in char_int: unkown char'),
		 nl(user_error))).
char_int(C,N) :-
	(N<0 -> N1 is N+256 ; N1=N), % hack for front-end bug
	(N1=0 -> atom_chars(C,[39,78,85,76]) % 'NUL
	       ; atom_chars(C,[39,N1])).

% relate Curry strings (list of chars) and Prolog strings (list of ints):
cp_string(String,Ints) :- map2(char_int,String,Ints).

% translate a Curry string (list of chars) into Prolog atom:
string2Atom(String,Atom) :-
	map2(char_int,String,Ints), atom_chars(Atom,Ints).

% translate a Prolog atom into a Curry string (list of chars):
atom2String(Atom,String) :-
	atom_chars(Atom,Ints), map2(char_int,String,Ints).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliaries for handling load paths:

% translate a string containing a path into list of these directories:
% (e.g., ".:pakcs/lib" -> ['.','pakcs/lib'])
pathString2loadPath(SCP,LP1) :-
	append(SDir1,[58|SDirs],SCP), % 58 = ':'
	!,
	atom_chars(Dir1,SDir1),
	pathString2loadPath(SDirs,LP),
	(Dir1='' -> LP1=LP ; LP1=[Dir1|LP]).
pathString2loadPath(SDir,LP) :-
	atom_chars(Dir,SDir),
	(Dir='' -> LP=[] ; LP=[Dir]).

% transform a path into a string (e.g., ['.','pakcs/lib'] -> ".:pakcs/lib"):
path2String([],[]).
path2String([D],DS) :- atom_chars(D,DS).
path2String([D1,D2|Ds],DS) :-
	atom_chars(D1,D1s),
	path2String([D2|Ds],D2S),
	append(D1s,[58|D2S],DS). % 58 = ':'

% set standard load path from an atom like '.:dir1':
setStandardLoadPath(LPA) :-
	atom_chars(LPA,LPS),
	pathString2loadPath(LPS,LP),
	(retract(standardLoadPath(_)) -> true ; true),
	asserta(standardLoadPath(LP)).

% compute the path for loading modules w.r.t. a main directory:
loadPath(MainDir,[MainDir|LP]) :-
	standardLoadPath(SP),
	(SP=[] -> getCurryPath(LCP) ; LCP=SP),
	getLibPath(LibPath),
	append(LCP,LibPath,LP).

getCurryPath(LCP) :-
	environ('CURRYPATH',CP) -> atom_chars(CP,SCP),
	                           pathString2loadPath(SCP,LCP)
				 ; LCP=[].

getLibPath(LP) :-
	environ('PAKCSLIBPATH',PLP),
	atom_chars(PLP,SPLP),
	pathString2loadPath(SPLP,LP), !.
getLibPath([]) :-
	write(user_error,'*** Warning: cannot determine load path (PAKCSLIBPATH undefined)!'),
	nl(user_error).

% findFileInPath(Path,F,PF): find a file name w.r.t. to path
% Path: a list of directory names (atoms)
% F: file name (an atom)
% FP: file name prefixed by directory found in path
findFileInPath(Path,File,PathFile) :-
	findFilePropertyInPath(Path,file_exists,File,PathFile).

findFilePropertyInPath([Dir|Dirs],Prop,File,PathFile) :-
	(atom_chars(File,[47|_])  % already absolute file name?
	 -> DirFile=File
	  ; appendAtom(Dir,'/',DirSlash),
	    appendAtom(DirSlash,File,DirFile)),
	Pred =.. [Prop,DirFile],
	(call(Pred)
	 -> PathFile = DirFile
	  ; findFilePropertyInPath(Dirs,Prop,File,PathFile)).

% findFileInLoadPath(F,PF): find a file name w.r.t. to load path
% F: file name (an atom)
% FP: file name prefixed by directory in load path
findFileInLoadPath(Prog,PathProg) :-
	split2dirbase(Prog,ProgDir,ProgBase),
	loadPath(ProgDir,LP),
	findFileInPath(LP,ProgBase,PathProg).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Further auxiliaries:

append([],Xs,Xs).
append([X|Xs],Ys,[X|Zs]) :- append(Xs,Ys,Zs).

% linear reverse:
rev(Xs,Ys) :- rev_acc(Xs,Ys,[]).
rev_acc([],Ys,Ys).
rev_acc([X|Xs],Ys,Zs) :- rev_acc(Xs,Ys,[X|Zs]).

% concatenate two atoms:
appendAtom(A1,A2,A3) :-
	atom_chars(A1,L1), atom_chars(A2,L2),
	append(L1,L2,L3),
	atom_chars(A3,L3).

member(E,[E|_]).
member(E,[_|Xs]) :- member(E,Xs).

% is there an identical member in a list?
memberEq(E,[F|_]) :- E==F, !.
memberEq(E,[_|Xs]) :- memberEq(E,Xs).

% delete first occurrence of an element in a list:
deleteFirst(E,[E|L],L) :- !.
deleteFirst(E,[F|L],[F|M]) :- deleteFirst(E,L,M).

% compute union (without dups) of two lists:
union([],Xs,Xs).
union([X|Xs],Ys,Zs) :- member(X,Ys), !, union(Xs,Ys,Zs).
union([X|Xs],Ys,[X|Zs]) :- union(Xs,Ys,Zs).

% compute difference (without dups) of two lists:
diff([],_,[]).
diff([X|Xs],Ys,Zs) :- member(X,Ys), !, diff(Xs,Ys,Zs).
diff([X|Xs],Ys,[X|Zs]) :- diff(Xs,Ys,Zs).

% map a unary predicate on a list:
map1(_,[]).
map1(P,[X|Xs]) :- C =.. [P,X], call(C), map1(P,Xs).

% map a binary predicate on two lists:
map2(_,[],[]).
map2(P,[X|Xs],[Y|Ys]) :- C =.. [P,X,Y], call(C), map2(P,Xs,Ys).

% map a unary predicate (which might be a partial application) on a list:
map1partial(_,[]).
map1partial(P,[X|Xs]) :-
	P =.. [Pred|PartialArgs],
	append(PartialArgs,[X],Args),
	C =.. [Pred|Args], call(C),
	map1partial(P,Xs).

% map a binary predicate (which might be a partial application) on two lists:
map2partial(_,[],[]).
map2partial(P,[X|Xs],[Y|Ys]) :-
	P =.. [Pred|PartialArgs],
	append(PartialArgs,[X,Y],Args),
	C =.. [Pred|Args], call(C),
	map2partial(P,Xs,Ys).


% retract all facts for a predicate:
retractAllFacts(P/N) :-
	length(Args,N),
	Pred =.. [P|Args],
	retractAll(Pred), !.

retractAll(L) :- retract(L), fail.
retractAll(_).


% put N commas before a string:
prefixComma(C1,N,C1) :- N=1, !.
prefixComma(C1,N,C2) :- N>1, N1 is N-1, prefixComma([44|C1],N1,C2).


% try to write a file (and immediately close it) and catch file errors:
tryWriteFile(File) :-
	on_exception(ErrorMsg,
	             (open(File,write,Stream), close(Stream)),
		     print_error(ErrorMsg)).


% fileExistsAndNewer(f1,f2) is true if file f1 exists and is newer than f2:
fileExistsAndNewer(File1,File2) :-
	file_exists(File1),
	file_property(File1,mod_time(MT1)),
	file_property(File2,mod_time(MT2)),
	MT1>MT2.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Error outputs, modified from Sicstus-Prolog, Bips/msgs.pl .

print_error(syntax_error(_Goal,between(AL,BL),Msg,Tokens,AfterError)) :- !,
        seen,told,
	format(user_error, "{SYNTAX ERROR: in lines ~d-~d}~n", [AL,BL]),
	errdisplay_list([**|Msg]), errdisplay(**), errnl,
	length(Tokens, Length),
	BeforeError is Length-AfterError,
        errdisplay_list(BeforeError, Tokens, Tokens1), errnl,
	errdisplay('** here **'), errnl,
	errdisplay_list(Tokens1), errnl,
        removeTempFiles,
        !,
        fail.
print_error(existence_error(Goal,_ArgNo,ObjType,Culprit,Message)) :- !,
	nonvar(Message),
        seen,told,
	if(
             Message = past_end_of_stream,
	     format(user_error,
	             '{EXISTENCE ERROR: ~q: attempt to read past end of stream}~n',
	             [Goal]),
	     format(user_error,
	             '{EXISTENCE ERROR: ~q ~q does not exist}~n',
		     [ObjType,Culprit])
          ),
          !,
          fail.

print_error(reserved(3)) :- write('Execution aborted'), nl, !, fail.
print_error(Error) :-
        nonvar(Error),
	format(user_error, '{ERROR: ~q}~n', [Error]),
        seen,told,
        !,
        fail.
print_error(_):-
  	format(user_error, '{GENERAL ERROR.}~n', []),
        seen,told,
        !,
        fail.

errdisplay(T) :- write(T).
errnl :- nl.

errdisplay_list([]).
errdisplay_list([Head|Tail]) :-
	errdisplay_token(Head),
	errdisplay(' '),
	errdisplay_list(Tail).

errdisplay_list(0, S, S) :- !.
errdisplay_list(I, [Head|Tail], S) :-
	errdisplay_token(Head),
	errdisplay(' '),
	J is I-1,
	errdisplay_list(J, Tail, S).

errdisplay_token(atom(X)) :- !, writeq(user_error, X).
errdisplay_token(var(_,X,_)) :- !, errdisplay_string(X).
errdisplay_token(number(X)) :- !, errdisplay(X).
errdisplay_token(string(X)) :- !, errdisplay_str([]), errdisplay_str(X).
errdisplay_token(X-Y) :- !, write(user_error, X-Y).
errdisplay_token(X) :- errdisplay(X).

errdisplay_string([]).
errdisplay_string([X|Xs]) :-
        put(user_error, X),
        errdisplay_string(Xs).

errdisplay_str([]) :- put(user_error, 0'").
errdisplay_str([0'"|Xs]) :- !,
        errdisplay('""'),
        errdisplay_str(Xs).
errdisplay_str([X|Xs]) :-
        put(user_error, X),
        errdisplay_str(Xs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
