%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Curry2Prolog run time system.
%
% required: environment variable PAKCSHOME

:- use_module(basics).
:- use_module(version).
:- use_module(library(system)).
:- use_module(library(sockets)). % for the parser server
:- use_module(compiler). % compiler from FlatCurry into Prolog
:- use_module(saveprog). % definition of saving a state
:- use_module(prim_standard). % implementation of standard external functions
:- use_module(prim_readshowterm). % for showing characters


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- dynamic numberOfCalls/1, numberOfExits/1, singlestep/0, tracemode/0,
	   spypoints/1, spymode/0, spyFail/0, printDepth/1, rtargs/1,
	   profiling/1, profile_data/3, currentprogram/1, lastload/1,
	   suspendmode/1, timemode/1, verbosemode/1, errormode/1,
	   servermode/1, serversocket/1, nextIOproof/0,
	   parser_warnings/1, freeVarsUpper/1, letBindings/1,
	   hasPrintedFailure/0, printAllFailures/0, pakcsrc/2,
	   errorAbort/0.

currentprogram("prelude").
lastload(""). % program in last load command
numberOfCalls(0). % number of function calls
numberOfExits(0). % number of function exits (including suspensions)
%nextIOproof. % will be asserted in an alternative proof to check IO non-det.
%errorAbort. % will be asserted in case of an abort causes by error/user
singlestep. % single step mode initially on in debug mode
tracemode.  % trace mode initially on in debug mode
spypoints([]). % list of spy points
%spymode. % initially no spy points
%spyFail. % show fail ports in spy mode
printDepth(0). % maximal print depth of terms
profiling(no). % show profiling statistics in debug mode
errormode(yes). % yes if messages on stderr should be shown in saved states
suspendmode(no). % yes if suspended goals should be shown
timemode(no).	 % yes if execution times should be shown
verbosemode(no). % yes if program should be executed in verbose mode
parser_warnings(yes). % no if the warnings of the parser should be suppressed
freeVarsUpper(yes). % yes if free variables must start with upper chars
letBindings([]). % list of top-level bindings
rtargs([]).	% run-time arguments from script call

servermode(no).       % should programs be parsed with the server?
serverhost(localhost). % the host with the parser server
serversocket(8765).

evalToken(eval). % token send through eval arguments for correct concurrency
worldToken('$world'). % token send through world arguments in I/O actions
%pakcsrc(name,value). % definitions from PAKCS rc file

% Read the parser configuration file to define some constants
readParserConfigure :-
        environ('PAKCSHOME',PH),
	appendAtom(PH,'/parser/config.txt',ConfigFile),
	readConfigFile(ConfigFile,Props),
	append(_,[prop(maxtuple,MT)|_],Props),
	atom_chars(MT,MTs),
	number_chars(MTA,MTs),
	retract(maxTupleArity(_)), asserta(maxTupleArity(MTA)),
	append(_,[prop(port,PNr)|_],Props),
	atom_chars(PNr,PNrs),
	number_chars(SN,PNrs),
	retract(serversocket(_)), asserta(serversocket(SN)).

% Read the PAKCS rc file to define some constants
readRcFile(ArgProps) :-
        ((environ('HOME',Home),	appendAtom(Home,'/.pakcsrc',HomeConfigFile),
	  file_exists(HomeConfigFile))
	 -> readConfigFile(HomeConfigFile,HomeProps)
	  ; HomeProps=[]),
        environ('PAKCSHOME',PH), appendAtom(PH,'/pakcsrc',ConfigFile),
	(file_exists(ConfigFile)
	 -> readConfigFile(ConfigFile,GlobalProps)
	  ; GlobalProps=[]),
	append(ArgProps,HomeProps,ArgHomeProps),
	append(ArgHomeProps,GlobalProps,AllProps),
	deletePropDups(AllProps,UniqueProps),
	map1(assertPakcsrc,UniqueProps),
	pakcsrc(verboserc,yes), !,
	writeNQ('>>> Reading RC files:'),
	(file_exists(HomeConfigFile)
	 -> writeNQ(' '), writeNQ(HomeConfigFile) ; true),
	(file_exists(ConfigFile)
	 -> writeNQ(' '), writeNQ(ConfigFile) ; true),
	nlNQ,
	writeNQ('Current configurations: '), nlNQ,
	writeRCvalues.
readRcFile(_).

deletePropDups([],[]).
deletePropDups([prop(Name,Value)|Props],[prop(Name,Value)|DProps]) :-
	deleteEqualProps(Name,Props,Props1),
	deletePropDups(Props1,DProps).

deleteEqualProps(_,[],[]).
deleteEqualProps(Name,[prop(Name,_)|Props],DProps) :- !,
	deleteEqualProps(Name,Props,DProps).
deleteEqualProps(Name,[Prop|Props],[Prop|DProps]) :-
	deleteEqualProps(Name,Props,DProps).

assertPakcsrc(prop(Name,Value)) :- assertz(pakcsrc(Name,Value)).

writeRCvalues :-
	pakcsrc(Name,Value),
	writeNQ(Name), writeNQ(' = '), writeNQ(Value), nl, fail.
writeRCvalues :- nlNQ.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Start main program:
main :-
	% just to be sure:
	(environ('PAKCSHOME',_) -> true ;
	 write(user_error,'ERROR: environment variable PAKCSHOME is unknown!'),
	 nl(user_error), fail),
	prolog_flag(argv,Args,Args), processDArgs(Args,Props),
	readParserConfigure,
	readRcFile(Props),
	processArgs(Args),
	printPakcsHeader,
	main_repeat.
main :- halt.  % halt if failure (in parameters) occurred


% get all run-time arguments of the form "-Dname=value" as property list:
processDArgs([],[]).
processDArgs([Arg|Args],[prop(Name,Val)|Props]) :-
	atom_chars(Arg,[45,68|Def]), !, % [45,68] = "-D"
	append(NameS,[61|ValS],Def), % 61 = '='
	atom_chars(Name,NameS),
	atom_chars(Val,ValS),
	processDArgs(Args,Props).
processDArgs([_|Args],Props) :-
	processDArgs(Args,Props).

% process the remaining run-time arguments:
processArgs([]).
processArgs([Arg|Args]) :-
	atom_chars(Arg,[45,68|_]), !, % [45,68] = "-D"
	processArgs(Args).
processArgs(['-quiet'|Args]) :-
	retract(quietmode(_)),
	asserta(quietmode(yes)), !,
	processArgs(Args).
processArgs(['-server'|Args]) :-
	retract(servermode(_)),
	newParserClient(PCStream),
	asserta(servermode(PCStream)), !,
	processArgs(Args).
processArgs(['-path',Path|Args]) :-
	setStandardLoadPath(Path), !,
	processArgs(Args).
processArgs(['-set',path,Path|Args]) :- !,
	atom_chars(Path,PathL),
	append("path ",PathL,OptionL),
	processSetOption(OptionL), !,
	processArgs(Args).
processArgs(['-set',printdepth,Num|Args]) :- !,
	atom_chars(Num,NumL),
	append("printdepth ",NumL,OptionL),
	processSetOption(OptionL), !,
	processArgs(Args).
processArgs(['-set',Option|Args]) :-
	atom_chars(Option,OptionL),
	processSetOption(OptionL), !,
	processArgs(Args).
processArgs(['-l',Prog|Args]) :-
	atom_chars(Prog,ProgL),
	append(":load ",ProgL,Load),
	(process(Load) -> true ; true), !,  % don't fail during load
	processArgs(Args).
processArgs([Arg|Args]) :-
	retract(rtargs(RTA)),
	append(RTA,[Arg],RTAs),
	assertz(rtargs(RTAs)),
	processArgs(Args).


% Compute the prompt of the interactive loop:
pakcs_prompt(Prompt) :-
        modulename(MN), atom_chars(MN,MNS),
	currentprogram(CPS),
	atom_chars(CP,CPS),
	split2dirbase(CP,_,BaseProg),
	(MN=BaseProg -> MCP=CPS
	 ; append(CPS,"(module: ",CPS1),
	   append(CPS1,MNS,CPS2), append(CPS2,[41],MCP)),
	append(MCP,"> ",PromptString),
	atom_chars(Prompt,PromptString),
	!.

% main read-eval loop of the environment:
main_repeat :-
	repeat,
	pakcs_prompt(Prompt), prompt(_,Prompt),
	readline(Input), %write('READLINE: '), write(Input), nl,
	(Input = -1 -> true
            ; removeBlanks(Input,ShortInput),
	      process(ShortInput)),
	servermode(SM),
	(SM=no -> true ; write2server(SM,'LOGOUT'), close(SM)),
	halt.

% process a command and exit (used in saved states):
process_and_exit(Cmd) :-
	(process(Cmd) -> true ; true),
	halt.

% Expand a command that may be shortened to its full name:
expandCommand(ShortCmd,Cmd) :-
	allCommands(AllCmds),
	findall(FullCmd,prefixOf(ShortCmd,AllCmds,FullCmd),FullCmds),
	(FullCmds=[Cmd] -> true ;
	 (FullCmds=[]
	  -> write(user_error,'ERROR: unknown command. Type :h for help'),
	     nl(user_error), fail
	   ; write(user_error,'ERROR: command not unique. Type :h for help'),
	     nl(user_error), fail)).

prefixOf(Prefix,[Full|_],Full) :- append(Prefix,_,Full).
prefixOf(Prefix,[_|FullS],Full) :- prefixOf(Prefix,FullS,Full).

% all possible commands:
allCommands(["analyze","cd","dir","eval","fork","help","interface",
	     "load","peval","quit","reload","save","set",
	     "show","type","xml"]).


% execute command (in repeat-fail loop):
process([]) :- !, fail.
process("?") :-	write('Type :h for help'), nl, !, fail.
process(S) :- append(":!",STail,S), !,
	removeBlanks(STail,ShellString),
	atom_chars(Cmd,ShellString),
	system(Cmd),
	!, fail.
process([58|Cs]) :- !, % 58=':'
	(append(ShortCmd,[32|Rest],Cs) -> true ; ShortCmd=Cs, Rest=[]),
	expandCommand(ShortCmd,Cmd),
	removeBlanks(Rest,Params),
	process_command(Cmd,Params).
process(S) :- append("let",STail,S), !,
	removeBlanks(STail,SStrip),
	(mainbinding(Var,Exp,SStrip,[])
	 -> true
          ; write('*** Syntax error'), nl, !, fail),
	typecheck(Exp,_Type),
	exp2Term(Exp,[],_Term,Vs),
	(Vs=[(V=_)|Bs]
	 -> write('*** Error: free variables in top-level binding: '),
	    writeVar(user_output,V), writeVars(user_output,Bs), nl, fail
	  ; true),
	retract(letBindings(Lets)), asserta(letBindings([Var=Exp|Lets])),
	!, fail.
process(Input) :- !,
	(mainexpr(Exp,FreeVars,Input,[]) -> true
                                 ; write('*** Syntax error'), nl, !, fail),
	%write('Type expression: '), writeq(Exp), nl,
	typecheck(Exp,Type),
	exp2Term(Exp,[],Term,Vs),
	%write('Translated expression: '), writeq(Term), nl,
	checkFreeVars(FreeVars,Vs),
	(verbosemode(yes) -> write('Goal: '),
	                     \+ \+ (bindFreeVars(Vs), writeCurry(Term)),
			     write(' :: '),
	                     numbersmallvars(97,_,Type), writetype(Type), nl
		           ; true),
	% print free goal variables if present:
	writeFreeVars(Vs),
	% redefine the prompt for the trace mode:
	prompt(_,''),
	retract(numberOfCalls(_)), retract(numberOfExits(_)),
	asserta(numberOfCalls(0)), asserta(numberOfExits(0)),
	retractAllFacts(profile_data/3),
	%(plprofiling(yes) -> profile_reset([user:_]) ; true),
	(retract(nextIOproof) -> true ; true), % clean fact for nextIOproof/0
	%statistics(runtime,_),
	%statistics(walltime,_),
	!,
	process_eval(Term,Type,Vs).


process_command("quit",[]) :- !.
process_command("help",[]) :- !,
	write('Commands (can be abbreviated to a prefix if unique):'), nl,
	write(':load <prog>      - compile and load program "<prog>.curry"'),nl,
	write(':reload           - repeat last :l command'),nl,
	write('<expression>      - evaluate <expression> to normal form'), nl,
	write('let <var>=<exp>   - define variable binding for subsequent goals'), nl,
	write(':type <expr>      - show the type of <expression>'),nl,
	write(':eval <function>  - show the evaluation mode of <function>'),nl,
	write(':analyze          - analyze program (see submenu)'),nl,
	write(':interface        - show interface of current program'),nl,
	write(':interface <prog> - show interface of program "<prog>.curry"'),nl,
	write(':show             - show source of currently loaded Curry program'), nl,
	write(':cd <dir>         - change current directory to <dir>'), nl,
	write(':dir              - show names of all Curry programs in directory'), nl,
	write(':!<command>       - execute <command> in shell'), nl,
	write(':save             - save current state to <prog>.state'), nl,
	write(':save <expr>      - save current state to <prog>.state with initial <expr>'), nl,
	write(':fork <expr>      - fork new process evaluating <expr> (of type "IO ()")'), nl,
	write(':xml              - translate current program into XML format'), nl,
	write(':peval            - partially evaluate current program'), nl,
	write(':set <option>     - set a command line option'), nl,
	write(':set              - help on :set command'), nl,
	write(':help             - show this message'), nl,
	write(':quit             - leave the PAKCS environment'), nl,
	nl, fail.

process_command("set",[]) :- !,
	write('Options for ":set" command:'), nl,
	write('+/-allfails       - show all failures if printfail is turned on'), nl,
	write('+/-compact        - reduce size of target program during compilation'), nl,
	write('+/-debug          - debug mode (compile with debugging information)'), nl,
	write('+/-error          - show messages on standard error in saved program states'), nl,
	write('+/-free           - free variables must start with uppercase character'), nl,
	write('+/-plprofile      - use Prolog profiler'), nl,
	write('+/-printfail      - show failures in top-level evaluation'), nl,
	write('+/-profile        - show profile data in debug mode'), nl,
	write('+/-suspend        - show suspended goals at end of suspended computation'), nl,
	write('+/-time           - show execution time'), nl,
	write('+/-verbose        - verbose mode (printing initial goals)'), nl,
	write('+/-warn           - show parser warnings'), nl,
	write('path <path>       - set current search path for loading modules'), nl,
	write('printdepth <n>    - set print depth to <n> (0 = unlimited)'), nl,
	nl,
	write('Options in debug mode:'), nl,
	write('+/-single         - single step mode'), nl,
	write('+/-spy            - spy mode'), nl,
	write('+/-trace          - trace mode'), nl,
	write('spy <function>    - set spy point on <function>'), nl,
	nl,
	write('Current settings: '), nl,
	(printAllFailures -> write('+') ; write('-')),
	write(allfails), write('  '),
	(compileWithCompact -> write('+') ; write('-')),
	write(compact),	write('  '),
	(compileWithDebug -> write('+') ; write('-')),
	write(debug),	write('  '),
	errormode(EM), (EM=yes -> write('+') ; write('-')),
	write(error),	write('  '),
	freeVarsUpper(FV), (FV=yes -> write('+') ; write('-')),
	write(free),	write('  '),
	(compileWithFailPrint -> write('+') ; write('-')),
	write(printfail), write('  '),
	nl,
	profiling(P), (P=yes -> write('+') ; write('-')),
	write(profile),	write('  '),
	plprofiling(PLP), (PLP=yes -> write('+') ; write('-')),
	write(plprofile), write('  '),
	suspendmode(BM), (BM=yes -> write('+') ; write('-')),
	write(suspend),	write('  '),
	timemode(T), (T=yes -> write('+') ; write('-')),
	write(time),	write('  '),
	verbosemode(V), (V=yes -> write('+') ; write('-')),
	write(verbose),	write('  '),
	parser_warnings(W), (W=yes -> write('+') ; write('-')),
	write(warn), write('  '),
	nl,
	loadPath('.',LP), path2String(LP,SP),
	atom_chars(AP,SP), write('path       = '), write(AP), nl,
	printDepth(PD), write('printdepth = '),
	(PD=0 -> write(PD) ; PD1 is PD-1, write(PD1)), nl,
	write('printfail  = '),
	(compileWithFailPrint
	 -> (printAllFailures -> write(all) ; write(first))
	  ; write(no)), nl,
	(compileWithDebug ->
	  (singlestep -> write('+') ; write('-')), write(single), write('  '),
	  (spymode    -> write('+') ; write('-')), write(spy), write('  '),
	  (tracemode  -> write('+') ; write('-')), write(trace), write('  '),
	  write('/ spy points: '), spypoints(SPs), write(SPs), nl
	 ; true),
	!, fail.

process_command("set",Option) :- !,
	processSetOption(Option),
	!, fail.

process_command("load",LTail) :- !,
	extractProgName(LTail,Prog),
	retract(lastload(_)), asserta(lastload(Prog)),
	parseProgram(Prog),
	append(Prog,".fcy",FlatProgL), atom_chars(FlatProgName,FlatProgL),
	append(Prog,".pl",ProgPL), atom_chars(LocalPrologFile,ProgPL),
	tryXml2Fcy(FlatProgName),
	(findFileInLoadPath(FlatProgName,PathFlatProgName)
	 -> true
	  ; write(user_error,'ERROR: FlatCurry file '),
	    write(user_error,FlatProgName),
	    write(user_error,' not found!'),
	    nl(user_error),
	    deletePrologTarget(LocalPrologFile),!, fail),
	append([46,47|Prog],".fcy",LocalFlatProgL),
	atom_chars(LocalFlatProg,LocalFlatProgL),
	(LocalFlatProg=PathFlatProgName
	 -> checkProgramHeader(LocalPrologFile), PrologFile=LocalPrologFile
	  ; getNewPrologFileName(PrologFile)),
	(fileExistsAndNewer(PrologFile,PathFlatProgName)
         -> writeErrNQ('Compiled Curry program \''),
	    writeErrNQ(PrologFile),
	    writeErrNQ('\' is up-to-date.'), nlErrNQ
	  ; writeErrNQ('Compiling \''),
	    writeErrNQ(PathFlatProgName),
	    writeErrNQ('\' into Prolog program \''),
	    writeErrNQ(PrologFile),
	    writeErrNQ('\'...'), nlErrNQ,
	    initializeCompilerState,
	    atom_chars(ProgName,Prog),
	    c2p(ProgName,PrologFile)),
	!,
	file_exists(PrologFile),
	on_exception(ErrorMsg,
                     (compile(PrologFile),
		      (LocalFlatProg=PathFlatProgName -> true
		       ; % remove temporary Prolog file:
			 appendAtom('rm -f ',PrologFile,RmCmd), system(RmCmd)),
		      retract(currentprogram(_)),
	              asserta(currentprogram(Prog)),
		      initializationsInProg(ProgInits),
		      call(ProgInits)),
	             print_error(ErrorMsg) ),
	retract(letBindings(_)), asserta(letBindings([])),
	(compileWithDebug -> retract(spypoints(_)),
	                     asserta(spypoints([])),
	                     singleOn, traceOn, spyOff
	                   ; true),
	!, fail.

process_command("reload",[]) :- !,
	lastload(Prog),
	(Prog="" -> write(user_error,'ERROR: no load command to repeat'),
	            nl(user_error), !, fail
                  ; true),
        process_command("load",Prog).

process_command("type",ExprInput) :- !,
	(mainexpr(Exp,FreeVars,ExprInput,[]) -> true
                                 ; write('*** Syntax error'), nl, !, fail),
	typecheck(Exp,Type),
	exp2Term(Exp,[],Term,Vs),
	checkFreeVars(FreeVars,Vs),
	\+ \+ (bindFreeVars(Vs), writeCurry(Term)),
	write(' :: '),
	numbersmallvars(97,_,Type), writetype(Type), nl,
	!, fail.

process_command("eval",FunL) :- !,
	atom_chars(FName,FunL),
	((transDefinedFunc(FName,IntName),evaluation(IntName,Rule))
           -> (Rule='External'(_)
	       -> write('Externally defined function')
	        ; numbersmallvars(97,_,Rule),
		  Rule = 'Rule'(Args,RHS),
	          writePattern(IntName,Args),
                  write(' = '), nl, tab(3), writeExp(3,RHS))
          ; write('Evaluation annotation unknown')),
	nl, !, fail.

process_command("analyze",[]) :- !,
	lastload(Prog),
	(Prog="" -> write('ERROR: no program loaded for analysis'), nl, !, fail
                  ; true),
	write('Select analysis (functions|icalls|igraph|help) > '),
	readline(Line),
	\+ Line = -1,
	removeBlanks(Line,Input),
	\+ Input = [],
	AllAnas = ["functions","icalls","igraph","help"],
	findall(FullCmd,prefixOf(Input,AllAnas,FullCmd),FullCmds),
	(FullCmds=[Ana] -> true ;
	 (FullCmds=[]
	  -> write(user_error,'ERROR: unknown analysis'),
	     nl(user_error), fail
	   ; write(user_error,'ERROR: analysis prefix not unique.'),
	     nl(user_error), fail)),
	standardLoadPath(DirList),
	path2String(DirList,LPathS),
	processAnalyze(Ana,LPathS,Prog),
	!, fail.

process_command("interface",[]) :- !,
	lastload(Prog),
	(Prog="" -> process_command("interface","prelude")
                  ; process_command("interface",Prog)).
process_command("interface",IFTail) :- !,
	extractProgName(IFTail,Prog),
	atom_chars(ProgA,Prog),
	(prog_exists(ProgA)
	 -> RealProg = ProgA
	  ; getLibPath(LibPath),
	    (findFilePropertyInPath(LibPath,prog_exists,ProgA,RealProg)
	     -> true
	      ; write('ERROR: program "'),
		write(ProgA), write('" does not exist!'), nl, fail)),
	!,
        environ('PAKCSHOME',PH),
	appendAtom(PH,'/tools/genint.state -int ',CP1),
	appendAtom(CP1,RealProg,GenIntCmd),
        %write('Executing: '), write(GenIntCmd), nl,
        system(GenIntCmd),
	!, fail.

process_command("xml",[]) :- !,
	lastload(Prog),
	(Prog="" -> write('ERROR: no program loaded for XML translation'), nl,
	            !, fail
                  ; true),
        environ('PAKCSHOME',PH),
        atom_chars(PH,PHL),
	append(PHL,"/tools/curry2xml.state ",XC1),
	append(XC1,Prog,XC2),
        atom_chars(XmlCmd,XC2),
        %write('Executing: '), write(XmlCmd), nl,
        system(XmlCmd),
	!, fail.

process_command("peval",[]) :- !,
	lastload(Prog),
	(Prog="" -> write('ERROR: no program loaded for partial evaluation'),
	            nl,
	            !, fail
                  ; true),
        environ('PAKCSHOME',PH),
        atom_chars(PH,PHL),
	append(PHL,"/tools/Peval/peval.state ",PC1),
	append(PC1,Prog,PC2),
        atom_chars(PevalCmd,PC2),
        %write('Executing: '), write(PevalCmd), nl,
        system(PevalCmd),
	!,
	append(Prog,"_pe",ProgPE), % name of partially evaluated program
	write('Loading partially evaluated program "'),
	atom_chars(ProgPEA,ProgPE), write(ProgPEA),
	write('"...'), nl,
	process_command("load",ProgPE).

process_command("dir",[]) :- !,
        system('ls *.curry'),
	fail.

process_command("show",[]) :- !,
	currentprogram(Prog),
	(findSourceProg(Prog,SourceFileName)
	 -> append("more ",SourceFileName,MP),
	    atom_chars(Cmd,MP),
	    system(Cmd)
	  ; write('No source program file available, generating source from FlatCurry...'),
	    nl, nl,
	    environ('PAKCSHOME',PH),
	    atom_chars(PH,PHL),
	    append(PHL,"/tools/genint.state -mod ",SP1),
	    append(SP1,Prog,SP2),
	    atom_chars(ShowProgCmd,SP2),
            %write('Executing: '), write(ShowProgCmd), nl,
	    system(ShowProgCmd)),
	!, fail.

process_command("cd",DirString) :- !,
	atom_chars(Dir,DirString),
	(file_exists(Dir)
	 -> (working_directory(_,Dir) -> true
              ; write(user_error,'ERROR: cd command failed!'), nl(user_error))
	  ; write(user_error,'ERROR: directory \''),
	    write(user_error,Dir),
	    write(user_error,'\' does not exist!'),
	    nl(user_error)),
	!, fail.

process_command("save",MainGoal) :- !,
	currentprogram(Prog),
	(Prog="prelude" -> write(user_error,'ERROR: no program loaded'),
	                   nl(user_error), fail
	                 ; true),
	append(Prog,".state",ProgSt),
	atom_chars(ProgStName,ProgSt),
	initializationsInProg(ProgInits),
	retractDynamicPreds,
	(retract(rtargs(_)) -> true ; true),
	(MainGoal=[]
	 -> saveprog_entry(ProgStName,(ProgInits,main_repeat))
	 ; % start saved program in non-verbose mode if initial goal provided:
	   retract(verbosemode(QM)), asserta(verbosemode(no)),
	   saveprog_entry(ProgStName,
			  (ProgInits,process_and_exit(MainGoal))),
	   retract(verbosemode(no)), asserta(verbosemode(QM))),
	environ('PAKCSHOME',PH),
	atom_chars(PH,PHL),
	append(PHL,"/bin/.makesavedstate ",CMD1),
	(pakcsrc(standalone,yes) -> append(CMD1,"-standalone ",CMD2)
	                          ; CMD2=CMD1),
	(errormode(yes) -> append(CMD2,"-error ",CMD3) ; CMD3=CMD2),
	append(CMD3,ProgSt,CMD4),
	atom_chars(Cmd,CMD4),
	%write('Executing: '), write(Cmd), nl,
	system(Cmd),
	call(ProgInits),
	!, fail.

process_command("fork",STail) :- !,
	process_fork(STail),
	!,
	fail.

process_command(_,_) :- !,
	write('ERROR: unknown command. Type :h for help'), nl, fail.

process(":single") :- !, checkDebugMode, single, fail.
process(":trace") :- !, checkDebugMode, tracing, fail.
process(":spy") :- !, checkDebugMode, spying, fail.
process(SP) :-
	append(":spy ",STail,SP), !, checkDebugMode,
	removeBlanks(STail,P),
	atom_chars(PN,P),
	spypoint(PN), fail.
process(":status") :- !, checkDebugMode, showDebugStatus, fail.
process(Input) :- Input=[58|_], !,
	write('ERROR: unknown command. Type :h for help'), nl, fail.


% perform analyze command:
processAnalyze("functions",LPath,Prog) :-
	(LPath=[] -> PathCmd=""
	 ; append("CURRYPATH=",LPath,PathCmd0),
	   append(PathCmd0," ; ",PathCmd)),
	environ('PAKCSHOME',PH),
        atom_chars(PH,PHL),
	append(PathCmd,PHL,CP0),
	append(CP0,"/tools/check.state ",CP1),
	append(CP1,Prog,CP2),
        atom_chars(AnaCmd,CP2),
        %write('Executing: '), write(AnaCmd), nl,
        system(AnaCmd).
processAnalyze("igraph",LPath,Prog) :-
	(LPath=[] -> PathCmd=""
	 ; append("CURRYPATH=",LPath,PathCmd0),
	   append(PathCmd0," ; ",PathCmd)),
	environ('PAKCSHOME',PH),
        atom_chars(PH,PHL),
	append(PathCmd,PHL,CP0),
	append(CP0,"/tools/ImportGraph.state ",CP1),
	append(CP1,Prog,CP2),
        atom_chars(AnaCmd,CP2),
        %write('Executing: '), write(AnaCmd), nl,
        system(AnaCmd).
processAnalyze("icalls",_LPath,Prog) :-
	(LPath=[] -> PathCmd=""
	 ; append("CURRYPATH=",LPath,PathCmd0),
	   append(PathCmd0," ; ",PathCmd)),
	environ('PAKCSHOME',PH),
        atom_chars(PH,PHL),
	append(PathCmd,PHL,CP0),
	append(CP0,"/tools/ImportCalls.state ",CP1),
	append(CP1,Prog,CP2),
        atom_chars(AnaCmd,CP2),
        %write('Executing: '), write(AnaCmd), nl,
        system(AnaCmd).
processAnalyze("help",_,_) :-
	write('Analysis names (can be abbreviated to a prefix if unique):'), nl,
	nl,
	write('functions - check properties of defined functions'),nl,
	write('icalls    - show all calls to imported functions/constructors'),nl,
	write('igraph    - visualize the module dependencies as a graph with DaVinci'),nl.

% process the various options of the ":set" command:
processSetOption("+error") :- !,
	retract(errormode(_)),
	asserta(errormode(yes)).
processSetOption("-error") :- !,
	retract(errormode(_)),
	asserta(errormode(no)).
processSetOption("+free") :- !,
	retract(freeVarsUpper(_)),
	asserta(freeVarsUpper(yes)).
processSetOption("-free") :- !,
	retract(freeVarsUpper(_)),
	asserta(freeVarsUpper(no)).
processSetOption("+plprofile") :- !,
	retract(plprofiling(_)),
	prolog_flag(compiling,_,profiledcode),
	asserta(plprofiling(yes)),
	use_module(library(gauge)), % for visualization of profile data
	(lastload("") -> true ; process(":r")).
processSetOption("-plprofile") :- !,
	retract(plprofiling(_)),
	prolog_flag(compiling,_,compactcode),
	asserta(plprofiling(no)),
	(lastload("") -> true ; process(":r")).
processSetOption("+profile") :- !,
	retract(profiling(_)),
	asserta(profiling(yes)).
processSetOption("-profile") :- !,
	retract(profiling(_)),
	asserta(profiling(no)).
processSetOption("+single") :- !, checkDebugMode, singleOn.
processSetOption("-single") :- !, checkDebugMode, singleOff.
processSetOption("+trace") :- !, checkDebugMode, traceOn.
processSetOption("-trace") :- !, checkDebugMode, traceOff.
processSetOption("+spy") :- !, checkDebugMode, spyOn.
processSetOption("-spy") :- !, checkDebugMode, spyOff.
processSetOption("+suspend") :- !,
	retract(suspendmode(_)),
	asserta(suspendmode(yes)).
processSetOption("-suspend") :- !,
	retract(suspendmode(_)),
	asserta(suspendmode(no)).
processSetOption("+time") :- !,
	retract(timemode(_)),
	asserta(timemode(yes)).
processSetOption("-time") :- !,
	retract(timemode(_)),
	asserta(timemode(no)).
processSetOption("+verbose") :- !,
	retract(verbosemode(_)),
	asserta(verbosemode(yes)).
processSetOption("-verbose") :- !,
	retract(verbosemode(_)),
	asserta(verbosemode(no)).
processSetOption("+warn") :- !,
	retract(parser_warnings(_)),
	asserta(parser_warnings(yes)),
	servermode(PCS),
	(PCS=no
	  -> true  % not connected to parser server, do nothing
	   ; clientWriteList(PCS,['WARNINGSON'])).
processSetOption("-warn") :- !,
	retract(parser_warnings(_)),
	asserta(parser_warnings(no)),
	(PCS=no
	  -> true  % not connected to parser server, do nothing
	   ; clientWriteList(PCS,['WARNINGSOFF'])).

processSetOption("+compact") :-
	(compileWithCompact -> true ; asserta(compileWithCompact)), !.
processSetOption("-compact") :-
	(compileWithCompact -> retract(compileWithCompact) ; true), !.

processSetOption("+allfails") :-
	(printAllFailures -> true ; asserta(printAllFailures)), !.
processSetOption("-allfails") :-
	(printAllFailures -> retract(printAllFailures) ; true), !.

processSetOption("+debug") :- compileWithDebug, !.
processSetOption("+debug") :- !,
	asserta(compileWithDebug),
	(lastload("") -> true ; process(":r")).
processSetOption("-debug") :- compileWithDebug, !,
	retract(compileWithDebug),
	(lastload("") -> true ; process(":r")).
processSetOption("-debug") :- !.

processSetOption("-printfail") :-
	compileWithFailPrint, !,
	retract(compileWithFailPrint),
	(lastload("") -> true ; process(":r")).
processSetOption("-printfail") :- !.
processSetOption("+printfail") :-
	compileWithFailPrint, !.
processSetOption("+printfail") :-
	asserta(compileWithFailPrint), !,
	(lastload("") -> true ; process(":r")).

processSetOption("path") :- !,
	setStandardLoadPath(''),
	loadPath('.',LP),
	write('Current search path for loading modules: '), nl,
	path2String(LP,SP),
	atom_chars(ASP,SP), write(ASP), nl.
processSetOption(Option) :-
	append("path ",OptTail,Option), !,
	removeBlanks(OptTail,P),
	atom_chars(AP,P),
	setStandardLoadPath(AP),
	loadPath('.',LP),
	write('Current search path for loading modules: '), nl,
	path2String(LP,SP),
	atom_chars(ASP,SP), write(ASP), nl.
processSetOption(Option) :-
	append("printdepth ",OptTail,Option), !,
	removeBlanks(OptTail,P),
	(number_chars(D,P) -> true
	       ; write('Illegal print depth number'), nl, fail),
	retract(printDepth(_)),	
	(D=0 -> D1=D ; D1 is D+1),
	asserta(printDepth(D1)).
processSetOption(Option) :-
	append("spy ",OptTail,Option), !,
	checkDebugMode,
	removeBlanks(OptTail,P),
	atom_chars(PN,P),
	spypoint(PN).
processSetOption(_) :- !,
	write('ERROR: unknown option. Type :set for help'), nl.

% fork an expression (arg1 is the expression (string), arg2=verbose/quiet):
process_fork(ExprString) :-
	% check the type of the forked expression (must be "IO ()"):
	removeBlanks(ExprString,ExprInput),
	(mainexpr(Exp,_,ExprInput,[]) -> true
                                 ; write('*** Syntax error'), nl, !, fail),
	typecheck(Exp,Type),
	(Type = 'TCons'('IO',['TCons'('()',[])]) -> true
	  ; write('*** Type error: Forked expression must be of type "IO ()"!'), nl,
	    !, fail),
	% start saved program in non-verbose mode:
	retract(verbosemode(QM)), asserta(verbosemode(no)),
	pid(PID),
	number_chars(PID,PIDS),
	append("/tmp/pakcs_fork_",PIDS,StateP),
	append(StateP,".state",StateS),
	atom_chars(StateName,StateS),
	removeBlanks(ExprString,MainGoal),
        saveprog_entry(StateName,process_and_exit(MainGoal)),
	retract(verbosemode(no)), asserta(verbosemode(QM)),
	append("(",StateS,FC0),
	append(FC0," ; rm -f ", FC1),
	append(FC1,StateS,FC2),
	append(FC2,") > /dev/null 2> /dev/null &",ForkCmdS),
	atom_chars(ForkCmd,ForkCmdS),
	system(ForkCmd),
	sleep(2). % wait a little bit for starting up the new process

isIoType(Type) :- var(Type), !, fail.
isIoType('TCons'('IO',_)) :- !.

% evaluate an expression containing a list of variables:
process_eval(Exp,Type,Vs) :-
	clearDynamicPreds,
	worldToken(World),
	(isIoType(Type) -> E=apply(Exp,World) ; E=Exp),
	(retract(hasPrintedFailure) -> true ; true),
	statistics(runtime,[RTime1,_]),
	statistics(walltime,[ETime1,_]),
	evalToken(Eval),
	on_exception(ErrorMsg,
	             %nf(E,V,eval,Done),
	             call_residue(nf(E,V,Eval,Done),Suspended),
		     (asserta(errorAbort),
		      ErrorMsg=debugger_abort
		       -> write('Execution aborted.'), nl, fail
		        ; print_error(ErrorMsg)) ),
	statistics(runtime,[RTime2,_]),
	statistics(walltime,[ETime2,_]),
	% don't print failures after backtracking:
	((hasPrintedFailure ; printAllFailures) -> true
	 ; asserta(hasPrintedFailure)),
	%statistics(runtime,[_,Time]),
	%statistics(walltime,[_,ETime]),
	(timemode(yes)
           -> write('Runtime: '), RTime is RTime2-RTime1, write(RTime),
	      write(' msec.'), nl,
	      write('Elapsed time: '), ETime is ETime2-ETime1, write(ETime),
	      write(' msec.'), nl
           ; true),
	numberOfCalls(NC), numberOfExits(NE),
	(NC>0 -> write('Number of function calls: '), write(NC), nl,
	         write('Number of function exits: '), write(NE), nl
               ; true),
	(var(Done)
         -> suspendmode(SM),
	    (SM=yes -> write('*** Goal suspended!')
	     ; write('*** Goal suspended (use ":set +suspend" for details)!')),
            (Vs=[] -> true ; nl, write('Bindings: '), %bindFreeVars(Vs),
                             writebinds(Vs), nl),
            writeSuspendedGoals(Suspended), nl, flush_output,
	    show_profile_data,
	    !, fail
         ; ((nonvar(V), V='$io'('()'), nextIOproof)
            -> flush_output, retract(nextIOproof),
               write('ERROR: non-determinism in I/O actions occurred!'),
               nl,
	       show_profile_data,
	       !, fail
            ; suspendmode(SM),
	      ((SM=yes;Suspended=[]) -> true
	       ; write('*** Warning: there are suspended constraints (for details: ":set +suspend")'), nl),
	      ((nonvar(V), V='$io'('()'))
               -> (Vs=[] -> true ; write('Bindings: '), bindFreeVars(Vs),
                                   writebinds(Vs), nl),
                  flush_output,
		  (profiling(yes) % no IO ND checking during profiling
		   -> show_profile_data, !, fail
		    ; (hasPrintedFailure -> true
		       ; asserta(hasPrintedFailure)), % don't print failures
		                                      % during IO ND checking
		      asserta(nextIOproof), fail)
	       ; write('Result: '), bindFreeVars(Vs),
                 writeCurry(V),
	         (Vs=[] ->true ; nl, write('Bindings: '), writebinds(Vs)),
	         writeSuspendedGoals(Suspended), write(' ? '),
                 flush_output,
	         readmore(More),
		 \+ More = ";",  % backtrack if user types ";"
		 show_profile_data,
		 !, More = -1))). % do not fail if user types end-of-file
process_eval(_,_,_) :- % no further message ("no more sols") in case of abort
	retract(errorAbort),
	!, fail.
process_eval(_,_,_) :- % ignore proof attempt for IO ND
	retract(nextIOproof),
	show_profile_data,
	!, fail.
process_eval(_,_,_) :-
	write('No more solutions.'), nl,
	show_profile_data,
	!, fail.

readmore(More) :-
	pakcsrc(moresolutionsonreturn,yes), !,
	readline(M),
	((M="" ; M="n" ; M= -1)
	 -> (M="" -> More=";" ; (M="n" -> More="" ; More=M))
          ; write('Type <return> for more or "n" > '),
	    readmore(More)).
readmore(More) :-
	readline(M),
	((M="" ; M=";" ; M= -1) -> M=More
                                 ; write('Type ";" for more or <return> > '),
			           readmore(More)).


writeSuspendedGoals(_) :-
	suspendmode(no), !.
writeSuspendedGoals([]) :- !.
writeSuspendedGoals(Suspended) :-
	nl, write('Suspended goals (in internal representation):'),
	writeSuspendedGoalList(Suspended).

writeSuspendedGoalList([]).
writeSuspendedGoalList([_-G|Suspended]) :-
	nl,
	tryWriteSuspGoal(G),
	writeSuspendedGoalList(Suspended).

% try to format suspended goals more nicely:
tryWriteSuspGoal(_:G) :-
	G =.. [Pred|Args],
	rev(Args,[_,_,Result|RArgs]),
	rev(RArgs,FunArgs),
	FunCall =.. [Pred|FunArgs],
	write('let '), writeCurry(Result), write(' = '),
	writeCurry(FunCall),
	!.
tryWriteSuspGoal(_:G) :-
	writeCurry(G).

% check whether all free vars start with uppercase latter, if required:
checkFreeVars(FreeVars,Vs) :-
	freeVarsUpper(yes),
	filterNonUpperFreeVars(FreeVars,Vs,NUVs),
	NUVs = [(V=_)|RVs],
	!,
	write(user_error,'ERROR: Expression contains undefined symbols: '),
	writeVar(user_error,V), writeVars(user_error,RVs), nl(user_error),
	write(user_error,'(Note: free variables must start with an uppercase letter in initial goals)'),
	nl(user_error),
	fail.
checkFreeVars(_,_).

filterNonUpperFreeVars(_,[],[]).
filterNonUpperFreeVars(FreeVars,[(V=_)|Bs],FBs) :-
	atom_chars(V,[_,C|Cs]), % first char always '_'
	atom_chars(RV,[C|Cs]),
	(member(RV,FreeVars) ; (64<C, C<91)), !,
	filterNonUpperFreeVars(FreeVars,Bs,FBs).
filterNonUpperFreeVars(FreeVars,[(V=VB)|Bs],[(V=VB)|FBs]) :-
	filterNonUpperFreeVars(FreeVars,Bs,FBs).


% write free variables, if there are any:
writeFreeVars([]).
writeFreeVars([(V=_)|Bs]) :-
	write('Free variables in goal: '),
	writeVar(user_output,V), writeVars(user_output,Bs), nl.

writeVars(_,[]).
writeVars(Str,[(V=_)|Bs]) :-
	write(Str,', '), writeVar(Str,V),
	writeVars(Str,Bs).

% write a variable (i.e., omit first char which is always '_'):
writeVar(Str,V) :-
	atom_chars(V,[_|Cs]), atom_chars(V1,Cs),
	write(Str,V1).

% show an output substitution in a readable way:
writebinds([]).
writebinds([(V=B)|Bs]) :-
	nl, writeVar(user_output,V), write('='),
	% output not very helpful, should be improved
	writeCurry(B),
	writebinds(Bs).

% instantiate unbound goal variables by the name of their variable
% in order to print the results in a nicer way:
bindFreeVars([]).
bindFreeVars([(V=B)|Bs]) :-
	(var(B) -> B=V ; true),
	bindFreeVars(Bs).


checkDebugMode :-
	compileWithDebug, currentprogram(P), P\=="prelude", !.
checkDebugMode :-
	write('ERROR: current program not loaded in debug mode, option not applicable!'),
	nl, !, fail.


% exists there a Curry source program with name Prog (of type string)?
findSourceProg(Prog,SourceFileName) :-
	(Ext = ".lcurry" ; Ext = ".curry"),
	append(Prog,Ext,LCProgL), atom_chars(LCurryProgName,LCProgL),
	findFileInLoadPath(LCurryProgName,PathProgName),
	!,
	atom_chars(PathProgName,SourceFileName).

% exists there a program with name Prog (of type atom)?
prog_exists(Prog) :-
	appendAtom(Prog,'.lcurry',LCurryProgName),
	file_exists(LCurryProgName), !.
prog_exists(Prog) :-
	appendAtom(Prog,'.curry',CurryProgName),
	file_exists(CurryProgName), !.
prog_exists(Prog) :-
	appendAtom(Prog,'.fcy',FLCProgName),
	file_exists(FLCProgName), !.
prog_exists(Prog) :-
	appendAtom(Prog,'_flat.xml',XMLProgName),
	file_exists(XMLProgName), !.
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% print failures:

failprint(Exp,E,E) :-
	\+ hasPrintedFailure,
	write('Failure due to irreducible expression: '),
	writeCurry(Exp), nl,
	(printAllFailures -> true ; asserta(hasPrintedFailure)),
	!,
	fail.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% show profile data, if desired:
show_profile_data :-
	%(plprofiling(yes) -> view([user:_]) ; true),
	profiling(yes),
	currentprogram(Prog),
	append(Prog,".profile",ProgP),
	atom_chars(ProgPName,ProgP),
	tryWriteFile(ProgPName),
	!,
	tell(ProgPName),
	write('Profile data:'), nl,
	%listing(profile_data/3),
	findall(pdata(F,N,Info),profile_data(F,N,Info),PD),
	qsort(pdata_compare,PD,SortedPD),
	map1(format_profile_entry,SortedPD),
	nl,
	told,
	append("more ",ProgP,MP),
	atom_chars(Cmd,MP),
        system(Cmd).
show_profile_data.

% comparison of profile data entries:
pdata_compare(pdata(F1,_,_),pdata(F2,_,_)) :- F1 @< F2.

% printing formatted profile data entries:
format_profile_entry(pdata(Func,_,port(C,F,E,R))) :-
	write(Func), write(': '),
	write('Calls='), write(C), write(' / '),
	write('Fails='), write(F), write(' / '),
	write('Exits='), write(E), write(' / '),
	write('Redos='), write(R), nl.

% predicates for storing profile information during debugging:
profileCall(P) :-
	functor(P,PF,PN),
	( retract(profile_data(PF,PN,port(C,F,E,R))) -> true
	; port(C,F,E,R)=port(0,0,0,0)),
	C1 is C+1,
	asserta(profile_data(PF,PN,port(C1,F,E,R))).

profileFail(P) :-
	functor(P,PF,PN),
	( retract(profile_data(PF,PN,port(C,F,E,R))) -> true
	; port(C,F,E,R)=port(0,0,0,0)),
	F1 is F+1,
	asserta(profile_data(PF,PN,port(C,F1,E,R))).

profileExit(P) :-
	functor(P,PF,PN),
	( retract(profile_data(PF,PN,port(C,F,E,R))) -> true
	; port(C,F,E,R)=port(0,0,0,0)),
	E1 is E+1,
	asserta(profile_data(PF,PN,port(C,F,E1,R))).

profileRedo(P) :-
	functor(P,PF,PN),
	( retract(profile_data(PF,PN,port(C,F,E,R))) -> true
	; port(C,F,E,R)=port(0,0,0,0)),
	R1 is R+1,
	asserta(profile_data(PF,PN,port(C,F,E,R1))).


% generic quicksort:
qsort(_,[],[]).
qsort(P,[E|R],Sorted) :-
	split(P,E,R,Ls,Hs),
	qsort(P,Ls,SLs),
	qsort(P,Hs,SHs),
	append(SLs,[E|SHs],Sorted).

split(_,_,[],[],[]).
split(P,E,[X|Xs],[X|Ls],Hs) :-
	Compare =.. [P,X,E],
	call(Compare),
	!,
	split(P,E,Xs,Ls,Hs).
split(P,E,[X|Xs],Ls,[X|Hs]) :-
	split(P,E,Xs,Ls,Hs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% call the front-end to parse a Curry program (argument is a string):

parseProgram(Prog) :-
	(findSourceProg(Prog,_)
	 -> (servermode(no)
	     -> % use parser in batch mode:
  	        environ('PAKCSHOME',TCP),
		atom_chars(TCP,TCPL), append(TCPL,"/bin/parsecurry -fcy",CL1),
		( parser_warnings(no) -> append(CL1," -nowarns",CL2)
		                       ; CL2 = CL1 ),
		( quietmode(yes) -> append(CL2," -quiet",CL3)
	                          ; CL3 = CL2 ),
		standardLoadPath(LP),
		(LP=[] -> CL4 = CL3
	                ; path2String(LP,LPS), append(LPS,[34],LPSQ),
	                  append(" -path ",[34|LPSQ],PLPS),
		          append(CL3,PLPS,CL4)),
		append(CL4,[32|Prog],LoadCmdL),
		atom_chars(LoadCmd,LoadCmdL),
		%write('Executing: '), write(LoadCmd), nl,
		system(LoadCmd)
	      ; % use parser in server mode:
	        parserServerProg(Prog))
	  ; true). % do not parse if source program does not exist

% communication with parser server:
parserServerProg(ProgString) :-
	servermode(PCS),
	atom_chars(ProgName,ProgString),
	split2dirbase(ProgName,ProgDir,ProgBase),
	loadPath(ProgDir,LP),
	working_directory(DIR,DIR),
	makeAbsoluteDirNames(DIR,LP,ALP),
	path2String(ALP,LPS),
	atom_chars(Path,LPS),
	clientWriteList(PCS,['FCY','PATH',Path]),
	file2client(PCS,ProgName,ProgBase).

% add an absolute directory name to relative dir names:
makeAbsoluteDirNames(_,[],[]).
makeAbsoluteDirNames(AbsDir,[Dir|Dirs],[ADir|ADirs]) :-
	(atom_chars(Dir,[47|_]) % already absolute file name
	 -> ADir = Dir
	  ; (Dir = '.' -> ADir = AbsDir
	                ; appendAtom(AbsDir,'/',AbsDirSlash),
	   	          appendAtom(AbsDirSlash,Dir,ADir))),
	makeAbsoluteDirNames(AbsDir,Dirs,ADirs).
	
file2client(PCS,F,Base) :-
	atom_chars(F,[47|_]), % file with absolute path name
	!,
	clientFileCmd(PCS,F,Base).
file2client(PCS,PF,Base) :-
	atom_chars(PF,PFS),
	append(PFS,".curry",FS),
	atom_chars(F,FS),
	append(PFS,".fcy",FflcS),
	atom_chars(Fflc,FflcS),
	(fileExistsAndNewer(Fflc,F) -> true
	  ; % delete .fcy file if it is not newer (due to bug in parser server)
	    appendAtom('rm -f ',Fflc,RMFLC),
	    %write(user_error,RMFLC), nl(user_error),
	    system(RMFLC)),
	clientFileCmd(PCS,PF,Base).

% open a new stream connection to the parser server:
newParserClient(Stream) :-
        socket('AF_INET',St),
        serverhost(Host),
        serversocket(Port),
        socket_connect(St,'AF_INET'(Host,Port),Stream).

% send a message X to the parser server via the Stream:
write2server(Stream,X) :- 
         %write(user_error,'send to server: "'), write(user_error,X),
	 %write(user_error,'"... '), nl(user_error), flush_output(user_error),
	write(Stream,X), nl(Stream), flush_output(Stream).

clientWriteList(_,[]).
clientWriteList(Str,[X|Xs]) :- 
	write2server(Str,X),
        readServerAnswer(Str,_Lines),
	 %atom_chars(Answer,_Lines), write(user_error,Answer),
	 %write(user_error,' received'), nl(user_error),
        clientWriteList(Str,Xs).

clientFileCmd(_,File,_) :-
	environ('PAKCSHOME',PH),
        appendAtom(PH,'/lib/prelude.curry',File),
	!. % don't parse the prelude
clientFileCmd(Str,File,Base) :- 
	write(user_error,'Parsing \''),
	write(user_error,File),
	write(user_error,'\'...'), nl(user_error),
	write2server(Str,'LOAD'),
        readServerAnswer(Str,_),
	write2server(Str,Base),
        readServerAnswer(Str,Lines),
	atom_chars(Input,Lines), write(user_error,Input).

% read and show an answer of the server (terminated by END_OF_OUTPUT):
readServerAnswer(Str,AllLines) :-
	readStreamLine(Str,Line),
	 %write(user_error,'ANSWER FROM SERVER: '),
	 %atom_chars(ALine,Line), write(user_error,ALine), nl(user_error),
	(Line="END_OF_OUTPUT" -> AllLines = [] %Line
	                       ; readServerAnswer(Str,MoreLines),
	                         append(Line,[10|MoreLines],AllLines)).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% write a (Prolog) term in Curry syntax:
% (used in curry2prolog for printing results or during debugging)

% remove all share structures in a term:
removeShares(T,T) :- var(T), !.
removeShares(share(M),UT) :- !,
	get_mutable(V,M),
	(V='$eval'(Exp) -> true ; Exp=V),
	removeShares(Exp,UT).
removeShares(T,UT) :-
	T =.. [F|Args],
	map2(removeShares,Args,UArgs),
	UT =.. [F|UArgs].

writeCurry(T) :- printDepth(D), removeShares(T,UT), writeCurryD(D,UT).

writeCurryD(D,_) :- D=1, !, write('...').
writeCurryD(_,T) :- var(T), !, write(T).
writeCurryD(_,T) :- number(T), !, write(T).
writeCurryD(_,T) :- isCharCons(T), !,
	show_term(T,_,TS,[]), % use ReadShowTerm.showTerm for showing chars
	string2Atom(TS,TA), write(TA).
writeCurryD(_,T) :- atom(T), !,
	(atom_chars(T,[95|_]) -> writeVar(user_output,T) % 95 = '_'
	 ; revTransFunctor(T,ExtName),
	   (isId(ExtName) -> write(ExtName)
                           ; write('('), write(ExtName), write(')'))).
writeCurryD(D,T) :-
	isCompleteList(T),
	!,
	(isString(T)
	    -> % use ReadShowTerm.showTerm for showing strings:
	       show_term(T,_,TS,[]), string2Atom(TS,TA), write(TA)
	     ; write('['), writeCurryList(D,T)).
writeCurryD(D,[T|Ts]) :- !,
	write('('), writeCurryConsList(D,[T|Ts]), write(')').
writeCurryD(D,apply(F,X)) :- !,
	D1 is D-1,
	write('('), writeCurryD(D1,F), write(' '),
	writeCurryD(D1,X), write(')').
writeCurryD(D,if_then_else(C,T,E)) :- !,
	D1 is D-1,
	write('(if '), writeCurryD(D1,C),
	write(' then '), writeCurryD(D1,T),
	write(' else '), writeCurryD(D1,E), write(')').
writeCurryD(D,'prelude..'(F,G)) :- !,
	D1 is D-1,
	write('('), writeCurryD(D1,F),
	write(' . '), writeCurryD(D1,G), write(')').
writeCurryD(D,T) :-
	D1 is D-1,
	T =.. [Cons,Arg1|Args],
	atom_chars(Cons,[40,44|_]), % tuple constructor
	!,
	write('('), writeCurryD(D1,Arg1), writeCurryTuple(D1,Args), write(')').
writeCurryD(D,T) :-
	D1 is D-1,
	T =.. [IntCons,Arg1,Arg2],
	revTransFunctor(IntCons,Cons),
	\+ isId(Cons),		% write as an infix operator:
	!,
	write('('), writeCurryD(D1,Arg1), write(' '),
	write(Cons), write(' '),
	writeCurryD(D1,Arg2), write(')').
writeCurryD(D,T) :-
	D1 is D-1,
	T =.. [IntCons|Args],
	revTransFunctor(IntCons,Cons),
	write('('), write(Cons),
	writeCurryArgs(D1,Args).

writeCurryTuple(_,[]).
writeCurryTuple(D,[A|As]) :-
	write(','),
	writeCurryD(D,A),
	writeCurryTuple(D,As).

writeCurryList(_,[]).
writeCurryList(D,[_|_]) :- D=1, !, write('...]').
writeCurryList(D,[A|As]) :-
	D1 is D-1,
	writeCurryD(D,A),
	(As=[] -> write(']') ; write(',')),
	writeCurryList(D1,As).

writeCurryConsList(D,_) :- D=1, !, write('...').
writeCurryConsList(D,T) :-
	nonvar(T), T = [A|As], !,
	D1 is D-1,
	writeCurryD(D,A),
	write(':'),
	writeCurryConsList(D1,As).
writeCurryConsList(D,T) :- writeCurryD(D,T).

writeCurryArgs(_,[]) :- write(')').
writeCurryArgs(D,[A|As]) :-
	write(' '),
	writeCurryD(D,A),
	writeCurryArgs(D,As).

% is the argument a complete list structure, i.e., ends with the empty list?
isCompleteList(Xs) :- var(Xs), !, fail.
isCompleteList([]).
isCompleteList([_|Xs]) :- isCompleteList(Xs).


isId(T) :- atom_chars(T,[F|_]),
	(65=<F, F=<90 ; 97=<F, F=<122).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% small pretty printer for types

writetype(T) :- writetype(T,top).
% the second argument is 'top' or 'nested':
% in case of 'nested', brackets are written around complex type expressions

writetype(A,_) :- atom(A), !, write(A).
writetype('FuncType'(S,T),top) :-
	(S='FuncType'(_,_) -> S_Tag=nested ; S_Tag=top),
	writetype(S,S_Tag), write(' -> '), writetype(T,top).
writetype('FuncType'(S,T),nested) :-
	(S='FuncType'(_,_) -> S_Tag=nested ; S_Tag=top),
	write('('), writetype(S,S_Tag), write(' -> '), writetype(T,top),
	write(')').
writetype('TCons'([],['TCons'('Char',[])]),_) :- !,% print "[Char]" as "String"
	write('String').
writetype('TCons'([],[T]),_) :- !,  % list type
	write('['), writetype(T,top), write(']').
writetype('TCons'(TC,[Type|Types]),_) :-
	atom_chars(TC,[40,44|_]), % tuple type constructor
	!,
	write('('), writetype(Type,top), writeTupleType(Types), write(')').
writetype('TCons'(TC,[]),_) :- writeTypeCons(TC), !.
writetype('TCons'(TC,Ts),top) :-
	writeTypeCons(TC), writetypes(Ts), !.
writetype('TCons'(TC,Ts),nested) :-
	write('('), writeTypeCons(TC),
	writetypes(Ts), write(')'), !.

writeTypeCons(TC) :-
	modulename(Mod), atom_chars(Mod,ModS),
	atom_chars(TC,TCS),
	(append(ModS,[46|NameS],TCS)
	 -> atom_chars(Name,NameS), write(Name)
	  ; write(TC)), !.

writeTupleType([]).
writeTupleType([T|Ts]) :-
	write(','),
	writetype(T,top),
	writeTupleType(Ts).

writetypes([]).
writetypes([T|Ts]) :- write(' '), writetype(T,nested), writetypes(Ts).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% small pretty printer for type expressions:

% provide a name for each variable
numbersmallvars(N1,N2,A) :-
	var(A), !,
	N2 is N1+1,
	atom_chars(A,[N1]).
numbersmallvars(N1,N2,T) :-
	T =.. [_|Ts],
	numbersmallvarsl(N1,N2,Ts).

numbersmallvarsl(N,N,[]).
numbersmallvarsl(N1,N3,[T|Ts]) :-
	numbersmallvars(N1,N2,T), numbersmallvarsl(N2,N3,Ts).

% replace each variable in a term by a unique integer
% (starting from the first arg):
% (this is similar to numbervars except that no special structure
% is put around the integer):
vars2integers(A,N1,N2) :-
	var(A), !,
	N2 is N1+1,
	A=N1.
vars2integers(T,N1,N2) :-
	T =.. [_|Ts],
	vars2integersl(Ts,N1,N2).

vars2integersl([],N,N).
vars2integersl([T|Ts],N1,N3) :-
	vars2integers(T,N1,N2), vars2integersl(Ts,N2,N3).


writeExp(_,'Var'(A)) :- writeVarIndex(A).
writeExp(_,'Lit'('Intc'(A))) :- write(A).
writeExp(_,'Lit'('Floatc'(A))) :- write(A).
writeExp(_,'Lit'('Charc'(N))) :-
	char_int(C,N),
	show_term(C,_,CS,[]), % use ReadShowTerm.showTerm for showing chars
	string2Atom(CS,CA), write(CA).
writeExp(Tab,'Apply'(S,T)) :-
	write('('), writeExp(Tab, S), write(' '), writeExp(Tab,T), write(')').
writeExp(Tab,'Or'(S,T)) :-
	write('(or'), nl, Tab1 is Tab+3, tab(Tab1),
	writeExp(Tab1,S), nl, tab(Tab1), writeExp(Tab1,T), write(')'),
	nl, tab(Tab).
writeExp(_,'Comb'(_,ICF,[])) :- !,
	revTransFunctor(ICF,CF),
	(isId(CF) -> write(CF) ; write('('), write(CF), write(')')).
writeExp(Tab,'Comb'(_,[40,44|_],[E1|Es])) :- % tuple constructor
	!,
	write('('), writeExp(Tab,E1), writeTupleExps(Tab,Es), write(')').
writeExp(Tab,'Comb'(_,ICF,[E1,E2])) :- !,
	write('('),
	revTransFunctor(ICF,CF),
	(isId(CF) -> write(CF), writeExps(Tab,[E1,E2])
                   ; writeExp(Tab,E1),
	             (CF='.' -> write(' . ') ; write(CF)),
	             writeExp(Tab,E2)),
	write(')'), !.
writeExp(Tab,'Comb'(_,ICF,Es)) :-
	write('('),
	revTransFunctor(ICF,CF),
	(isId(CF) -> write(CF) ; write('('), write(CF), write(')')),
	writeExps(Tab,Es), write(')'), !.
writeExp(Tab,'Case'('Rigid',E,Cs)) :-
	write('case '), writeExp(Tab,E), write(' of'), nl,
	Tab1 is Tab+3, tab(Tab1), writeCases(Tab1,Cs), nl, tab(Tab).
writeExp(Tab,'Case'('Flex',E,Cs)) :-
	write('fcase '), writeExp(Tab,E), write(' of '), nl,
	Tab1 is Tab+3, tab(Tab1), writeCases(Tab1,Cs), nl, tab(Tab).
writeExp(Tab,'GuardedExpr'(Vars,Con,E)) :-
	writeExp(Tab,Con), write(' => '),
	writeExp(Tab,E),
	(Vars=[] -> true ; write(' where '), writePatVars(',',Vars),
                           write(' free')).
writeExp(Tab,'Constr'(Vars,E)) :-
	write('('),
	(Vars=[] -> true
                  ; write('let '), writePatVars(',',Vars), write(' free in ')),
	writeExp(Tab,E), write(')').
writeExp(Tab,'Choice'(E)) :-
	write('choice'), nl, Tab1 is Tab+3,
	tab(Tab1), writeExp(Tab1,E).


writeTupleExps(_,[]).
writeTupleExps(Tab,[E|Es]) :-
	write(','), writeExp(Tab,E), writeTupleExps(Tab,Es).

writeExps(_,[]).
writeExps(Tab,[E|Es]) :- write(' '), writeExp(Tab,E), writeExps(Tab,Es).

writeCases(_,[]).
writeCases(Tab,['Branch'('Pattern'(C,Vars),E)|Cs]) :-
	writePattern(C,Vars), write(' -> '),
	writeExp(Tab,E), nl, tab(Tab),
	writeCases(Tab,Cs).
writeCases(Tab,['Branch'('LPattern'(C),E)|Cs]) :-
	writePattern(C,[]), write(' -> '),
	writeExp(Tab,E), nl, tab(Tab),
	writeCases(Tab,Cs).

writePattern(P,[]) :- isCharCons(P), !,
	show_term(P,_,PS,[]), % use ReadShowTerm.showTerm for showing chars
	string2Atom(PS,PA), write(PA).
writePattern(IC,[]) :- !,
	revTransFunctor(IC,C),
	write(C).
writePattern(IC,Vars) :-
	atom_chars(IC,[40,44|_]), % tuple constructor
	!,
	write('('), writePatVars(',',Vars), write(')').
writePattern(IC,[V1,V2]) :- !,
	revTransFunctor(IC,C),
	(isId(C) -> write(C), write(' '), writeVarIndex(V1), write(' '),
	            writeVarIndex(V2)
                  ; writeVarIndex(V1), write(' '), write(C),
	            write(' '), writeVarIndex(V2)).
writePattern(IC,Vars) :-
	revTransFunctor(IC,C),
	(isId(C) -> write(C) ; write('('), write(C), write(')')),
	write(' '), writePatVars(' ',Vars).

writePatVars(_,[]).
writePatVars(_,[V]) :- writeVarIndex(V).
writePatVars(Separator,[V1,V2|Vs]) :-
	writeVarIndex(V1),
	write(Separator), writePatVars(Separator,[V2|Vs]).

writeVarIndex(I) :- write('x'), write(I).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% predicates for debugging:

% assert new spy points:
spypoint(FName) :-
	nonvar(FName),
	transDefinedFunc(FName,IntName),
	retract(spypoints(SPs)), asserta(spypoints([IntName|SPs])),
	spyOn, traceOff, singleOff,
	write('Spy mode turned on'), nl, !.
spypoint(FName) :-
	write('ERROR: Cannot place spy point on '), write(FName), nl.

singleOn :- singlestep -> true ; assert(singlestep).
singleOff :- singlestep -> retract(singlestep) ; true.
traceOn :- tracemode -> true ; assert(tracemode).
traceOff :- tracemode -> retract(tracemode) ; true.
spyOn  :- spymode -> true ; assert(spymode).
spyOff :- spymode -> retract(spymode) ; true.

spythis(_) :- tracemode, !.
spythis(Call) :- 
	spymode,
	functor(Call,FName,_),
	spypoints(SPs),
	member(FName,SPs),
	singleOn, traceOn, spyOff,
	!.

traceCall(P,Skip) :-
	retract(numberOfCalls(NC)), NC1 is NC+1, asserta(numberOfCalls(NC1)),
	% include this to avoid infinite loops (e.g., in a WWW version):
	%(NC1>10000 -> write('Execution aborted: more than 10000 reduction steps'),
        %            nl, abort ; true),
	(profiling(yes) -> profileCall(P) ; true),
	(spythis(P) -> write('Call: '), writeCurry(P), nl,
	               (singlestep -> call_singlestepmenu(Answer) ; true)
                     ; true),
	(Answer==eval ->
           evalToken(Eval),
	   (on_exception(ErrorMsg,nf(P,NF,Eval,_Done),
		                  print_error(ErrorMsg))
             -> writeCurry(NF), nl
              ; write('*** Evaluation failed.'), nl), singleOn, traceOn
	  ; Skip=Answer).
traceCall(P,_) :-
	(profiling(yes) -> profileFail(P) ; true),
	((spythis(P) ; spyFail)
	   ->  write('Fail: '), writeCurry(P), nl,
	       (singlestep -> call_singlestepmenu(_) ; true)
            ; true),
	!, fail.

traceExit(P,R,E,Skip) :-
	(nonvar(Skip) % have we reached the exit of the skipped call?
          -> singleOn, traceOn ; true),
	retract(numberOfExits(NE)), NE1 is NE+1, asserta(numberOfExits(NE1)),
	(profiling(yes) -> profileExit(P) ; true),
	(spythis(P) ->
           write('Exit: '), writeCurry(P),
	   (var(E) -> write(' (*suspended*)'), nl
            ; write(' (HNF: '), writeCurry(R),
              write(')'), nl),
           (singlestep -> exit_singlestepmenu ; true)
          ; true).
traceExit(P,_,_,_) :-
	(profiling(yes) -> profileRedo(P) ; true),
	(spythis(P) -> write('Redo: '), writeCurry(P), nl,
	              (singlestep -> exit_singlestepmenu ; true)
                    ; true),
	!, fail.


% single step menu for call ports:
call_singlestepmenu(Skip) :-
        write('(g)o (t)race (n)otrace (s)kip (l)eap (e)val (a)bort <return>(single step) >'),
        get0(C),nl,
        call_debug_option(C,Skip),        % check input
	% if input trace/notrace -> show menu again
        (member(C,[110,116]) -> call_singlestepmenu(Skip) ; true), !.

call_debug_option(103,_) :- !, skip(10),   % g: turn off single step mode
        singleOff, write('Single step mode off.'), nl.
call_debug_option(116,_) :- !, skip(10),   % t: turn on trace mode
        traceOn, write('Trace mode on.'), nl.
call_debug_option(110,_) :- !, skip(10),   % n: turn off trace mode
        traceOff, write('Trace mode off.'), nl.
call_debug_option(115,skip) :- !, skip(10),   % s: skip over function call
        singleOff, traceOff.
call_debug_option(108,skip) :- !, skip(10),   % l: jump to next spy point
        singleOff, traceOff, spyOn.
call_debug_option(97,_) :- !, skip(10),   % a: abort
        raise_exception(debugger_abort). %abort.
call_debug_option(101,eval) :- !, skip(10),   % e: evaluate to normal form
	singleOff, traceOff.
call_debug_option(10,_) :- !. % <return>: do the next single step
call_debug_option(_,S) :- write('ERROR: wrong option!'), nl,
	skip(10), call_singlestepmenu(S).

% single step menu for exit ports:
exit_singlestepmenu :-
        write('(g)o (t)race (n)otrace (l)eap (a)bort <return>(single step) >'),
        get0(C), nl,
        exit_debug_option(C),        % check input
	% if input trace/notrace -> show menu again
        (member(C,[110,116]) -> exit_singlestepmenu ; true), !.

exit_debug_option(103) :- !, skip(10),   % g: turn off single step mode
        singleOff, write('Single step mode off.'), nl.
exit_debug_option(116) :- !, skip(10),   % t: turn on trace mode
        traceOn, write('Trace mode on.'), nl.
exit_debug_option(110) :- !, skip(10),   % n: turn off trace mode
        traceOff, write('Trace mode off.'), nl.
exit_debug_option(108) :- !, skip(10),   % l: jump to next spy point
        singleOff, traceOff, spyOn.
exit_debug_option(97) :- !, skip(10),   % a: abort
        raise_exception(debugger_abort). %abort.
exit_debug_option(10) :- !. % do the next single step
exit_debug_option(_) :- write('ERROR: wrong option!'), nl,
	skip(10), exit_singlestepmenu.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Translation of internal <-> external names:

% translate a functor (operation or constructor symbol) from
% external into internal representation (fails if functor unknown):
transDefinedFunc(F,FI) :- translate(F,FI), !.
% allow access to non-exported names in the interpreter shell:
transDefinedFunc(F,F) :- functiontype(F,_), !.
transDefinedFunc(F,ModF) :-
	modulename(Mod), atom_chars(Mod,ModS),
	atom_chars(F,Fs),
	append(ModS,[46|Fs],ModFs),
	atom_chars(ModF,ModFs),
	functiontype(ModF,_), !.
% allow qualified access to prelude names in the interpreter shell:
transDefinedFunc(F,FN) :-
	atom_chars(F,FS),
	append("prelude.",FNS,FS),
	atom_chars(FN,FNS),
	functiontype(FN,_), !.

isTupleCons(S) :- atom_chars(S,[40,44|_]). % tuple constructor

% translate a functor (operation or constructor symbol) from
% internal into external representation:
revTransFunctor(IntName,Name) :-
	translate(Name,IntName), !.
revTransFunctor(Name,Name).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Parser for expressions:
%
% current restrictions:
% - no lambda abstractions

hasfixity('.',5,right) :- !. % hack for list cons operator
hasfixity('prelude..',9,right) :- !. % hack for function composition operator
hasfixity(O,P,A) :- opFixity(O,P,A), !.
hasfixity(_,9,left). % non declared opids have fixity "infixl 9"

% parser for (let) bindings:
mainbinding(X,E) --> id(XS), skipblanks, "=", skipblanks,
	             expr(E), {atom_chars(X,XS)}.

% parser for expressions:
mainexpr(E,Vars) --> skipblanks, expr(E), wherefree(Vars).

wherefree(Vars) --> "where", !, skipblanks, varList(Vars), "free" ; {Vars=[]}.

varList([V|Vs]) --> id(Name), skipblanks,  {atom_chars(V,Name)},
	            (",", skipblanks, varList(Vs)
                     ; skipblanks, {Vs=[]}).

expr(comb(if_then_else,[E1,E2,E3])) -->
	"if", skipblanks, expr(E1), "then", skipblanks, expr(E2),
	"else", skipblanks, expr(E3), !. 
expr(E) --> exprlist(EL), {resolvePrios(EL,E)}.

% resolve the priorities between operators:
resolvePrios([int(N)],int(N)) :- !.
resolvePrios([float(N)],float(N)) :- !.
resolvePrios([char(N)],char(N)) :- !.
resolvePrios(Es,E) :-
	append3(Es1,[opid(Op)],Es2,Es),
	minprio(Es1,P1),
	minprio(Es2,P2),
	transDefinedFunc(Op,TOp),
	hasfixity(TOp,PO,Ass),
	(Ass=non -> P1>PO, P2>PO
         ; (Ass=left -> P1>=PO, P2>PO
                      ; P1>PO, P2>=PO)),
	E = comb(TOp,[E1,E2]),
	!,
	resolvePrios(Es1,E1),
	resolvePrios(Es2,E2).
resolvePrios([opid(-)|Es],E) :- % special case for unary minus
	minprio(Es,P),
	P>6,
	E = comb((-),[int(0),E2]), !,
	resolvePrios(Es,E2).
resolvePrios([var(F)|Args],comb(TF,TArgs)) :-
	transDefinedFunc(F,TF), !,
	vars2combs(Args,TArgs).
resolvePrios([var(V)],var(UV)) :- !,
	(atom_chars(V,[95|_])  % 95 = '_'
	 -> V=UV
	  ; appendAtom('_',V,UV)).
resolvePrios([E],E) :- \+ E=opid(_), !.
resolvePrios([E1,E2|Es],E) :- \+ E1=opid(_), !,
	var2comb(E1,TE1),
	var2comb(E2,TE2),
	resolvePrios([apply(TE1,TE2)|Es],E).

% compute minimal priority of an expression list:
minprio([opid(Op)],PO) :- !, transDefinedFunc(Op,TOp), hasfixity(TOp,PO,_).
minprio([_],10) :- !. % anything else has highest priority
minprio([E|Es],P) :- minprio([E],PE), minprio(Es,PEs), min(PE,PEs,P).

min(X,Y,Z) :- X<Y -> Z=X ; Z=Y.

append3(X,Y,Z,L) :- append(XY,Z,L), append(X,Y,XY).


var2comb(var(F),comb(TF,[])) :- transDefinedFunc(F,TF), !.
var2comb(var(V),var(UV)) :- !,
	(atom_chars(V,[95|_])  % 95 = '_'
	 -> V=UV
	  ; appendAtom('_',V,UV)).
var2comb(E,E) :- \+ E=opid(_).

vars2combs([],[]).
vars2combs([E|Es],[TE|TEs]) :- var2comb(E,TE), vars2combs(Es,TEs).


exprlist([BE|E]) --> bexpr(BE), !, (exprlist(E) ; {E=[]}).

getBinding(Var,[(Var=E)|_],E).
getBinding(Var,[_|Bs],E) :- getBinding(Var,Bs,E).

% basic expressions:
bexpr(Exp) --> id(S), {atom_chars(Name,S),
		       letBindings(Lets), getBinding(Name,Lets,Exp)}.
bexpr(var(Name)) --> id(S), {atom_chars(Name,S)}.
bexpr(var(Name)) --> "(", skipblanks, opid(S), ")", skipblanks,
                     {(S="=" ; S="|") -> !, fail ; atom_chars(Name,S)}.
bexpr(opid(Name)) --> opid(S),
	              {(S="=" ; S="|") -> !, fail ; atom_chars(Name,S)}.
bexpr(opid(Name)) --> "`", id(S), "`", skipblanks, {atom_chars(Name,S)}.
bexpr(Num) --> numberconst(S),
	       {number_chars(N,S), (integer(N) -> Num=int(N) ; Num=float(N))}.
bexpr(char(C)) --> "'", escape(C), "'", skipblanks. % escape characters
bexpr(char(C)) --> "'", [C], "'", skipblanks. % simple characters
bexpr(comb('()',[])) --> "()", !, skipblanks.
bexpr(E) --> "(", skipblanks, exprlist(EL), ")", skipblanks,
             {transformOpSection(EL,E)}, !.
bexpr(E) --> "(", skipblanks, tupleElems(Es), {elems2tuple(Es,E)}.
bexpr(comb([],[])) --> "[]", !, skipblanks. % emptylist

% arithmetic sequences:
bexpr(comb(enumFrom,[E])) --> "[", skipblanks,
             expr(E), "..", skipblanks, "]", skipblanks, !.
bexpr(comb(enumFromTo,[E1,E2])) --> "[", skipblanks,
             expr(E1), "..", skipblanks, expr(E2), "]", skipblanks, !.
bexpr(comb(enumFromThen,[E1,E2])) --> "[", skipblanks,
             expr(E1), ",", skipblanks, expr(E2), "..", skipblanks,
	     "]", skipblanks, !.
bexpr(comb(enumFromThenTo,[E1,E2,E3])) --> "[", skipblanks,
             expr(E1), ",", skipblanks, expr(E2), "..", skipblanks,
	     expr(E3), "]", skipblanks, !.

bexpr(E) --> "[", skipblanks, listelems(E). % lists
bexpr(E) --> "\"", stringelems(E).  % strings

tupleElems([E|Es]) --> expr(E), (  ",", skipblanks, tupleElems(Es)
                                 ; ")", skipblanks, {Es=[]}).

transformOpSection([E,opid(Op)],comb(TOp,[TE])) :- % left section
	!,
	transDefinedFunc(Op,TOp),
	var2comb(E,TE).
transformOpSection([opid(-),E],comb((-),[int(0),TE])) :-
	!, % (-3) is not a section
	var2comb(E,TE).
transformOpSection([opid(Op),E],
		   comb(flip,[comb(TOp,[]),TE])) :- % right section
	transDefinedFunc(Op,TOp),
	var2comb(E,TE).

elems2tuple([E],E) :- !.
elems2tuple(Exprs,comb(TupleCons,Exprs)) :-
	length(Exprs,N),
	Cons=")",
	prefixComma(Cons,N,ConsComma),
	atom_chars(TupleCons,[40|ConsComma]).

listelems(comb((.),[E,Es])) --> expr(E), (",", skipblanks, listelems(Es)
                                        ; "]", skipblanks, {Es=comb([],[])}).

stringelems(comb((.),[char(C),Es])) --> escape(C), !, stringelems(Es).
stringelems(comb([],[])) --> "\"", skipblanks, !.
stringelems(comb((.),[char(C),Es])) --> [C], stringelems(Es).


id([C|Cs]) --> [C],
        { C >= "A", C =< "Z"
         ;C >= "a", C =< "z" },!,
        idrest(Cs), !, {\+ member([C|Cs],["where","free","let"])}.

idrest([C|Cs]) --> [C],
        { C >= "0", C =< "9"
         ;C >= "A", C =< "Z"
         ;C >= "a", C =< "z"
         ;[C] == "."  % here we accept dots to support module qualifications
         ;[C] == "_"
         ;[C] == "'" },!,
        idrest(Cs).
idrest([]) --> skipblanks.

numberconst([C|Cs]) --> [C], 
        { C >= "0", C =< "9" },
        numberconstrest(Cs).
numberconstrest([C|Cs]) --> [C], 
        { C >= "0", C =< "9" },
        numberconstrest(Cs).
numberconstrest([46|Cs]) --> ".", floatconstrest(Cs).
numberconstrest([]) --> skipblanks.
floatconstrest([C|Cs]) --> [C], 
        { C >= "0", C =< "9" },
        floatconstrest(Cs).
floatconstrest([]) --> skipblanks.

opid([C|Cs]) --> specialchar([C]), opidrest(Cs).
opidrest([C|Cs]) --> specialchar([C]), !, opidrest(Cs).
opidrest([]) --> skipblanks.

specialchar("~") --> "~".
specialchar("!") --> "!".
specialchar("@") --> "@".
specialchar("#") --> "#".
specialchar("$") --> "$".
specialchar("%") --> "%".
specialchar("^") --> "^".
specialchar("&") --> "&".
specialchar("*") --> "*".
specialchar("+") --> "+".
specialchar("-") --> "-".
specialchar("=") --> "=".
specialchar("<") --> "<".
specialchar(">") --> ">".
specialchar("?") --> "?".
specialchar(".") --> ".".
specialchar("/") --> "/".
specialchar("|") --> "|".
specialchar("\\") --> "\\".
specialchar(":") --> ":".

escape(34) --> "\\\"".
escape(92) --> "\\\\".
escape(10) --> "\\n".

skipblanks --> " ", !, skipblanks.
skipblanks --> {true}.


removeBlanks(L1,L5) :- % remove blanks at the beginning and end of L1
	skipblanks(L1,L2),
	rev(L2,L3),
	skipblanks(L3,L4),
	rev(L4,L5).

% extract the program name from a given input, i.e., remove all blanks
% and delete possible suffix ".curry" or ".lcurry":
extractProgName(S,P) :-
	removeBlanks(S,S1),
	(append(P,".curry",S1) -> true ;
	 append(P,".lcurry",S1) -> true ; P=S).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5

readline(Input) :-
	get0(C),
	(C=10 -> Input = []
               ; (C= -1 -> Input= -1 ; readline(Cs), Input=[C|Cs])).

% read a line from a stream:
readStreamLine(Str,[]) :-
	at_end_of_stream(Str), !, close(Str).
readStreamLine(Str,Line) :-
	get0(Str,C),
        (C=10 -> Line=[]
               ; Line=[C|Rest], readStreamLine(Str,Rest)).

  
% convert an expression (of Flat-Curry) into a Prolog term and return
% list of assignments (variable -> Prolog variable):

exp2Term(var(V),Vs,V1,Vs1) :- !, addVar(V,Vs,V1,Vs1).
exp2Term(int(I),Vs,I,Vs) :- !.
exp2Term(float(I),Vs,I,Vs) :- !.
exp2Term(char(I),Vs,I_Atom,Vs) :- !, char_int(I_Atom,I).
exp2Term(comb(Name,Exprs),Vs,Term,Vs1) :- !,
	exp2Terms(Exprs,Vs,Terms,Vs1),
	Term =.. [Name|Terms].
exp2Term(apply(F,X),Vs,apply(FT,XT),Vs2) :- !,
	exp2Term(F,Vs,FT,Vs1), exp2Term(X,Vs1,XT,Vs2).
exp2Term(_Expr,_,_,_) :- write('*** Syntax error'), nl,
	!, fail.

exp2Terms([],Vs,[],Vs).
exp2Terms([E|Es],Vs,[T|Ts],Vs2) :-
	exp2Term(E,Vs,T,Vs1),
	exp2Terms(Es,Vs1,Ts,Vs2).

% convert variables into Prolog variables:
addVar(V,[],NV,[(V=NV)]).
addVar(V,[(X=NX)|Vs],NX,[(X=NX)|Vs]) :- V==X, !.
addVar(V,[(X=NX)|Vs],NV,[(X=NX)|Vs1]) :- addVar(V,Vs,NV,Vs1).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check the header of a generated (Prolog) program and delete it
% if it is not appropriate:
checkProgramHeader(PrologProg) :-
	file_exists(PrologProg),
	% catch read errors:
	on_exception(_Exc,readProgramHeader(PrologProg,Header),
		     (deletePrologTarget(PrologProg),fail)),
	(checkHeaderParams(Header) -> true ; deletePrologTarget(PrologProg)),
	!.
checkProgramHeader(_).

% delete an existing Prolog target file:
deletePrologTarget(PrologProg) :-
	file_exists(PrologProg),
	writeErrNQ('Deleting old target file \''),
	writeErrNQ(PrologProg),
	writeErrNQ('\'.'), nlErrNQ,
	appendAtom('rm -f ',PrologProg,RmCmd),
	system(RmCmd), !.
deletePrologTarget(_).

% check header parameters for current settings:
checkHeaderParams(Header) :-
	compilerVersion(CV),
	atom_chars(CV,CVL),
	member([37|CVL],Header), % 37 = '%'
	(compileWithCompact -> member("COMPACT",Header)
	                   ; \+ member("COMPACT",Header)),
	(compileWithDebug -> member("DEBUG",Header)
	                   ; \+ member("DEBUG",Header)),
	(compileWithFailPrint -> member("FAILPRINT",Header)
	                       ; \+ member("FAILPRINT",Header)),
	(plprofiling(yes) -> member("PLPROFILE",Header)
	                       ; \+ member("PLPROFILE",Header)).

% read the header of a generated (Prolog) program:
readProgramHeader(Prog,Header) :-
	open(Prog,read,ProgStream),
	readStreamLine(ProgStream,Line),
	close(ProgStream),
	(Line = -1 -> Header=[]
	            ; split2words(Line,Header)).

split2words([],[]).
split2words([32|Cs],Words) :- !, split2words(Cs,Words).
split2words([C|Cs],Words) :- split2words_word(Cs,[C],Words).

split2words_word([],Word,[Word]).
split2words_word([32|Cs],Word,[Word|Words]) :- !,
	split2words(Cs,Words).
split2words_word([C|Cs],Word,Words) :-
	append(Word,[C],WordC), % slow, but doesn't matter here
	split2words_word(Cs,WordC,Words).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type check an expression (of Flat-Curry):

typecheck(Exp,Type) :-	typecheck(Exp,[],Type,_).

typecheck(var(V),Vs,Type,Vs1) :- !, addVarType(V,Vs,Type,Vs1).
typecheck(int(_),Vs,'TCons'('Int',[]),Vs) :- !.
typecheck(float(_),Vs,'TCons'('Float',[]),Vs) :- !.
typecheck(char(_),Vs,'TCons'('Char',[]),Vs) :- !.
typecheck(comb(Name,[]),Vs,FType,Vs) :- !,
	(functiontype(Name,FType)
	  -> true
           ; write('ERROR: type of function "'), write(Name),
             write('" is unknown'), nl, !, fail).
typecheck(comb(Name,Exprs),Vs,Type,Vs1) :-
	comb2apply([comb(Name,[])|Exprs],ApplyExpr),
	typecheck(ApplyExpr,Vs,Type,Vs1).
typecheck(apply(F,X),Vs,Type,Vs2) :-
	typecheck(F,Vs,TF,Vs1), !,
	typecheck(X,Vs1,TX,Vs2), !,
	(unify_type(TF,'FuncType'(TA,Type)) -> true ;
          write('ERROR: Type error in application: '),
 	  exp2Term(apply(F,X),[],apply(FT,XT),_),
	  writeCurry(apply(FT,XT)), nl,
	  write('*** term           : '), writeCurry(FT), nl,
	  numbersmallvars(97,_,TF),
	  write('*** type           : '), writetype(TF), nl,
	  write('*** is not of functional type'), nl,
	  !, fail),
	(unify_type(TA,TX) -> true ;
          write('ERROR: Type error in application: '),
 	  exp2Term(apply(F,X),[],Term,_),
	  writeCurry(Term), nl,
	  numbersmallvars(97,_,(TA,TX)),
	  write('*** required argument type : '), writetype(TA), nl,
	  write('*** does not match         : '), writetype(TX), nl,
	  !, fail).


% translate an expression list into a nested apply:
comb2apply([Expr],Expr) :- !.
comb2apply(Exprs,apply(AE,E)) :-
	append(Es,[E],Exprs),
	comb2apply(Es,AE).


% add type information for a variable:
addVarType(V,[],Type,[type(V,Type)]). % new type for first variable occurrence
addVarType(V,[type(X,TX)|VTs],TX,[type(X,TX)|VTs]) :- V==X, !.
addVarType(V,[type(X,TX)|VTs],TV,[type(X,TX)|VTs1]) :-
	addVarType(V,VTs,TV,VTs1).

% unify types: instantiate occurrences of type variables by the other type:
unify_type(T1,T2) :- var(T1), var(T2), !, T1=T2.
unify_type(T1,T2) :- var(T1), !, tvar_occurs_not(T1,T2), T1=T2.
unify_type(T1,T2) :- var(T2), !, tvar_occurs_not(T2,T1), T1=T2.
unify_type('FuncType'(T1,T2),'FuncType'(T3,T4)) :-
	unify_type(T1,T3), unify_type(T2,T4).
unify_type('TCons'(TC,TArgs1),'TCons'(TC,TArgs2)) :-
	unify_types(TArgs1,TArgs2).

unify_types([],[]) :- !.
unify_types([T1|Ts1],[T2|Ts2]) :- unify_type(T1,T2), unify_types(Ts1,Ts2).

tvar_occurs_not(TV,T) :- var(T), !, TV\==T.
tvar_occurs_not(TV,'FuncType'(T1,T2)) :-
	tvar_occurs_not(TV,T1), tvar_occurs_not(TV,T2).
tvar_occurs_not(TV,'TCons'(_,TArgs)) :-
	map1partial(tvar_occurs_not(TV),TArgs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Read a configuration file:
% contains comment lines (start with #)
% or lines of the form "property=value"
readConfigFile(File,Props) :-
	see(File),
	readLines(Lines),
	seen,
	extractProperties(Lines,Props).

extractProperties([],[]).
extractProperties([[35|_]|Ls],Ps) :- % ignore comment lines starting with #
	!,
	extractProperties(Ls,Ps).
extractProperties([L|Ls],[prop(N,V)|Ps]) :-
	append(Ns,[61|Vs],L), % 61 = '='
	!,
	atom_chars(N,Ns),
	atom_chars(V,Vs),
	extractProperties(Ls,Ps).
extractProperties([_|Ls],Ps) :-
	extractProperties(Ls,Ps).

readLines(Lines) :-
	readline(L),
	(L = -1 -> Lines=[] ; readLines(Ls), Lines = [L|Ls]).

