%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% transformation of Flat-Curry programs into Prolog programs
%
% main predicate: (initializeCompilerState,) c2p(<progname>)

% required: environment variable PAKCSHOME

:- use_module(basics).
:- use_module(library(system)).
:- use_module(external). % specification of external functions
:- use_module(readFlcFromFcy).

:- dynamic runtimeEfficient/0, compileWithSharing/0, numberOfShares/1,
	   maxTupleArity/1, includePrelude/0,
	   bugInFlcFile/0, libraryInclusions/1,
	   allFunctions/1, externalFuncs/1, currentFunction/1,
	   newFunctionCounter/1, newAuxFunctions/1,
	   dynamicPredNames/1, loadProgInits/1, costCenters/1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Parameters for the transformation:

compilerVersion('PAKCS1.6V2'). % included in headers
% If you are interested in faster execution but larger code sizes,
% then uncomment the following fact. In this case the compilation time
% might increase.
%runtimeEfficient.
compileWithSharing.	% remove if it should be compiled without sharing
numberOfShares(0).   % count number of shared variables in a compiled program
includePrelude.      % uncomment this if prelude should not be included
                     % (in this case, only the main program will be translated
                     %  which is useful for compiler debugging)
maxTupleArity(7).    % The maximal arity of tuples

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

libraryInclusions([]).  % list of necessary library inclusions
allModules([]).  % list of all modules (main + imported)
allFunctions([]).     % list of all defined functions
externalFuncs([]).    % list of all external functions
currentFunction(unknown). % name of currently translated function (for errors)
newFunctionCounter(0). % counter for generating new function names
newAuxFunctions([]). % list of generated auxiliary functions
                     % (for nested case/or)
dynamicPredNames([]). % list of pred names/file names for dynamic predicates
loadProgInits(true).  % Initialization directives to be performed
                      % after loading the program

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
addLibraryInclusion(Lib) :-
	retract(libraryInclusions(Libs)),
	(member(Lib,Libs) -> NewLibs=Libs ; NewLibs=[Lib|Libs]),
	asserta(libraryInclusions(NewLibs)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% for profiling:
costCenters(['']).     % name of all cost centers

% add a cost center:
addCostCenter(CC) :-
	costCenters(CCs), member(CC,CCs), !.
addCostCenter(CC) :-
	retract(costCenters(CCs)),
	append(CCs,[CC],NCCs),
	asserta(costCenters(NCCs)).

getCostCenterOfName(NameS,CC) :-
	append(_,[123|NCC],NameS), % 123={
	append(CCS,[125|_],NCC), % 125=}
	append([123|CCS],[125],CCD),
	atom_chars(CC,CCD), !.
getCostCenterOfName(_,'').

deleteCostCenterInName(NameCC,Name) :-
	atom_chars(NameCC,NameCCS),
	append(Name0,[123|NCC],NameCCS), % 123={
	append(_,[125|Name1],NCC), % 125=}
	append(Name0,Name1,NameS),
	atom_chars(Name,NameS), !.
deleteCostCenterInName(Name,Name).

addCostCenterOfFunc('Func'(Name,_,_,_)) :-
	getCostCenterOfName(Name,CC),
	addCostCenter(CC).

% write a cost center name without decoration:
writeCostCenter(Stream,CC) :-
	atom_chars(CC,CCS),
	append([123|CCNS],[125],CCS),
	atom_chars(CCN,CCNS),
	write(Stream,' '), write(Stream,CCN).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% directory containing the run-time modules:
moduleDir(MD) :-
        environ('PAKCSHOME',TCP),
        appendAtom(TCP,'/curry2prolog/',MD).

pleaseReport :-
	write(user_error,'*** Please report this error to the PAKCS developers'),
	nl(user_error),
	write(user_error,'*** if you have used the standard compiler!'),
	nl(user_error).

% external functions that are not explicitly defined in any module:
internalExternalFunctions([((>>)/3,'>>'),
	('readFileContents'/2,'readFileContents'),
	('writeFileContents'/3,'writeFileContents'),
	(findall/1,findall),(findfirst/1,findfirst)]).

% functions of the prelude that should not be translated (apart
% from the supported externally defined functions):
notTranslatedFuncs([(>>)/2,(&)/2,try/1,tryone/1,findall/1,findfirst/1,
	localVars/2,localInPar/2,
	cond/2,condList/2,condList2/2,committedChoice/1,
	showConstructor/1,showList/1,showListElements/1,
	basicserverloop/2,openPortOnce/3,
	openPortSocketOnce/4,basicStdInLoop/1,sendPortOnce/3,
	sendPortOnceEval/3,connectPortExt/2,afterExt/3,
	printToScreen/1,unsafePrintExt/2]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

c2p(Prog) :-
	appendAtom(Prog,'.pl',PrologFile),
	c2p(Prog,PrologFile).

c2p(Prog,PrologFile) :-
	(retract(libraryInclusions(_)) -> true ; true),
	asserta(libraryInclusions([])),
	(compileWithCompact -> compactFlatCurry(Prog); true),
	(includePrelude
	 -> readProg(Prog,FlatMainProg),
	    'Prog'(MainMod,_,_,_,_,_) = FlatMainProg,
	    readFurtherProgsWithImports(Prog,FlatMainProg,[],[],
				'Prog'(MainMod,[],[],[],[],[]),FlatProg)
	  ; readProg(Prog,FlatProg)),
	generateProg(FlatProg,PrologFile), !.
c2p(Prog,PrologFile) :-
	write(user_error,'*** Error during compilation of program "'),
	write(user_error,Prog),
	write(user_error,'"!'), nl(user_error),
	appendAtom('rm -f ',PrologFile,RmCmd),
	system(RmCmd).

% compact a FlatCurry file by eliminating all unused functions:
compactFlatCurry(ProgAtom) :-
	standardLoadPath(LP),
	(LP=[] -> SetLP=""
	  ; path2String(LP,LPS),
	    append("CURRYPATH=",LPS,SLP1),
	    append(SLP1,"; export CURRYPATH; ",SetLP)),
	environ('PAKCSHOME',PH),
        atom_chars(PH,PHL),
	append(SetLP,PHL,PC0),
	append(PC0,"/tools/optimize/CompactFlat.state ",PC1),
	atom_chars(ProgAtom,ProgString),
	append(PC1,ProgString,PC2),
	append(PC2," >&2",PC3),
        atom_chars(CompactCmd,PC3),
        %write('Executing: '), write(CompactCmd), nl,
        system(CompactCmd),
	!.

% read a program and all the imported modules:
readProgsWithImports([],_,FlatProg,FlatProg) :- !.
readProgsWithImports([ProgName|Imports],ProcessedModNames,ProcessedFlatProg,
		     FlatProg) :-
	stripPath(ProgName,PureProgName),
	member(PureProgName,ProcessedModNames),
	!,
	%write('*** already read: '), write(ProgName), nl,
	readProgsWithImports(Imports,ProcessedModNames,ProcessedFlatProg,
			     FlatProg).
readProgsWithImports([ProgName|Imports],ProcessedModNames,ProcessedFlatProg,
		     FlatProg) :-
	%write('*** read next: '), write(ProgName), nl,
	readProg(ProgName,FlatMainProg),
	readFurtherProgsWithImports(ProgName,FlatMainProg,Imports,
				    ProcessedModNames,ProcessedFlatProg,
				    FlatProg).

readFurtherProgsWithImports(ProgName,FlatMainProg,Imports,
			    ProcessedModNames,ProcessedFlatProg,FlatProg) :-
	split2dirbase(ProgName,ProgDirName,_),
	FlatMainProg = 'Prog'(_,MainImportStrings,_,_,_,_),
	map2(atom_chars,MainRImports,MainImportStrings),
	map2partial(addDirName(ProgDirName),MainRImports,MainImports),
	union(Imports,MainImports,AllImports),
	joinFlatProg(ProcessedFlatProg,FlatMainProg,NewProcessedFlatProg),
	stripPath(ProgName,PureProgName),
	readProgsWithImports(AllImports,[PureProgName|ProcessedModNames],
			     NewProcessedFlatProg,FlatProg).


% add a directory name to a file name (if not '.' or absolute file name):
addDirName(_,Name,Name) :-
	atom_chars(Name,[47|_]), % already absolute file name
	!.
addDirName(DirName,ProgName,AbsName) :-
	appendAtom(ProgName,'.fcy',FcyProgName),
	(DirName = '.'
	 -> FlatProgName = FcyProgName
	  ; appendAtom(DirName,'/',DirNameSlash),
	    appendAtom(DirNameSlash,FcyProgName,FlatProgName)),
	(file_exists(FlatProgName)
	 -> stripFcySuffix(FlatProgName,AbsName)
	  ; % FlatCurry file does not exist in DirName:
	    getLibPath(LibPath),
	    (findFileInPath(LibPath,FcyProgName,LibFcyProgName)
	     ->	stripFcySuffix(LibFcyProgName,AbsName)
	      ; AbsName=ProgName)).

% strip suffix '.fcy' from a file name (an atom):
stripFcySuffix(FcyName,Name) :-
	atom_chars(FcyName,FcyNameS),
	append(NameS,".fcy",FcyNameS),
	atom_chars(Name,NameS).

% strip any path name from a file name (an atom):
stripPath(Name,ProgName) :-
	atom_chars(Name,SName),
	rev(SName,RSName),
	(append(RSProgName,[47|_],RSName) % 47='/'
	  -> rev(RSProgName,SProgName), atom_chars(ProgName,SProgName)
	   ; ProgName = Name),
	!.


% initialize the state of the compiler, i.e., clean all tables and
% set all parameters to their default values:
initializeCompilerState :-
	(retract(compileWithSharing) -> true ; true),
	asserta(compileWithSharing),
	(retract(includePrelude) -> true ; true),
	asserta(includePrelude),
	(retract(numberOfShares(_)) -> true ; true),
	asserta(numberOfShares(0)),
	(retract(allFunctions(_)) -> true ; true),
	asserta(allFunctions([])),
	(retract(externalFuncs(_)) -> true ; true),
	asserta(externalFuncs([])),
	(retract(newFunctionCounter(_)) -> true ; true),
	asserta(newFunctionCounter(0)),
	(retract(newAuxFunctions(_)) -> true ; true),
	asserta(newAuxFunctions([])),
	(retract(dynamicPredNames(_)) -> true ; true),
	asserta(dynamicPredNames([])),
	(retract(loadProgInits(_)) -> true ; true),
	asserta(loadProgInits(true)),
	(retract(costCenters(_)) -> true ; true),
	asserta(costCenters([''])),
	% remove all non-standard infix declarations to avoid syntax errors
	% in compiled program:
	op(0,xfx,(+#)),
	op(0,xfx,(-#)),
	op(0,xfx,(*#)),
	op(0,xfx,(/#)),
	op(0,xfx,(=#)),
	op(0,xfx,(/=#)),
	op(0,xfx,(<#)),
	op(0,xfx,(<=#)),
	op(0,xfx,(>#)),
	op(0,xfx,(>=#)),
	op(0,xfx,(+.)),
	op(0,xfx,(-.)),
	op(0,xfx,(*.)),
	op(0,xfx,(/.)),
	op(0,xfx,(<.)),
	op(0,xfx,(<=.)),
	op(0,xfx,(>.)),
	op(0,xfx,(>=.)).


generateProg(FlatProg,PrologFile) :-
	(compileWithDebug
	 -> write(user_error,'...including code for debugging'),
	    nl(user_error)
	  ; true),
	(compileWithFailPrint
	 -> write(user_error,'...including code for failure printing'),
	    nl(user_error)
	  ; true),
	tryWriteFile(PrologFile),
	tell(PrologFile),
	writePrologHeader,
	writeClause((:- prolog_flag(single_var_warnings,_,off))),
	writeClause((:- prolog_flag(redefine_warnings,_,off))),
	nl,
	writeProg(FlatProg),
	nl,
	writeLibraryInclusions(FlatProg),
	nl,
	loadProgInits(ProgInits),
	writeClause(initializationsInProg(ProgInits)),
	nl,
	writeClause((:- dynamic dynamicPredInfo/2)),
	dynamicPredNames(DynPreds),
	map1(writeDynamicInfoClause,DynPreds),
	nl,
	write('%%%%% Number of shared variables: '),
	numberOfShares(SC), write(SC), nl,
	told, !.
generateProg(_,PrologFile) :-
	told,
	write(user_error,'*** Error during compiling, no program generated!'),
	nl(user_error),
	appendAtom('rm -f ',PrologFile,RmCmd),
	system(RmCmd).

% write head of generated Prolog file:
writePrologHeader :-
	write('%'),
	compilerVersion(CV),
	write(CV),
	(compileWithCompact -> write(' COMPACT') ; true),
	(compileWithDebug -> write(' DEBUG') ; true),
	(compileWithFailPrint -> write(' FAILPRINT') ; true),
	(plprofiling(yes) -> write(' PLPROFILE') ; true),
	nl, nl.

% check for necessary inclusion of standard libraries (ports, CLP,...):
writeLibraryInclusions('Prog'(Mod,_,_,_,_,_)) :-
	libraryInclusions(Libs),
	(member(prim_dynamic,Libs)
	 -> retract(loadProgInits(ProgInits1)),
	    asserta(loadProgInits((initializeDynamic,ProgInits1)))
	  ; true),
	((member(prim_ports,Libs), \+ Mod="cpns_demon")
	 -> % if Ports are used, ensure to start CPNS demon (but not for
	    % the CPNS demon program itself!):
	    environ('PAKCSHOME',PH),
	    appendAtom(PH,'/cpns/start',StartCPNS),
	    retract(loadProgInits(ProgInits2)),
	    asserta(loadProgInits((system(StartCPNS),ProgInits2)))
	  ; true),
	moduleDir(MD),
	map1partial(writeLibraryInclusion(MD),Libs).

writeLibraryInclusion(Dir,Lib) :-
	appendAtom(Dir,Lib,DirLib),
	writeClause((:- ensure_loaded(DirLib))).


% try translating XML file into .fcy file:
tryXml2Fcy(FlatFile) :-
	atom_chars(FlatFile,FlatFileS),
	append(ProgS,".fcy",FlatFileS),
	append(ProgS,".curry",CurryFileS),
	atom_chars(CurryFile,CurryFileS),
	append(ProgS,"_flat.xml",XmlFileS),
	atom_chars(XmlFile,XmlFileS),
	\+ file_exists(CurryFile),
	\+ file_exists(FlatFile),
	file_exists(XmlFile),
	% translate XML file into .fcy file:
	environ('PAKCSHOME',PH),
	atom_chars(PH,PHL),
	append(PHL,"/bin/curry2xml -fcy ",CP1),
	append(CP1,ProgS,CP2),
	atom_chars(Cmd,CP2),
	%write('Executing: '), write(Cmd), nl,
	system(Cmd),
	!.
tryXml2Fcy(_).

% read a flat program:
readProg(Prog,FlatProg) :-
	appendAtom(Prog,'.fcy',ProgFile),
	readProgFile(ProgFile,FlatProg).

readProgFile(Prog,FlatProg) :-
	split2dirbase(Prog,ProgDir,ProgBase),
	loadPath(ProgDir,LP),
	readProgInLoadPath(LP,ProgBase,FlatProg).

readProgInLoadPath([],FlatFile,_) :-
	write('ERROR: FlatCurry file '), write(FlatFile),
	write(' not found!'), nl,
	!, fail.
readProgInLoadPath([Dir|Dirs],ProgFile,FlatProg) :-
	(atom_chars(ProgFile,[47|_])  % already absolute file name?
	 -> DirProgFile=ProgFile
	  ; appendAtom(Dir,'/',DirSlash),
	    appendAtom(DirSlash,ProgFile,DirProgFile)),
	tryXml2Fcy(DirProgFile),
	(file_exists(DirProgFile)
	 -> readFlcFromFcy(DirProgFile,FlatProg)
	  ; readProgInLoadPath(Dirs,ProgFile,FlatProg)).


% add the predefined objects not explicitly contained in the FlatPrelude:
addPredefinedObjects('Prog'(ModName,Imps,Types,Funcs,Ops,Trans),
                     'Prog'(ModName,Imps,Types,Funcs,Ops,Trans)) :-
        % predefined types already contained?
        member('Type'("Int",[],[]),Types),
        !.
addPredefinedObjects('Prog'(ModName,Imps,Types,Funcs,Ops,Trans),
                     'Prog'(ModName,NImps,NTypes,Funcs,Ops,NTrans)) :-
        % delete "prelude" from import list due to bug in flc printer:
        write(user_error,'*** Warning: Adding predefined objects!'), nl(user_error),
        deleteFirst("prelude",Imps,NImps),
        NTypes = [
               'Type'("Int",[],[]),
               'Type'("Float",[],[]),
               'Type'("Char",[],[]),
               'Type'("Bool",[],['Cons'("True",0,[]),'Cons'("False",0,[])]),
               'Type'("(,)",[0,1],['Cons'("(,)",2,['TVar'(0),'TVar'(1)])]),
               'Type'("(,,)",[0,1,2],
		      ['Cons'("(,,)",3,['TVar'(0),'TVar'(1),'TVar'(2)])]),
               'Type'("(,,,)",[0,1,2,3],
		      ['Cons'("(,,,)",4,
			      ['TVar'(0),'TVar'(1),'TVar'(2),'TVar'(3)])]),
               'Type'("(,,,,)",[0,1,2,3,4],
		      ['Cons'("(,,,,)",5,
			      ['TVar'(0),'TVar'(1),'TVar'(2),
			       'TVar'(3),'TVar'(4)])]),
               'Type'("(,,,,,)",[0,1,2,3,4,5],
		      ['Cons'("(,,,,,)",6,
			      ['TVar'(0),'TVar'(1),'TVar'(2),
			       'TVar'(3),'TVar'(4),'TVar'(5)])]),
               'Type'("(,,,,,,)",[0,1,2,3,4,5],
		      ['Cons'("(,,,,,,)",7,
			      ['TVar'(0),'TVar'(1),'TVar'(2),
			       'TVar'(3),'TVar'(4),'TVar'(5),'TVar'(6)])]),
               'Type'("[]",[0],['Cons'("[]",0,[]),
                                'Cons'(":",2,
				      ['TVar'(0),'TCons'("[]",['TVar'(0)])])]),
               'Type'("IO",[0],['Cons'("$io",1,['TVar'(0)])]),
               'Type'("()",[],['Cons'("()",0,[])]),
               'Type'("Success",[],['Cons'("success",0,[])]) | Types],
        NTrans = [
               'Trans'("Int","Int"),
               'Trans'("Float","Float"),
               'Trans'("Char","Char"),
               'Trans'("True","True"),
               'Trans'("False","False"),
               'Trans'("Bool","Bool"),
               'Trans'("[]","[]"),
               'Trans'("[]","[]"),
               'Trans'("(,)","(,)"),
               'Trans'("(,)","(,)"),
               'Trans'("(,,)","(,,)"),
               'Trans'("(,,)","(,,)"),
               'Trans'("(,,,)","(,,,)"),
               'Trans'("(,,,)","(,,,)"),
               'Trans'("(,,,,)","(,,,,)"),
               'Trans'("(,,,,)","(,,,,)"),
               'Trans'("(,,,,,)","(,,,,,)"),
               'Trans'("(,,,,,)","(,,,,,)"),
               'Trans'("(,,,,,,)","(,,,,,,)"),
               'Trans'("(,,,,,,)","(,,,,,,)"),
               'Trans'("Success","Success"),
               'Trans'("success","success"),
               %'Trans'("$io","$io"),
               'Trans'("IO","IO"),
               'Trans'("()","()"),
               'Trans'("()","()"),
               'Trans'("Ports.internalPort","Ports.internalPort")
                | Trans].


% join two flat programs:
joinFlatProg('Prog'(ModName1,Imps1,Types1,Funcs1,Ops1,Trans1),
             'Prog'(_ModName2,Imps2,Types2,Funcs2,Ops2,Trans2),
             'Prog'(ModName1,Imps3,Types3,Funcs3,Ops3,Trans3)) :-
	append(Imps1,Imps2,Imps3),
	append(Types1,Types2,Types3),
	append(Funcs1,Funcs2,Funcs3),
	append(Ops1,Ops2,Ops3),
	append(Trans1,Trans2,Trans3).

dropExtPName((F/N,_),F/N).

writeProg('Prog'(Mod,_,Types,OrgFuncs,Ops,Trans)) :-
	atom_chars(ModName,Mod),
	writeClause(modulename(ModName)), nl,
	map1(writeOpClause,Ops), nl,
	% redundant checking, could be omitted if fron-end is correct:
	(retract(bugInFlcFile) -> true ; true),
	map1(check_flcFunction,OrgFuncs),
	\+ bugInFlcFile,
	(plprofiling(yes) -> map1(addCostCenterOfFunc,OrgFuncs) ; true),
	costCenters([_|CCs]),
	(CCs=[] -> true
	 ; write(user_error,'...including code for profiling cost centers:'),
	   map1partial(writeCostCenter(user_error),CCs),
	   nl(user_error)),
	computeAllExternalFunctions(OrgFuncs,ExtFuncs),
	map2(dropExtPName,ExtFuncs,ExtFuncsNameArity),
	(retract(externalFuncs(_)) -> true ; true),
	asserta(externalFuncs(ExtFuncsNameArity)),
	map2(elimNestedOrCases,OrgFuncs,AllProgFuncs),
	newAuxFunctions(TmpAuxFuns), rev(TmpAuxFuns,RevAuxFuns),
	append(AllProgFuncs,Types,AllFunData),
	computeCorrectType(AllFunData,RevAuxFuns,AuxFuns),
	append(AllProgFuncs,AuxFuns,AllProgAuxFuncs),
	map2(transformFunction2FirstOrder,AllProgAuxFuncs,AllProgAuxFuncs1st),
	deleteFuncDecls(AllProgAuxFuncs1st,Funcs),
	computeAllFunctions(Funcs),
	getConstructors(Types,ConsList),
	map1(writeFunc,Funcs),
	transhnf(Funcs,ExtFuncs),
	allFunctions(AllFuncs),
	append(AllFuncs,ConsList,AllConsFuncs),
	transapply(AllConsFuncs),
	trans_constrEq(ConsList),
	trans_boolEq(ConsList),
	transnf(ConsList),
	(compileWithSharing -> transpropshar(ConsList) ; true),
	writeClause(translate(':','.')),
	writeClause(translate('.','prelude..')),
	deleteTypeConsInTrans(Types,Trans,DTCTrans),
	map1(writeTransClause,DTCTrans), nl,
	map1(writeFTypeClause,AllProgAuxFuncs1st), nl,
	map1(writeDTypeClause,Types), nl,
	map1(writeEvalClause,Funcs), nl,
	!,
	\+ bugInFlcFile,
	nl.

% delete all type constructors from translation table:
deleteTypeConsInTrans([],Trans,Trans).
deleteTypeConsInTrans(['Type'(TCName,_,_)|Types],Trans,DTrans) :-
	deleteFirst('Trans'(_,TCName),Trans,Trans1), !,
	deleteTypeConsInTrans(Types,Trans1,DTrans).
deleteTypeConsInTrans([_|Types],Trans,DTrans) :-
	deleteTypeConsInTrans(Types,Trans,DTrans).


% a call to this predicate sets the global error flat so that no compiled
% program is produced:
setFlcBug :- bugInFlcFile -> true ; assertz(bugInFlcFile).

% check for deprecated functions and print warning:
checkForDeprecatedFunction(Name) :-
	member(Name,['Term.readTerm','Term.readQTerm','Term.showTerm',
		     'Term.showQTerm','System.getDate','System.showDate',
		     'FlatTools.writeFLC','FlatCurryTools.writeFLC',
		     'HTML.Form','Ports.openSocketConnectPort']), !,
	write(user_error,'*** Warning: do not use deprecated function "'),
	write(user_error,Name),
	write(user_error,'".'),
	nl(user_error).
checkForDeprecatedFunction(_).

% check for unsupported functions and print error:
checkForUnspportedFunction(Name) :-
	member(Name,[try,solveAll,solveAll2,once,tryone,one,condSearch,
		     best,browse,browseList,commit]),
	currentFunction(QFunc), atom_chars(QFunc,QFuncS),
	append(_,[46|_],QFuncS), % are we outside the prelude?
	!,
	write(user_error,'ERROR in "'),
	write(user_error,QFunc),
	write(user_error,'": "'),
	write(user_error,Name),
	write(user_error,'" is not yet implemented!'),
	nl(user_error),
	setFlcBug.
checkForUnspportedFunction(_).

% check for ocurrences of constructors of Dynamic type and print error:
checkForDynamicConstructor(Name) :-
	member(Name,['Dynamic.dynamic','Dynamic.persistent']),
	currentFunction(QFunc),
	!,
	write(user_error,'ERROR in "'),
	write(user_error,QFunc),
	write(user_error,'": Function "'),
	write(user_error,Name),
	write(user_error,'" is not allowed in this context!'),
	nl(user_error),
	setFlcBug.
checkForDynamicConstructor(_).

% check for maximal arity of tuples and print warning:
checkForTupleArity([40,44|Name]) :-
	length([44|Name],TA),
	maxTupleArity(MA),
	TA>MA, !,
	write(user_error,'*** Warning: arity of '), write(user_error,TA),
	write(user_error,'-tupel too large. Reconfigure your system.'),
	nl(user_error).
checkForTupleArity(_).

% write clauses with operator fixity declaration:
writeOpClause('Op'(Name,Type,Prec)) :-
	atom_chars(Atom,Name),
	transOpType(Type,FlatType),
	writeClause(opFixity(Atom,Prec,FlatType)).
transOpType('InfixOp',non).
transOpType('InfixlOp',left).
transOpType('InfixrOp',right).

% delete external functions from the list of functions:
deleteFuncDecls([],[]).
deleteFuncDecls([Func|Funcs],DelFuncs) :-
	flcFunc2FA(Func,Name/Arity),
	notTranslatedFuncs(NTFuncs),
	member(Name/Arity,NTFuncs),
	!,
	deleteFuncDecls(Funcs,DelFuncs).
deleteFuncDecls([Func|Funcs],DelFuncs) :-
	flcFunc2FA(Func,Name/Arity),
	externalFuncs(Exts),
	member(Name/Arity,Exts),
	!,
	deleteFuncDecls(Funcs,DelFuncs).
deleteFuncDecls([Func|Funcs],[Func|DelFuncs]) :-
	deleteFuncDecls(Funcs,DelFuncs).

% compute all defined functions in this program:
computeAllFunctions(DFs) :-
	map2(flcFunc2FA,DFs,FAs),
	externalFuncs(EFs),
	append(FAs,EFs,AllFuncs),
	retract(allFunctions(_)),
	asserta(allFunctions(AllFuncs)).

% compute all external functions in this program:
computeAllExternalFunctions([],IEFs) :- internalExternalFunctions(IEFs).
computeAllExternalFunctions(['Func'(Name,_,Type,'External'(EStr))|Funcs],
			    [(F/Arity,EName)|EFs]) :- !,
	argnum(Type,Arity),
	flatName2Atom(Name,F), %write(user_error,F/Arity),
	(append("prelude.",ES,EStr) -> true ; ES=EStr),
	atom_chars(EName,ES),
	computeAllExternalFunctions(Funcs,EFs).
computeAllExternalFunctions([_|Funcs],EFs) :-
	computeAllExternalFunctions(Funcs,EFs).


% transform Flat-Curry representation of a function into name/arity pair:
flcFunc2FA('Func'(Name,Arity,_,_),F/Arity) :- flatName2Atom(Name,F).

% transform a FlatCurry name (string) into atom:
flatName2Atom(":",'.') :- !. % translate Curry list cons into Prolog list cons
flatName2Atom(".",'prelude..') :- !. % to avoid confusion, rename Curry's compose
flatName2Atom(Name,Atom) :- atom_chars(Atom,Name).
	

getConstructors([],Cs) :-
	TupleCons = ['()'/0,'(,)'/2,'(,,)'/3,'(,,,)'/4,'(,,,,)'/5],
	(includePrelude
             -> Cs=['$stream'/1] % all standard constructors defined in prelude
              ; Cs=['True'/0,'False'/0,success/0,[]/0,'.'/2|TupleCons]).
getConstructors(['Type'(_,_,DataCons)|Types],AllCons) :-
	getDataCons(DataCons,DCs),
	getConstructors(Types,TCs),
	append(DCs,TCs,AllCons).

getDataCons([],[]).
getDataCons(['Cons'(Name,Arity,_)|DataCons],[Con/Arity|DCs]) :-
	flatName2Atom(Name,Con),
	getDataCons(DataCons,DCs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pre-transformation of defined functions:
% possibly add missing arguments in case of higher-order equations,
% i.e., in this implementation all functions are defined by first-order
% rules

% compute argument number (of uncurryfied function) from the functions type:
argnum('FuncType'(_,T2),N) :-
	argnum(T2,N2), N is N2+1.
argnum('TVar'(_),0).
argnum('TCons'(TC,_),N) :- TC="IO" -> N=1 ; N=0.

% compute maximum variable index in a term (list):
maxVarIndex('Var'(V),V).
maxVarIndex('Lit'(_),-1).
maxVarIndex('Comb'(_,_,Es),M) :- map2(maxVarIndex,Es,Ms), maxList(Ms,M).
maxVarIndex('Apply'(E1,E2),M) :-
	maxVarIndex(E1,M1), maxVarIndex(E2,M2), max(M1,M2,M).
maxVarIndex('Constr'(Vs,E),M) :-
	maxList(Vs,Ms), maxVarIndex(E,ME), max(Ms,ME,M).
maxVarIndex('Or'(E1,E2),M) :-
	maxVarIndex(E1,M1), maxVarIndex(E2,M2), max(M1,M2,M).
maxVarIndex('Case'(_,E,Cs),M) :-
	maxVarIndex(E,ME), map2(maxVarCase,Cs,Ms), maxList(Ms,MC),
	max(MC,ME,M).
maxVarIndex('GuardedExpr'(Vs,C,E),M) :-
	maxList(Vs,Ms), maxVarIndex(C,MC), maxVarIndex(E,ME),
	max(Ms,MC,MSC), max(MSC,ME,M).
maxVarIndex('Choice'(E),M) :- maxVarIndex(E,M).

maxVarCase('Branch'('Pattern'(_,Vs),E),M) :-
	maxList(Vs,Ms), maxVarIndex(E,ME), max(Ms,ME,M).
maxVarCase('Branch'('LPattern'(_),E),M) :-
	maxVarIndex(E,M).

% maximum of a list of naturals:
maxList([],-1).
maxList([X|Xs],M) :- maxList(Xs,Ms), max(X,Ms,M).

max(X,Y,Z) :- X>Y -> Z=X ; Z=Y.

% for simplicity, we generate negative variable indices here:
genVarIndices(_,0,[]).
genVarIndices(Max,N,[I|Is]) :-
	N>0, I is Max+1, N1 is N-1,
	genVarIndices(I,N1,Is).


% transform higher-order defining rules into first-order rules
% by eta-expansion:
transformFunction2FirstOrder('Func'(Name,Arity,Type,'Rule'(Args,RHS)),
		          'Func'(Name,NewArity,Type,'Rule'(NewArgs,NewRHS))) :-
	!,
	argnum(Type,NewArity),
	NNewArgs is NewArity-Arity,
	(NNewArgs<0
	  -> write(user_error,'*** WARNING: Function "'),
	     flatName2Atom(Name,FName), write(user_error,FName),
	     write(user_error,'" has illegal type!'), nl(user_error)
	   ; true),
	(NNewArgs=<0
	 -> NewArgs=Args, NewRHS=RHS % don't add any arguments
	  ; maxList(Args,MaxArgs), maxVarIndex(RHS,MaxRHS),
	    max(MaxArgs,MaxRHS,MaxIndex),
	    genVarIndices(MaxIndex,NNewArgs,AddArgs),
	    append(Args,AddArgs,NewArgs),
	    addArgsToRHS(RHS,AddArgs,NewRHS)).
	%freeVarsInExp(NewRHS,Vs),
	%diff(Vs,NewArgs,FVs),
	%(FVs=[] -> true ; write(user_error,NewRHS),nl(user_error)).
	%write(user_error,NewRHS),nl(user_error),
	%write(user_error,Vs),nl(user_error).
transformFunction2FirstOrder('Func'(Name,_Arity,Type,'External'(EName)),
			  'Func'(Name,NewFArity,Type,'External'(EName))) :- !,
	argnum(Type,NewFArity).
transformFunction2FirstOrder(Arg,Arg) :-
	write(user_error,'ERROR: Illegal argument in transformFunction2FirstOrder: '),
	write(user_error,Arg),
	nl(user_error).

% add new variables to right-hand side, i.e., perform eta-expansion:
addArgsToRHS('Case'(CT,C,Cases),AddArgs,'Case'(CT,C,NewCases)) :- !,
	addArgsToBranches(Cases,AddArgs,NewCases).
addArgsToRHS('Or'(E1,E2),AddArgs,'Or'(NE1,NE2)) :- !,
	addArgsToRHS(E1,AddArgs,NE1),
	addArgsToRHS(E2,AddArgs,NE2).
addArgsToRHS('Choice'(E),AddArgs,'Choice'(NE)) :- !,
	addArgsToRHS(E,AddArgs,NE).
addArgsToRHS('GuardedExpr'(Vs,Con,Exp),AddArgs,'GuardedExpr'(Vs,Con,NExp)) :-
	!,
	addArgsToRHS(Exp,AddArgs,NExp).
addArgsToRHS(Exp,[],Exp) :- !.
addArgsToRHS(Exp,[A|As],NewExp) :- !,
	addArgsToRHS('Apply'(Exp,'Var'(A)),As,NewExp).

addArgsToBranches([],_,[]).
addArgsToBranches(['Branch'(Pat,Exp)|Cs],AddArgs,
	       ['Branch'(Pat,NewExp)|NCs]) :-
	addArgsToRHS(Exp,AddArgs,NewExp),
	addArgsToBranches(Cs,AddArgs,NCs).


% eliminate nested or/case expressions in right-hand sides of function
% definitions:
elimNestedOrCases('Func'(Name,Arity,Type,'Rule'(Args,RHS)),
		  'Func'(Name,Arity,Type,'Rule'(Args,NewRHS))) :-
	!,
	elimNestedCasesInRHS(RHS,NewRHS).
elimNestedOrCases('Func'(Name,Arity,Type,'External'(EName)),
		  'Func'(Name,Arity,Type,'External'(EName))) :- !.
elimNestedOrCases(Arg,Arg) :-
	write(user_error,'ERROR: Illegal argument in elimNestedOrCases: '),
	write(user_error,Arg),
	nl(user_error).

% remove nested cases in right-hand side:
elimNestedCasesInRHS('Case'(CT,CE,Cases),'Case'(CT,NCE,NewCases)) :- !,
	elimCasesInExp(CE,NCE),
	elimNestedCasesInBranches(Cases,NewCases).
elimNestedCasesInRHS('Or'(E1,E2),'Or'(NE1,NE2)) :- !,
	elimNestedCasesInRHS(E1,NE1),
	elimNestedCasesInRHS(E2,NE2).
elimNestedCasesInRHS('Choice'(E),'Choice'(NE)) :- !,
	elimNestedCasesInRHS(E,NE).
elimNestedCasesInRHS(Exp,NewExp) :- elimCasesInExp(Exp,NewExp).

elimNestedCasesInBranches([],[]).
elimNestedCasesInBranches(['Branch'(Pat,Exp)|Cs],
	                  ['Branch'(Pat,NewExp)|NCs]) :-
	elimNestedCasesInRHS(Exp,NewExp),
	elimNestedCasesInBranches(Cs,NCs).


elimCasesInExp('Var'(V),'Var'(V)).
elimCasesInExp('Lit'(L),'Lit'(L)).
elimCasesInExp('Comb'(CT,CF,Args),'Comb'(CT,CF,NewArgs)) :-
	map2(elimCasesInExp,Args,NewArgs).
elimCasesInExp('Apply'(E1,E2),'Apply'(NE1,NE2)) :-
	elimCasesInExp(E1,NE1),
	elimCasesInExp(E2,NE2).
elimCasesInExp('Constr'(Vs,E),'Constr'(Vs,NE)) :-
	elimCasesInExp(E,NE).
elimCasesInExp('Or'(E1,E2),'Comb'('FuncCall',FNew,FVExps)) :-
	% elimCasesInExp(E1,NE1), no need to do that.
	% elimCasesInExp(E2,NE2),
	freeVarsInExp('Or'(E1,E2),FVs),
	map2(var2Exp,FVs,FVExps),
	genAuxFuncName(FNew),
	length(FVs,FNArity),
        elimNestedOrCases('Func'(FNew,FNArity,'TVar'(0), % here I am too lazy
			                       % to compute the correct type
			      'Rule'(FVs,'Or'(E1,E2))),ElimF),
        addAuxFunction(ElimF).
elimCasesInExp('Case'(CT,CE,Branches),
	       'Comb'('FuncCall',FNew,[NCE|FVExps])) :-
	elimCasesInExp(CE,NCE),
	% elimCasesInBranches(Branches,NewBranches),
	freeVarsInBranches(Branches,FVs),
	map2(var2Exp,FVs,FVExps),
	genAuxFuncName(FNew),
	length(FVs,FVLength),
	FNArity is FVLength+1,
	newVarIndex(0,FVs,CVI),
	elimNestedOrCases('Func'(FNew,FNArity,'TVar'(0), % here I am too lazy
			                       % to compute the correct type
		         'Rule'([CVI|FVs],'Case'(CT,'Var'(CVI),Branches))),
                         ElimF),
        addAuxFunction(ElimF).
elimCasesInExp('GuardedExpr'(Vs,E1,E2),'GuardedExpr'(Vs,NE1,NE2)) :-
	elimCasesInExp(E1,NE1),
	elimCasesInExp(E2,NE2).
elimCasesInExp('Choice'(E),'Choice'(NE)) :-
	elimCasesInExp(E,NE).

elimCasesInBranches([],[]).
elimCasesInBranches(['Branch'(Pat,Exp)|Cs],
	            ['Branch'(Pat,NewExp)|NCs]) :-
	elimCasesInExp(Exp,NewExp),
	elimCasesInBranches(Cs,NCs).

var2Exp(I,'Var'(I)).

% find a new variable index not occurring in a list of variables:
newVarIndex(Index,Vars,NewIndex) :-
	member(Index,Vars)
	 -> I1 is Index+1, newVarIndex(I1,Vars,NewIndex)
	  ; NewIndex=Index.

% generate new name for auxiliary function:
genAuxFuncName(FName) :-
	retract(newFunctionCounter(FC)),
	FC1 is FC+1,
	asserta(newFunctionCounter(FC1)),
	number_chars(FC,FCS),
	append("newfun_",FCS,FName).

% add new auxiliary function:
addAuxFunction(AF) :-
	retract(newAuxFunctions(AFs)),
	asserta(newAuxFunctions([AF|AFs])).


% compute the correct types of a list of functions w.r.t. to a list of already
% type correct functions:
computeCorrectType(_,[],[]).
computeCorrectType(AllFunData,['Func'(Name,Arity,_,'Rule'(Args,RHS))|Funs],
		   [NewFun|NewFuns]) :-
	map2(var2vartype,Args,TEnv),
	typeExpr(RHS,AllFunData,TEnv,EType),
	tenvtype2funtype(TEnv,EType,Type),
	freevars2tvars(Type,0,_),
	NewFun='Func'(Name,Arity,Type,'Rule'(Args,RHS)),
	computeCorrectType([NewFun|AllFunData],Funs,NewFuns).
computeCorrectType(AllFunData,['Func'(Name,Arity,Type,'External'(EName))|Funs],
		   [NewFun|NewFuns]) :-
	NewFun='Func'(Name,Arity,Type,'External'(EName)),
	computeCorrectType([NewFun|AllFunData],Funs,NewFuns).

tenvtype2funtype([],T,T).
tenvtype2funtype([type(_,T)|Ts],RT,'FuncType'(T,FT)) :-
	tenvtype2funtype(Ts,RT,FT).

var2vartype(V,type(V,_)).

getTypeFromTypeEnv([type(V,T)|_],V,T) :- !.
getTypeFromTypeEnv([_|TEnv],V,T) :- getTypeFromTypeEnv(TEnv,V,T).

typeExpr('Var'(V),_,TEnv,Type) :- getTypeFromTypeEnv(TEnv,V,Type), !.
typeExpr('Lit'('Intc'(_)),_,_,'TCons'("Int",[])) :- !.
typeExpr('Lit'('Floatc'(_)),_,_,'TCons'("Float",[])) :- !.
typeExpr('Lit'('Charc'(_)),_,_,'TCons'("Char",[])) :- !.
typeExpr('Comb'(_,CF,Args),Funs,TE,T) :-
	getTypeOfFunction(Funs,CF,FunType),
	typeExprs(Args,Funs,TE,FunType,T), !.
typeExpr('Apply'(E1,E2),Funs,TE,T2) :-
	typeExpr(E1,Funs,TE,FType),
	typeExpr(E2,Funs,TE,T1),
	(FType='FuncType'(T1,T2) -> true ;
	   write(user_error,'*** Illegal FlatCurry file: Type error (Apply):'),
	   nl(user_error),
	   write(user_error,'*** Term: '), write(user_error,E1),
	   nl(user_error),
	   write(user_error,'*** Inferred type: '),
	   write(user_error,FType), nl(user_error),
	   write(user_error,'*** Expected type: '),
	   write(user_error,'FuncType'(T1,T2)), nl(user_error),
	   !, fail), !.
typeExpr('Constr'(Vs,E),Funs,TE,T) :-
	map2(var2vartype,Vs,TVs),
	append(TVs,TE,VTE),
	typeExpr(E,Funs,VTE,T), !.
typeExpr('Or'(E1,E2),Funs,TE,T) :-
	typeExpr(E1,Funs,TE,T1),
	typeExpr(E2,Funs,TE,T2),
	(T1=T2 -> T=T1 ;
	   write(user_error,'*** Illegal FlatCurry file: Type error (Or):'),
	   nl(user_error),
	   write(user_error,'*** Term: '), write(user_error,'Or'(E1,E2)),
	   nl(user_error),
	   write(user_error,'*** Inferred type of first argument:  '),
	   write(user_error,T1), nl(user_error),
	   write(user_error,'*** Inferred type of second argument: '),
	   write(user_error,T2), nl(user_error),
	   !, fail), !.
typeExpr('Case'(_,CE,Branches),Funs,TE,T) :-
	typeExpr(CE,Funs,TE,CT),
	typeBranches(Branches,CT,Funs,TE,T), !.
typeExpr('GuardedExpr'(Vs,E1,E2),Funs,TE,T) :-
	map2(var2vartype,Vs,TVs),
	append(TVs,TE,VTE),
	typeExpr(E1,Funs,VTE,SType),
	typeExpr(E2,Funs,VTE,T),
	(SType='TCons'("Success",[]) -> true ;
	   write(user_error,'*** Illegal FlatCurry file: Type error (GuardedExpr):'),
	   nl(user_error),
	   write(user_error,'*** Term: '), write(user_error,E1),
	   nl(user_error),
	   write(user_error,'*** Inferred type: '),
	   write(user_error,SType), nl(user_error),
	   write(user_error,'*** Expected type: '),
	   write(user_error,'TCons'("Success",[])), nl(user_error),
	   !, fail), !.
typeExpr('Choice'(E),Funs,TE,T) :-
	typeExpr(E,Funs,TE,T), !.

typeExprs([],_,_,Type,Type).
typeExprs([E|Es],Funs,TE,'FuncType'(T1,T2),ResultType) :-
	typeExpr(E,Funs,TE,TA1),
	(TA1=T1 -> true ;
	   write(user_error,'*** Illegal FlatCurry file: Type error (FunArgs):'),
	   nl(user_error),
	   write(user_error,'*** Term: '), write(user_error,E),
	   nl(user_error),
	   write(user_error,'*** Inferred type: '),
	   write(user_error,TA1), nl(user_error),
	   write(user_error,'*** Expected type: '),
	   write(user_error,T1), nl(user_error),
	   !, fail),
	typeExprs(Es,Funs,TE,T2,ResultType).

typeBranches([],_,_,_,_).
typeBranches(['Branch'('LPattern'(_),Exp)|Bs],CT,Funs,TE,T) :-
	typeExpr(Exp,Funs,TE,T),
	typeBranches(Bs,CT,Funs,TE,T).
typeBranches(['Branch'('Pattern'(Cons,Vs),Exp)|Bs],CT,Funs,TE,T) :-
	map2(var2vartype,Vs,TVs),
	append(TVs,TE,VTE),
	getTypeOfFunction(Funs,Cons,ConsType),
	map2(varIndex2VarExp,Vs,Vars),
	typeExprs(Vars,Funs,VTE,ConsType,CT), % type pattern arguments
	typeExpr(Exp,Funs,VTE,T),
	typeBranches(Bs,CT,Funs,TE,T).

getTypeOfFunction([],Name,_) :- % usually, this case should not occur!
	write(user_error,'WARNING: Type of function '),
	atom_chars(AName,Name), write(user_error,AName),
	write(user_error,' not found!'), nl(user_error).
getTypeOfFunction(['Func'(Name,_,FType,_)|_],Name,Type) :- !,
	tvars2freevars(FType,[],Type,_).
getTypeOfFunction(['Type'(TC,TVs,Conss)|_],Name,Type) :-
	getTypeOfCons(Conss,Name,ArgTypes), !,
	map2(addTVar,TVs,TTVs),
	constype2funtype(ArgTypes,'TCons'(TC,TTVs),FType),
	tvars2freevars(FType,[],Type,_).
getTypeOfFunction([_|Funs],Name,FunType) :-
	getTypeOfFunction(Funs,Name,FunType).

addTVar(V,'TVar'(V)).
varIndex2VarExp(V,'Var'(V)).

constype2funtype([],T,T).
constype2funtype([T|Ts],RT,'FuncType'(T,FT)) :-	constype2funtype(Ts,RT,FT).

getTypeOfCons(['Cons'(Name,_,ArgTypes)|_],Name,ArgTypes) :- !.
getTypeOfCons([_|Conss],Name,ArgTypes) :- getTypeOfCons(Conss,Name,ArgTypes).


% transform all 'TVar's into Prolog variables:
tvars2freevars('TVar'(I),TVs,V,NTVs) :-
	addTVar2FreeVar(I,TVs,V,NTVs).
tvars2freevars('FuncType'(T1,T2),TVs,'FuncType'(NT1,NT2),NTVs) :-
	tvars2freevars(T1,TVs,NT1,NTVs1),
	tvars2freevars(T2,NTVs1,NT2,NTVs).
tvars2freevars('TCons'(TC,Ts),TVs,'TCons'(TC,NTs),NTVs) :-
	tvars2freevarsOnList(Ts,TVs,NTs,NTVs).

tvars2freevarsOnList([],TVs,[],TVs).
tvars2freevarsOnList([T|Ts],TVs,[NT|NTs],NTVs) :-
	tvars2freevars(T,TVs,NT,TVs1),
	tvars2freevarsOnList(Ts,TVs1,NTs,NTVs).

addTVar2FreeVar(I,[],V,[(I,V)]). % new entry for first variable occurrence
addTVar2FreeVar(I,[(I,V)|TVs],V,[(I,V)|TVs]) :- !.
addTVar2FreeVar(I,[TV|TVs],V,[TV|NTVs]) :- addTVar2FreeVar(I,TVs,V,NTVs).

	       
% transform all Prolog variables into numbered 'TVar's:
freevars2tvars(V,I,I1) :-
	var(V), !,
	V='TVar'(I),
	I1 is I+1.
freevars2tvars('TVar'(_),I,I).
freevars2tvars('FuncType'(T1,T2),I,I2) :-
	freevars2tvars(T1,I,I1),
	freevars2tvars(T2,I1,I2).
freevars2tvars('TCons'(_,Ts),I,J) :-
	freevars2tvarsOnList(Ts,I,J).

freevars2tvarsOnList([],I,I).
freevars2tvarsOnList([T|Ts],I,J) :-
	freevars2tvars(T,I,I1),
	freevars2tvarsOnList(Ts,I1,J).


% compute the free variables in an expression (where variables are represented
% as integers):
freeVarsInExp('Var'(V),[V]).
freeVarsInExp('Lit'(_),[]).
freeVarsInExp('Comb'(_,_,Terms),V) :-
	freeVarsInExps(Terms,V).
freeVarsInExp('Apply'(F,X),V) :-
	freeVarsInExp(F,V1),
	freeVarsInExp(X,V2),
	union(V1,V2,V).
freeVarsInExp('Constr'(CVars,Exp),V) :-
	freeVarsInExp(Exp,VE),
	diff(VE,CVars,V).
freeVarsInExp('Or'(E1,E2),V) :-
	freeVarsInExp(E1,V1),
	freeVarsInExp(E2,V2),
	union(V1,V2,V).
freeVarsInExp('Case'(_,CE,Branches),V) :-
	freeVarsInExp(CE,V1),
	freeVarsInBranches(Branches,V2),
	union(V1,V2,V).
freeVarsInExp('GuardedExpr'(GVars,Con,Exp),V) :-
	freeVarsInExp(Con,V1),
	freeVarsInExp(Exp,V2),
	union(V1,V2,V12),
	diff(V12,GVars,V).
freeVarsInExp('Choice'(Exp),V) :-
	freeVarsInExp(Exp,V).

freeVarsInExps([],[]).
freeVarsInExps([E|Es],V) :-
	freeVarsInExp(E,V1),
	freeVarsInExps(Es,V2),
	union(V1,V2,V).

freeVarsInBranches([],[]).
freeVarsInBranches(['Branch'('Pattern'(_,PVs),E)|Bs],V) :-
	freeVarsInExp(E,V1),
	diff(V1,PVs,V2),
	freeVarsInBranches(Bs,V3),
	union(V2,V3,V).
freeVarsInBranches(['Branch'('LPattern'(_),E)|Bs],V) :-
	freeVarsInExp(E,V1),
	freeVarsInBranches(Bs,V2),
	union(V1,V2,V).


% does a (logical) variable occures in an expression (where variables
% are represented as logical variables)?
occursInExp(V,'Var'(X)) :- X==V.
occursInExp(V,'Comb'(_,_,Terms)) :- occursInExps(V,Terms).
occursInExp(V,'Apply'(F,X)) :- occursInExp(V,F) ; occursInExp(V,X).
occursInExp(V,'Constr'(_,Exp)) :- occursInExp(V,Exp).
occursInExp(V,'Or'(E1,E2)) :- occursInExp(V,E1) ; occursInExp(V,E2).
occursInExp(V,'Case'(_,CE,Branches)) :-
	occursInExp(V,CE) ; occursInBranches(V,Branches).
occursInExp(V,'GuardedExpr'(_,Con,Exp)) :-
	occursInExp(V,Con) ; occursInExp(V,Exp).
occursInExp(V,'Choice'(Exp)) :-	occursInExp(V,Exp).

occursInExps(V,[E|Es]) :- occursInExp(V,E) ; occursInExps(V,Es).

occursInBranches(V,['Branch'(_,E)|Bs]) :-
	occursInExp(V,E) ; occursInBranches(V,Bs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliaries for handling dynamic data:

% Add new initialization directives for dynamic data (used for saving states):
addDynamicNameInfo(PredName,FileName) :-
	retract(dynamicPredNames(Ds)),
	asserta(dynamicPredNames([(PredName,FileName)|Ds])), !.

% Write initial info clause for dynamic predicates:
writeDynamicInfoClause((PredName,'')) :- !,
	writeClause(dynamicPredInfo(PredName,'')).
writeDynamicInfoClause((PredName,FileName)) :-
	evalToken(Eval),
	writeClause((dynamicPredInfo(PredName,RF) :-
		       nf(FileName,R,Eval,E1), boolEq(R,R,_,E1,_),
		       string2Atom(R,RA), checkDynamicAccessMethod(RA,RF),
		       retract((dynamicPredInfo(PredName,_) :- _)),
		       asserta(dynamicPredInfo(PredName,RF)))).

% check the access method for persistent predicates (currently: "file:")
% and remove it:
checkDynamicAccessMethod(AccName,Name) :-
	atom_chars(AccName,AccNameS),
	append("file:",NameS,AccNameS),
	!,
	atom_chars(Name,NameS).
checkDynamicAccessMethod(AccName,AccName) :-
	appendAtom('Illegal access method for persistent predicate "',AccName,Err1),
	appendAtom(Err1,'"! (must be "file:")',Err2),
	raise_exception(Err2).


% retract all facts about dynamic predicates occurring in a dynamic
% initialization directive:
retractDynamicPreds :-
	dynamicPredInfo(Name/Arity,_),
	retractAllFacts(Name/Arity),
	fail.
retractDynamicPreds.

% clear all non-persistent facts about dynamic predicates occurring
% in a dynamic initialization directive:
clearDynamicPreds :-
	dynamicPredInfo(Name/Arity,Dir),
	clearDynamicPred(Name/Arity,Dir),
	fail.
clearDynamicPreds.

clearDynamicPred(Name/Arity,'') :- retractAllFacts(Name/Arity), !.
clearDynamicPred(Name/Arity,_) :- retractDeadDynamicFacts(Name/Arity), !.


% is the rhs of a rule a dynamic declaration?
isDynamicRuleDecl([],'Comb'('FuncCall',"Dynamic.dynamic",[]),'') :- !.
isDynamicRuleDecl([],'Comb'('FuncCall',"Dynamic.persistent",[S]),S) :- !.
isDynamicRuleDecl(Args,'Apply'(Exp,'Var'(A)),S) :-
	append(As,[A],Args),
	isDynamicRuleDecl(As,Exp,S).


% check the type of dynamic predicates (i.e., result type Dynamic and
% monomorphism restriction):
checkDynamicType(PredName,'FuncType'(T1,T2)) :- !,
	checkDynamicTypeForMonomorphism(PredName,T1),
	checkDynamicType(PredName,T2).
checkDynamicType(_,'TCons'("Dynamic.Dynamic",[])) :- !.
checkDynamicType(PredName,_) :-
	write(user_error,'ERROR: Dynamic predicate "'),
	write(user_error,PredName),
	write(user_error,'" has not result type "Dynamic"!'),  nl(user_error),
	setFlcBug.

checkDynamicTypeForMonomorphism(PredName,'FuncType'(T1,T2)) :-
	checkDynamicTypeForMonomorphism(PredName,T1),
	checkDynamicTypeForMonomorphism(PredName,T2).
checkDynamicTypeForMonomorphism(PredName,'TCons'(_,Ts)) :-
	map1partial(checkDynamicTypeForMonomorphism(PredName),Ts).
checkDynamicTypeForMonomorphism(PredName,'TVar'(_)) :-
	write(user_error,'ERROR: Type of dynamic predicate "'),
	write(user_error,PredName),
	write(user_error,'" contains type variable!'),  nl(user_error),
	setFlcBug.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of a defined function
writeFunc('Func'(_,_,_,'External'(_))). % no code for external functions

% special code for dynamic predicates:
writeFunc('Func'(Name,FArity,Type,'Rule'(Args,Exp))) :-
	isDynamicRuleDecl(Args,Exp,DynDeclArg), !,
	flatName2Atom(Name,FName),
	checkDynamicType(FName,Type),
	atom_chars(FName,FNameS),
	append(_,[46|UQNameS],FNameS),
	append("$DYN_",UQNameS,DynNameS),
	atom_chars(DynName,DynNameS),
        (DynDeclArg = ''
	  -> DynType = 'Dynamic.Temporary'
	   ; exp2Term([],DynDeclArg,ENameTerm),
	     DynType = 'Dynamic.Persistent'),
        length(Xs,FArity),
	length(Ys,FArity),
	PYs =.. [DynName,DynType,0,0|Ys],
	append(Xs,['Dynamic.Dynamic'(PYs),E0,E],HeadArgs),
	Head =.. [FName|HeadArgs],
	genNfArgs(Xs,Ys,E0,E,Body),
	writeClause((Head :- Body)),
	DArity is FArity+2,
	writeClause((:- dynamic DynName/DArity)),
	(DynType = 'Dynamic.Temporary'
	 -> addDynamicNameInfo(DynName/DArity,'')
	  ; addDynamicNameInfo(DynName/DArity,ENameTerm)),
	nl, !.

writeFunc('Func'(Name,FArity,_Type,'Rule'(FlatArgs,FlatExp))) :-
	flatName2Atom(Name,FName),
	retract(currentFunction(_)),
	asserta(currentFunction(FName)),
	% only for debugging:
	  %write(user_error,'*** Translating rule: '),
	  %write(user_error,FName), write(user_error,' '),
	  %write(user_error,FlatArgs), write(user_error,' = '),
	  %write(user_error,FlatExp), nl(user_error),
	write(':- block '), writeq(FName), write('(?,'),
	writeNTimes(FArity,'?,'), write('-,?).'), nl,
	flatargs2var(FlatArgs,Env,Args),
	flatexp2var(Env,FlatExp,RHS),
	(compileWithDebug ->
	   % generate clause for debugging:
	   append(Args,[R,_E0,E],AllArgs),
	   appendAtom(FName,'_DB',FNameDB),
	   LHS =.. [FName|Args],
	   LHSDB =.. [FName|AllArgs],
	   RHSDB =.. [FNameDB|AllArgs],
	   writeClause((LHSDB :- traceCall(LHS,Skip), RHSDB, traceExit(LHS,R,E,Skip))),
	   transExp(FNameDB,'',Args,Args,RHS)
	 ; transExp(FName,'',Args,Args,RHS)), nl.

genNfArgs([],[],E,E,true).
genNfArgs([X|Xs],[Y|Ys],E0,E,(nf(X,Y,E0,E1),Goal)) :-
	genNfArgs(Xs,Ys,E1,E,Goal).

% assign to each FlatCurry argument (an integer) in an argument list a new
% logic variable and unify second argument with this assignment, e.g.,
% flatargs2var([1,2,3],E,L) --> E=[(1,X),(2,Y),(3,Z)],L=[X,Y,Z]
flatargs2var([],[],[]).
flatargs2var([I|Is],[(I,Var)|E],[Var|Vars]) :- flatargs2var(Is,E,Vars).

% replace in a FlatCurry expression all variable indices by logic variables:
flatexp2var(Env,'Var'(I),'Var'(V)) :- !,
	getVarInEnv(I,Env,V).
flatexp2var(_,'Lit'(L),'Lit'(L)) :- !.
flatexp2var(Env,'Comb'(CT,CF,Exps),'Comb'(CT,CF,VarExps)) :- !,
	map2partial(flatexp2var(Env),Exps,VarExps).
flatexp2var(Env,'Apply'(E1,E2),'Apply'(VE1,VE2)) :- !,
	flatexp2var(Env,E1,VE1),
	flatexp2var(Env,E2,VE2).
flatexp2var(Env,'Constr'(VarIs,Exp),'Constr'(Vars,VarExp)) :- !,
	flatargs2var(VarIs,VarEnv,Vars),
	append(VarEnv,Env,Env1),
	flatexp2var(Env1,Exp,VarExp).
flatexp2var(Env,'Or'(E1,E2),'Or'(VE1,VE2)) :- !,
	flatexp2var(Env,E1,VE1),
	flatexp2var(Env,E2,VE2).
flatexp2var(Env,'Case'(CT,Exp,Cases),'Case'(CT,VarExp,VarCases)) :- !,
	flatexp2var(Env,Exp,VarExp),
	map2partial(flatcases2var(Env),Cases,VarCases).
flatexp2var(Env,'GuardedExpr'(VarIs,E1,E2),'GuardedExpr'(Vars,VE1,VE2)) :- !,
	flatargs2var(VarIs,VarEnv,Vars),
	append(VarEnv,Env,Env1),
	flatexp2var(Env1,E1,VE1),
	flatexp2var(Env1,E2,VE2).
flatexp2var(Env,'Choice'(Exp),'Choice'(VarExp)) :- !,
	flatexp2var(Env,Exp,VarExp).
flatexp2var(_,Expr,Expr) :-
	write(user_error,'ERROR in FlatCurry file: Unknown expression "'),
	write(user_error,Expr),
	write(user_error,'" in function "'),
	currentFunction(FName), write(user_error,FName),
	write(user_error,'"!'), nl(user_error),
	pleaseReport,
	setFlcBug,
	write('\%ERROR in FlatCurry file: Unknown expression "'),
	write(Expr), write('" in function "'),
	write(FName), write('"!'), nl.

flatcases2var(Env,'Branch'('Pattern'(L,VarIs),Exp),
	          'Branch'('Pattern'(L,Vars),VarExp)) :-
	flatargs2var(VarIs,VarEnv,Vars),
	append(VarEnv,Env,Env1),
	flatexp2var(Env1,Exp,VarExp).
flatcases2var(Env,'Branch'('LPattern'(L),Exp),
	          'Branch'('LPattern'(L),VarExp)) :-
	flatexp2var(Env,Exp,VarExp).

getVarInEnv(_,[],_) :-
	write(user_error,'ERROR in FlatCurry file: '),
	write(user_error,'undeclared variable in function "'),
	currentFunction(FName), write(user_error,FName),
	write(user_error,'"!'), nl(user_error),
	pleaseReport,
	setFlcBug,
	write('\%ERROR in FlatCurry file: '),
	write('undeclared variable in function "'),
	write(FName), write('"!'), nl.
getVarInEnv(I,[(I,Var)|_],Var) :- !.
getVarInEnv(I,[_|Env],Var) :- getVarInEnv(I,Env,Var).

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% write clauses with translation, type, and evaluation information

writeTransClause('Trans'(NameS,InternalNameS)) :-
	atom_chars(Name,NameS),
	atom_chars(InternalName,InternalNameS),
	writeClause(translate(Name,InternalName)).

writeFTypeClause('Func'(Name,_FArity,FlatType,_)) :-
	flatName2Atom(Name,FName),
	replaceTVarByLVar([],FlatType,_,FlatTypeP),
	writeClause(functiontype(FName,FlatTypeP)).

writeDTypeClause('Type'(TypeName,TypeArgs,ConsExprs)) :-
	map2(index2tvar,TypeArgs,TypeArgExps),
	ResultType = 'TCons'(TypeName,TypeArgExps),
	writeDTypeClauses(ResultType,ConsExprs).

index2tvar(I,'TVar'(I)). % transform tvar index into type expression

writeDTypeClauses(_,[]).
writeDTypeClauses(ResultType,['Cons'(ConsName,_,ArgTypes)|Cs]) :-
	flatName2Atom(ConsName,Cons),
	append(ArgTypes,[ResultType],TypeL),
	typelist2flattype(TypeL,CType),
	replaceTVarByLVar([],CType,_,CTypeP),
	writeClause(functiontype(Cons,CTypeP)),
	writeDTypeClauses(ResultType,Cs).

typelist2flattype([Type],Type) :- !.
typelist2flattype([T1|T2L],'FuncType'(T1,T2)) :-
	!,
	typelist2flattype(T2L,T2).


% remove all "TVar" constructors in a type expression by logic variables
% and replace FlatCurry names by atoms for better readability of
% the generated Prolog programs:
replaceTVarByLVar(Env,'TVar'(I),Env,TVar) :-
	getTVarInEnv(I,Env,TVar), !.
replaceTVarByLVar(Env,'TVar'(I),[(I,TVar)|Env],TVar).
replaceTVarByLVar(Env,'FuncType'(T1,T2),NewEnv,'FuncType'(TT1,TT2)) :-
	replaceTVarByLVar(Env,T1,Env1,TT1),
	replaceTVarByLVar(Env1,T2,NewEnv,TT2).
replaceTVarByLVar(Env,'TCons'(Name,Types),NewEnv,'TCons'(Cons,TTypes)) :-
	flatName2Atom(Name,Cons),
	replaceTVarByLVarL(Env,Types,NewEnv,TTypes).

replaceTVarByLVarL(Env,[],Env,[]).
replaceTVarByLVarL(Env,[Type|Types],NewEnv,[TType|TTypes]) :-
	replaceTVarByLVar(Env,Type,Env1,TType),
	replaceTVarByLVarL(Env1,Types,NewEnv,TTypes).

getTVarInEnv(I,[(I,TVar)|_],TVar) :- !.
getTVarInEnv(I,[_|Env],TVar) :- getTVarInEnv(I,Env,TVar).


% write the clauses containing the evaluation annotation:
writeEvalClause('Func'(Name,_FArity,_Type,'External'(ExtName))) :-
	flatName2Atom(Name,FName),
	atom_chars(ExtAtom,ExtName),
	writeClause(evaluation(FName,'External'(ExtAtom))).
writeEvalClause('Func'(Name,_FArity,_Type,'Rule'(Args,RHS))) :-
	flatName2Atom(Name,FName),
	transformEval(RHS,TransRHS),
	writeClause(evaluation(FName,'Rule'(Args,TransRHS))).

% basically, replace the character lists for function/constructor names
% by atoms:
transformEval('Var'(A),'Var'(A)).
transformEval('Lit'(L),'Lit'(L)).
transformEval('Apply'(E1,E2),'Apply'(TE1,TE2)) :-
	transformEval(E1,TE1), transformEval(E2,TE2).
transformEval('Or'(E1,E2),'Or'(TE1,TE2)) :-
	transformEval(E1,TE1), transformEval(E2,TE2).
transformEval('GuardedExpr'(Vs,E1,E2),'GuardedExpr'(Vs,TE1,TE2)) :-
	transformEval(E1,TE1), transformEval(E2,TE2).
transformEval('Constr'(Vs,E),'Constr'(Vs,TE)) :-
	transformEval(E,TE).
transformEval('Choice'(E),'Choice'(TE)) :-
	transformEval(E,TE).
transformEval('Comb'(CT,Name,Args),'Comb'(CT,Atom,TArgs)) :-
	flatName2Atom(Name,Atom),
	map2(transformEval,Args,TArgs).
transformEval('Case'(CT,E,Cases),'Case'(CT,TE,TCases)) :-
	transformEval(E,TE),
	map2(transformEvalC,Cases,TCases).


transformEvalC('Branch'('Pattern'(Name,Vars),E),
	       'Branch'('Pattern'(Cons,Vars),TE)) :-
	flatName2Atom(Name,Cons),
	transformEval(E,TE).
transformEvalC('Branch'('LPattern'(Lit),E),
	       'Branch'('LPattern'(C),TE)) :-
	transCaseLit2Cons(Lit,C),
	transformEval(E,TE).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for evaluation to head normal form

transhnf(Funcs,ExtFuncs) :-
	writeClause((:- block hnf(?,?,-,?))),
	writeClause((hnf(T,H,E0,E) :- var(T), !, H=T, E0=E)),
	(compileWithSharing
	 -> costCenters(CCs), map1(genShareHnfClause,CCs)
	  ; true),
	map1(transhnf_clause,Funcs),
	map1partial(transExthnf,ExtFuncs),
	writeClause(hnf(T,T,E,E)), nl.

genShareHnfClause(Suffix) :-
	appendAtom(share,Suffix,Share),
	Share_M =.. [Share,M],
	appendAtom(propagateShare,Suffix,PropShare),
	PropShare_HV_R =.. [PropShare,HV,R],
	writeClause((hnf(Share_M,R,E0,E) :- !, get_mutable(V,M),
                        (V='$eval'(Expr)
                         -> R=Expr, E0=E
                          ; hnf(V,HV,E0,E1),
			    ((nonvar(HV),
			     (functor(HV,'$io',1);functor(HV,success,0)))
			      -> R=HV % no sharing for IO-act./constraints
                               ; PropShare_HV_R),
			    update_mutable('$eval'(R),M), E1=E))).

% generate hnf clause for an external function w.r.t. specified evaluation
% of arguments:
transExthnf((success/0,_)) :- !. % success is treated as a constructor
transExthnf((FName/0,success)) :- !,
	% if somebody defines FName as (External success)
	writeClause((hnf(FName,success,E,E) :- !)).
transExthnf((FName/2,PName)) :- % =:=...
	atom_chars(PName,PNameS),
	append("=:=",CCS,PNameS),
	!,
	append("constrEq",CCS,CENameS),
	atom_chars(ConstrEq,CENameS),
	FCall =.. [FName,A,B],
	ConstrEq_A_B_E0_E =.. [ConstrEq,A,B,E0,E],
	(compileWithFailPrint
	 -> writeClause((hnf(FCall,success,E0,E) :- !,
			        (ConstrEq_A_B_E0_E ; failprint(FCall,E0,E))))
	  ; writeClause((hnf(FCall,success,E0,E) :- !, ConstrEq_A_B_E0_E))
	).
transExthnf((FName/FArity,PName)) :-
	deleteCostCenterInName(PName,PNameWOCC),
	(externalSpec(PNameWOCC,EvalArgs,Lib)
	 -> addLibraryInclusion(Lib)
	  ; write(user_error,'ERROR: Specification of external function "'),
	    write(user_error,PName),
	    write(user_error,'" not found!'), nl(user_error),
	    setFlcBug, fail),
	(length(EvalArgs,FArity) -> true
	  ; write(user_error,'ERROR: Inconsistent number of arguments in external function "'),
	    write(user_error,PName),
	    write(user_error,'"!'), nl(user_error),
	    setFlcBug, fail),
	functor(FCall,FName,FArity),
	FCall =.. [_|Args],
	atom_chars(FName,FNameS),
	getCostCenterOfName(FNameS,CC),
	genEvalLiterals(EvalArgs,Args,CC,E0,E1,EvalLits,HArgs,LastGoal),
	append(HArgs,[R,E1,E],PrimArgs),
	extName2PrologExtName(PName,PrimName),
	PrimCall =.. [PrimName|PrimArgs],
	(compileWithDebug ->
	      LastGoal = (traceCall(FCall,Skip),
                            PrimCall,
	                  traceExit(FCall,R,E,Skip))
            ; LastGoal = PrimCall),
	 writeClause((hnf(FCall,R,E0,E) :- !, EvalLits)).
transExthnf(_,_) :- !.

% translate Flat external name into external name of Prolog predicate:
extName2PrologExtName(apply,apply) :- !.
extName2PrologExtName(EName,PName) :- % ==...  -->  boolEq...
	atom_chars(EName,ENameS),
	append("==",CCS,ENameS), !,
	append("boolEq",CCS,PNameS),
	atom_chars(PName,PNameS).
extName2PrologExtName(EName,PredName) :-
	appendAtom('prim_',EName,PredName).

genEvalLiterals([],[],_,E,E,LastGoal,[],LastGoal).
genEvalLiterals([hnf|Evals],[Arg|Args],CC,E0,E,
                (hnf(Arg,HArg,E0,E1),Lits),[HArg|HArgs],LastGoal) :-
	genEvalLiterals(Evals,Args,CC,E1,E,Lits,HArgs,LastGoal).
genEvalLiterals([nf|Evals],[Arg|Args],CC,E0,E,
                (Nf_Arg_HArg,Lits),[HArg|HArgs],LastGoal) :-
	appendAtom(nf,CC,Nf),
	Nf_Arg_HArg =.. [Nf,Arg,HArg,E0,E1],
	genEvalLiterals(Evals,Args,CC,E1,E,Lits,HArgs,LastGoal).
genEvalLiterals([gnf|Evals],[Arg|Args],CC,E0,E,
		(Nf_Arg_HArg,BoolEq_HArg,Lits),[HArg|HArgs],LastGoal) :-
	appendAtom(nf,CC,Nf),
	Nf_Arg_HArg =.. [Nf,Arg,HArg,E0,E1],
	appendAtom(boolEq,CC,BoolEq),
	BoolEq_HArg =.. [BoolEq,HArg,HArg,_,E1,E2],
	genEvalLiterals(Evals,Args,CC,E2,E,Lits,HArgs,LastGoal).
genEvalLiterals([none|Evals],[Arg|Args],CC,E0,E,Lits,[Arg|HArgs],LastGoal) :-
	genEvalLiterals(Evals,Args,CC,E0,E,Lits,HArgs,LastGoal).

transhnf_clause('Func'(_,_,_,'External'(_))).
				% hnf-clauses for external functions
                                % are generated by transExthnf
transhnf_clause('Func'(Name,_,_,'Rule'(FlatArgs,_))) :-
	flatName2Atom(Name,FName),
	flatargs2var(FlatArgs,_,Args),
	LHS =.. [FName|Args],
	append(Args,[H,E0,E],RArgs),
	F_Call =.. [FName|RArgs],
	writeClause((hnf(LHS,H,E0,E) :- !, F_Call)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for higher-order application

transapply(Funcs) :-
	writeClause((:- block apply(-,?,?,?,?), apply(?,?,?,-,?))),
	map1(transapply_clause,Funcs),
	nl.

transapply_clause(FName/Arity) :-
	Arity>0, !,
	transapply_partial_clauses(FName,0,Arity),
	Arity1 is Arity-1,
	functor(F_Call,FName,Arity1),
	F_Call =.. [_|Args],
	append(Args,[X],ArgsX),
	F_Call_X =.. [FName|ArgsX],
	writeClause((apply(F_Call,X,H,E0,E) :- hnf(F_Call_X,H,E0,E))).
transapply_clause(_).

transapply_partial_clauses(FName,K,Arity) :-
	K < Arity-1, !,
	functor(F_Call,FName,K),
	F_Call =.. [_|Args],
	append(Args,[X],ArgsX),
	F_Call_X =.. [FName|ArgsX],
	writeClause(apply(F_Call,X,F_Call_X,E,E)),
	K1 is K+1,
	transapply_partial_clauses(FName,K1,Arity).
transapply_partial_clauses(_,_,_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for solving equational constraints

trans_constrEq(ConsList) :-
	costCenters(CCs), map1partial(trans_constrEq(ConsList),CCs).

trans_constrEq(ConsList,Suffix) :-
	appendAtom(constrEq,Suffix,ConstrEq),
	appendAtom(constrEqHnf,Suffix,ConstrEqHnf),
	ConstrEqMode =.. [ConstrEq,?,?,-,?],
	writeClause((:- block ConstrEqMode)),
	ConstrEq_A_B_E0_E =.. [ConstrEq,A,B,E0,E],
	ConstrEqHnf_HA_HB_E2_E =.. [ConstrEqHnf,HA,HB,E2,E],
	(compileWithDebug ->
            writeClause((ConstrEq_A_B_E0_E :- traceCall(=:=(A,B),Skip),
	                                 hnf(A,HA,E0,E1), hnf(B,HB,E1,E2),
	                                 ConstrEqHnf_HA_HB_E2_E,
					 traceExit(=:=(A,B),success,E,Skip)))
          ; writeClause((ConstrEq_A_B_E0_E :- hnf(A,HA,E0,E1),hnf(B,HB,E1,E2),
	                                 ConstrEqHnf_HA_HB_E2_E))),
	nl,
	ConstrEqHnfMode =.. [ConstrEqHnf,?,?,-,?],
	writeClause((:- block ConstrEqHnfMode)),
	%writeClause((constrEq_hnf(X,H,E0,E) :- var(X), !, bind(X,H,E0,E))),
	%writeClause((constrEq_hnf(H,X,E0,E) :- var(X), !, bind(X,H,E0,E))),
	ConstrEqHnf_X_H_E0_E =.. [ConstrEqHnf,X,H,E0,E],
	ConstrEqHnf_H_X_E0_E =.. [ConstrEqHnf,H,X,E0,E],
	appendAtom(bindTryNf,Suffix,BindTryNf),
	BindTryNf_X_H_E0_E =.. [BindTryNf,X,H,E0,E],
	writeClause((ConstrEqHnf_X_H_E0_E :- var(X), !, BindTryNf_X_H_E0_E)),
	writeClause((ConstrEqHnf_H_X_E0_E :- var(X), !, BindTryNf_X_H_E0_E)),
	% to do: extend for inclusion of suffix
	%(runtimeEfficient -> map1(trans_constrEq_hnf,ConsList) ; true),
	ConstrEqHnf_T1_T2_E0_E =.. [ConstrEqHnf,T1,T2,E0,E],
	writeClause((ConstrEqHnf_T1_T2_E0_E :- integer(T1), !, T1=T2, E0=E)),
	writeClause((ConstrEqHnf_T1_T2_E0_E :- float(T1), !, {T1=T2}, E0=E)), %CLP
	ConstrEqHnf_A_B_E0_E =.. [ConstrEqHnf,A,B,E0,E],
	appendAtom(genConstrEqHnfBody,Suffix,GenConstrEqHnfBody),
	GenConstrEqHnfBody_1_NA =.. [GenConstrEqHnfBody,1,NA,A,B,EqBody],
	writeClause((ConstrEqHnf_A_B_E0_E :-
		        functor(A,FA,NA), functor(B,FB,NB),
		        FA==FB, NA==NB, GenConstrEqHnfBody_1_NA,
		        hnf(EqBody,_,E0,E))),
	nl,
	GenConstrEqHnfBody_N_NA_Succ =.. [GenConstrEqHnfBody,N,NA,_,_,'success'],
	writeClause((GenConstrEqHnfBody_N_NA_Succ :- N>NA,!)),
	appendAtom('=:=',Suffix,Eq),
	Eq_ArgA_ArgB =.. [Eq,ArgA,ArgB],
	GenConstrEqHnfBody_N_NA_Eq =.. [GenConstrEqHnfBody,N,NA,A,B,Eq_ArgA_ArgB],
	writeClause((GenConstrEqHnfBody_N_NA_Eq :- N=NA, !,
		       arg(N,A,ArgA), arg(N,B,ArgB))),
	appendAtom('&',Suffix,ConcAnd),
	ConcAnd_Eq =.. [ConcAnd,Eq_ArgA_ArgB,RemBody],
	GenConstrEqHnfBody_N_NA_And =.. [GenConstrEqHnfBody,N,NA,A,B,ConcAnd_Eq],
	GenConstrEqHnfBody_N1_NA =.. [GenConstrEqHnfBody,N1,NA,A,B,RemBody],
	writeClause((GenConstrEqHnfBody_N_NA_And :-
		       arg(N,A,ArgA), arg(N,B,ArgB),
		       N1 is N+1, GenConstrEqHnfBody_N1_NA)),
	nl,
	% optimization: try to normalize the term before binding:
	BindTryNf_X_T_E0_E =.. [BindTryNf,X,T,E0,E],
	appendAtom(nf,Suffix,Nf),
	Nf_T_NT_E0_E1 =.. [Nf,T,NT,E0,E1],
	appendAtom(bindDirect,Suffix,BindDirect),
	appendAtom(bind,Suffix,Bind),
	BindDirect_X_NT_E1_E =.. [BindDirect,X,NT,E1,E],
	Bind_X_T_E0_E =.. [Bind,X,T,E0,E],
	writeClause((BindTryNf_X_T_E0_E :- Nf_T_NT_E0_E1,
	                       (nonvar(E1) -> BindDirect_X_NT_E1_E
                                            ; Bind_X_T_E0_E))),
	BindDirect_X_T_E0_E =.. [BindDirect,X,T,E0,E],
	appendAtom(occursNot,Suffix,OccursNot),
	OccursNot_X_T =.. [OccursNot,X,T],
	writeClause((BindDirect_X_T_E0_E :- var(T), !, X=T, E0=E)),
	writeClause((BindDirect_X_T_E0_E :- OccursNot_X_T, X=T, E0=E)),
	nl,
	Bind_X_T_E0_E =.. [Bind,X,T,E0,E],
	writeClause((Bind_X_T_E0_E :- var(T), !, X=T, E0=E)),
	%(runtimeEfficient -> map1(transbind,ConsList) ; true),
	writeClause((Bind_X_T_E0_E :- integer(T), !, X=T, E0=E)),
	writeClause((Bind_X_T_E0_E :- float(T), !, {X=T}, E0=E)), %CLP
	Bind_A_B_E0_E =.. [Bind,A,B,E0,E],
	appendAtom(occursNotArgs,Suffix,OccursNotArgs),
	OccursNotArgs_1_NB_A_B =.. [OccursNotArgs,1,NB,A,B],
	appendAtom(bindArgs,Suffix,BindArgs),
	BindArgs_1_NB_A_B_E0_E =.. [BindArgs,1,NB,A,B,E0,E],
	writeClause((Bind_A_B_E0_E :-
		        functor(B,FB,NB), OccursNotArgs_1_NB_A_B,
		        functor(A,FB,NB), BindArgs_1_NB_A_B_E0_E)),
	nl,
	OccursNotArgs_N_NA_A_B =.. [OccursNotArgs,N,NA,A,B],
	OccursNotArgs_N1_NA_A_B =.. [OccursNotArgs,N1,NA,A,B],
	OccursNot_A_ArgB =.. [OccursNot,A,ArgB],
	writeClause((OccursNotArgs_N_NA_A_B :- N>NA,!)),
	writeClause((OccursNotArgs_N_NA_A_B :- arg(N,B,ArgB),
		        OccursNot_A_ArgB,
		        N1 is N+1, OccursNotArgs_N1_NA_A_B)),
	nl,
	BindArgs_N_NA_A_B_E0_E =.. [BindArgs,N,NA,A,B,E0,E],
	Bind_ArgA_HArgB_E1_E2 =.. [Bind,ArgA,HArgB,E1,E2],
	BindArgs_N1_NA_A_B_E2_E =.. [BindArgs,N1,NA,A,B,E2,E],
	writeClause((BindArgs_N_NA_A_B_E0_E :- N>NA,!, E0=E)),
	writeClause((BindArgs_N_NA_A_B_E0_E :-
		        arg(N,A,ArgA), arg(N,B,ArgB),
		        hnf(ArgB,HArgB,E0,E1), Bind_ArgA_HArgB_E1_E2,
		        N1 is N+1, BindArgs_N1_NA_A_B_E2_E)),
	nl,
	OccursNot_X_Y =.. [OccursNot,X,Y],
	writeClause((OccursNot_X_Y :- var(Y), !, X\==Y)),
	map1partial(transocc_not(OccursNot),ConsList),
	writeClause(OccursNot_X_Y),
	nl.

trans_constrEq_hnf(Cons/Arity) :-
	functor(CX,Cons,Arity),
	functor(CY,Cons,Arity),
	CX =.. [_|Xs],
	CY =.. [_|Ys],
	(Arity=0 -> writeClause((constrEq_hnf(CX,CY,E,E) :- !))
         ; gen_constrEq_hnf_body(Xs,Ys,Body),
	   writeClause((constrEq_hnf(CX,CY,E0,E) :- !, hnf(Body,_,E0,E)))).

gen_constrEq_hnf_body([X],[Y],(X=:=Y)).
gen_constrEq_hnf_body([X1,X2|Xs],[Y1,Y2|Ys],&((X1=:=Y1),Body)) :-
	gen_constrEq_hnf_body([X2|Xs],[Y2|Ys],Body).


transbind(Cons/Arity) :-
	functor(CX,Cons,Arity),
	functor(CY,Cons,Arity),
	CX =.. [_|Xs],
	CY =.. [_|Ys],
	gen_bind_body(Xs,Ys,E0,E,HnfBody),
	gen_bind_occ_body(X,Ys,(X=CX,HnfBody),occursNot,Body),
	writeClause((bind(X,CY,E0,E) :- !, Body)).

gen_bind_body([],[],E,E,true).
gen_bind_body([X|Xs],[Y|Ys],E0,E,(hnf(Y,HY,E0,E1),bind(X,HY,E1,E2),Body)) :-
	gen_bind_body(Xs,Ys,E2,E,Body).

gen_bind_occ_body(_,[],RestBody,_,RestBody).
gen_bind_occ_body(X,[Y|Ys],RestBody,OccursNot,(OccursNot_X_Y,Body)) :-
	OccursNot_X_Y =.. [OccursNot,X,Y],
	gen_bind_occ_body(X,Ys,RestBody,OccursNot,Body).

transocc_not(OccursNot,Cons/Arity) :-
	functor(CY,Cons,Arity),
	CY =.. [_|Ys],
	gen_bind_occ_body(X,Ys,true,OccursNot,Body),
	OccursNot_X_CY =.. [OccursNot,X,CY],
	writeClause((OccursNot_X_CY :- !, Body)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for (Boolean) strict equality test
	
trans_boolEq(ConsList) :-
	costCenters(CCs), map1partial(trans_boolEq(ConsList),CCs).

trans_boolEq(_ConsList,Suffix) :-
	appendAtom(boolEq,Suffix,BoolEq),
	appendAtom(boolEqHnf,Suffix,BoolEqHnf),
	BoolEqMode =.. [BoolEq,?,?,?,-,?],
	writeClause((:- block BoolEqMode)),
	BoolEq_A_B_R_E0_E =.. [BoolEq,A,B,R,E0,E],
	BoolEqHnf_HA_HB_R_E2_E =.. [BoolEqHnf,HA,HB,R,E2,E],
	(compileWithDebug ->
            writeClause((BoolEq_A_B_R_E0_E :- traceCall(==(A,B),Skip),
	                                    hnf(A,HA,E0,E1), hnf(B,HB,E1,E2),
	                                    BoolEqHnf_HA_HB_R_E2_E,
					    traceExit(==(A,B),R,E,Skip)))
	  ; writeClause((BoolEq_A_B_R_E0_E :- hnf(A,HA,E0,E1),hnf(B,HB,E1,E2),
	                                    BoolEqHnf_HA_HB_R_E2_E))),
	nl,
	BoolEqHnfMode1 =.. [BoolEqHnf,-,?,?,?,?],
	BoolEqHnfMode2 =.. [BoolEqHnf,?,-,?,?,?],
	BoolEqHnfMode3 =.. [BoolEqHnf,?,?,?,-,?],
	writeClause((:- block BoolEqHnfMode1, BoolEqHnfMode2, BoolEqHnfMode3)),
	% to do: extend for inclusion of suffix
	%(runtimeEfficient -> map1(trans_boolEq_hnf,ConsList) ; true),
	BoolEqHnf_A_B_R_E0_E =.. [BoolEqHnf,A,B,R,E0,E],
	writeClause((BoolEqHnf_A_B_R_E0_E :-
		       number(A), !, (A==B->R='True';R='False'), E0=E)),
	appendAtom(genBoolEqHnfBody,Suffix,GenBoolEqHnfBody),
	GenBoolEqHnfBody_1_NA =.. [GenBoolEqHnfBody,1,NA,A,B,SeqBody],
	writeClause((BoolEqHnf_A_B_R_E0_E :-
		       functor(A,FA,NA),functor(B,FB,NB),
		       ((FA==FB,NA==NB,GenBoolEqHnfBody_1_NA)
		          -> hnf(SeqBody,R,E0,E)
		           ; R='False',E0=E))),
	nl,
	GenBoolEqHnfBody_N_NA_True =.. [GenBoolEqHnfBody,N,NA,_,_,'True'],
	writeClause((GenBoolEqHnfBody_N_NA_True :- N>NA,!)),
	appendAtom('==',Suffix,Eq),
	Eq_ArgA_ArgB =.. [Eq,ArgA,ArgB],
	appendAtom('&&',Suffix,BoolAnd),
	BoolAnd_Eq =.. [BoolAnd,Eq_ArgA_ArgB,RemBody],
	GenBoolEqHnfBody_N_NA_And =.. [GenBoolEqHnfBody,N,NA,A,B,BoolAnd_Eq],
	GenBoolEqHnfBody_N1_NA =.. [GenBoolEqHnfBody,N1,NA,A,B,RemBody],
	writeClause((GenBoolEqHnfBody_N_NA_And :-
		       arg(N,A,ArgA), arg(N,B,ArgB),
		       N1 is N+1, GenBoolEqHnfBody_N1_NA)),
	nl.

trans_boolEq_hnf(Cons/Arity) :-
	functor(CX,Cons,Arity),
	functor(CY,Cons,Arity),
	CX =.. [_|Xs],
	CY =.. [_|Ys],
	gen_boolEq_hnf_body(Xs,Ys,Body),
	writeClause((boolEq_hnf(CX,CY,R,E0,E) :- !, hnf(Body,R,E0,E))).

gen_boolEq_hnf_body([],[],'True').
gen_boolEq_hnf_body([X|Xs],[Y|Ys],&&(==(X,Y),Body)) :-
	gen_boolEq_hnf_body(Xs,Ys,Body).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for evaluation to normal form
	
transnf(ConsList) :-
	costCenters(CCs), map1partial(transnf(ConsList),CCs).

transnf(_ConsList,Suffix) :-
	appendAtom(nf,Suffix,Nf),
	appendAtom(nfHnf,Suffix,NfHnf),
	NfMode =.. [Nf,?,?,-,?],
	writeClause((:- block NfMode)),
	Nf_A_V_E0_E =.. [Nf,A,V,E0,E],
	NfHnf_HA_V_E1_E =.. [NfHnf,HA,V,E1,E],
	writeClause((Nf_A_V_E0_E :- hnf(A,HA,E0,E1), NfHnf_HA_V_E1_E)),
	nl,
	NfHnfMode =.. [NfHnf,?,?,-,?],
	writeClause((:- block NfHnfMode)),
	NfHnf_T_V_E0_E =.. [NfHnf,T,V,E0,E],
	writeClause((NfHnf_T_V_E0_E :- var(T), !, V=T, E0=E)),
	% to do: extend for inclusion of suffix
	%(runtimeEfficient -> map1(transnf_hnf,ConsList) ; true),
	NfHnf_A_A_E0_E =.. [NfHnf,A,A,E0,E],
	writeClause((NfHnf_A_A_E0_E :- number(A), !, E0=E)),
	NfHnf_A_R_E0_E =.. [NfHnf,A,R,E0,E],
	appendAtom(nfHnfArgs,Suffix,NfHnfArgs),
	NfHnfArgs_1_NA_A_R_E0_E =.. [NfHnfArgs,1,NA,A,R,E0,E],
	writeClause((NfHnf_A_R_E0_E :-
		       functor(A,FA,NA), functor(R,FA,NA),
		       NfHnfArgs_1_NA_A_R_E0_E)),
	nl,
	NfHnfArgs_N_NA_A_R_E0_E =.. [NfHnfArgs,N,NA,A,R,E0,E],
	writeClause((NfHnfArgs_N_NA_A_R_E0_E :- N>NA,!,E0=E)),
	Nf_ArgA_ArgR_E0_E1 =.. [Nf,ArgA,ArgR,E0,E1],
	NfHnfArgs_N1_NA_A_R_E1_E =.. [NfHnfArgs,N1,NA,A,R,E1,E],
	writeClause((NfHnfArgs_N_NA_A_R_E0_E :-
		       arg(N,A,ArgA), arg(N,R,ArgR),
		       Nf_ArgA_ArgR_E0_E1,
		       N1 is N+1, NfHnfArgs_N1_NA_A_R_E1_E)),
	nl.

transnf_hnf(Cons/Arity) :-
	functor(CX,Cons,Arity),
	functor(CY,Cons,Arity),
	CX =.. [_|Xs],
	CY =.. [_|Ys],
	gen_nf_hnf_body(Xs,Ys,E0,E,Body),
	writeClause((nf_hnf(CX,CY,E0,E) :- !, Body)).

gen_nf_hnf_body([],[],E,E,true).
gen_nf_hnf_body([X|Xs],[Y|Ys],E0,E,(nf(X,Y,E0,E1),Body)) :-
	gen_nf_hnf_body(Xs,Ys,E1,E,Body).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of clauses for sharing propagation
	
transpropshar(ConsList) :-
	costCenters(CCs), map1partial(transpropshar(ConsList),CCs).

transpropshar(_ConsList,Suffix) :-
	appendAtom(propagateShare,Suffix,PropShare),
	PropShare_X_X =.. [PropShare,X,X],
	writeClause((PropShare_X_X :- var(X), !)),
	%(runtimeEfficient -> map1(transpropshar_clause,ConsList) ; true),
	IP='Ports.internalPort'(_,_,_,_),
	PropShare_IP_IP =.. [PropShare,IP,IP],
	writeClause((PropShare_IP_IP :- !)),
	Str='$stream'(_),
	PropShare_Str_Str =.. [PropShare,Str,Str],
	writeClause((PropShare_Str_Str :- !)),
	writeClause((PropShare_X_X :- number(X),!)),
	PropShare_A_R =.. [PropShare,A,R],
	appendAtom(propagateShareArgs,Suffix,PropShareArgs),
	PropShareArgs_1_NA_A_R =.. [PropShareArgs,1,NA,A,R],
	writeClause((PropShare_A_R :-
		       functor(A,FA,NA), functor(R,FA,NA),
		       PropShareArgs_1_NA_A_R)),
	nl,
	appendAtom(propagateShareArg,Suffix,PropShareArg),
	PropShareArg_ArgA_ArgR =.. [PropShareArg,ArgA,ArgR],
	PropShareArgs_N_NA_A_R =.. [PropShareArgs,N,NA,A,R],
	PropShareArgs_N1_NA_A_R =.. [PropShareArgs,N1,NA,A,R],
	writeClause((PropShareArgs_N_NA_A_R :- N>NA,!)),
	writeClause((PropShareArgs_N_NA_A_R :-
		       arg(N,A,ArgA), arg(N,R,ArgR),
		       PropShareArg_ArgA_ArgR,
		       N1 is N+1, PropShareArgs_N1_NA_A_R)),
	nl,
	PropShareArg_A_A =.. [PropShareArg,A,A],
	writeClause((PropShareArg_A_A :- nonvar(A),
		              functor(A,'Ports.internalPort',4), !)),
	PropShareArg_A_SA =.. [PropShareArg,A,SA],
	appendAtom(makeShare,Suffix,MakeShare),
	MakeShare_A_SA =.. [MakeShare,A,SA],
	writeClause((PropShareArg_A_SA :- MakeShare_A_SA)),
	nl,
	MakeShare_X_X =.. [MakeShare,X,X],
	appendAtom(share,Suffix,Share),
	Share_M =.. [Share,M],
	MakeShare_X_ShM =.. [MakeShare,X,Share_M],
	writeClause((MakeShare_X_X :- var(X), !)),
	writeClause((MakeShare_X_X :- X=Share_M, !)),
	writeClause((MakeShare_X_ShM :- create_mutable(X,M))),
	nl.

transpropshar_clause('Ports.internalPort'/4) :- !.
transpropshar_clause('$stream'/1) :- !.
transpropshar_clause(Cons/Arity) :-
	functor(CX,Cons,Arity),
	functor(CY,Cons,Arity),
	CX =.. [_|Xs],
	CY =.. [_|Ys],
	gen_propshar_body(Xs,Ys,Body),
	writeClause((propagateShare(CX,CY) :- !, Body)).

gen_propshar_body([],[],true).
gen_propshar_body([X|Xs],[Y|Ys],(propSharArg(X,Y),Body)) :-
	gen_propshar_body(Xs,Ys,Body).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of expressions:
% here we assume that all cases and ors occur only outermost and not
% at nested inner positions
transExp(FName,Aux,Patterns,Vars,'Case'(CType,CaseExp,Cases)) :- !,
	(CaseExp='Var'(X)
	 -> % case expression is variable:
	    selectVar(X,Vars,I,VarsX),
	    % check whether the case variable occurs in any case branch
	    (occursInBranches(X,Cases) -> NewVars=Vars ; NewVars=VarsX),
	    CaseTerm=X, UMShares=[], ShareCases=Cases,
	    number_chars(I,IS), atom_chars(IA,IS) % transform number into atom
	  ; % case expression is complex (not a variable)
	    (compileWithSharing
	        -> exp2ShareExpr(CaseExp,UMShares,SCExp),
		   map2partial(replaceShareVars(UMShares),Vars,NewVars),
		   exp2Term(NewVars,SCExp,CaseTerm),
		   map2partial(insertShareInBranch(UMShares),Cases,ShareCases)
	         ; exp2Term(Vars,CaseExp,CaseTerm),
		   UMShares=[], NewVars=Vars, ShareCases=Cases),
	    IA='ComplexCase'  % naming needs to be improved!
	),
	atom_chars(FName,FNameS), getCostCenterOfName(FNameS,CC),
	map2partial(addSuffix2MakeShare(CC),UMShares,UShares),
	((compileWithFailPrint, Aux\=='') ->
	    append(Patterns,[OLHS,H,E0,E],LArgs)
	  ; append(Patterns,[H,E0,E],LArgs)),
	LHS =.. [FName|LArgs],
	appendAtom(FName,Aux,FNaux),
	appendAtom(FNaux,'_',FNameUS),
	appendAtom(FNameUS,IA,F_I),
	(compileWithFailPrint ->
	    (Aux='' -> append(NewVars,[FP_Call,H,E1,E],F_I_Args)
	             ; append(NewVars,[OLHS,H,E1,E],F_I_Args))
	  ; append(NewVars,[H,E1,E],F_I_Args)),
	F_I_Call =.. [F_I,Y|F_I_Args],
	(compileWithFailPrint ->
	   (Aux=''
	    -> (compileWithDebug
	        -> append(FName_WODBS,"_DB",FNameS),
		   atom_chars(FName_WODB,FName_WODBS),
		   FP_Call =.. [FName_WODB|Patterns]
	         ; FP_Call =.. [FName|Patterns])
	     ; FP_Call=OLHS),
	   FailCall = failprint(FP_Call,E1,E),
	   writeClauseWithInitGoals(LHS,UShares,
			    (hnf(CaseTerm,Y,E0,E1), (F_I_Call ; FailCall)))
	 ; writeClauseWithInitGoals(LHS,UShares,
				    (hnf(CaseTerm,Y,E0,E1), F_I_Call))
	), nl,
	length(NewVars,LenNewVars), NumVars is LenNewVars+1,
	write(':- block '), writeq(F_I),
	(CType='Rigid'
	 -> write('(-,'),
	    writeNTimes(NumVars,'?,'),
	    (compileWithFailPrint -> write('?,') ; true),
	    write('?,?),'),
	    writeq(F_I), write('(?,'),
	    writeNTimes(NumVars,'?,'), write('-,?).'), nl
	  ; % CType='Flex'
	    write('(?,'),
	    writeNTimes(NumVars,'?,'),
	    (compileWithFailPrint -> write('?,') ; true),
	    write('-,?).'), nl),
	transCases(F_I,NewVars,ShareCases).

transExp(FName,Aux,Patterns,Vars,'Or'(Exp1,Exp2)) :- !,
	(((Exp1 = 'Case'(_,_,_) ; Exp2 = 'Case'(_,_,_)),
	  pakcsrc(warnoverlapping,yes))
	 -> write(user_error,'WARNING: function "'),
	    currentFunction(FuncName), write(user_error,FuncName),
	    write(user_error,'" is non-deterministic due to non-trivial overlapping rules!'),
	    nl(user_error)
	  ; true),
	transExp(FName,Aux,Patterns,Vars,Exp1),
	transExp(FName,Aux,Patterns,Vars,Exp2).

transExp(FName,Aux,Patterns,Vars,'Choice'(Exp)) :- !,
	write(user_error,'ERROR: "'),
	write(user_error,FName),
	write(user_error,' eval choice" not yet supported!'),
	nl(user_error),
	transExp(FName,Aux,Patterns,Vars,Exp),
	setFlcBug.

transExp(FName,Aux,Patterns,Vars,Exp) :-
	((compileWithFailPrint, Aux\=='')
	  -> append(Patterns,[_,H,E0,E],LArgs)
	   ; append(Patterns,[H,E0,E],LArgs)),
	LHS =.. [FName|LArgs],
	(compileWithSharing
	 -> exp2ShareExpr(Exp,UMShares,SExp),
	    map2partial(replaceShareVars(UMShares),Vars,NewVars),
	    exp2Term(NewVars,SExp,Term)
	  ; exp2Term(Vars,Exp,Term), UMShares=[]),
	atom_chars(FName,FNameS), getCostCenterOfName(FNameS,CC),
	map2partial(addSuffix2MakeShare(CC),UMShares,UShares),
	(var(Term) -> writeClauseWithInitGoals(LHS,UShares,hnf(Term,H,E0,E))
         ; functor(Term,TF,TA),
	   externalFuncs(EFs),
	   (member(TF/TA,EFs)
             % check for not yet implemented special cases
            -> writeClauseWithInitGoals(LHS,UShares,hnf(Term,H,E0,E))
             ; (isConstructorRooted(Exp) ->
	          % right-hand side is constructor-rooted:
                  H=Term,
		  (UShares=[]
		   -> E0=E, writeClause(LHS)
		    ; writeClauseWithInitGoals(LHS,UShares,E0=E))
                ; % right-hand side is operation-rooted:
	          Term =.. [TF|TArgs],
	          append(TArgs,[H,E0,E],TArgsH),
	          TFCall =.. [TF|TArgsH],
                  writeClauseWithInitGoals(LHS,UShares,TFCall) ))).

isConstructorRooted('Lit'(_)).
isConstructorRooted('Comb'('ConsCall',_,_)).
% we implement "success" as a constructor of type "Success":
isConstructorRooted('Comb'('FuncCall',"success",[])).
isConstructorRooted('Constr'(_,Exp)) :- % since we ignore Constr in the code:
	isConstructorRooted(Exp).

	
transCases(_,_,[]).
transCases(F_I,VarsX,['Branch'('Pattern'(Name,ConsVars),Exp)|Cases]) :-
	flatName2Atom(Name,Cons),
	Pattern =.. [Cons|ConsVars],
	append(ConsVars,VarsX,NewVars),
	(number(Cons) -> number_chars(Cons,ConsS),
	                 atom_chars(ConsA,ConsS)  % transform number into atom
		       ; ConsA=Cons),
	appendAtom('_',ConsA,Aux),
	transExp(F_I,Aux,[Pattern|VarsX],NewVars,Exp),
	transCases(F_I,VarsX,Cases).
transCases(F_I,VarsX,['Branch'('LPattern'(Lit),Exp)|Cases]) :-
	transCaseLit2Cons(Lit,Cons),
	(number(Cons) -> number_chars(Cons,ConsS),
	                 atom_chars(ConsA,ConsS)  % transform number into atom
		       ; ConsA=Cons),
	appendAtom('_',ConsA,Aux),
	transExp(F_I,Aux,[Cons|VarsX],VarsX,Exp),
	transCases(F_I,VarsX,Cases).

% extract constructor of literal:
transCaseLit2Cons('Intc'(I),I) :- !.
transCaseLit2Cons('Floatc'(F),F) :- !.
transCaseLit2Cons('Charc'(N),C) :- !, char_int(C,N).
transCaseLit2Cons(Lit,Lit) :-
	write(user_error,'ERROR in FlatCurry file: Illegal argument "'),
	write(user_error,Lit),
	write(user_error,'" in case branch!'), nl(user_error),
	pleaseReport,
	setFlcBug.

% translate FlatCurry expression into corresponding Prolog term:
% (first argument are the currently visible variables, only used
% to check inconsistencies in FlatCurry)
exp2Term(Vars,'Var'(V),V) :- !,
	(memberEq(V,Vars) -> true
	  ; write(user_error,'ERROR in FlatCurry file in function "'),
	    currentFunction(FuncName), write(user_error,FuncName),
	    write(user_error,'":'), nl(user_error),
	    write(user_error,'variable "'),
	    write(user_error,V),
	    write(user_error,'" does not occur in left-hand side.'),
	    nl(user_error),
	    pleaseReport,
	    setFlcBug,
	    write('\%ERROR in FlatCurry file: variable "'),
	    write(V),
	    write('" does not occur in left-hand side!'), nl).
exp2Term(_,'Lit'('Intc'(I)),I) :- !.
exp2Term(_,'Lit'('Floatc'(F)),F) :- !.
exp2Term(_,'Lit'('Charc'(N)),C) :- !, char_int(C,N).
exp2Term(_,'Lit'('Ident'(S)),A) :- !,
	write(user_error,'ERROR in FlatCurry file in function "'),
	currentFunction(FuncName), write(user_error,FuncName),
	write(user_error,'":'), nl(user_error),
	flatName2Atom(S,A),
	write(user_error,'Expression "'),
	write(user_error,'Lit'('Ident'(A))),
	write(user_error,'" should not occur in FlatCurry expressions.'),
	nl(user_error),
	pleaseReport,
	setFlcBug,
	write('\%ERROR in FlatCurry file: Expression "'),
	write('Lit'('Ident'(A))),
	write('" should not occur in FlatCurry expressions.'),
	nl.
exp2Term(Vars,'Comb'(_CombType,NameS,Exprs),Term) :- !,
	flatName2Atom(NameS,Name),
	checkForDeprecatedFunction(Name),
	checkForUnspportedFunction(Name),
	checkForDynamicConstructor(Name),
	checkForTupleArity(NameS),
	map2partial(exp2Term(Vars),Exprs,Terms),
	Term =.. [Name|Terms].

% The following translation does not work for top-level constructor
% partial applications:
%exp2Term(Vars,'Apply'(F,X),Term) :- F='Comb'('PartCall',_,_), !,
%	exp2Term(Vars,F,FT), exp2Term(Vars,X,XT),
%	FT =.. [Func|Terms], append(Terms,[XT],TermsXT),
%	Term =.. [Func|TermsXT].
exp2Term(Vars,'Apply'(F,X),apply(FT,XT)) :- !,
	exp2Term(Vars,F,FT), exp2Term(Vars,X,XT).
exp2Term(Vars,'GuardedExpr'(LocalVars,Con,Exp),cond(CT,ET)) :- !,
	append(Vars,LocalVars,NewVars),
	exp2Term(NewVars,Con,CT), exp2Term(NewVars,Exp,ET).
exp2Term(Vars,'Constr'(LocalVars,Exp),ET) :- !,
	append(Vars,LocalVars,NewVars),
	exp2Term(NewVars,Exp,ET).
exp2Term(_,Expr,'***unknown expression***') :-
	write(user_error,'ERROR in FlatCurry file: Unknown expression "'),
	write(user_error,Expr),
	write(user_error,'" in FlatCurry file!'), nl(user_error),
	pleaseReport,
	setFlcBug.


% replace multiple variables by new variables connected to the old
% ones via a makeShare goal:
exp2ShareExpr(Exp,Shares,SExp) :-
	getSharedVars(Exp,Shares),
	insertShare(Exp,Shares,SExp).

insertShareInBranch(Shares,'Branch'(Pattern,Exp),'Branch'(Pattern,SExp)) :-
        insertShare(Exp,Shares,SExp).

insertShare('Var'(V),Shares,Share) :- !,
	varToShare(V,Shares,Share).
insertShare('Lit'(L),_,'Lit'(L)) :- !.
insertShare('Comb'(CT,Name,Exprs),Shares,'Comb'(CT,Name,SExprs)) :- !,
	insertShares(Exprs,Shares,SExprs).
insertShare('Apply'(F,X),Shares,'Apply'(SF,SX)) :- !,
	insertShare(F,Shares,SF), insertShare(X,Shares,SX).
insertShare('GuardedExpr'(Vars,Con,Exp),Shares,
		  'GuardedExpr'(NewVars,SCon,SExp)) :- !,
        map2partial(replaceShareVars(Shares),Vars,NewVars),
	insertShare(Con,Shares,SCon),
	insertShare(Exp,Shares,SExp).
insertShare('Constr'(Vars,Exp),Shares,'Constr'(NewVars,SExp)) :- !,
        map2partial(replaceShareVars(Shares),Vars,NewVars),
	insertShare(Exp,Shares,SExp).
insertShare('Choice'(Exp),Shares,'Choice'(SExp)) :- !,
	insertShare(Exp,Shares,SExp).
insertShare('Case'(CType,Exp,Branches),Shares,'Case'(CType,SExp,SBranches)) :-
	!,
	insertShare(Exp,Shares,SExp),
	map2partial(insertShareInBranch(Shares),Branches,SBranches).
insertShare('Or'(Exp1,Exp2),Shares,'Or'(SExp1,SExp2)) :-
	!,
	insertShare(Exp1,Shares,SExp1),
	insertShare(Exp2,Shares,SExp2).
insertShare(Exp,_,Exp) :-
	write(user_error,'INTERNAL ERROR in "insertShare"!'),
	nl(user_error),
	write(user_error,'Unknown expression in FlatCurry file:'),
	nl(user_error),
	write(user_error,Exp),
	nl(user_error),
	pleaseReport,
	setFlcBug.

insertShares([],_,[]).
insertShares([T|Ts],Shares,[ST|STs]) :-
	insertShare(T,Shares,ST),
	insertShares(Ts,Shares,STs).

varToShare(V,[],'Var'(V)).
varToShare(V,[makeShare(W,NV)|_],'Var'(NV)) :- V==W, !.
varToShare(V,[_|Shares],Share) :- varToShare(V,Shares,Share).

replaceShareVars([],V,V).
replaceShareVars([makeShare(V,NV)|_],W,NV) :- V==W, !.
replaceShareVars([_|UVs],W,NV) :- replaceShareVars(UVs,W,NV).


% add suffix to "makeShare":
addSuffix2MakeShare(Suffix,makeShare(V,NV),MSs) :-
	appendAtom(makeShare,Suffix,MakeShare),
	MSs =.. [MakeShare,V,NV].


% get the variables in a term which occurs more than once and must be shared:
getSharedVars(Exp,Shares) :-
	countVarsInTerm(Exp,[],Vars),
	varsIntoShares(Vars,Shares).

varsIntoShares([],[]).
varsIntoShares([V/C|Vs],[makeShare(V,_)|Shs]) :-
	C>1, !,
	  % for statistics:
	  retract(numberOfShares(SC)),
	  SC1 is SC+1,
	  asserta(numberOfShares(SC1)),
	varsIntoShares(Vs,Shs).
varsIntoShares([_|Vs],Shs) :-
	varsIntoShares(Vs,Shs).

countVarsInTerm('Var'(V),Vars,IVars) :- !, incVarCount(V,Vars,IVars).
countVarsInTerm('Lit'(_),Vars,Vars) :- !.
countVarsInTerm('Comb'(_,_,Terms),Vars,IVars) :- !,
	countVarsInTerms(Terms,Vars,IVars).
countVarsInTerm('Apply'(F,X),Vars,IVars) :- !,
	countVarsInTerm(F,Vars,FVars),
	countVarsInTerm(X,FVars,IVars).
countVarsInTerm('GuardedExpr'(_,Con,Exp),Vars,IVars) :- !,
	countVarsInTerm(Con,Vars,CVars),
	countVarsInTerm(Exp,CVars,IVars).
countVarsInTerm('Constr'(_,Exp),Vars,IVars) :- !,
	countVarsInTerm(Exp,Vars,IVars).
countVarsInTerm('Choice'(Exp),Vars,IVars) :- !,
	countVarsInTerm(Exp,Vars,IVars).
countVarsInTerm(Exp,Vars,Vars) :- !,
	write(user_error,'INTERNAL ERROR in "countVarsInTerm"!'),
	nl(user_error),
	write(user_error,'Unknown expression in FlatCurry file:'),
	nl(user_error),
	write(user_error,Exp),
	nl(user_error),
	pleaseReport,
	setFlcBug.

countVarsInTerms([],Vars,Vars).
countVarsInTerms([X|Xs],Vars,IVars) :-
	countVarsInTerm(X,Vars,XVars),
	countVarsInTerms(Xs,XVars,IVars).

incVarCount(V,[],[V/1]).
incVarCount(V,[X/T|Vars],[X/T1|Vars]) :- V==X, !, T1 is T+1.
incVarCount(V,[X/T|Vars],[X/T|IVars]) :- incVarCount(V,Vars,IVars).



% select a variable in a list of variables.
% return position in the list and list of variables without this variable
selectVar(X,[Y|Ys],1,Ys) :- X==Y, !.
selectVar(X,[Y|Ys],XNr,[Y|Zs]) :- selectVar(X,Ys,Nr,Zs), XNr is Nr+1.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check correctness of FlatCurry rules and expressions.
% This checking is redundant and should only ensure
% that the front-end produces syntactically correct code.

check_flcFunction('Func'(Name,_,_,Rule)) :-
	flatName2Atom(Name,FName),
	retract(currentFunction(_)),
	asserta(currentFunction(FName)),
        check_flcRule(Rule), !.
check_flcFunction(_) :-
	write(user_error,'INTERNAL ERROR in FlatCurry file "'),
	write(user_error,'" in function "'),
	currentFunction(FName), write(user_error,FName),
	write(user_error,'"!'), nl(user_error),
	pleaseReport,
	setFlcBug.

check_flcRule('Rule'(Args,Expr)) :-
	map1(integer,Args),
	check_flcExpr(Expr).
check_flcRule('External'(Name)) :-
	check_flcString(Name).

check_flcString(S) :- map1(integer,S).

check_flcCaseType('Rigid').
check_flcCaseType('Flex').

check_flcCombType('FuncCall').
check_flcCombType('ConsCall').
check_flcCombType('PartCall').

check_flcExpr('Var'(_)) :- !.
check_flcExpr('Lit'(L)) :- !, check_flcLit(L).
check_flcExpr('Comb'(CT,Name,Es)) :- !,
	check_flcCombType(CT),
	check_flcString(Name),
	map1(check_flcExpr,Es).
check_flcExpr('Apply'(E1,E2)) :- !,
	check_flcExpr(E1), check_flcExpr(E2).
check_flcExpr('Constr'(Xs,E)) :- !,
	map1(integer,Xs),
	check_flcExpr(E).
check_flcExpr('Or'(E1,E2)) :- !,
	check_flcExpr(E1), check_flcExpr(E2).
check_flcExpr('Case'(CT,E,Cs)) :- !,
	check_flcCaseType(CT),
	check_flcExpr(E),
	map1(check_flcCase,Cs).
check_flcExpr('GuardedExpr'(Xs,C,E)) :- !,
	map1(integer,Xs),
	check_flcExpr(C), check_flcExpr(E).
check_flcExpr('Choice'(E)) :- !,
	check_flcExpr(E).
check_flcExpr(E) :-
	write(user_error,'ERROR in FlatCurry file: Illegal expression "'),
	write(user_error,E),
	write(user_error,'" in function "'),
	currentFunction(FName), write(user_error,FName),
	write(user_error,'"!'), nl(user_error),
	write(user_error,'> Probably, this is due to a compiler bug in transforming'),
	nl(user_error),
	write(user_error,'> do/let/where expressions. Suggested solution:'),
	nl(user_error),
	write(user_error,'> Simplify dependencies in the do/let/where expression.'),
	nl(user_error),
	%pleaseReport,
	setFlcBug.

check_flcLit('Intc'(_)) :- !.
check_flcLit('Floatc'(_)) :- !.
check_flcLit('Charc'(_)) :- !.
check_flcLit(Lit) :-
	write(user_error,'ERROR in FlatCurry file: Illegal literal "'),
	write(user_error,Lit),
	write(user_error,'" in function "'),
	currentFunction(FName), write(user_error,FName),
	write(user_error,'"!'), nl(user_error),
	pleaseReport,
	setFlcBug.

check_flcCase('Branch'('Pattern'(Name,Xs),E)) :- !,
	check_flcString(Name),
	map1(integer,Xs),
	check_flcExpr(E).
check_flcCase('Branch'('LPattern'(C),E)) :- !,
	check_flcLit(C),
	check_flcExpr(E).
check_flcCase(Branch) :-
	write(user_error,'ERROR in FlatCurry file: Illegal case pattern "'),
	write(user_error,Branch),
	write(user_error,'" in function "'),
	currentFunction(FName), write(user_error,FName),
	write(user_error,'"!'), nl(user_error),
	pleaseReport,
	setFlcBug.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliaries:

% generate N new free variables:
genVars(N,[]) :- N=0, !.
genVars(N,[_|Xs]) :- N1 is N-1, genVars(N1,Xs).

writeNTimes(0,_) :- !.
writeNTimes(N,S) :- N>0, write(S), N1 is N-1, writeNTimes(N1,S).

% write a clause and puts some initialization goals at the beginning
% of the right-hand side:
writeClauseWithInitGoals(LHS,Init,RHS) :-
	rev(Init,RInit),
	writeClauseWithRevInitGoals(LHS,RInit,RHS).
writeClauseWithRevInitGoals(LHS,[],RHS) :-
	writeClause((LHS :- RHS)).
writeClauseWithRevInitGoals(LHS,[G|Gs],RHS) :-
	writeClauseWithRevInitGoals(LHS,Gs,(G,RHS)).

writeClause(C) :-
	deleteLastTrue(C,C1),
	\+ \+ (numbervars(C1,0,_), writeq(C1)), write('.'), nl.

% delete a possibly generated last literal "true" in the clause:
deleteLastTrue((Head :- true),Head) :- !.
deleteLastTrue((Head :- B),(Head :- B1)) :- deleteLastTrueInBody(B,B1).
deleteLastTrue(Head,Head).

deleteLastTrueInBody((L,true),L) :- !.
deleteLastTrueInBody((L,B),(L,B1)) :- deleteLastTrueInBody(B,B1).


% split a file name into directory name (possibly '.') and base name:
% (all parameters are atoms)
split2dirbase(File,Dir,Base) :-
	atom_chars(File,FileS),
	rev(FileS,RevFileS),
	append(RevBaseS,[47|RevDirS],RevFileS),  % 47 = '/'
	!,
	rev(RevDirS,DirS), atom_chars(Dir,DirS),
	rev(RevBaseS,BaseS), atom_chars(Base,BaseS).
split2dirbase(File,'.',File).

% check whether an atom consists only of alphanum chars:
isAlphaNumAtom(A) :- atom_chars(A,AS), isAlphaNumString(AS).
isAlphaNumString([]).
isAlphaNumString([C|Cs]) :-
	(65=<C, C=<90 ; C=95 ; 97=<C, C=<122),
	isAlphaNumString(Cs).

% replace all dots by underscores in an atom:
replaceDotByUnderscore(DA,UA) :-
	atom_chars(DA,DAS),
	replaceDotsByUnderscores(DAS,UAS),
	atom_chars(UA,UAS).
replaceDotsByUnderscores(DAS,UAS) :-
	append(S1,[46|S2],DAS), !,
	replaceDotsByUnderscores(S2,S3),
	append(S1,[95|S3],UAS).
replaceDotsByUnderscores(S,S).

% set compilation mode (i.e., fastcode: yes or no)
setFastCode(yes) :- prolog_flag(compiling,_,fastcode), !.
setFastCode(no) :- prolog_flag(compiling,_,compactcode), !.
setFastCode(_).


