%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Definitions of builtins of module Dynamic:
%

:- use_module(library(system)).

:- dynamic dynamicTime/1, dynamicVersion/3.

dynamicTime(1).
%dynamicVersion(DBdir,P/N,V). % loaded database version of persistent predicate

setDynamicVersion(Dir,P/N,V) :- % set info about database version
	(retract(dynamicVersion(Dir,P/N,_)) -> true ; true),
	asserta(dynamicVersion(Dir,P/N,V)).


?- block 'prim_Dynamic.assertFact'(-,?,?,?,?),
         'prim_Dynamic.assertFact'(?,?,?,-,?).
'prim_Dynamic.assertFact'('Dynamic.Dynamic'(P),_,'$io'('()'),E0,E) :-
	retract(dynamicTime(DT)),
	DT1 is DT+1,
	assertz(dynamicTime(DT1)),
	P =.. [Pred,DType,_,_|Args],
	NP =.. [Pred,DT1,0|Args],
	assertz(NP),
	logDynamicEvent(DType,assertz,NP), E0=E.

?- block 'prim_Dynamic.retractFact'(-,?,?,?,?),
         'prim_Dynamic.retractFact'(?,?,?,-,?).
'prim_Dynamic.retractFact'('Dynamic.Dynamic'(P),_,'$io'(B),E0,E) :-
	P =.. [Pred,DType,_,_|Args],
	NP =.. [Pred,Start,Stop|Args],
	reloadNewestDynamicVersionOf(Pred),
	retract(dynamicTime(DT)),
	DT1 is DT+1,
	assertz(dynamicTime(DT1)),
	call(NP),
	isInTimeInterval(DT,Start,Stop),
	!,
	NPA =.. [Pred,Start,DT|Args],
	retract(NP), !, logDynamicEvent(DType,retract,NP),
	assertz(NPA),
	B='True', E0=E.
'prim_Dynamic.retractFact'('Dynamic.Dynamic'(_),_,'$io'('False'),E,E).

?- block 'prim_Dynamic.getDynamicKnowledge'(?,?,-,?).
'prim_Dynamic.getDynamicKnowledge'(_,'$io'('Dynamic.isKnownAtTime'(DT)),E0,E) :-
        reloadNewestDynamicVersions,
        dynamicTime(DT), E0=E.

?- block 'prim_Dynamic.isKnownAtTime'(?,?,?,-,?).
'prim_Dynamic.isKnownAtTime'(DT,'Dynamic.Dynamic'(P),success,E0,E) :-
	P =.. [Pred,_,_,_|Args],
	NP =.. [Pred,Start,Stop|Args],
	call(NP),
	isInTimeInterval(DT,Start,Stop), E0=E.

isInTimeInterval(DT,Start,Stop) :-
	DT>=Start, (Stop=0 -> true; DT=<Stop).


% log dynamic events (assert/retract):
logDynamicEvent('Dynamic.Temporary',_,_) :- !.
logDynamicEvent('Dynamic.Persistent',Op,Pred) :-
	Pred =.. [P,_,_|Args],
	StorePred =.. [P,0,0|Args],
	OpPred =.. [Op,StorePred],
	dynamicPredInfo(P/_,Dir),
	logDynamic(OpPred,Dir).

logDynamic(Term,Dir) :-
	%writeq(user_error,Term), nl(user_error),
	dir2tmplockfile(Dir,LockFile),
	lockWithFile(LockFile),
	dir2logfile(Dir,LogFile),
	open(LogFile,append,Stream),
	writeq(Stream,Term), put(Stream,46), nl(Stream),
	flush_output(Stream), close(Stream),
	addChangePid(Dir),
	unlockWithFile(LockFile).


?- block 'prim_Dynamic.startTransaction'(?,?,-,?).
'prim_Dynamic.startTransaction'(_,'$io'(DT1),E0,E) :-
	lockAllPersistentDynamics,
        reloadNewestDynamicVersions,
	retract(dynamicTime(DT)),
	DT1 is DT+1,
	assertz(dynamicTime(DT1)),
	logAllDynamicPreds(startTransaction(DT1)),
	E0=E, !.

?- block 'prim_Dynamic.commitTransaction'(-,?,?,?,?),
         'prim_Dynamic.commitTransaction'(?,?,?,-,?).
'prim_Dynamic.commitTransaction'(TDT,_,'$io'('()'),E0,E) :-
	logAllDynamicPreds(commitTransaction(TDT)),
	unlockAllPersistentDynamics,
	E0=E.

?- block 'prim_Dynamic.abortTransaction'(?,?,-,?).
'prim_Dynamic.abortTransaction'(_,_,_,_) :-
	logAllDynamicPreds(abortTransaction),
	getDynamicDirs(Dirs), map1partial(addChangePid(0),Dirs),
	unlockAllPersistentDynamics,
	!, fail.


% get directories of all dynamic predicates:
getDynamicDirs(Dirs) :-
	findall(Dir,isDynamicDir(Dir),Dirs).

isDynamicDir(Dir) :- dynamicPredInfo(_,Dir), \+ Dir=''.


% write a term (e.g., transaction marker) to event log of all
% persistent dynamic predicates occurring
% in a dynamic initialization directive:
logAllDynamicPreds(Term) :-
	getDynamicDirs(Dirs),
	map1partial(logDynamic(Term),Dirs).

% initialize dynamic predicates:
initializeDynamic :-
	dynamicPredInfo(Name/Arity,Dir),
	initializeDynamic(Dir,Name/Arity),
	fail.
initializeDynamic.

initializeDynamic('',P/N) :- !,	retractAllFacts(P/N).
initializeDynamic(Dir,P/N) :-
	file_exists(Dir), file_property(Dir,type(directory)), !,
	checkDynSpec(Dir,P/N),
	setDynamicVersion(Dir,P/N,0), !.
initializeDynamic(Dir,_) :-
	file_exists(Dir), !,
	appendAtom('ERROR: Directory "',Dir,E1),
	appendAtom(E1,'" for dynamic entries cannot be created!',ErrMsg),
	write(user_error,ErrMsg), nl(user_error), !, fail.
initializeDynamic(Dir,P/N) :-
	createDynamicDir(Dir,P/N),
	setDynamicVersion(Dir,P/N,1).

checkDynSpec(Dir,P/N) :-
	dir2specfile(Dir,SpecFile),
	file_exists(SpecFile),
	see(SpecFile), read(Spec), seen,
	dynamicSpec(P/N,Spec), !.
checkDynSpec(Dir,_) :-
	appendAtom('ERROR: Directory "',Dir,E1),
	appendAtom(E1,'" contains no matching dynamic entries!',ErrMsg),
	write(user_error,ErrMsg), nl(user_error), !, fail.

% replay all events written to log file:
replayLogFile(Dir,P/N) :-
	dir2tmplockfile(Dir,LockFile),
	lockWithFile(LockFile),
	(pakcsrc(dynamicmessages,yes)
	 -> write(user_error,'>>> Restoring persistent dynamic data in directory \''),
	    write(user_error,Dir), write(user_error,'\'...'),
	    nl(user_error) ; true),
	statistics(runtime,[RT1,_]),
	dir2dbfile(Dir,DBFile),
	dir2pofile(Dir,POFile),
	(file_exists(POFile) -> load_files(POFile) % for faster loading
	                      ; consult(DBFile)),
	dir2logfile(Dir,LogFile),
	open(LogFile,read,Stream),
	(at_end_of_stream(Stream) -> close(Stream)
	  ; repeat,
	    on_exception(Exc,readAndReplay(Stream),
                ((print_error(Exc)->true;true),
  	         write(user_error,'ERROR during reading of '),
		 write(user_error,LogFile),
		 write(user_error,' (restored state might be incomplete)'),
		 nl(user_error))),
	    close(Stream),
	    appendAtom(DBFile,'.bak',DBBakFile),
	    tell(DBBakFile),
	    writeq((:- dynamic P/N)), put(46), nl,
	    listing(P/N),
	    told,
	    try_save_predicates(P/N,POFile),
	    appendAtom('mv ',DBBakFile,C1), appendAtom(C1,' ',C2),
	    appendAtom(C2,DBFile,Cmd1), system(Cmd1),
	    appendAtom(LogFile,'.bak',LogBakFile),
	    appendAtom('mv ',LogFile,Cmd2), appendAtom(Cmd2,' ',Cmd3),
	    appendAtom(Cmd3,LogBakFile,Cmd4), system(Cmd4),
	    appendAtom('touch ',LogFile,Cmd5), system(Cmd5),
	    readDynamicVersion(Dir,Version),
	    NewVersion is Version+1,
	    writeDynamicVersion(Dir,NewVersion)),
	readDynamicVersion(Dir,CurrentVersion),
	setDynamicVersion(Dir,P/N,CurrentVersion),
	writeChangePids(Dir,[]),
	(pakcsrc(dynamicmessages,yes)
	 -> write(user_error,'>>> ...restored in '),
	    statistics(runtime,[RT2,_]),
	    RT is RT2-RT1,
	    write(user_error,RT),
	    write(user_error,' msec'), nl(user_error) ; true),
	unlockWithFile(LockFile),
	!.

readAndReplay(Stream) :- at_end_of_stream(Stream), !.
readAndReplay(Stream) :-
	read(Stream,Event),
	(Event=startTransaction(T)
	 -> readAndReplayTransaction(Stream,T,[])
	  ; call(Event)),
	!, fail.

readAndReplayTransaction(Stream,T,_) :- at_end_of_stream(Stream), !,
	(pakcsrc(dynamicmessages,yes)
	 -> write(user_error,'>>> Warning: ignoring incomplete transaction '),
	    write(user_error,T), nl(user_error)
	  ; true).
readAndReplayTransaction(Stream,T,Evs) :-
	read(Stream,Event),
	(Event=commitTransaction(T)
	 -> callRevList(Evs)
	  ; (Event=abortTransaction
	     -> true
	      ; readAndReplayTransaction(Stream,T,[Event|Evs]))),
	!.

callRevList([]).
callRevList([C|Cs]) :- callRevList(Cs), call(C).

% in order to ignore nested transactions:
startTransaction(_).
commitTransaction(_).
abortTransaction.


% lock all persistent data:
lockAllPersistentDynamics :-
	getDynamicDirs(Dirs), map1(lockDynamicDir,Dirs).

lockDynamicDir(Dir) :-
	dir2lockfile(Dir,LockFile), lockWithFile(LockFile).

lockWithFile(LockFile) :-
	((file_exists(LockFile), pakcsrc(dynamicmessages,yes))
	 -> write(user_error,'>>> Waiting for removing lock file \''),
	    write(user_error,LockFile), write(user_error,'\'...'),
	    nl(user_error) ; true),
	appendAtom('lockfile -1 ',LockFile,LockCmd),
	system(LockCmd), !.

% unlock all persistent data:
unlockAllPersistentDynamics :-
	getDynamicDirs(Dirs), map1(unlockDynamicDir,Dirs).

unlockDynamicDir(Dir) :-
	dir2lockfile(Dir,LockFile), unlockWithFile(LockFile).

unlockWithFile(LockFile) :-
	appendAtom('rm -f ',LockFile,UnlockCmd),
	system(UnlockCmd), !.


% reload newest versions of all persistent data:
reloadNewestDynamicVersions :-
	dynamicVersion(Dir,P/N,Version),
	reloadNewestDynamicVersion(Dir,P/N,Version),
	fail.
reloadNewestDynamicVersions.

% reload newest version of a particular dynamic predicate, if necessary:
reloadNewestDynamicVersionOf(Pred) :-
	dynamicVersion(Dir,Pred/N,Version), !, % is Pred persistent?
	reloadNewestDynamicVersion(Dir,Pred/N,Version).
reloadNewestDynamicVersionOf(_).

reloadNewestDynamicVersion(Dir,P/N,Version) :-
	readDynamicVersion(Dir,FileVersion),
	Version < FileVersion, !,
	retractAllFacts(P/N), % remove existing facts
	replayLogFile(Dir,P/N), !.
reloadNewestDynamicVersion(Dir,P/N,_) :-
	readChangePids(Dir,Pids), pid(Pid),
	((Pids=[] ; Pids=[Pid]) -> true
	  ; retractAllFacts(P/N), % remove existing facts
	    replayLogFile(Dir,P/N)),
	!.


% create new directory for dynamic data:
createDynamicDir(Dir,P/N) :-
	(pakcsrc(dynamicmessages,yes)
	 -> write(user_error,'>>> Creating new persistent dynamic data in directory \''),
	    write(user_error,Dir), write(user_error,'\'...'),
	    nl(user_error) ; true),
	appendAtom('mkdir -p ',Dir,Cmd1), system(Cmd1),
	dir2logfile(Dir,LogFile),
	appendAtom('touch ',LogFile,Cmd2), system(Cmd2),
	dir2dbfile(Dir,DBFile),
	appendAtom('touch ',DBFile,Cmd3), system(Cmd3),
	dir2specfile(Dir,SpecFile),
	dynamicSpec(P/N,Spec),
	open(SpecFile,write,SpecStr),
	writeq(SpecStr,Spec), put(SpecStr,46), nl(SpecStr),
	flush_output(SpecStr), close(SpecStr).


% read version number from file:
readDynamicVersion(Dir,Version) :-
	dir2versionfile(Dir,VersFile),
	file_exists(VersFile), !,
	open(VersFile,read,VersStr),
	read(VersStr,Version),
	close(VersStr).
readDynamicVersion(_,1).

% write version number into file:
writeDynamicVersion(Dir,Version) :-
	dir2versionfile(Dir,VersFile),
	open(VersFile,write,VersStr),
	writeq(VersStr,Version), put(VersStr,46), nl(VersStr),
	flush_output(VersStr), close(VersStr).


% read changePIDs from file:
readChangePids(Dir,Pids) :-
	dir2changepidsfile(Dir,CPFile),
	file_exists(CPFile), !,
	open(CPFile,read,Stream),
	read(Stream,Pids),
	close(Stream).
readChangePids(_,[]).

% write changePIDs into file:
writeChangePids(Dir,Pids) :-
	dir2changepidsfile(Dir,CPFile),
	open(CPFile,write,Stream),
	writeq(Stream,Pids), put(Stream,46), nl(Stream),
	flush_output(Stream), close(Stream).

addChangePid(Dir) :- pid(Pid), addChangePid(Pid,Dir).

addChangePid(Pid,Dir) :-
	readChangePids(Dir,Pids),
	(member(Pid,Pids) -> true ; writeChangePids(Dir,[Pid|Pids])),
	!.


% define specification data stored in directory:
dynamicSpec(P/N,dynamicPredicate(P,N)).

% retract all dynamic facts that are dead at current dynamic time:
retractDeadDynamicFacts(P/N) :-
	length(Args,N),
	Pred =.. [P|Args],
	Args = [_,Stop|_],
	dynamicTime(DT),
	call(Pred),
	Stop>0, DT>Stop,
	retract(Pred),
	fail. 
retractDeadDynamicFacts(_).


% File for writing all change events to persistent predicates:
dir2logfile(Dir,LogFile) :- appendAtom(Dir,'/eventlog',LogFile).

% Lockfile for transactions:
dir2lockfile(Dir,LockFile) :- appendAtom(Dir,'/LOCK',LockFile).

% Lockfile for change event logging and reconstructing the database
% (i.e., replaying all change events):
dir2tmplockfile(Dir,TmpLockFile) :- appendAtom(Dir,'/LOCK.TMP',TmpLockFile).

% File to store all facts in Prolog format:
dir2dbfile(Dir,DBFile) :- appendAtom(Dir,'/database.pl',DBFile).

% File to store all facts in faster load format:
dir2pofile(Dir,POFile) :- appendAtom(Dir,'/dbpred.po',POFile).

% File to store the specification of the dynamic predicate:
dir2specfile(Dir,SpecFile) :- appendAtom(Dir,'/spec',SpecFile).

% File to store the version number of the database:
dir2versionfile(Dir,VersFile) :- appendAtom(Dir,'/version',VersFile).

% File to store the PIDs of all processes that performed change events:
dir2changepidsfile(Dir,CPFile) :- appendAtom(Dir,'/changepids',CPFile).

