%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Implementation of readFlatCurryFile of module Flat
% (reading FlatCurry programs):

?- block 'prim_Flat.readFlatCurryFile'(-,?,?,?,?),
         'prim_Flat.readFlatCurryFile'(?,?,?,-,?).
'prim_Flat.readFlatCurryFile'(FileString,_,'$io'(CurryFlatProg),E0,E) :-
	cp_string(FileString,FileName),
	atom_chars(FileNameAtom,FileName),
	readProgFile(FileNameAtom,FlatProg),
	!,
	flc_transFlatProg(FlatProg,CurryFlatProg),
	!, E0=E.

flc_transFlatProg('Prog'(ModName,Imports,Types,Funcs,Ops,Trans),
	          'Flat.Prog'(FModName,FImports,CTypes,CFuncs,COps,CTrans)) :-
	cp_string(FModName,ModName),
	map2(cp_string,FImports,Imports),
	map2(flc_transTypeDecl,Types,CTypes),
	flc_transFuncs(Funcs,CFuncs),
	map2(flc_transOpDecl,Ops,COps),
	map2(flc_transTable,Trans,CTrans).

flc_transTypeDecl('Type'(Name,TParams,ConsDecls),
		  'Flat.Type'(FName,TParams,CConsDecls)) :-
	cp_string(FName,Name),
	map2(flc_transConsDecl,ConsDecls,CConsDecls).

flc_transConsDecl('Cons'(Name,Arity,Types),'Flat.Cons'(FName,Arity,CTypes)) :-
	cp_string(FName,Name),
	map2(flc_transTypeExpr,Types,CTypes).

flc_transTypeExpr('TVar'(N),'Flat.TVar'(N)).
flc_transTypeExpr('FuncType'(T1,T2),'Flat.FuncType'(CT1,CT2)) :-
	flc_transTypeExpr(T1,CT1), flc_transTypeExpr(T2,CT2).
flc_transTypeExpr('TCons'(Name,Ts),'Flat.TCons'(FName,CTs)) :-
	cp_string(FName,Name),
	map2(flc_transTypeExpr,Ts,CTs).


flc_transOpDecl('Op'(Name,Fixity,Prio),'Flat.Op'(FName,CFixity,Prio)) :-
	cp_string(FName,Name),
	flc_transFixity(Fixity,CFixity).

flc_transFixity('InfixOp','Flat.InfixOp').
flc_transFixity('InfixlOp','Flat.InfixlOp').
flc_transFixity('InfixrOp','Flat.InfixrOp').


flc_transFuncs([],[]).
flc_transFuncs([F|Fs],[TF|TFs]) :-
	flc_transFuncDecl(F,TF),
	flc_transFuncs(Fs,TFs).

flc_transFuncDecl('Func'(Name,Arity,Type,Rule),
          'Flat.Func'(FName,Arity,CType,CRule)) :-
	cp_string(FName,Name),
	flc_transTypeExpr(Type,CType),
	flc_transRule(Rule,CRule).

flc_transRule('Rule'(Args,Expr),'Flat.Rule'(Args,CExpr)) :-
	flc_transExpr(Expr,CExpr).
flc_transRule('External'(Name),'Flat.External'(FName)) :-
	cp_string(FName,Name).

flc_transCaseType('Rigid','Flat.Rigid').
flc_transCaseType('Flex','Flat.Flex').

flc_transCombType('FuncCall','Flat.FuncCall').
flc_transCombType('ConsCall','Flat.ConsCall').
flc_transCombType('PartCall','Flat.PartCall').

flc_transExpr('Var'(N),'Flat.Var'(N)) :- !.
flc_transExpr('Lit'(L),'Flat.Lit'(CL)) :- !, flc_transLit(L,CL).
flc_transExpr('Comb'(CT,Name,Es),'Flat.Comb'(CCT,FName,CEs)) :- !,
	flc_transCombType(CT,CCT),
	cp_string(FName,Name),
	map2(flc_transExpr,Es,CEs).
flc_transExpr('Apply'(E1,E2),'Flat.Apply'(CE1,CE2)) :- !,
	flc_transExpr(E1,CE1), flc_transExpr(E2,CE2).
flc_transExpr('Constr'(Xs,E),'Flat.Constr'(Xs,CE)) :- !,
	flc_transExpr(E,CE).
flc_transExpr('Or'(E1,E2),'Flat.Or'(CE1,CE2)) :- !,
	flc_transExpr(E1,CE1), flc_transExpr(E2,CE2).
flc_transExpr('Case'(CT,E,Cs),'Flat.Case'(CCT,CE,CCs)) :- !,
	flc_transCaseType(CT,CCT),
	flc_transExpr(E,CE),
	map2(flc_transCase,Cs,CCs).
flc_transExpr('GuardedExpr'(Xs,C,E),'Flat.GuardedExpr'(Xs,CC,CE)) :- !,
	flc_transExpr(C,CC), flc_transExpr(E,CE).
flc_transExpr('Choice'(E),'Flat.Choice'(CE)) :- !,
	flc_transExpr(E,CE).
flc_transExpr(E,E) :-
	write(user_error,'*** Internal error in FlatCurry file: Illegal expression "'),
	write(user_error,E),
	write(user_error,'"!'), nl(user_error),
	pleaseReport, fail.

flc_transLit('Intc'(I),'Flat.Intc'(I)) :- !.
flc_transLit('Floatc'(F),'Flat.Floatc'(F)) :- !.
flc_transLit('Charc'(N),'Flat.Charc'(C)) :- char_int(C,N), !.
flc_transLit(Lit,Lit) :-
	write(user_error,'*** Internal error in FlatCurry file: Illegal literal "'),
	write(user_error,Lit),
	write(user_error,'"!'), nl(user_error),
	pleaseReport, fail.

flc_transCase('Branch'('Pattern'(Name,Xs),E),
	      'Flat.Branch'('Flat.Pattern'(FName,Xs),CE)) :- !,
	cp_string(FName,Name),
	flc_transExpr(E,CE).
flc_transCase('Branch'('LPattern'(C),E),
	      'Flat.Branch'('Flat.LPattern'(LC),CE)) :- !,
	flc_transLit(C,LC),
	flc_transExpr(E,CE).
flc_transCase(Branch,Branch) :-
	write(user_error,'*** Internal error in FlatCurry file: Illegal case pattern "'),
	write(user_error,Branch),
	write(user_error,'"!'), nl(user_error),
	pleaseReport, fail.

flc_transTable('Trans'(C1,C2),'Flat.Trans'(FC1,FC2)) :-
	cp_string(FC1,C1),
	cp_string(FC2,C2).

