------------------------------------------------------------------------------
--- Library for HTML and CGI programming.
--- <A HREF="http://www.informatik.uni-kiel.de/~mh/publications/papers/PADL01.html">
--- This paper</A> contains a description of the basic ideas
--- behind this library.
---
--- @author Michael Hanus, Bernd Brassel
--- @version July 2004
------------------------------------------------------------------------------

module HTML(HtmlExp(..),HtmlForm(..),FormParam(..),CookieParam(..),CgiRef,
            form,Form,answerText,
            cookieForm,getCookies,
            htxt,htxts,hempty,nbsp,h1,h2,h3,h4,h5,
            par,emphasize,bold,italic,code,
            center,blink,teletype,pre,verbatim,address,href,anchor,
            ulist,olist,litem,dlist,table,hrule,breakline,image,
            style,textstyle,blockstyle,
            button,resetbutton,textfield,password,textarea,checkbox,
            radio_main,radio_main_off,radio_other,
            selection,selectionInitial,multipleSelection,
            hiddenfield,htmlQuote,addAttr,
            showHtmlExps,showHtmlExp,showHtmlDoc,showHtmlDocCSS,runcgi,
            getUrlParameter,urlencoded2string,string2urlencoded,
            showLatexExps,showLatexExp,showLatexDoc,showLatexDocs,
            showLatexDocsWithPackages,showLatexDocWithPackages,
            germanLatexDoc) where

import System
import Read
import Char
import ReadShowTerm
import List
import Time -- for cookies

infixl 0 `addAttr`

--- The data type for representing HTML expressions.
data HtmlExp =
   HtmlText String                               -- text string
 | HtmlStruct String [(String,String)] [HtmlExp] -- (type attrs contents)
 | HtmlElem   String [(String,String)]           -- single element like HR, BR
 | HtmlCRef HtmlExp CgiRef
        -- an input element (first arg) with a cgi reference
 | HtmlEvent  HtmlExp (CgiEnv -> IO HtmlForm)
        -- an input element (first arg) with an associated event handler


--- The (abstract) data type for representing references to cgi input elements.
data CgiRef = CgiRef String

--- The type for representing cgi environments
--- (i.e., cgi references to cgi input elements).
type CgiEnv = CgiRef -> String

------------------------------------------------------------------------------
--- The data type for representing HTML forms (active web pages).
--- The constructor arguments are the title, the parameters, and
--- the contents (body) of the web page.
data HtmlForm = HtmlForm String [FormParam] [HtmlExp]
              | HtmlAnswer String String -- content type / content

--- The possible parameters of an HTML form.
--- The parameters of a cookie (FormCookie) are its name and value and
--- optional parameters (expiration date, domain, path (e.g., the path "/"
--- makes the cookie valid for all documents on the server), security) which
--- are collected in a list.
--- @cons FormCookie name value params - a cookie to be sent to the
---                                      client's browser
--- @cons FormCSS s - a URL for a CSS file for this form
--- @cons FormJScript s - a URL for a Javascript file for this form
--- @cons FormTarget s - a name of a target frame where the output of the
---                      script should be represented (should only be used
---                      for scripts running in a frame)
--- @cons BodyAttr  ps - optional attribute for the body element (more than
---                      one occurrence is allowed)
data FormParam = FormCookie  String String [CookieParam]
               | FormCSS     String
               | FormJScript String
               | FormTarget  String
               | BodyAttr    (String,String)

--- The possible parameters of a cookie.
data CookieParam = CookieExpire ClockTime
                 | CookieDomain String
                 | CookiePath   String
                 | CookieSecure

--- A basic HTML form for active web pages.
--- @param title - the title of the form
--- @param hexps - the form's body (list of HTML expressions)
--- @return an HTML form
form :: String -> [HtmlExp] -> HtmlForm
form title hexps = HtmlForm title [BodyAttr ("bgcolor","#ffffff")] hexps

--- A basic HTML form for active web pages
--- (deprecated, included only for backward compatibility)
--- @param title - the title of the form
--- @param hexps - the form's body (list of HTML expressions)
--- @return an HTML form
Form :: String -> [HtmlExp] -> HtmlForm
Form title hexps = HtmlForm title [] hexps

--- An HTML form with simple cookies.
--- The cookies are sent to the client's browser together with this form.
--- @param title - the title of the form
--- @param cookies - the cookies as a list of name/value pairs
--- @param hexps - the form's body (list of HTML expressions)
--- @return an HTML form
cookieForm :: String -> [(String,String)] -> [HtmlExp] -> HtmlForm
cookieForm t cs he = HtmlForm t (map (\(n,v)->FormCookie n v []) cs) he

-- Shows the cookie in standard syntax:
formatCookie (name,value,params) =
  "Set-Cookie: " ++ name ++ "=" ++ string2urlencoded value ++
  concatMap (\p->"; "++formatCookieParam p) params

-- Formats a cookie parameter:
formatCookieParam :: CookieParam -> String
formatCookieParam (CookieExpire e) = "expires="++toCookieDateString e
formatCookieParam (CookieDomain d) = "domain="++d
formatCookieParam (CookiePath   p) = "path="++p
formatCookieParam CookieSecure = "secure"

-- Formats a clock time into a date string for cookies:
toCookieDateString :: ClockTime -> String
toCookieDateString time =
 let (y,mo,d,_,_,_) = toDate time
  in show d ++ "-" ++ shortMonths!!(mo-1) ++ "-" ++ show y ++ " " ++
     toTimeString time
  where shortMonths = ["Jan","Feb","Mar","Apr","May","Jun",
                       "Jul","Aug","Sep","Oct","Nov","Dec"]


--- A textual result instead of an HTML form as a result for active web pages.
--- @param txt - the contents of the result page
--- @return an HTML answer form
answerText :: String -> HtmlForm
answerText = HtmlAnswer "text/plain"


------------------------------------------------------------------------------
-- some useful abbreviations:

--- Basic text as HTML expression.
--- The text may contain special HTML chars (like &lt;,&gt;,&amp;,&quot;)
--- which will be quoted so that they appear as in the parameter string.
htxt   :: String -> HtmlExp
htxt s = HtmlText (htmlQuote s)

--- A list of strings represented as a list of HTML expressions.
--- The strings may contain special HTML chars that will be quoted.
htxts :: [String] -> [HtmlExp]
htxts = map htxt

--- An empty HTML expression.
hempty :: HtmlExp
hempty = HtmlText ""

--- Non breaking Space
nbsp   :: HtmlExp
nbsp = HtmlText "&nbsp;"

--- Header 1
h1      :: [HtmlExp] -> HtmlExp
h1 hexps = HtmlStruct "h1" [] hexps

--- Header 2
h2      :: [HtmlExp] -> HtmlExp
h2 hexps = HtmlStruct "h2" [] hexps

--- Header 3
h3      :: [HtmlExp] -> HtmlExp
h3 hexps = HtmlStruct "h3" [] hexps

--- Header 4
h4      :: [HtmlExp] -> HtmlExp
h4 hexps = HtmlStruct "h4" [] hexps

--- Header 5
h5      :: [HtmlExp] -> HtmlExp
h5 hexps = HtmlStruct "h5" [] hexps

--- Paragraph
par      :: [HtmlExp] -> HtmlExp
par hexps = HtmlStruct "p" [] hexps

--- Emphasize
emphasize      :: [HtmlExp] -> HtmlExp
emphasize hexps = HtmlStruct "em" [] hexps

--- Boldface
bold      :: [HtmlExp] -> HtmlExp
bold hexps = HtmlStruct "b" [] hexps

--- Italic
italic      :: [HtmlExp] -> HtmlExp
italic hexps = HtmlStruct "i" [] hexps

--- Program code
code      :: [HtmlExp] -> HtmlExp
code hexps = HtmlStruct "code" [] hexps

--- Centered text
center      :: [HtmlExp] -> HtmlExp
center hexps = HtmlStruct "center" [] hexps

--- Blinking text
blink      :: [HtmlExp] -> HtmlExp
blink hexps = HtmlStruct "blink" [] hexps

--- Teletype font
teletype      :: [HtmlExp] -> HtmlExp
teletype hexps = HtmlStruct "tt" [] hexps

--- Unformatted input, i.e., keep spaces and line breaks and
--- don't quote special characters.
pre      :: [HtmlExp] -> HtmlExp
pre hexps = HtmlStruct "pre" [] hexps

--- Verbatim (unformatted), special characters (&lt;,&gt;,&amp;,&quot;)
--- are quoted.
verbatim  :: String -> HtmlExp
verbatim s = HtmlStruct "pre" [] [HtmlText (htmlQuote s)]

--- Address
address       :: [HtmlExp] -> HtmlExp
address hexps = HtmlStruct "address" [] hexps

--- Hypertext reference
href           :: String -> [HtmlExp] -> HtmlExp
href ref hexps = HtmlStruct "a" [("href",ref)] hexps

--- An anchor for hypertext reference inside a document
anchor           :: String -> [HtmlExp] -> HtmlExp
anchor anc hexps = HtmlStruct "a" [("name",anc)] hexps

--- Unordered list
--- @param items - the list items where each item is a list of HTML expressions
ulist       :: [[HtmlExp]] -> HtmlExp
ulist items = HtmlStruct "ul" [] (map litem items)

--- Ordered list
--- @param items - the list items where each item is a list of HTML expressions
olist :: [[HtmlExp]] -> HtmlExp
olist items = HtmlStruct "ol" [] (map litem items)

--- A single list item (usually not explicitly used)
litem hexps = HtmlStruct "li" [] hexps

--- Description list
--- @param items - a list of (title/description) pairs (of HTML expressions)
dlist       :: [([HtmlExp],[HtmlExp])] -> HtmlExp
dlist items = HtmlStruct "dl" [] (concatMap ditem items)
 where
   ditem (hexps1,hexps2) = [HtmlStruct "dt" [] hexps1,
                            HtmlStruct "dd" [] hexps2]

--- Table with a matrix of items (each item is a list of HTML exprs)
table :: [[[HtmlExp]]] -> HtmlExp
table items = HtmlStruct "table" []
 (map (\row->HtmlStruct "tr" []
                 (map (\item -> HtmlStruct "td" [] item) row)) items)

--- Horizontal rule
hrule :: HtmlExp
hrule = HtmlElem "hr" []

--- Break a line
breakline :: HtmlExp
breakline = HtmlElem "br" []

--- Image
--- @param src - the URL of the image
--- @param alt - the alternative text shown instead of the image
image :: String -> String -> HtmlExp
image src alt = HtmlElem "img" [("src",src),("alt",htmlQuote alt)]


-------------- styles:
--- Provide a style for HTML elements.
--- The style argument is the name of a style class defined in an
--- external style sheet.
--- @param st - name of a style class
--- @param hexps - list of HTML expressions
style :: String -> [HtmlExp] -> HtmlExp
style st hexps = HtmlStruct "span" [("class",st)] hexps

--- Provide a style for a basic text.
--- The style argument is the name of a style class defined in an
--- external style sheet.
--- @param st - name of a style class
--- @param txt - a string (special characters will be quoted)
textstyle :: String -> String -> HtmlExp
textstyle st txt = HtmlStruct "span" [("class",st)] [htxt txt]

--- Provide a style for a block of HTML elements.
--- The style argument is the name of a style class defined in an
--- external style sheet. This element is used (in contrast to "style")
--- for larger blocks of HTML elements since a line break is placed
--- before and after these elements.
--- @param st - name of a style class
--- @param hexps - list of HTML expressions
blockstyle :: String -> [HtmlExp] -> HtmlExp
blockstyle st hexps = HtmlStruct "div" [("class",st)] hexps


-------------- forms and input fields:
--- Submit button with a label string and an event handler
button :: String -> (CgiEnv -> IO HtmlForm) -> HtmlExp
button label handler =
    HtmlEvent
       (HtmlElem "input" [("type","submit"),("name","EVENT"),
                          ("value",htmlQuote label)])
       handler

--- Reset button with a label string
resetbutton :: String -> HtmlExp
resetbutton label =
    HtmlElem "input" [("type","reset"),("value",htmlQuote label)]

--- Input text field with a reference and an initial contents
textfield :: CgiRef -> String -> HtmlExp
textfield (CgiRef ref) contents = -- ref argument will be instantiated
    HtmlCRef
       (HtmlElem "input" [("type","text"),("name",ref),
                          ("value",htmlQuote contents)])
       (CgiRef ref)

--- Input text field (where the entered text is obscured) with a reference
password :: CgiRef -> HtmlExp
password (CgiRef ref) = -- ref argument will be instantiated
    HtmlCRef
       (HtmlElem "input" [("type","password"),("name",ref)])
       (CgiRef ref)

--- Input text area with a reference, height/width, and initial contents
textarea :: CgiRef -> (Int,Int) -> String -> HtmlExp
textarea (CgiRef ref) (height,width) contents =
    HtmlCRef
       (HtmlStruct "textarea" [("name",ref),
                                ("rows",show height),("cols",show width)]
                               [htxt contents])
       (CgiRef ref)

--- A checkbox with a reference and a value.
--- The value is returned of checkbox is on, otherwise "" is returned.
checkbox :: CgiRef -> String -> HtmlExp
checkbox (CgiRef ref) value = -- ref argument will be instantiated
    HtmlCRef
       (HtmlElem "input" [("type","checkbox"),("name",ref),
                          ("value",htmlQuote value)])
       (CgiRef ref)

--- A main button of a radio (initially "on") with a reference and a value.
--- The value is returned of this button is on.
--- A complete radio button suite always consists of a main button
--- (radio_main) and some further buttons (radio_others) with the
--- same reference. Initially, the main button is selected
--- (or nothing is selected if one uses radio_main_off instead of radio_main).
--- The user can select another button but always at most one button
--- of the radio can be selected. The value corresponding to the
--- selected button is returned in the environment for this radio reference.
radio_main :: CgiRef -> String -> HtmlExp
radio_main (CgiRef ref) value = -- ref argument will be instantiated
    HtmlCRef
       (HtmlElem "input" [("type","radio"),("name",ref),
                          ("value",htmlQuote value),("checked","yes")])
       (CgiRef ref)

--- A main button of a radio (initially "off") with a reference and a value.
--- The value is returned of this button is on.
radio_main_off :: CgiRef -> String -> HtmlExp
radio_main_off (CgiRef ref) value = -- ref argument will be instantiated
    HtmlCRef
       (HtmlElem "input" [("type","radio"),("name",ref),
                          ("value",htmlQuote value)])
       (CgiRef ref)

--- A further button of a radio (initially "off") with a reference (identical
--- to the main button of this radio) and a value.
--- The value is returned of this button is on.
radio_other :: CgiRef -> String -> HtmlExp
radio_other (CgiRef ref) value = -- ref argument will be instantiated
    HtmlElem "input" [("type","radio"),("name",ref),("value",htmlQuote value)]


--- A selection button with a reference and a list of name/value pairs.
--- The names are shown in the selection and the value is returned
--- for the selected name.
selection :: CgiRef -> [(String,String)] -> HtmlExp
selection (CgiRef ref) menue = -- ref argument will be instantiated
    HtmlCRef
       (HtmlStruct "select" [("name",ref)]
          ((concat . map \(n,v)->[HtmlStruct "option" [("value",v)] [htxt n]])
           menue))
       (CgiRef ref)

--- A selection button with a reference, a list of name/value pairs,
--- and a preselected item in this list.
--- The names are shown in the selection and the value is returned
--- for the selected name.
--- @param ref - a CGI reference
--- @param nvs - list of name/value pairs
--- @param sel - the index of the initially selected item in the list nvs
--- @return an HTML expression representing the selection button
selectionInitial :: CgiRef -> [(String,String)] -> Int -> HtmlExp
selectionInitial (CgiRef ref) sellist sel =
    HtmlCRef (HtmlStruct "select" [("name",ref)] (selOption sellist sel))
             (CgiRef ref)
 where
   selOption [] _ = []
   selOption ((n,v):nvs) i =
      HtmlStruct "option"
                 ([("value",v)] ++ if i==0 then [("selected","yes")] else [])
                 [htxt n] : selOption nvs (i-1)

--- A selection button with a reference and a list of name/value/flag pairs.
--- The names are shown in the selection and the value is returned
--- if the corresponding name is selected. If flag is True, the
--- corresonding name is initially selected. If more than one name
--- has been selected, all values are returned in one string
--- where the values are separated by '\n' characters.
multipleSelection :: CgiRef -> [(String,String,Bool)] -> HtmlExp
multipleSelection (CgiRef ref) sellist =
   HtmlCRef (HtmlStruct "select" [("name",ref),("multiple","yes")]
                                 (map selOption sellist))
            (CgiRef ref)
 where
   selOption (n,v,flag) =
      HtmlStruct "option"
                 ([("value",v)] ++ if flag then [("selected","yes")] else [])
                 [htxt n]

--- A hidden field to pass a value referenced by a fixed name.
--- This function should be used with care since it may cause
--- conflicts with the CGI-based implementation of this library.
hiddenfield :: String -> String -> HtmlExp
hiddenfield name value =
    HtmlElem "input" [("type","hidden"),("name",name),("value",value)]


------------------------------------------------------------------------------
--- Quotes special characters (<,>,&,", umlauts) in a string
--- as HTML special characters.

htmlQuote :: String -> String
htmlQuote [] = []
htmlQuote (c:cs) | c=='<' = "&lt;"   ++ htmlQuote cs
                 | c=='>' = "&gt;"   ++ htmlQuote cs
                 | c=='&' = "&amp;"  ++ htmlQuote cs
                 | c=='"' = "&quot;" ++ htmlQuote cs
                 | c=='\228' = "&auml;" ++ htmlQuote cs
                 | c=='\246' = "&ouml;" ++ htmlQuote cs
                 | c=='\252' = "&uuml;" ++ htmlQuote cs
                 | c=='\196' = "&Auml;" ++ htmlQuote cs
                 | c=='\214' = "&Ouml;" ++ htmlQuote cs
                 | c=='\220' = "&Uuml;" ++ htmlQuote cs
                 | c=='\223' = "&szlig;"++ htmlQuote cs
                 | otherwise = c : htmlQuote cs

------------------------------------------------------------------------------
--- Adds an attribute (name/value pair) to an HTML element.

addAttr :: HtmlExp -> (String,String) -> HtmlExp
addAttr (HtmlText s) _ = HtmlText s  -- strings have no attributes
addAttr (HtmlStruct tag attrs hexps) attr =
    HtmlStruct tag (attrs++[attr]) hexps
addAttr (HtmlElem tag attrs) attr =
    HtmlElem tag (attrs++[attr])
addAttr (HtmlEvent hexp handler) attr =
    HtmlEvent (addAttr hexp attr) handler
addAttr (HtmlCRef  hexp cref) attr =
    HtmlCRef (addAttr hexp attr) cref


------------------------------------------------------------------------------
--- Transforms a list of HTML expressions into string representation.
showHtmlExps :: [HtmlExp] -> String
showHtmlExps [] = ""
showHtmlExps (he:hes) = showWithLnPrefix he ++ showHtmlExps hes
 where
   showWithLnPrefix hexp = let s = getText hexp
                            in if s/="" && isSpace (head s)
                               then "\n" ++ tail s
                               else showHtmlExp hexp

-- get the string contents of an HTML expression:
getText :: HtmlExp -> String
getText (HtmlText s)       = s
getText (HtmlStruct _ _ _) = ""
getText (HtmlElem   _ _)   = ""
getText (HtmlEvent  he _)  = getText he
getText (HtmlCRef   he _)  = getText he

-- get the (last) tag of an HTML expression:
getTag :: HtmlExp -> String
getTag (HtmlText _)         = ""
getTag (HtmlStruct tag _ _) = tag
getTag (HtmlElem   tag _)   = tag
getTag (HtmlEvent  he _)    = getTag he
getTag (HtmlCRef   he _)    = getTag he

-- is this a tag where a line break can be safely added?
tagWithLn t = t/="" &&
              t `elem` ["br","p","li","ul","ol","hr",
                        "h1","h2","h3","h4","h5","h6",
                        "title","head","body"]


--- Transforms a single HTML expression into string representation.
showHtmlExp :: HtmlExp -> String
showHtmlExp (HtmlText s) = s
showHtmlExp (HtmlStruct tag attrs hexps) =
  let maybeLn = if tagWithLn tag then "\n" else ""
   in maybeLn ++ showHtmlOpenTag tag attrs ++
      showHtmlExps hexps ++
      "</" ++ tag ++ ">" ++ maybeLn
showHtmlExp (HtmlElem tag attrs) =
  let maybeLn = if tagWithLn tag then "\n" else ""
   in maybeLn ++ showHtmlOpenTag tag attrs ++ maybeLn
showHtmlExp (HtmlEvent hexp _) = showHtmlExp hexp
showHtmlExp (HtmlCRef  hexp _) = showHtmlExp hexp

showHtmlOpenTag tag attrs =
  "<" ++ tag ++ concat (map ((" "++) . attr2string) attrs) ++ ">"
  where
    attr2string (attr,value) = attr ++ "=\"" ++ encodeQuotes value ++ "\""

    -- encode double quotes as "&quot;":
    encodeQuotes [] = []
    encodeQuotes (c:cs) | c=='"'    = "&quot;" ++ encodeQuotes cs
                        | otherwise = c : encodeQuotes cs


------------------------------------------------------------------------------
--- Transforms HTML expressions into string representation of complete
--- HTML document with title.
--- @param title - the title of the HTML document
--- @param hexps - the body (list of HTML expressions) of the document
--- @return string representation of the HTML document
showHtmlDoc :: String -> [HtmlExp] -> String
showHtmlDoc title html =
  showHtmlExp (HtmlStruct "html" []
                  [HtmlStruct "head" []
                       [HtmlStruct "title" [] [HtmlText title]],
                   HtmlStruct "body" [("bgcolor","#ffffff")] html])


--- Transforms HTML expressions into string representation of complete
--- HTML document with title and a URL for a style sheet file.
--- @param title - the title of the HTML document
--- @param css - the URL for a CSS file for this document
--- @param hexps - the body (list of HTML expressions) of the document
--- @return string representation of the HTML document
showHtmlDocCSS :: String -> String -> [HtmlExp] -> String
showHtmlDocCSS title css html =
  showHtmlExp (HtmlStruct "html" []
                  [HtmlStruct "head" []
                       [HtmlStruct "title" [] [HtmlText title],
                        HtmlStruct "link" [("rel","stylesheet"),
                                      ("type","text/css"),("href",css)] []],
                   HtmlStruct "body" [("bgcolor","#ffffff")] html])


------------------------------------------------------------------------------
--- Gets the parameter attached to the URL of the script.
--- For instance, if the script is called with URL
--- "http://.../script.cgi?parameter", then "parameter" is
--- returned by this I/O action.
--- Note that an URL parameter should be "URL encoded" to avoid
--- the appearance of characters with a special meaning.
--- Use the functions "urlencoded2string" and "string2urlencoded"
--- to decode and encode such parameters, respectively.

getUrlParameter :: IO String
getUrlParameter = getEnviron "QUERY_STRING"

--- Translates urlencoded string into equivalent ASCII string.
urlencoded2string :: String -> String
urlencoded2string [] = []
urlencoded2string (c:cs)
  | c == '+'  = ' ' : urlencoded2string cs
  | c == '%'  = chr (readHex (take 2 cs)) : urlencoded2string (drop 2 cs)
  | otherwise = c : urlencoded2string cs

--- Translates arbitrary strings into equivalent urlencoded string.
string2urlencoded :: String -> String
string2urlencoded [] = []
string2urlencoded (c:cs)
  | isAlphaNum c = c : string2urlencoded cs
  | c == ' '     = '+' : string2urlencoded cs
  | otherwise = let oc = ord c
    in '%' : int2hex(oc `div` 16) : int2hex(oc `mod` 16) : string2urlencoded cs
 where
   int2hex i = if i<10 then chr (ord '0' + i)
                       else chr (ord 'A' + i - 10)


------------------------------------------------------------------------------
--- Gets the cookies sent from the browser for the current CGI script.
--- The cookies are represented in the form of name/value pairs since
--- no other components are important here.
getCookies :: IO [(String,String)]
getCookies =
 do cookiestring <- getEnviron "HTTP_COOKIE"
    return $ parseCookies cookiestring

-- translate a string of cookies (of the form "NAME1=VAL1; NAME2=VAL")
-- into a list of name/value pairs:
parseCookies :: String -> [(String,String)]
parseCookies str = if str=="" then [] else
  let (c1,cs) = break (==';') str
   in parseCookie c1 :
      parseCookies (dropWhile (==' ') (if cs=="" then "" else tail cs))
 where
  parseCookie s = let (name,evalue) = break (=='=') s in
      (name,if evalue=="" then "" else urlencoded2string (tail evalue))


------------------------------------------------------------------------------
--- Transforms an HTML form (possibly containing input fields)
--- into string representation of complete HTML document
--- and shows the result as an HTML document on stdout.
--- The first argument is the URL of this executable.

runcgi :: String -> IO HtmlForm -> IO ()
runcgi url hformact =
  --putStrLn "Content-type: text/html\n" >> --debug
  getEnviron "CONTENT_LENGTH" >>= \clen ->
  getNChar (readNat clen) >>= \cont ->
  let cenv = parseCgiEnv cont
  in  --putStrLn (showHtmlExps [pre [htxt cont,hrule]]) >> --debug
      --putStrLn (showHtmlExps [pre [par (env2html cenv),hrule]]) >> --debug
      if cenv == [] -- initial form?
      then runcgiformInEnv url hformact cenv
      else let handler = readQTerm (urlencoded2string (getFormEvent "" cenv))
           in  runcgiformInEnv url (handler (cgiGetValue cenv)) cenv

runcgiformInEnv :: String -> IO HtmlForm -> [(String,String)] -> IO ()
runcgiformInEnv url hformact cenv =
  hformact >>= \cform ->
  setCookies cform >>= \hform ->
  showCgiFormInEnv url hform cenv

showCgiFormInEnv :: String -> HtmlForm -> [(String,String)] -> IO ()
showCgiFormInEnv url (HtmlForm ftitle fparams fhexp) cenv =
  putStrLn "Content-type: text/html" >>
  getEnviron "QUERY_STRING" >>= \qstr ->
  putStrLn "" >>  -- end of HTTP header
  let (title,params,hexps,firsthandler) =
         htmlForm2html (HtmlForm ftitle fparams fhexp) (getMaxFieldNr cenv + 1)
  in showform (changeAssoc cenv "DEFAULT" ("EVENT_"++firsthandler))
              (if qstr=="" then url else url++"?"++qstr)
              (HtmlForm title params hexps)
showCgiFormInEnv _ (HtmlAnswer ctype cont) _ = do
  putStrLn ("Content-type: "++ctype)
  putStrLn "" -- end of HTTP header
  putStrLn cont

-- extract and set the cookies contained in a form:
setCookies :: HtmlForm -> IO HtmlForm
setCookies (HtmlAnswer ctype cont) = return (HtmlAnswer ctype cont)
setCookies (HtmlForm title params hexp) =
  do if cookies==[]
       then done
       else putStrLn "Cache-control: no-cache=\"set-cookie\""
     foldr (>>) done (map (putStrLn . formatCookie) cookies)
     return (HtmlForm title otherparams hexp)
 where
   (cookies,otherparams) = splitFormParams params

   splitFormParams [] = ([],[])
   splitFormParams (FormCookie n v ps : fps) =
     let (cs,ops) = splitFormParams fps in ((n,v,ps):cs,ops)
   splitFormParams (FormCSS s : fps) =
     let (cs,ops) = splitFormParams fps in (cs,FormCSS s : ops)
   splitFormParams (FormJScript s : fps) =
     let (cs,ops) = splitFormParams fps in (cs,FormJScript s : ops)
   splitFormParams (FormTarget s : fps) =
     let (cs,ops) = splitFormParams fps in (cs,FormTarget s : ops)
   splitFormParams (BodyAttr s : fps) =
     let (cs,ops) = splitFormParams fps in (cs,BodyAttr s : ops)

-- get the EVENT_ definition of the cgi environment
-- (or "DEFAULT" value if it is not there):
getFormEvent :: String -> [(String,String)] -> String
getFormEvent deflt [] = deflt
getFormEvent deflt ((tag,val):tvs) =
   if tag == "DEFAULT" then getFormEvent (drop 6 val) tvs else
   if take 6 tag == "EVENT_" then urlencoded2string (drop 6 tag)
                             else getFormEvent deflt tvs

-- compute the maximal field number of all "FIELD_nr" in a CGI environment:
getMaxFieldNr :: [(String,String)] -> Int
getMaxFieldNr [] = 0
getMaxFieldNr ((name,_):env) =
  if take 6 name == "FIELD_"
  then max (readNat (drop 6 name)) (getMaxFieldNr env)
  else getMaxFieldNr env
 
max x y = if x>y then x else y


-- get the value assigned to a name in a given cgi environment
cgiGetValue :: [(String,String)] -> CgiRef -> String
cgiGetValue cenv (CgiRef ref) =
    concat (intersperse "\n" (findall (\s->s=:=lookupAssoc cenv ref)))
 where
   -- non-deterministic definition: for instance, in case of multiple
   -- selections, it returns more than one value
   lookupAssoc ((tag,val):_) r | (tag==r)=:=True = val
   lookupAssoc ((_,_):tvs) r = lookupAssoc tvs r

-- transform HTML form into HTML document (by instantiating CgiRefs
-- (starting with the second argument) and modifying event handlers):
-- (Result: title/HTML document/form params/encoded first handler)
htmlForm2html :: HtmlForm -> Int -> (String,[FormParam],[HtmlExp],String)
htmlForm2html (HtmlForm title params html) crefnr =
  let (transhtml, fh) = translateHandlers (fst (numberCgiRefs html crefnr))
  in  (title, params, transhtml, fh)


-- instantiate all CgiRefs with a unique tag in HTML expressions:
numberCgiRefs :: [HtmlExp] -> Int -> ([HtmlExp],Int)
-- arguments: HTMLExps, number for cgi-refs
-- result: translated HTMLExps, new number for cgi-refs
numberCgiRefs [] i = ([],i)
numberCgiRefs (HtmlText s : hexps) i =
   let (nhexps,j) = numberCgiRefs hexps i
   in (HtmlText s : nhexps, j)
numberCgiRefs (HtmlStruct tag attrs hexps1 : hexps2) i =
   let (nhexps1,j) = numberCgiRefs hexps1 i
       (nhexps2,k) = numberCgiRefs hexps2 j
   in (HtmlStruct tag attrs nhexps1 : nhexps2, k)
numberCgiRefs (HtmlElem tag attrs : hexps) i =
   let (nhexps,j) = numberCgiRefs hexps i
   in (HtmlElem tag attrs : nhexps, j)
numberCgiRefs (HtmlEvent (HtmlElem tag attrs) handler : hexps) i =
   let (nhexps,j) = numberCgiRefs hexps i
   in (HtmlEvent (HtmlElem tag attrs) handler : nhexps, j)
numberCgiRefs (HtmlCRef hexp (CgiRef ref) : hexps) i
  | ref =:= ("FIELD_"++show i)
  = let ([nhexp],j) = numberCgiRefs [hexp] (i+1)
        (nhexps,k) = numberCgiRefs hexps j
    in (nhexp : nhexps, k)


-- translate all event handlers into their internal form:
-- (assumption: all CgiRefs have already been instantiated and eliminated)
-- the result is the translated HTML expression list (without HtmlEvents)
-- and the string encoding of the first event handler (for the default handler)
translateHandlers :: [HtmlExp] -> ([HtmlExp],String)
translateHandlers [] = ([],"")
translateHandlers (HtmlText s : hexps) =
   let (nhexps,fh) = translateHandlers hexps
   in  (HtmlText s : nhexps, fh)
translateHandlers (HtmlStruct tag attrs hexps1 : hexps2) =
   let (nhexps1,fh1) = translateHandlers hexps1
       (nhexps2,fh2) = translateHandlers hexps2
   in  (HtmlStruct tag attrs nhexps1 : nhexps2, if fh1=="" then fh2 else fh1)
translateHandlers (HtmlElem tag attrs : hexps) =
   let (nhexps,fh) = translateHandlers hexps
   in  (HtmlElem tag attrs : nhexps, fh)
translateHandlers (HtmlEvent (HtmlElem tag attrs) handler : hexps) =
   let (nhexps,_) = translateHandlers hexps
       fh = string2urlencoded (showQTerm handler)
   in  (HtmlElem tag (changeAssoc attrs "name" ("EVENT_" ++ fh)) : nhexps, fh)


-- show a HTML form as their Ascii representation:
showform cenv url (HtmlForm title params html) =
   putStr (showHtmlExp
               (HtmlStruct "html" []
                   [HtmlStruct "head" []
                        ([HtmlStruct "title" [] [HtmlText title]] ++
                         concatMap param2html params),
                    HtmlStruct "body" bodyattrs
                        [HtmlStruct "form"
                           ([("method","POST"),("action",url)]++targetattr)
                           ( --[par (env2html cenv),hrule] ++
                            cenv2hidden cenv ++ html)]]))
 where
  param2html (FormCSS css) =
     [HtmlStruct "link" [("rel","stylesheet"),("type","text/css"),("href",css)]
                 []]
  param2html (FormJScript js) =
     [HtmlStruct "script" [("type","text/javascript"),("src",js)] []]
  param2html (FormTarget _) = []
  -- no rule for FormCookie since they have been already processed
  param2html (BodyAttr _) = []
  -- no rule for BodyAttr since it is considered later

  bodyattrs = [ps | (BodyAttr ps) <- params]

  target = [s | (FormTarget s) <- params]

  targetattr = if target==[] then [] else [("target",head target)]


-- translate cgi environment into HTML (for debugging purposes):
env2html env = concat (map (\(n,v)->[htxt (n++": "++v),breakline]) env)


-- translate environment into hidden fields (without EVENT field!):
-- (note: the field values are urlencoded to avoid problems
--  with passing special characters; moreover, the names of fields
--  containing urlencoded values are prefixed by "U")
cenv2hidden env = concat (map pair2hidden env)
 where
   pair2hidden (n,v)
     | take 6 n == "EVENT_" = []
     | take 6 n == "FIELD_" = [hiddenfield ('U':n) (string2urlencoded v)]
     | otherwise            = [hiddenfield n v]

-- get n chars from stdin:
getNChar n = if n<=0 then return ""
                     else getChar >>= \c ->
                          getNChar (n-1) >>= \cs -> return (c:cs)

-- translate a string of cgi environment bindings into list of binding pairs:
parseCgiEnv :: String -> [(String,String)]
parseCgiEnv s | s == ""   = []
              | otherwise = map ufield2field
                             (map (\(n,v)->(n,urlencoded2string v))
                                  (map (splitChar '=') (split (=='&') s)))
 where
   ufield2field (n,v) = if take 7 n == "UFIELD_"
                        then (tail n, urlencoded2string v)
                        else (n,v)

   -- split a string at particular character:
   splitChar c xs = let (ys,zs) = break (==c) xs
                    in if zs==[] then (ys,zs) else (ys,tail zs)

   -- split a string at all positions of a particular character:
   split p xs =
    let (ys,zs) = break p xs
    in if zs==[] then [ys]
                 else ys : split p (tail zs)


------------------------------------------------------------------------------
-- association lists (list of tag/value pairs):

-- change an associated value (or add association, if not there):
changeAssoc :: [(tt,tv)] -> tt -> tv -> [(tt,tv)]
changeAssoc [] tag val = [(tag,val)]
changeAssoc ((tag1,val1):tvs) tag val =
   if tag1 == tag then (tag,val) : tvs
                  else (tag1,val1) : changeAssoc tvs tag val


------------------------------------------------------------------------------
--- Transforms HTML expressions into LaTeX string representation.

showLatexExps :: [HtmlExp] -> String
showLatexExps hexps = concat (map showLatexExp hexps)

--- Transforms an HTML expression into LaTeX string representation.
showLatexExp :: HtmlExp -> String
showLatexExp (HtmlText s) = "{" ++ specialchars2tex s ++ "}"
showLatexExp (HtmlStruct tag attrs htmlexp)
 | tag=="html" = showLatexExps htmlexp
 | tag=="head" = ""                    -- ignore header
 | tag=="body" = showLatexExps htmlexp
 | tag=="form" = showLatexExps htmlexp
 | tag=="h1"   = "\\section*{" ++ showLatexExps htmlexp ++ "}\n"
 | tag=="h2"   = "\\subsection*{" ++ showLatexExps htmlexp ++ "}\n"
 | tag=="h3"   = "\\subsubsection*{" ++ showLatexExps htmlexp ++ "}\n"
 | tag=="h4"   = "\\paragraph*{" ++ showLatexExps htmlexp ++ "}\n"
 | tag=="h5"   = "\\subparagraph*{" ++ showLatexExps htmlexp ++ "}\n"
 | tag=="p"    = showLatexExps htmlexp ++ "\\par\n"
 | tag=="b"    = "{\\bf " ++ showLatexExps htmlexp ++ "}"
 | tag=="em"   = "\\emph{" ++ showLatexExps htmlexp ++ "}"
 | tag=="i"    = "{\\it " ++ showLatexExps htmlexp ++ "}"
 | tag=="tt"   = "{\\tt " ++ showLatexExps htmlexp ++ "}"
 | tag=="center" = latexEnvironment "center" (showLatexExps htmlexp)
 | tag=="pre"  = latexEnvironment "verbatim" (showLatexExps htmlexp)
 | tag=="font" = showLatexExps htmlexp  -- ignore font changes
 | tag=="address" = showLatexExps htmlexp
 | tag=="blink"   = showLatexExps htmlexp
 | tag=="a"    = showLatexExps htmlexp ++
                 -- add href attribute as footnote, if present:
                 maybe ""
                       (\url->"\\footnote{\\tt "++specialchars2tex url++"}\n")
                       (findHtmlAttr "href" attrs)
 | tag=="ul"   = latexEnvironment "itemize"  (showLatexExps htmlexp)
 | tag=="ol"   = latexEnvironment "enumerate" (showLatexExps htmlexp)
 | tag=="li"   = "\\item\n" ++ showLatexExps htmlexp ++ "\n"
 | tag=="dl"   = latexEnvironment "description" (showLatexExps htmlexp)
 -- tables will be set using the longtable environment, 
 -- (The package longtable is added by default to every latex document)
 | tag=="table" = attrLatexEnv "longtable" (latexTabFormat htmlexp) 
                                           (showLatexTableContents htmlexp)
 | tag=="tr"   = let cells = map showLatexExp htmlexp
                  in concat (intersperse " & " cells) ++ "\\\\\n"
 | tag=="td"   = showLatexExps htmlexp
 | otherwise   = "{\\tt<"++tag++">}" ++ showLatexExps htmlexp ++
                "{\\tt</"++tag++">}"
showLatexExp (HtmlElem tag attrs)
 | tag=="br"  = "\\\\\n"
 | tag=="p"   = "\\\\[1ex]\n"
 | tag=="li"  = "\\item\n"
 | tag=="hr"  = "\\vspace{2ex}\\hrule\n"
 | tag=="a"   = ""
 | tag=="dt"  = "\\item["
 | tag=="dd"  = "]~\\\\\n"
 | tag=="img" = "{" ++  maybe "{\\tt<IMAGE>}" specialchars2tex
                             (findHtmlAttr "alt" attrs)
                    ++ "}"
 | tag=="input" && maybe "" id (findHtmlAttr "type" attrs) == "hidden" = ""
 | otherwise = "{\\tt<"++tag++">}"

-- create latex environment of name "env" with content "content"
latexEnvironment :: String -> String -> String
latexEnvironment env content = attrLatexEnv env "" content

-- create latex environment of name "env" with content "content"
-- adding the parameters "attr"
attrLatexEnv :: String -> String -> String -> String
attrLatexEnv env attr content
 = "\\begin{"++env++"}"++attr++"\n"
 ++content
 ++"\n\\end{"++env++"}\n"

-- yield the format of a table, e.g. {lll} from list of html rows.
-- for longtables we set the chunksize big enough 
-- to avoid having to rerun latex for inaccurat tables.
latexTabFormat :: [HtmlExp] -> String
latexTabFormat rows = "{" ++ replicate breadth 'l' ++ "}"
  ++ "\\setcounter{LTchunksize}{"++show (length rows+5)++"}%"
  where
    breadth = foldl max 0 (map getBreadth rows)

-- retrieve the breadth of an Html row
getBreadth :: HtmlExp -> Int
getBreadth row = case row of
                     HtmlStruct "tr" _ tds -> length tds
                     _ -> error "getBreadth: no row given"

-- tranlate expressions inside tables
showLatexTableContents :: [HtmlExp] -> String
showLatexTableContents hexps = concatMap showLatexTableContent hexps

-- tranlate expressions inside tables
showLatexTableContent :: HtmlExp -> String
showLatexTableContent (HtmlText s) = "{" ++ specialchars2tex s ++ "}"
showLatexTableContent (HtmlStruct tag attrs htmlexp)
 | tag=="html" = showLatexTableContents htmlexp
 | tag=="head" = ""                    -- ignore header
 | tag=="body" = showLatexTableContents htmlexp
 | tag=="form" = showLatexTableContents htmlexp
 | tag=="p"    = showLatexTableContents htmlexp ++ "\\par\n"
 | tag=="b"    = "{\\bf " ++ showLatexTableContents htmlexp ++ "}"
 | tag=="em"   = "\\emph{" ++ showLatexTableContents htmlexp ++ "}"
 | tag=="i"    = "{\\it " ++ showLatexTableContents htmlexp ++ "}"
 | tag=="tt"   = "{\\tt " ++ showLatexTableContents htmlexp ++ "}"
 | tag=="font" = showLatexTableContents htmlexp  -- ignore font changes
 | tag=="address" = showLatexTableContents htmlexp
 | tag=="blink"   = showLatexTableContents htmlexp
 | tag=="a"    = showLatexTableContents htmlexp ++
                 -- add href attribute as footnote, if present:
                 maybe ""
                       (\url->"\\footnote{\\tt "++specialchars2tex url++"}\n")
                       (findHtmlAttr "href" attrs)
 | tag=="tr"   = let cells = map showLatexTableContent htmlexp
                  in concat (intersperse " & " cells) ++ "\\\\\n"
 | tag=="td"   = showLatexTableContents htmlexp
 | otherwise   = "{\\tt<"++tag++">}" ++ showLatexTableContents htmlexp ++
                "{\\tt</"++tag++">}"
showLatexTableContent (HtmlElem tag attrs)
 | tag=="br"  = "\\\\\n"
 | tag=="p"   = "\\\\[1ex]\n"
 | tag=="hr"  = "\\vspace{2ex}\\hrule\n"
 | tag=="a"   = ""
 | tag=="img" = "{" ++  maybe "{\\tt<IMAGE>}" specialchars2tex
                             (findHtmlAttr "alt" attrs)
                    ++ "}"
 | tag=="input" && maybe "" id (findHtmlAttr "type" attrs) == "hidden" = ""
 | otherwise = "{\\tt<"++tag++">}"

-- find a specific tag field in a list of HTML attributes:
findHtmlAttr :: String -> [(String,String)] -> Maybe String
findHtmlAttr _    [] = Nothing
findHtmlAttr atag ((t,f):attrs) =
  if atag==t then Just f
             else findHtmlAttr atag attrs


-- convert special characters into TeX representation:
specialchars2tex [] = []
specialchars2tex (c:cs)
  | c==chr 228  = "\\\"a"  ++ specialchars2tex cs
  | c==chr 246  = "\\\"o"  ++ specialchars2tex cs
  | c==chr 252  = "\\\"u"  ++ specialchars2tex cs
  | c==chr 196  = "\\\"A"  ++ specialchars2tex cs
  | c==chr 214  = "\\\"O"  ++ specialchars2tex cs
  | c==chr 220  = "\\\"U"  ++ specialchars2tex cs
  | c==chr 223  = "\\ss{}" ++ specialchars2tex cs
  | c=='~'      = "{\\char126}" ++ specialchars2tex cs
  | c=='\\'     = "{\\char92}" ++ specialchars2tex cs
  | c=='<'      = "{$<$}" ++ specialchars2tex cs
  | c=='>'      = "{$>$}" ++ specialchars2tex cs
  | c=='_'      = "\\_" ++ specialchars2tex cs
  | c=='#'      = "\\#" ++ specialchars2tex cs
  | c=='$'      = "\\$" ++ specialchars2tex cs
  | c=='%'      = "\\%" ++ specialchars2tex cs
  | c=='{'      = "\\{" ++ specialchars2tex cs
  | c=='}'      = "\\}" ++ specialchars2tex cs
  | c=='&'      = let (special,rest) = break (==';') cs
                  in  if null rest
                      then "\\&" ++ specialchars2tex special -- wrong format
                      else htmlspecial2tex special ++
                           specialchars2tex (tail rest)
  | otherwise   = c : specialchars2tex cs

htmlspecial2tex special
  | special=="Auml"   =  "{\\\"A}"
  | special=="Euml"   =  "{\\\"E}"
  | special=="Iuml"   =  "{\\\"I}"
  | special=="Ouml"   =  "{\\\"O}"
  | special=="Uuml"   =  "{\\\"U}"
  | special=="auml"   =  "{\\\"a}"
  | special=="euml"   =  "{\\\"e}"
  | special=="iuml"   =  "{\\\"\\i}"
  | special=="ouml"   =  "{\\\"o}"
  | special=="uuml"   =  "{\\\"u}"
  | special=="szlig"  =  "{\\ss}"
  | special=="Aacute" =  "{\\\'A}"
  | special=="Eacute" =  "{\\\'E}"
  | special=="Iacute" =  "{\\\'I}"
  | special=="Oacute" =  "{\\\'O}"
  | special=="Uacute" =  "{\\\'U}"
  | special=="aacute" =  "{\\\'a}"
  | special=="eacute" =  "{\\\'e}"
  | special=="iacute" =  "{\\\'\\i}"
  | special=="oacute" =  "{\\\'o}"
  | special=="uacute" =  "{\\\'u}"
  | special=="Agrave" =  "{\\`A}"
  | special=="Egrave" =  "{\\`E}"
  | special=="Igrave" =  "{\\`I}"
  | special=="Ograve" =  "{\\`O}"
  | special=="Ugrave" =  "{\\`U}"
  | special=="agrave" =  "{\\`a}"
  | special=="egrave" =  "{\\`e}"
  | special=="igrave" =  "{\\`\\i}"
  | special=="ograve" =  "{\\`o}"
  | special=="ugrave" =  "{\\`u}"
  | special=="Acirc"  =  "{\\^A}"
  | special=="Ecirc"  =  "{\\^E}"
  | special=="Icirc"  =  "{\\^I}"
  | special=="Ocirc"  =  "{\\^O}"
  | special=="Ucirc"  =  "{\\^U}"
  | special=="acirc"  =  "{\\^a}"
  | special=="ecirc"  =  "{\\^e}"
  | special=="icirc"  =  "{\\^\\i}"
  | special=="ocirc"  =  "{\\^o}"
  | special=="ucirc"  =  "{\\^u}"
  | special=="Oslash" =  "{\\O}"
  | special=="oslash" =  "{\\o}"
  | special=="amp"    =  "{\\&}"
  | special=="ntilde" =  "{\\~n}"
  | special=="otilde" =  "{\\~o}"
  | special=="ccedil" =  "{\\c{c}}"
  | special=="nbsp"   =  "~"
  | special=="quot"   =  "\""
  | special=="lt"     =  "{$<$}"
  | special=="gt"     =  "{$>$}"
  | otherwise = "\\&"++special++";"

------------------------------------------------------------------------------
--- Transforms HTML expressions into a string representation of a complete
--- LaTeX document.

showLatexDoc :: [HtmlExp] -> String
showLatexDoc htmlexps = showLatexDocs [htmlexps]

--- Transforms HTML expressions into a string representation of a complete
--- LaTeX document.
--- The variable "packages" holds the packages to add to the latex document
--- e.g. "ngerman"

showLatexDocWithPackages :: [HtmlExp] -> [String] -> String
showLatexDocWithPackages hexps packages 
  = showLatexDocsWithPackages [hexps] packages

--- Transforms a list of HTML expressions into a string representation
--- of a complete LaTeX document where each list entry appears
--- on a separate page.

showLatexDocs :: [[HtmlExp]] -> String
showLatexDocs htmlexps_list = showLatexDocsWithPackages htmlexps_list []


--- Transforms a list of HTML expressions into a string representation
--- of a complete LaTeX document where each list entry appears
--- on a separate page.
--- The variable "packages" holds the packages to add to the latex document
--- (e.g., "ngerman").

showLatexDocsWithPackages :: [[HtmlExp]] -> [String] -> String
showLatexDocsWithPackages htmlexps_list packages =
 "\\documentclass[12pt]{article}\n"++
 concatMap (\p->"\\usepackage{"++p++"}\n") packages++
 -- Package longtable is added by default.
 "\\usepackage{longtable}"++
 "\\nonstopmode\n"++
 "\\setlength{\\topmargin}{-1.5cm}\n"++
 "\\setlength{\\oddsidemargin}{0.0cm}\n"++
 "\\setlength{\\evensidemargin}{0.0cm}\n"++
 "\\setlength{\\marginparwidth}{0.0cm}\n"++
 "\\setlength{\\marginparsep}{0.0cm}\n"++
 "\\setlength{\\textwidth}{16.5cm}\n"++
 "\\setlength{\\textheight}{24.0cm}\n"++
 "\\pagestyle{empty}\n"++
 "\\begin{document}\n\\sloppy\n"++
 "\\addtolength{\\baselineskip}{0.3ex}\n"++
 "\\setlength{\\parindent}{0.0ex}\n\\addtolength{\\parskip}{1.5ex}\n"++
 concat (intersperse "\\newpage\n" (map showLatexExps htmlexps_list))++
 "\\end{document}\n"

--- show german latex document
germanLatexDoc :: [HtmlExp] -> String
germanLatexDoc hexps = showLatexDocWithPackages hexps ["ngerman"]

------------------------------------------------------------------------------

{-

The installation of a cgi script written with this library
can be done by the command

  makecurrycgi prog /home/joe/public_html/prog.cgi initial_form

where "prog" is the name of the Curry program with the cgi script,
"/home/joe/public_html/prog.cgi" is the desired location of the
compiled cgi script, and "initial_form" is the Curry expression
(of type IO HtmlForm) computing the HTML form (where makecurrycgi
is a shell script stored in <PAKCS>/bin).

Alternatively, a cgi script can be also installed by the following
sequence of single actions (which are combined by the command above):

1. Define a main function, e.g.,

     main = runcgi "prog.cgi" initial_form

   where prog.cgi is the name of the compiled program to be installed.

2. Start curry2prolog, load the script and create saved state via

     :save main

3. Terminate curry2prolog and execute "makecurrycgi prog".

4. Move or copy the generated executable "prog.cgi" to the desired
   directory accessible by the web server
   (e.g., to /home/joe/public_html/prog.cgi).

-}
