------------------------------------------------------------------------------
--- Library for distributed programming with ports.
--- <A HREF="http://www.informatik.uni-kiel.de/~mh/publications/papers/PPDP99.html">
--- This paper</A> contains a description of the basic ideas
--- behind this library.
---
--- @author Michael Hanus
--- @version October 2001
------------------------------------------------------------------------------

module Ports(Port,openPort,send,doSend,openNamedPort,connectPort,
             after,ping,timeoutOnStream,openProcessPort,openSocketConnectPort,
             SP_Msg(..),
             choiceSPEP,
             CPNS_Message(..),cpns_start,cpns_stop) where

import Time
import System(system)

--- The internal constructor for the port datatype is not visible to the user.

data Port a = internalPort String Int Int a


--- Opens an internal port for communication.
--- @param p - a free variable which will be constrainted
---            with the port messages
--- @param s - a free variable which will be instantiated
---            to the stream of incoming messages
openPort    :: Port a -> [a] -> Success
openPort external


--- Sends a message to a port.
send        :: a -> Port a -> Success
send external

--- I/O action that sends a message to a port.
doSend       :: a -> Port a -> IO ()
doSend msg p = doSolve (send msg p)


--- Opens the standard input stream (currently only supported in TasteCurry).
openStdIn :: IO [String]
openStdIn external


--- A constrained which is satisfied after some amount of time
--- (currently only supported in TasteCurry).
--- @param n - the satisfaction time in milliseconds
after       :: Int -> Success
after external

--- Checks whether port p is still reachable.
--- @param n - the time to wait for reachability in milliseconds
--- @param p - a port to be checked for reachability
--- @return Nothing if port p is unreachable within n milliseconds,
---         or (Just m) if port p could be contacted within m milliseconds
ping :: Int -> Port _ -> IO (Maybe Int)
ping external

--- Checks for instantiation of a stream within some amount of time.
--- @param n - the time to wait for instantiation in milliseconds
--- @param str - the stream to be checked for instantiation
---              (usually the stream of incoming messages at some port)
--- @return (Just str) if str is instantiated within n milliseconds,
---         or Nothing otherwise
timeoutOnStream :: Int -> [a] -> Maybe [a]
timeoutOnStream external


--- A "stream port" is an adaption of the port concept to model the
--- communication with bidirectional streams, i.e., a stream port is
--- a port connection to a bidirectional stream (e.g., opened by
--- openProcessPort or openSocketConnectPort) where the communication
--- is performed via the following stream port messages.
---
--- @cons SP_Put s     - write the argument s on the output stream
--- @cons SP_GetLine s - unify the argument s with the next text line of the
---                      input stream
--- @cons SP_GetChar c - unify the argument c with the next character of the
---                      input stream
--- @cons SP_EOF b     - unify the argument b with True if we are at the end
---                      of the input stream, otherwise with False
--- @cons SP_Close     - close the input/output streams

data SP_Msg =
     SP_Put     String -- write the argument on the output stream
   | SP_GetLine String -- unify the argument with the next text line of the
                       -- input stream
   | SP_GetChar Char   -- unify the argument with the next character of the
                       -- input stream
   | SP_EOF     Bool   -- unify the argument with True if we are at the end
                       -- of the input stream, otherwise with False
   | SP_Close          -- close the input/output streams


--- Opens a new connection to a process that executes a shell command.
--- @param cmd - the shell command to be executed
--- @return the output/input stream (represented as a stream port)
---         that is connected to the standard input/output of the process
---         performing the execution of cmd.
openProcessPort :: String -> IO (Port SP_Msg)
openProcessPort external

--- Opens a new connection to a Unix socket.
--- This function will not be supported in the future.
--- Use function <code>IOExts.connectToSocket</code>!
--- @param snr - the socket number of the connection
--- @param host - the host name of the connection
--- @return the output/input stream (represented as a stream port)
---         that is connected to the socket snr@host
openSocketConnectPort :: Int -> String -> IO (Port SP_Msg)
openSocketConnectPort external


--- Opens an external port with a symbolic name.
--- @param portname - the symbolic name under which the port is accessible
---                   (any string without occurrences of '@')
--- @return the stream of incoming messages at this port
openNamedPort :: String -> IO [_]
openNamedPort name =
  do stream <- openPortOnSocket socketnr portnr       -- open new port
     cpns_client "localhost" (CPNS_Put name socketnr portnr) -- register this port
     return stream
  where socketnr,portnr free

--- Connects to an external port.
--- @param portname - the symbolic name of the external port
---                   (must be either of the form "name@machine" or "name"
---                    where the latter is a shorthand for "name@localhost")
--- @return a port with the symbolic name portname
connectPort   :: String -> IO (Port _)
connectPort nameAtMachine =
  do let (name,atMachine) = break (=='@') nameAtMachine
         machine = if atMachine=="" then "localhost" else tail atMachine
     cpns_client machine (CPNS_Get name snr pnr)  -- get remote socket/port no.
     if snr==0
      then error ("connectPort: Port \""++name++"@"++machine++
                  "\" is not registered!")
      else done
     port <- connectPortAtSocket snr pnr machine
     return port
  where snr,pnr free


--- This function implements a committed choice over the receiving
--- of messages via a stream port and an external port.
---
--- <EM>Note that the implementation of choiceSPEP works only with
--- Sicstus-Prolog 3.8.5 or higher (due to a bug in previous versions
--- of Sicstus-Prolog).</EM>
---
--- @param sp - a stream port sp
--- @param ms - a stream of messages received via an external port
--- @return (Left s) if s is an input line received
---                  at the stream port (via SP_GetLine) or
--- 
---         (Right ms) if the stream ms is instantiated
---                    with at least one new message at the head

choiceSPEP :: Port SP_Msg -> [Msg] -> Either String [Msg]
choiceSPEP external


------------------------------------------------------------------------------

-- The following predefined actions are not intended for application programs.
-- They are the basis to implement ports with symbolic names
-- via a name server (see below).


-- (openPortOnSocket snr pnr) is an action which opens an external port
-- on socket number snr with internal port number pnr and returns
-- the stream of incoming messages.
-- snr and pnr are allowed to be unbound: in this case they will be bound to the
-- numbers associated to a free port

openPortOnSocket :: Int -> Int -> IO [_]
openPortOnSocket external


-- The internal function that reads a port stream lazily.
basicServerLoop :: Port a -> [a]
basicServerLoop external


-- (connectPortAtSocket snr pnr host) is an action which returns a port that has
-- been opened at <host> with socket number <snr> and port number <pnr>

connectPortAtSocket :: Int -> Int -> String -> IO (Port _)
connectPortAtSocket external


-- Using these two primitive actions, we can implement "ports with
-- symbolic names" via a name server, called Curry Port Name Server (CPNS):

-- If we connect to a port with symbolic name pn, we first connect
-- to the CPNS of the host named by pn to get the physical socket
-- number of this port. In order to connect to CPNS from any
-- machine in the world, the CPNS demon always listens at the following
-- port:
-- (Note that this must be identical for all machines running
-- Distributed Curry! If this port is occupied by another process
-- on a host, you cannot run Distributed Curry on it.)

cpns_serversocket = 8766

cpns_startuplockfile = "/tmp/PAKCS_CPNSD.LOCK"

-- Now the implementation of the Curry Port Name Server is straightforward:

--- Type of messages to be processed by the Curry Port Name Server.
--- Although these messages are not intended for public use, this
--- type must be public so that any machine in the world can send
--- these messages.
---
--- @cons CPNS_Put name sn pn - assign the values sn and pn to name
--- @cons CPNS_Get name sn pn - instantiate sn and pn with the
---                             values assigned to name
--- @cons CPNS_Close          - terminate the CPNS demon

data CPNS_Message = CPNS_Put String Int Int
                  | CPNS_Get String Int Int
                  | CPNS_Close

--- Starts the "Curry Port Name Server" (CPNS) running on the local machine.
--- The CPNS is responsible to resolve symbolic names for ports
--- into physical socket numbers so that a port can be reached
--- under its symbolic name from any machine in the world.

cpns_start :: IO ()
cpns_start = catchFail startup
                       (putStrLn "FAILURE occurred during startup!" >>
                        delete_startuplockfile)
 where
   delete_startuplockfile = do
     putStrLn ("Removing startup lock file \""++cpns_startuplockfile++"\"...")
     system ("rm -f "++cpns_startuplockfile)
     done

   startup =
     putStrLn "Starting Curry Port Name Server..." >>
     openPortOnSocket cpns_serversocket 0 >>= \msgs ->
     delete_startuplockfile >>
     putStrLn "Curry Port Name Server is ready." >>
     cpns_loop (\_->(0,0)) msgs

-- The implementation of the CPNS demon is a loop processing the
-- incoming messages.
-- The assignment from names to socket numbers (the local state of the
-- server) is implemented as a function (which results in a primitive
-- linear search for entries, but currently the number of entries
-- is usually very small on each host):

cpns_loop :: (String->(Int,Int)) -> [CPNS_Message] -> IO ()
cpns_loop _   (CPNS_Close   : _) = putStrLn "CPNS demon terminated."
cpns_loop n2i (CPNS_Get n sn pn : s) | (sn,pn) =:= (n2i n) = cpns_loop n2i s
cpns_loop n2i (CPNS_Put n sn pn : s) =
  do time <- getClockTime
     putStrLn ("Register port \""++n++"\": socket "++show sn++
               " / number "++show pn ++ " at " ++ toDateString time)
     cpns_loop new_n2i s
  where new_n2i m = if m==n then (sn,pn) else n2i m


-- cpns_client sends a message string to the CPNS port of the given host:

cpns_client host msg =
  do alive <- cpns_alive host
     if not alive
      then error ("Curry port name server at host \""++host++
                  "\" is not reachable!")
      else connectPortAtSocket cpns_serversocket 0 host >>= doSend msg


-- cpns_alive checks whether the CPNS of the given host is reachable
-- within 5 seconds:

cpns_alive :: String -> IO Bool
cpns_alive host =
  do port <- connectPortAtSocket cpns_serversocket 0 host
     pingresult <- ping 5000 port
     return $ maybe False (const True) pingresult


--- Terminates the "Curry Port Name Server" running on the local machine.
cpns_stop :: IO ()
cpns_stop = cpns_client "localhost" CPNS_Close


--  end of module Ports
