------------------------------------------------------------------------------
--- Library for converting ground terms to strings and vice versa.
---
--- @author Michael Hanus
--- @version April 2004
------------------------------------------------------------------------------

module ReadShowTerm(showTerm,showQTerm,readsTerm,readTerm,readQTerm,readsQTerm,
                    readTermFile,readTermListFile,
                    writeTermFile,writeTermListFile) where

import Char(isSpace)

--- Transforms a ground(!) term into a string representation
--- in standard prefix notation.
--- Thus, showTerm suspends until its argument is ground.
--- This function is similar to the prelude function <code>show</code>
--- but can read the string back with <code>readTerm</code>
--- (provided that the constructor names are unique without the module
--- qualifier).
showTerm :: _ -> String
showTerm external

--- Transforms a ground(!) term into a string representation
--- in standard prefix notation.
--- Thus, showTerm suspends until its argument is ground.
--- Note that this function differs from the prelude function <code>show</code>
--- since it prefixes constructors with their module name
--- in order to read them back with <code>readTerm</code>.
showQTerm :: _ -> String
showQTerm external

--- Transform a string containing a term in standard prefix notation
--- without module qualifiers into the corresponding data term.
--- In case of a successful parse, the result is a one element list
--- containing a pair of the data term and the remaining unparsed string.

readsTerm :: String -> [(_,String)]
readsTerm external

--- Transforms a string containing a term in standard prefix notation
--- without module qualifiers into the corresponding data term.

readTerm :: String -> _
readTerm s = case result of
  [(term,tail)]
     -> if all isSpace tail then term
        else error ("ReadShowTerm.readTerm: no parse, unmatched string after term: "++tail)
  [] ->  error "ReadShowTerm.readTerm: no parse"
  _  ->  error "ReadShowTerm.readTerm: ambiguous parse"
 where result = readsTerm s

--- Transforms a string containing a term in standard prefix notation
--- with qualified constructor names into the corresponding data term.
--- In case of a successful parse, the result is a one element list
--- containing a pair of the data term and the remaining unparsed string.

readsQTerm :: String -> [(_,String)]
readsQTerm external

--- Transforms a string containing a term in standard prefix notation
--- with qualified constructor names into the corresponding data term.

readQTerm :: String -> _
readQTerm s = case result of
  [(term,tail)] -> if all isSpace tail then term
                   else error "ReadShowTerm.readQTerm: no parse"
  [] ->  error "ReadShowTerm.readQTerm: no parse"
  _  ->  error "ReadShowTerm.readQTerm: ambiguous parse"
 where result = readsQTerm s


--- Reads a file containing a string representation of a term
--- in standard prefix notation (without module qualifiers)
--- and returns the corresponding data term.

readTermFile :: String -> IO _
readTermFile file = readFile file >>= return . readTerm

--- Reads a file containing lines with string representations of terms
--- (without module qualifiers) of the same type
--- and returns the corresponding list of data terms.

readTermListFile :: String -> IO [_]
readTermListFile file = readFile file >>= return . map readTerm . lines

--- Writes a ground term into a file in standard prefix notation
--- (without module qualifiers).
--- @param filename - The name of the file to be written.
--- @param term - The term to be written to the file as a string.

writeTermFile :: String -> _ -> IO ()
writeTermFile filename term = writeFile filename (showTerm term)

--- Writes a list of ground terms into a file (without module qualifiers).
--- Each term is written into a separate line which might be useful
--- to modify the file with a standard text editor.
--- @param filename - The name of the file to be written.
--- @param terms - The list of terms to be written to the file.

writeTermListFile :: String -> [_] -> IO ()
writeTermListFile filename terms =
    writeFile filename (unlines (map showTerm terms))

