------------------------------------------------------------------------------
--- Library to access parts of the system environment.
---
--- @author Michael Hanus
--- @version January 2000
------------------------------------------------------------------------------

module System(Date(..),getDate,showDate,getCPUTime,getElapsedTime,
              getArgs,getEnviron,getHostname,getPID,getProgName,system,
              getLoadPath,findFileInLoadPath,sleep,evalTime,evalSpace) where

import Directory(doesFileExist)

--- A date is presented in the following form:
--- (Date year month day hour minute second)
--- This type will not be supported in the future.
--- Use type Time.ClockTime!

data Date = Date Int Int Int Int Int Int

--- Returns the current date and time.
--- This function will not be supported in the future.
--- Use function Time.getClockTime!

getDate :: IO Date
getDate external

--- Transform a date into a readable form.
--- This function will not be supported in the future.
--- Use function Time.toDateString!

showDate :: Date -> String
showDate (Date y mo d h mi s) = digit2 mo ++"/"++ digit2 d ++"/"++ show y
                                ++ " " ++
                                digit2 h ++":"++ digit2 mi ++":"++ digit2 s
  where digit2 n = if n<10 then ['0',chr(ord '0' + n)]
                           else show n

--- Returns the current cpu time of the process in milliseconds.

getCPUTime :: IO Int
getCPUTime external
 
--- Returns the current elapsed time of the process in milliseconds.

getElapsedTime :: IO Int
getElapsedTime external
 
--- Returns the list of the program's command line arguments.
--- The program name is not included.

getArgs :: IO [String]
getArgs external
 
--- Returns the value of an environment variable.
--- The empty string is returned for undefined environment variables.

getEnviron :: String -> IO String
getEnviron external
 
--- Returns the hostname of the machine running this process.

getHostname :: IO String
getHostname external

--- Returns the process identifier of the current Curry process.

getPID :: IO Int
getPID external


--- Returns the name of the current program, i.e., the name of the
--- main module currently executed.

getProgName :: IO String
getProgName external


--- Executes a shell command and return with the exit code of the command.
--- An exit status of zero means successful execution.

system :: String -> IO Int
system external


--- Returns the current path (list of directory names) that is
--- used for loading modules.
getLoadPath :: IO [String]
getLoadPath = do
  pakcshome <- getEnviron "PAKCSHOME"
  pakcspath <- getEnviron "PAKCSLIBPATH"
  currypath <- getEnviron "CURRYPATH"
  return $ ["."] ++
           (if null currypath then [] else splitPath currypath) ++
           (if null pakcspath
            then if null pakcshome then [] else [pakcshome++"/lib"]
            else splitPath pakcspath)
 where
  -- split a path string into list of directory names:
  splitPath xs = let (ys,zs) = break (==':') xs
                  in if null zs then [ys]
                                else ys : splitPath (tail zs)


--- Adds a directory name to a file by looking up the
--- current load path (CURRYPATH).
findFileInLoadPath :: String -> IO String
findFileInLoadPath prog =
  if head prog == '/' -- Is prog an absolute file name?
  then do
    exfile <- doesFileExist prog
    if exfile then return prog
              else error ("File \""++prog++"\" not found!")
  else do
    loadpath <- getLoadPath
    findFirstFile loadpath
 where
   findFirstFile [] = error ("File \""++prog++"\" not found!")
   findFirstFile (dir:dirs) = do
     exfile <- doesFileExist (dir++"/"++prog)
     if exfile then return (dir++"/"++prog)
               else findFirstFile dirs


--- The evaluation of the action (sleep n) puts the Curry process
--- asleep for n seconds.

sleep :: Int -> IO ()
sleep external

--- Evaluates the argument to normal form (and return the normal form)
--- and print the time needed for this evaluation on standard error.

evalTime :: a -> a
evalTime external

--- Evaluates the argument to normal form (and return the normal form)
--- and print the time and space needed for this evaluation on standard error.
--- During the evaluation, the garbage collector is turned off.

evalSpace :: a -> a
evalSpace external

