------------------------------------------------------------------------------
--- A pretty printer for AbstractCurry programs.
---
--- This library defines a function "showProg" that shows
--- an AbstractCurry program in standard Curry syntax.
---
--- @author Martin Engelke, Bernd Brassel, Michael Hanus
--- @version June 2004
------------------------------------------------------------------------------

module AbstractCurryPrinter(showProg,
                            showTypeDecls,
                            showTypeDecl,
                            showTypeExpr,
                            showFuncDecl,
                            showExpr) where

import AbstractCurry
import List

-------------------------------------------------------------------------------
-- Functions to print an AbstractCurry program in standard Curry syntax
-------------------------------------------------------------------------------

--- Shows an AbstractCurry program in standard Curry syntax.
showProg :: CurryProg -> String
showProg (CurryProg _ imports typedecls funcdecls opdecls)
   =    showImports imports
     ++ showOpDecls opdecls
     ++ showTypeDecls typedecls
     ++ prefixInter showFuncDecl funcdecls "\n\n"
     ++ "\n"

showImports :: [String] -> String
showImports imports 
    =  prefixInter showImport (filter (/="prelude") imports) "\n" 
    ++ (if imports==["prelude"] then "" else "\n\n")

showImport :: String -> String
showImport imp
   = if imp /= "prelude" then "import " ++ imp else ""

showOpDecls :: [COpDecl] -> String
showOpDecls opdecls 
    =  prefixInter showOpDecl   opdecls   "\n" 
    ++ (if opdecls == [] then "" else "\n\n")

showOpDecl :: COpDecl -> String
showOpDecl (COp (_,name) fixity precedence)
   = showFixity fixity ++ " " ++ show precedence ++ " " 
   ++ if isInfixOpName name then name else '`':name++"`"

showFixity :: CFixity -> String
showFixity CInfixOp  = "infix"
showFixity CInfixlOp = "infixl"
showFixity CInfixrOp = "infixr"

--- Shows a list of AbstractCurry type declarations in standard Curry syntax.
showTypeDecls :: [CTypeDecl] -> String
showTypeDecls typedecls 
    =  prefixInter showTypeDecl typedecls "\n\n"
    ++ (if typedecls == [] then "" else "\n\n")

--- Shows an AbstractCurry type declaration in standard Curry syntax.
showTypeDecl :: CTypeDecl -> String
showTypeDecl (CTypeSyn (_,name) _ indexlist typeexpr)
   = "type " ++ name
             ++ (prefixMap (showTypeExpr False) (map CTVar indexlist) " ")
     ++ " = " ++ showTypeExpr False typeexpr
showTypeDecl (CType (_,name) _ indexlist consdecls)
   = "data " ++ name
             ++ (prefixMap (showTypeExpr False) (map CTVar indexlist) " ")
     ++ "\n"++showBlock ("= "++(combineMap showConsDecl consdecls "\n| "))

showConsDecl :: CConsDecl -> String
showConsDecl (CCons (_,name) _ _ typelist)
   = name ++ (prefixMap (showTypeExpr True) typelist " ")

--- Shows an AbstractCurry type expression in standard Curry syntax.
--- If the first argument is True, the type expression is enclosed
--- in brackets.
showTypeExpr :: Bool -> CTypeExpr -> String
showTypeExpr _ (CTVar (_,name)) = name
showTypeExpr nested (CFuncType domain range) =
   maybeShowBrackets nested (showTypeExpr (isCFuncType domain) domain ++
                             " -> " ++ showTypeExpr False range)
       
showTypeExpr nested (CTCons (mod,name) typelist)
   | mod=="prelude" && name == "untyped" = "-"
   | otherwise  = maybeShowBrackets (nested && not (null typelist))
                                    (showTypeCons mod name typelist)

isCFuncType t = case t of
                  CFuncType _ _ -> True
                  _ -> False

--- Shows an AbstractCurry function declaration in standard Curry syntax.
showFuncDecl :: CFuncDecl -> String
showFuncDecl (CFunc (_,name) _ _ typeexpr (CRules evalannot rules))
   = (if evalannot == CFlex then ""
         else bolName ++ " eval " ++ (showEvalAnnot evalannot)++"\n")
   ++(if isUntyped typeexpr then "\n" 
         else bolName ++ " :: " ++ (showTypeExpr False typeexpr)++"\n")
   ++(if funcIsInfixOp then  rulePrints
      else name ++ (prefixInter showRule rules ("\n"++name)))
   where
     funcIsInfixOp = isInfixOpName name
     bolName = if funcIsInfixOp then "("++name++")" else name
     rulePrints = concat $ intersperse "\n" 
                    $ map (insertName . (span (/=' ')) . tail . showRule) rules
     insertName (fstArg,rest) = fstArg++" "++name++rest
showFuncDecl (CFunc (_,name) _ _ typeexpr (CExternal _))
   = bolName ++ " :: " ++ (showTypeExpr False typeexpr) ++"\n"++
     bolName ++ " external"
   where
     bolName = if isInfixOpName name then "("++name++")" else name

showLocalFuncDecl :: CFuncDecl -> String
showLocalFuncDecl = showFuncDecl 
--showLocalFuncDecl (CFunc name _ _ typeexpr (CRules _ rules))
--   =    name ++ " :: " ++ (showTypeExpr False typeexpr) ++ "\n"
--     ++ name ++ " " ++ showBlock (prefixMap showRule rules ("\n " ++ name ++ " "))
--showLocalFuncDecl (CFunc name _ _ typeexpr (CExternal _))
--   =    name ++ " external\n"
--     ++ name ++ " :: " ++ (showTypeExpr False typeexpr)

showRule :: CRule -> String
showRule (CRule pattlist crhslist localdecls)
   =    (prefixMap showPattern pattlist " ")
     ++ showCrhsList crhslist
     ++ (case localdecls of
                (_:_) -> "\n" ++ showBlock ("where" ++ (prefixMap showLocalDecl localdecls "\n"))
                _     -> ""
        )

showEvalAnnot :: CEvalAnnot -> String
showEvalAnnot CFlex = "flex"
showEvalAnnot CRigid = "rigid"
showEvalAnnot CChoice = "choice"

showCrhsList :: [(CExpr,CExpr)] -> String
showCrhsList [] = ""
showCrhsList ((g,r):cs)
   | cs == [] && g == CSymbol ("prelude","success") 
   =  " = " ++ showExpr r
   | otherwise 
   = "\n" ++ showBlock (combineMap showCrhs ((g,r):cs) "\n")

showCrhs :: (CExpr,CExpr) -> String
showCrhs (cond,expr)
   | otherwise
   = "| " ++ showExpr cond ++ "\n= " ++ showExpr expr

showLocalDecl :: CLocalDecl -> String
showLocalDecl (CLocalFunc funcdecl)
   = showLocalFuncDecl funcdecl
showLocalDecl (CLocalPat pattern expr localdecls)
   =    (showPattern pattern) ++ " = " ++ (showExpr expr)
     ++ (case localdecls of
                (_:_) -> "\n" ++ showBlock ("where" ++ (prefixMap showLocalDecl localdecls "\n"))
                _     -> ""
        )
showLocalDecl (CLocalVar index)
   = (showPattern (CPVar index)) ++ " free"

--- Shows an AbstractCurry expression in standard Curry syntax.
showExpr :: CExpr -> String
showExpr (CVar (_,name)) = name
showExpr (CLit lit) = showLiteral lit
showExpr (CSymbol (_,name)) = if isInfixOpName name then "("++name++")" else name
showExpr (CApply func arg) = showApplication (CApply func arg)
showExpr (CLambda patts expr) = showLambdaOrSection patts expr
showExpr (CLetDecl localdecls expr)
   = "let\n" ++ showBlock ((combineMap showLocalDecl localdecls "\n") 
     ++ "\nin " ++ (showBoxedExpr expr))
showExpr (CDoExpr stmts)
   = "\n    do\n" ++ showBlock (combineMap showStatement stmts "\n")
showExpr (CListComp expr stmts)
   =    "[ " ++ (showBoxedExpr expr) ++ " | "
     ++ (combineMap showStatement stmts ", ") ++ "]"
showExpr (CCase expr branches)
   =    "case " ++ (showBoxedExpr expr) ++ " of\n"
     ++ showBlock (combineMap showBranchExpr branches "\n")

-- show a lambda expression as a left/right section, if possible:
showLambdaOrSection patts expr = case patts of
  [CPVar pvar] -> case expr of
     (CApply (CApply (CSymbol (_,name)) lexpr) (CVar var))
      -> if isInfixOpName name
         then if pvar==var
              then "(" ++ showBoxedExpr lexpr ++ " " ++ name ++ ")"
              else if lexpr == (CVar pvar)
                   then "(" ++ name ++ " " ++ showExpr (CVar var) ++ ")"
                   else showLambda patts expr
         else showLambda patts expr
     (CApply (CApply (CSymbol (_,name)) (CVar var)) rexpr)
      -> if isInfixOpName name && pvar==var
         then "(" ++ name ++ " " ++ showBoxedExpr rexpr ++ ")"
         else showLambda patts expr
     _ -> showLambda patts expr
  _ -> showLambda patts expr

showLambda patts expr = "\\" ++ (combineMap showPattern patts " ")
                        ++ " -> " ++ (showExpr expr)


showStatement :: CStatement -> String
showStatement (CSExpr expr) = showExpr expr
showStatement (CSPat pattern expr)
   = (showPattern pattern) ++ " <- " ++ (showExpr expr)
showStatement (CSLet localdecls)
   = case localdecls of
       (decl:[]) -> "let " ++ showLocalDecl decl
       _         -> "let\n" ++ showBlock (combineMap showLocalDecl localdecls "\n")

showPattern :: CPattern -> String
showPattern (CPVar (_,name)) = name
showPattern (CPLit lit) = showLiteral lit
showPattern (CPComb (mod,name) pattlist)
   = showCons mod name pattlist

showBranchExpr :: CBranchExpr -> String
showBranchExpr (CBranch pattern expr)
   = (showPattern pattern) ++ " -> " ++ (showExpr expr)

showLiteral :: CLiteral -> String
showLiteral (CIntc i) = show i
showLiteral (CFloatc f) = show f
showLiteral (CCharc c) = "'"++showCCharc (CCharc c)++"'"

showCCharc :: CLiteral -> String
showCCharc (CCharc c) | c=='\n' = "\\n"
                      | c=='\r' = "\\r"
                      | c=='\\' = "\\\\"
                      | otherwise = [c]

showBlock :: String -> String
showBlock text
   = combineMap id (map ((++) "    ") (filter ((/=) "") (lines text))) "\n"


showTypeCons :: String -> String -> [CTypeExpr] -> String
showTypeCons _ name [] = name
showTypeCons mod name (t:ts)
   | mod == "prelude" = showPreludeTypeCons name (t:ts)
   | otherwise        = name ++ (prefixMap (showTypeExpr True) (t:ts) " ")

showPreludeTypeCons :: String -> [CTypeExpr] -> String
showPreludeTypeCons name typelist
  | name == "[]" = "[" ++ (showTypeExpr False (head typelist)) ++ "]"
  | isTuple name = "(" ++ (combineMap (showTypeExpr False) typelist ",") ++ ")"
  | otherwise    = name ++ (prefixMap (showTypeExpr True) typelist " ")



showApplication :: CExpr -> String
showApplication appl
   = case (applicationHead appl) of
       (CSymbol name) -> showSymbolApplication name appl
       _              -> showSimpleApplication appl

applicationHead :: CExpr -> CExpr
applicationHead expr
   = case expr of
       (CApply func _) -> applicationHead func
       _               -> expr

showSymbolApplication :: (String,String) -> CExpr -> String
showSymbolApplication (mod,name) appl
   | mod == "prelude" && name == ":" 
   = showListApplication appl
   | isInfixOpName name 
   = showInfixApplication name appl
   | mod == "prelude" && name == "if_then_else" 
   = showITEApplication appl
   | isTuple name
   = showTupleApplication appl
   | otherwise        
   = showSimpleApplication appl

showListApplication :: CExpr -> String
showListApplication appl
   | isStringList appl
     = "\"" ++ (showCharListApplication appl) ++ "\""
   | isClosedList appl
     = "[" ++ (showConsListApplication appl) ++ "]"
   | otherwise
     = "(" ++ (showSimpleListApplication appl) ++ ")"

showCharListApplication :: CExpr -> String
showCharListApplication (CApply (CApply _ (CLit c)) tail)
   = case tail of
       (CSymbol _) -> showCCharc c
       _           -> showCCharc c ++ showCharListApplication tail

showConsListApplication :: CExpr -> String
showConsListApplication (CApply (CApply _ head) tail)
   = case tail of
       (CSymbol _) -> showBoxedExpr head
       _             -> (showBoxedExpr head) ++ "," ++ (showConsListApplication tail)

showSimpleListApplication :: CExpr -> String
showSimpleListApplication (CApply (CApply _ head) tail)
   = case tail of
       (CSymbol _) -> showBoxedExpr head ++ ":[]"
       _             -> (showBoxedExpr head) ++ ":" ++ (showBoxedExpr tail)

showInfixApplication :: String -> CExpr -> String
showInfixApplication infixop (CApply func arg2) 
   = case func of 
       (CApply _ arg1) -> (showBoxedExpr arg1) ++ " " ++ infixop 
                          ++ " " ++ (showBoxedExpr arg2)
       _ -> "(" ++ infixop ++ ") " ++ (showExpr arg2)

showITEApplication :: CExpr -> String
showITEApplication (CApply (CApply (CApply (CSymbol _) condExpr) thenExpr) elseExpr)
   =    "if " ++ (showExpr condExpr) ++ " then "
     ++ (showExpr thenExpr) ++ " else "
     ++ (showExpr elseExpr) 

showTupleApplication :: CExpr -> String
showTupleApplication appl
   = "(" ++ (p_showTuple appl) ++ ")"
   where
   p_showTuple (CApply (CSymbol _) arg)
      = showExpr arg
   p_showTuple (CApply (CApply e1 e2) arg)
      = (p_showTuple (CApply e1 e2)) ++ "," ++ (showExpr arg)

showSimpleApplication :: CExpr -> String
showSimpleApplication appl
   = case appl of
     CApply func arg -> showSimpleApplication func ++ " " ++ showBoxedExpr arg
     _               -> showBoxedExpr appl


showCons :: String -> String -> [CPattern] -> String
showCons _ name [] = name
showCons mod name (p:ps)
   | mod == "prelude" = showPreludeCons name (p:ps)
   | otherwise        = "(" ++ name ++ (prefixMap showPattern (p:ps) " ") ++ ")"

showPreludeCons :: String -> [CPattern] -> String
showPreludeCons name pattlist
   | name == ":"  = "(" ++ (showPattern (head pattlist)) ++ ":" ++ (showPattern (head (tail pattlist))) ++ ")"
   | isTuple name = "(" ++ (combineMap showPattern pattlist ",") ++ ")"
   | otherwise    = "(" ++ name ++ (prefixMap showPattern pattlist " ") ++ ")"


showBoxedExpr :: CExpr -> String
showBoxedExpr expr
   | isSimpleExpr expr = showExpr expr
   | otherwise         = "(" ++ showExpr expr ++ ")"

-------------------------------------------------------------------------------
--- composition functions for AbstractCurryPrinter
-------------------------------------------------------------------------------

prefixMap :: (a -> String) -> [a] ->  String -> String
prefixMap f xs s
   = concatMap ((++)s) (map f xs)

prefixInter :: (a -> String) -> [a] ->  String -> String
prefixInter f xs s
   = concat $ intersperse s (map f xs)

combineMap :: (a -> String) -> [a] ->  String -> String
combineMap _ [] _ = ""
combineMap f (x:xs) s
   = (f x) ++ (prefixMap f xs s)


dropTags :: String -> String
dropTags (x:xs) = case x of
                    '\"' -> dropTags $ tail $ dropWhile (/='\"') xs
                    '>'  -> xs
                    _    -> dropTags xs


-------------------------------------------------------------------------------
--- tests for various properties of AbstractCurry constructs
-------------------------------------------------------------------------------

isInfixOpName :: String -> Bool
isInfixOpName = all (`elem` infixIDs)

isStringList :: CExpr -> Bool
isStringList (CSymbol (mod,name))
   = mod == "prelude" && name == "[]"
isStringList (CVar _) = False
isStringList (CApply head tail)
   = case head of 
       (CApply _ (CLit (CCharc _))) -> isStringList tail
       _                            -> False

isClosedList :: CExpr -> Bool
isClosedList expr
   = case expr of
       (CApply (CApply _ _) tail)
          -> isClosedList tail
       (CSymbol (mod,name))
          -> mod == "prelude" && name == "[]"
       _  -> False

isSimpleExpr :: CExpr -> Bool
isSimpleExpr expr
   = case expr of
       (CVar _)      -> True
       (CLit _)      -> True
       (CSymbol (_, name)) -> not $ isInfixOpName name
       (CApply f _)  -> case (applicationHead f) of
                          (CSymbol ("prelude",name)) ->    name == ":"
                                                      || name == "[]"
                                                      || name == "()"
                                                      || isTuple name
                          _                        -> False
       _             -> False

isUntyped :: CTypeExpr -> Bool
isUntyped typeexpr
   = case typeexpr of
       (CTCons (mod,name) []) -> mod == "prelude" && name == "untyped"
       _                    -> False

isTuple :: String -> Bool
isTuple [] = False
isTuple (x:xs) = (x == '(') && (p1_isTuple xs)
   where
   p1_isTuple [] = False
   p1_isTuple (z:[]) = z == ')'
   p1_isTuple (z1:z2:zs) = (z1 == ',') && (p1_isTuple (z2:zs))

------------------------------------------------------------------------------
--- constants used by AbstractCurryPrinter
------------------------------------------------------------------------------

infixIDs :: String
infixIDs =  "~!@#$%^&*+-=<>?./|\\:"


-- enclose string with brackets, if required by first argument:
maybeShowBrackets nested s =
   (if nested then "(" else "") ++ s ++ (if nested then ")" else "")
