----------------------------------------------------------------------
--- The Curry string classifier is a simple tool to process strings containing 
--- Curry source code. The source string is classified into five categories:
---
--- (1) code - the part where the actual program is defined
---
--- (2) big comment - parts enclosed in {- ... -}
---
--- (3) small comment - from "--" to the end of a line
---
--- (4) text - a string, i.e. text enclosed in "..."
---
--- (5) letter - the given string is the representation of a character
---
--- For an example to use the state scanner cf. addtypes, the tool 
--- to add function types to a given program.
---
--- @author Bernd Brassel
--- @version September 2004
----------------------------------------------------------------------

module CurryStringClassifier(Tokens(..),scan,plainCode,unscan) where
  
import Char(isDigit)

--- The different categories to classify the source code.
data Tokens = SmallComment String Tokens
            | BigComment String Tokens
            | Text String Tokens
            | Letter String Tokens
            | Code String Tokens
            | End

--- Divides the given string into the five categories.
scan :: String -> Tokens
scan = stateScan 1 (Code x End) x where x free


stateScan _ token x ""  | x=:="" = token
stateScan _ (Code co End) x [c] | x=:=[c] = maybeCode co End
stateScan _ (Text t End) x [c] 
  | x=:="" = if c=='\"' then Text t End
                        else error "File ended while scanning string."
stateScan _ (BigComment _ End) x [_]
  | x=:="" = error "File ended while expecting -}"
stateScan line (Code co End) x (c:c':cs)
  | c=='\"' = guard (x=:="")
                    (maybeCode co (stateScan line (Text y End) y (c':cs))) 
  | c=='-' && c'=='-' 
  = let (comment,rest) = span (/='\n') cs
     in guard (x=:="") (maybeCode co 
                       (SmallComment comment 
                       (stateScan line (Code y End) y rest)))
  | c=='{' && c'=='-'
  = guard (x=:="") (maybeCode co (stateScan line (BigComment y End) y cs))
  | c'=='\'' && elem c (infixIDs++delimiters) 
  = guard (x=:=[c]) $ maybeCode co $ 
     case cs of
       '\\':l:'\'':rest -> Letter ['\\',l] (stateScan line (Code y End) y rest)
       '\\':a:b:d:'\'':rest -> 
         if all isDigit [a,b,d] 
           then Letter ['\\',a,b,d] (stateScan line (Code y End) y rest)
           else error $ "Improperly terminated character found in line "
                        ++ show line
       l:'\'':rest      -> Letter [l] (stateScan line (Code y End) y rest)
       _ -> error $ "Improperly terminated character in line "++show line
  | c=='\n'
  = guard (x=:=c:y) (stateScan (line+1) (Code co End) y (c':cs))
  | otherwise
  = guard (x=:=c:y) (stateScan line (Code co End) y (c':cs))
  where
    y free
stateScan line (BigComment co End) x (c:c':cs)
  | c=='-' && c'=='}' 
  = guard (x=:="") (BigComment co (stateScan line (Code y End) y cs))
  | c=='\n'
  = guard (x=:=c:y) (stateScan (line+1) (BigComment co End) y (c':cs))
  | otherwise 
  = guard (x=:=c:y) (stateScan line (BigComment co End) y (c':cs))
  where
   y free
stateScan line (Text t End) x (c:c':cs)
  | c == '\"' = guard (x=:="") (Text t (stateScan line (Code y End) y (c':cs)))
  | c == '\\' = guard (x=:=c:c':y) (stateScan line (Text t End) y cs)
  | elem c toBeEscaped = error $ "unescaped "++c:" encountered in line "
                                 ++ show line
  | otherwise = guard (x=:=c:y) (stateScan line (Text t End) y (c':cs))
  where
   y free

guard c x | c = x

infixIDs :: String
infixIDs =  "~!@#$%^&*+-=<>?./|\\:"

delimiters :: String
delimiters = " ([{,}])"

toBeEscaped = "\\\n\r\t\""

maybeCode s ts = if s=="" then ts else Code s ts

--- Yields the program code without comments 
--- (but with the line breaks for small comments).

plainCode :: Tokens -> String
plainCode (Code s ts) = s++plainCode ts
plainCode (Text s ts) = '\"':s++'\"':plainCode ts
plainCode (Letter s ts) = '\'':s++'\'':plainCode ts
plainCode (BigComment _ ts) = plainCode ts
plainCode (SmallComment _ ts) = plainCode ts
plainCode End = ""


--- Inverse function of scan, i.e., unscan (scan x) = x.
--- unscan is used to yield a program after changing the list of tokens.

unscan :: Tokens -> String
unscan (Code s ts) = s++unscan ts
unscan (Text s ts) = '\"':s++'\"':unscan ts
unscan (Letter s ts) = '\'':s++'\'':unscan ts
unscan (BigComment s ts) = "{-"++s++"-}"++unscan ts
unscan (SmallComment s ts) = "--"++s++unscan ts 
unscan End = ""
