------------------------------------------------------------------------------
--- Some tools to support meta-programming in Curry based on FlatCurry.
---
--- This library contains
--- <UL>
--- <LI> a show function for a string representation of FlatCurry programs
---   (function "showFlatProg")
---
--- <LI> a function for showing FlatCurry expressions in (almost) Curry syntax
---   (function "showCurryExpr")
--- </UL>
---
--- Note that the previously contained function "writeFLC"
--- is no longer supported. Use Flat2Fcy.writeFCY instead
--- and change file suffix into ".fcy"!
---
--- @author Michael Hanus
--- @version February 2002
------------------------------------------------------------------------------

module FlatCurryTools(showFlatProg,showCurryType,showCurryExpr,showCurryVar)
   where

import FlatCurry
import List
import Char

--- Shows a FlatCurry program term as a string (with some pretty printing).
showFlatProg :: Prog -> String
showFlatProg (Prog modname imports types funcs ops) =
     " (Prog " ++ flatShowString modname
     ++ (if imports==[] then "\n  []" else
         "\n  [" ++ flatShowListElems flatShowString imports ++ "]")
     ++ (if types==[] then "\n  []" else
         "\n  [" ++ flatShowListElems flatShowType types ++ "\n ]")
     ++ "\n  [" ++ flatShowListElems flatShowFunc funcs ++ "\n  ]"
     ++ "\n " ++ flatShowList flatShowOp ops
     ++ "\n )\n"

flatShowVisibility Public  = " Public "
flatShowVisibility Private = " Private "

flatShowFixity InfixOp = " InfixOp "
flatShowFixity InfixlOp = " InfixlOp "
flatShowFixity InfixrOp = " InfixrOp "

flatShowOp (Op name fix prec) =
 "(Op " ++ flatShowQName name ++ flatShowFixity fix ++ show prec ++ ")"

flatShowType (Type name vis tpars consdecls) =
  "\n  (Type " ++ flatShowQName name ++ flatShowVisibility vis
               ++ flatShowList show tpars
               ++ flatShowList flatShowCons consdecls ++ ")"
flatShowType (TypeSyn name vis tpars texp) =
  "\n  (TypeSyn " ++ flatShowQName name ++ flatShowVisibility vis
                  ++ flatShowList show tpars
                  ++ flatShowTypeExpr texp ++ ")"

flatShowCons (Cons cname arity vis types) =
  "(Cons " ++ flatShowQName cname ++ " " ++ show arity
           ++ flatShowVisibility vis
           ++ flatShowList flatShowTypeExpr types ++ ")"

flatShowFunc (Func name arity vis ftype rl) =
  "\n  (Func " ++ flatShowQName name ++ " " ++ show arity ++ " "
               ++ flatShowVisibility vis
               ++ flatShowTypeExpr ftype
               ++ flatShowRule rl ++ ")"

flatShowRule (Rule params expr) =
  " (Rule " ++ flatShowList show params
            ++ flatShowExpr expr ++ ")"
flatShowRule (External name) =
  " (External " ++ flatShowString name ++ ")"

flatShowTypeExpr (FuncType t1 t2) =
  "(FuncType " ++ flatShowTypeExpr t1 ++ " " ++ flatShowTypeExpr t2 ++ ")"
flatShowTypeExpr (TCons tc ts) =
  "(TCons " ++ flatShowQName tc
            ++ flatShowList flatShowTypeExpr ts ++ ")"
flatShowTypeExpr (TVar n) = "(TVar " ++ show n ++ ")"


flatShowCombType FuncCall = "FuncCall"
flatShowCombType ConsCall = "ConsCall"
flatShowCombType (PartCall n) = "(PartCall " ++ show n ++ ")"

flatShowExpr (Var n) = "(Var " ++ show n ++ ")"
flatShowExpr (Lit l) = "(Lit " ++ flatShowLit l ++ ")"
flatShowExpr (Comb ctype cf es) =
  "(Comb " ++ flatShowCombType ctype ++ " "
           ++ flatShowQName cf ++ flatShowList flatShowExpr es ++ ")"
flatShowExpr (Free xs e) =
  "(Free " ++ flatShowList show xs ++ flatShowExpr e ++ ")"
flatShowExpr (Or e1 e2) =
  "(Or " ++ flatShowExpr e1 ++ " " ++ flatShowExpr e2 ++ ")"
flatShowExpr (Case Rigid e bs) =
  "(Case Rigid " ++ flatShowExpr e ++ flatShowList flatShowBranch bs ++ ")"
flatShowExpr (Case Flex e bs) =
  "(Case Flex " ++ flatShowExpr e ++ flatShowList flatShowBranch bs ++ ")"

flatShowLit (Intc   i) = "(Intc " ++ show i ++ ")"
flatShowLit (Floatc f) = "(Floatc " ++ show f ++ ")"
flatShowLit (Charc  c) =
 if ord c >= 32  &&  ord c < 127
 then "(Charc '" ++ [c] ++ "')"
 else "(Charc (chr " ++ show (ord c) ++ "))"

flatShowBranch (Branch p e) = "(Branch " ++ flatShowPattern p
                                         ++ flatShowExpr e ++ ")"

flatShowPattern (Pattern qn xs) =
      "(Pattern " ++ flatShowQName qn
                  ++ flatShowList show xs ++ ")"
flatShowPattern (LPattern lit) = "(LPattern " ++ flatShowLit lit ++ ")"


-- format a finite list of elements:
flatShowList :: (a->String) -> [a] -> String
flatShowList format elems = " [" ++ flatShowListElems format elems ++ "] "

flatShowListElems :: (a->String) -> [a] -> String
flatShowListElems format elems = concat (intersperse "," (map format elems))

-- format a qualified name:
flatShowQName (mod,name) = "(\""++mod++"\",\""++name++"\")"

-- format a string:
flatShowString s = "\""++s++"\""


------------------------------------------------------------------------------
--- Shows a FlatCurry type in Curry syntax.
---
--- @param trans - a translation function from qualified type names
---                to external type names
--- @param nested - True iff brackets must be written around complex types
--- @param texpr - the FlatCurry type expression to be formatted
--- @return the String representation of the formatted type expression

showCurryType :: ((String,String) -> String) -> Bool -> TypeExpr -> String

showCurryType _ _ (TVar i) = [chr (97+i)]
showCurryType tf nested (FuncType t1 t2) =
  showBracketsIf nested
    (showCurryType tf (isFuncType t1) t1 ++ " -> " ++
     showCurryType tf False t2)
showCurryType tf nested (TCons tc ts)
 | ts==[]  = tf tc
 | tc==("prelude","[]") && (head ts == TCons ("prelude","Char") [])
   = "String"
 | tc==("prelude","[]")
  = "[" ++ showCurryType tf False (head ts) ++ "]" -- list type
 | take 2 (snd tc) == "(,"                          -- tuple type
  = "(" ++ concat (intersperse "," (map (showCurryType tf False) ts)) ++ ")"
 | otherwise
  = showBracketsIf nested
    (tf tc ++ concatMap (\t->' ':showCurryType tf True t) ts)

isFuncType (TVar _)       = False
isFuncType (FuncType _ _) = True
isFuncType (TCons _ _)  = False


------------------------------------------------------------------------------
--- Shows a FlatCurry expressions in (almost) Curry syntax.
---
--- @param trans - a translation function from qualified functions names
---                to external function names
--- @param nested - True iff brackets must be written around complex terms
--- @param indent - the indentation used in  case expressions and if-then-else
--- @param expr - the FlatCurry expression to be formatted
--- @return the String representation of the formatted expression

showCurryExpr :: ((String,String) -> String) -> Bool -> Int -> Expr -> String

showCurryExpr _ _ _ (Var n) = showCurryVar n

showCurryExpr _ _ _ (Lit l) = showCurryLit l

showCurryExpr tf _ _ (Comb _ cf []) = tf cf
showCurryExpr tf nested b (Comb _ cf [e]) =
  showBracketsIf nested (tf cf ++ " "
                            ++ showCurryExpr tf True b e)
showCurryExpr tf nested b (Comb ct cf [e1,e2])
 | cf==("prelude","apply")
  = showBracketsIf nested
       (showCurryExpr tf True b e1 ++ " " ++ showCurryExpr tf True b e2)
 | isAlpha (head (snd cf))
  = showBracketsIf nested
    (tf cf ++" "++ showCurryElems (showCurryExpr tf True b) [e1,e2])
 | isFiniteList (Comb ct cf [e1,e2])
  = "[" ++
    concat (intersperse "," (showCurryFiniteList tf b (Comb ct cf [e1,e2])))
        ++ "]"
 | snd cf == "(,)" -- pair constructor?
  = "(" ++ showCurryExpr tf False b e1 ++ "," ++
           showCurryExpr tf False b e2 ++ ")"
 | otherwise
  = showBracketsIf nested
              (showCurryExpr tf True b e1 ++ " " ++ tf cf ++ " " ++
               showCurryExpr tf True b e2 )
showCurryExpr tf nested b (Comb _ cf (e1:e2:e3:es))
 | cf==("prelude","if_then_else") && es==[]
  = showBracketsIf nested
        ("\n" ++
         sceBlanks b ++ " if "   ++ showCurryExpr tf False (b+2) e1 ++ "\n" ++
         sceBlanks b ++ " then " ++ showCurryExpr tf False (b+2) e2 ++ "\n" ++
         sceBlanks b ++ " else " ++ showCurryExpr tf False (b+2) e3)
 | take 2 (snd cf) == "(,"  -- tuple constructor?
  = "(" ++
    concat (intersperse "," (map (showCurryExpr tf False b) (e1:e2:e3:es)))
        ++ ")"
 | otherwise
  = showBracketsIf nested
       (tf cf ++ " "
        ++ showCurryElems (showCurryExpr tf True b) (e1:e2:e3:es))

showCurryExpr tf nested b (Free [] e) = showCurryExpr tf nested b e

showCurryExpr tf nested b (Free (x:xs) e) =
  showBracketsIf nested
    ("let " ++ concat (intersperse "," (map showCurryVar (x:xs))) ++
     " free in " ++ showCurryExpr tf False b e)

showCurryExpr tf nested b (Or e1 e2) =
  showBracketsIf nested
    (showCurryExpr tf True b e1 ++ " ? " ++ showCurryExpr tf True b e2)

showCurryExpr tf nested b (Case ctype e cs) =
  showBracketsIf nested
    ((if ctype==Rigid then "case " else "fcase ") ++
     showCurryExpr tf True b e ++ " of\n " ++
     showCurryElems (showCurryCase tf (b+2)) cs ++ sceBlanks b)


showCurryVar i = "v" ++ show i

showCurryLit (Intc   i) = show i
showCurryLit (Floatc f) = show f
showCurryLit (Charc  c) = show c

showCurryCase tf b (Branch (Pattern l vs) e) =
  sceBlanks b ++ showPattern (tf l) vs
              ++ " -> " ++ showCurryExpr tf False b e ++ "\n"
 where
   showPattern c [] = c
   showPattern c [x] = c ++ " " ++ showCurryVar x
   showPattern c [x1,x2] =
     if isAlpha (head c)
     then c ++ " " ++ showCurryVar x1 ++ " " ++ showCurryVar x2
     else if c=="(,)" -- pair constructor?
          then "(" ++ showCurryVar x1 ++ "," ++ showCurryVar x2 ++ ")"
          else showCurryVar x1 ++ " " ++ c ++ " " ++ showCurryVar x2
   showPattern c (x1:x2:x3:xs) =
     if take 2 c == "(,"  -- tuple constructor?
     then "("++ concat (intersperse "," (map showCurryVar (x1:x2:x3:xs))) ++")"
     else c ++ " " ++ showCurryElems showCurryVar (x1:x2:x3:xs)

showCurryCase tf b (Branch (LPattern l) e) =
  sceBlanks b ++ showCurryLit l ++ " "
              ++ " -> " ++ showCurryExpr tf False b e ++ "\n"

showCurryFiniteList _ _ (Comb _ ("prelude","[]") []) = []
showCurryFiniteList tf b (Comb _ ("prelude",":") [e1,e2]) =
  showCurryExpr tf False b e1 : showCurryFiniteList tf b e2

-- format a finite list of elements (with a format function for list elems):
showCurryElems :: (a->String) -> [a] -> String
showCurryElems format elems =
   concat (intersperse " " (map format elems))

showBracketsIf nested s = if nested then '(' : s ++ ")" else s

sceBlanks b = take b (repeat ' ')

-- Is the expression a finite list (with an empty list at the end)?
isFiniteList :: Expr -> Bool
isFiniteList (Var _) = False
isFiniteList (Lit _) = False
isFiniteList (Comb _ name args)
  | name==("prelude","[]") && args==[] = True
  | name==("prelude",":") && length args == 2 = isFiniteList (args!!1)
  | otherwise = False
isFiniteList (Free _ _) = False
isFiniteList (Or _ _) = False
isFiniteList (Case _ _ _) = False



------------------------------------------------------------------------------
