-----------------------------------------------------------------------------
-- Show all calls to imported functions in a module
--
-- Michael Hanus, November 2003
-----------------------------------------------------------------------------

import FlatCurry
import Char
import List
import Sort
import System
import Directory

m1 = showAllImportedCalls "ImportCalls"

-- Check arguments and call main function:
main = do
  args <- getArgs
  if length args /= 1
   then putStrLn $ "ERROR: Illegal arguments: " ++
                   concat (intersperse " " args) ++ "\n" ++
                   "Usage: importcalls <module_name>"
   else showAllImportedCalls (stripCurrySuffix (head args))

stripCurrySuffix n = let rn = reverse n in
  if take 6 rn == "yrruc." then reverse (drop 6 rn) else
  if take 7 rn == "yrrucl." then reverse (drop 7 rn) else n


showAllImportedCalls prog = do
  impcalls <- getAllImpCalls prog
  putStrLn ("Calls to imported functions/constructors in module \""++prog++"\":\n")
  mapIO_ (\(mod,imps)->putStrLn (mod++":\n"++concatMap showName imps++"\n"))
         impcalls
 where
   showName name = if isAlpha (head name)
                   then name++" "
                   else '(':name++") "

getAllImpCalls prog = do
  (Prog mod imps _ funs _) <- readCurrentFlatCurry prog
  return $ calls2impCalls imps
             (mergeSort (\(_,n1) (_,n2)->leqString n1 n2)
                        (filter (\(m,_)->m/=mod) (allFunCalls funs)))

calls2impCalls :: [String] -> [(String,String)] -> [(String,[String])]
calls2impCalls [] _ = []
calls2impCalls (mod:mods) funs = 
 (mod, map snd (filter (\(m,_)->m==mod) funs)) : calls2impCalls mods funs

-- Computes the list of called functions in a list of function declarations
allFunCalls :: [FuncDecl] -> [(String,String)]
allFunCalls [] = []
allFunCalls (Func _ _ _ _ (Rule _ e) : funs) =
     union (funsInExpr e) (allFunCalls funs)
allFunCalls (Func _ _ _ _ (External _) : funs) = allFunCalls funs

-- Gets a list of all functions (and partially applied constructors)
-- called in an expression:
funsInExpr :: Expr -> [(String,String)]
funsInExpr (Var _) = []
funsInExpr (Lit _) = []
funsInExpr (Comb _ f es) =
  if f `elem` [("prelude","commit"),("prelude","apply"),("prelude","cond")]
  then nub (concatMap funsInExpr es)
  else nub (f : concatMap funsInExpr es)
funsInExpr (Free _ e) = funsInExpr e
funsInExpr (Or e1 e2) = union (funsInExpr e1) (funsInExpr e2)
funsInExpr (Case _ e bs) = union (funsInExpr e)
                                 (nub (concatMap funsInBranch bs))
                     where funsInBranch (Branch _ be) = funsInExpr be


----------------- Auxiliaries:

-- Read a FlatCurry program (parse only if necessary):
readCurrentFlatCurry :: String -> IO Prog
readCurrentFlatCurry modname = do
  progname <- findFileInLoadPath modname
  flcexists <- doesFileExist (progname++".fcy")
  if not flcexists
    then readFlatCurry progname
    else do ctime <- getSourceModificationTime progname
            ftime <- getModificationTime (progname++".fcy")
            if ctime>ftime
             then readFlatCurry progname
             else readFlatCurryFile (progname++".fcy")

getSourceModificationTime progname = do
  lexists <- doesFileExist (progname++".lcurry")
  if lexists then getModificationTime (progname++".lcurry")
             else getModificationTime (progname++".curry")

-- add a directory name for a Curry source file by looking up the
-- current load path (CURRYPATH):
findFileInLoadPath prog =
 if head prog == '/'  -- absolute file name?
   then findFirstFile [""]
   else do loadpath <- getLoadPath
           findFirstFile (map (++"/") loadpath)
 where
   findFirstFile [] = error ("Curry file for module \""++prog++"\" not found!")
   findFirstFile (dir:dirs) =
     do exc <- doesFileExist (dir++prog++".curry")
        exl <- doesFileExist (dir++prog++".lcurry")
        if exc||exl then return (dir++prog)
                    else findFirstFile dirs
