-----------------------------------------------------------------
-- A tool to visualize the module dependencies of a program
-- (without the prelude which is used everywhere) as a graph with DaVinci:

module ImportGraph where

import FlatCurry
import DaVinci
import List
import System
import Directory

main = do
  args <- getArgs
  if length args /= 1
   then putStrLn $ "ERROR: Illegal arguments: " ++
                   concat (intersperse " " args) ++ "\n" ++
                   "Usage: importgraph <module_name>"
   else analyzeAndShow (head args)

-- Analyze and show module dependencies for a program:
analyzeAndShow :: String -> IO ()
analyzeAndShow prog = do
  aimps <- readAndProcessModuleWithImports analyseImports prog
  showDeps aimps

analyseImports :: Prog -> (String,[String])
analyseImports (Prog mod imps _ _ _) = (mod,imps)

--------------------------------------------------------------------------
-- Read and process the imported interfaces of a module.
-- First argument is a processing action and second argument
-- is a list of modules to be processed.
readAndProcessModuleWithImports :: (Prog->a) -> String -> IO [a]
readAndProcessModuleWithImports process modname = do
  -- try to read the interface files directly:
  mbaimps <- tryReadAndProcessModuleWithImports process modname
  maybe (parseAndProcessModuleWithImports process modname) return mbaimps

parseAndProcessModuleWithImports process modname = do
  putStrLn $ ">>>>> Interfaces too old, parsing module \""++modname++"\"..."
  readFlatCurryInt modname
  collectMods [modname] []
 where
  collectMods [] _ = return []
  collectMods (mod:mods) implist =
    if mod `elem` implist
    then collectMods mods implist
    else
      do putStrLn $ "Reading interface of module '"++mod++"'..."
         filename <- findFileInLoadPath (mod++".fint")
         prog <- readFlatCurryFile filename
         results <- collectMods (mods++getImports prog) (mod:implist)
         return (process prog : results)

-- Read and process the imported interfaces of a module.
-- First argument is a processing action and second argument
-- is a list of modules to be processed.
-- Nothing is returned if some interface files do not exists or are not
-- up-to-date
tryReadAndProcessModuleWithImports :: (Prog->a) -> String -> IO (Maybe [a])
tryReadAndProcessModuleWithImports process modname = collectMods [modname] []
 where
  collectMods [] _ = return (Just [])
  collectMods (mod:mods) implist =
    if mod `elem` implist
    then collectMods mods implist
    else do
      putStrLn $ "Reading interface of module '"++mod++"'..."
      mbprog <- readFlatIntIfPossible mod
      maybe (return Nothing)
            (\prog->collectMods (mods++getImports prog) (mod:implist) >>=
                    \mbresults -> return
                              (maybe Nothing
                                     (\results->Just (process prog : results))
                                     mbresults))
             mbprog

getImports :: Prog -> [String]
getImports (Prog _ imps _ _ _) = imps

-- Read a FlatCurry interface for a module if it exists and is up-to-date:
readFlatIntIfPossible :: String -> IO (Maybe Prog)
readFlatIntIfPossible modname = do
  progname <- findSourceFileInLoadPath modname
  fintexists <- doesFileExist (progname++".fint")
  if not fintexists
    then return Nothing
    else do ctime <- getSourceModificationTime progname
            ftime <- getModificationTime (progname++".fint")
            if ctime>ftime
             then return Nothing
             else readFlatCurryFile (progname++".fint") >>= return . Just

--- Adds a directory name where a module is stored by looking up the
--- current load path (CURRYPATH).
findSourceFileInLoadPath mod =
  if head mod == '/' -- Is mod an absolute file name?
  then do
    exfile <- doesSourceFileExist mod
    if exfile then return mod
              else error ("Module \""++mod++"\" not found!")
  else do
    loadpath <- getLoadPath
    findFirstFile loadpath
 where
   findFirstFile [] = error ("Module \""++mod++"\" not found!")
   findFirstFile (dir:dirs) = do
     exfile <- doesSourceFileExist (dir++"/"++mod)
     if exfile then return (dir++"/"++mod)
               else findFirstFile dirs

doesSourceFileExist mod = do
  excfile <- doesFileExist (mod++".curry")
  exlfile <- doesFileExist (mod++".lcurry")
  return (excfile || exlfile)

getSourceModificationTime progname = do
  lexists <- doesFileExist (progname++".lcurry")
  if lexists then getModificationTime (progname++".lcurry")
             else getModificationTime (progname++".curry")

--------------------------------------------------------------------------
-- Show a list of module dependencies as a graph with DaVinci:
showDeps :: [(String,[String])] -> IO ()
showDeps deps = dvDisplay (deps2dvGraph deps)

deps2dvGraph :: [(String,[String])] -> DvGraph
deps2dvGraph deps = graph2dvGraph $
 Graph (map Node (delete "prelude" (nub (map fst deps ++ concatMap snd deps))))
       (concatMap (\(m,is)->map (Edge m) (delete "prelude" is)) deps)

--------------------------------------------------------------------------
-- A graph is a list of nodes and edges:
data Graph = Graph [Node] [Edge]

-- A node is identified by a string (we assume that these strings are unique):
data Node = Node String

-- An edge consists of a source and a target node (identified by their ids):
data Edge = Edge String String

-- Convert this representation into the daVinci
-- representation by grouping edges to their source node:
graph2dvGraph :: Graph -> DvGraph
graph2dvGraph (Graph nodes edges) =
 dvNewGraph 
   (map (\(dvid,label)->dvNodeWithEdges dvid label
                                        (map (\(Edge _ n2)->createEdge n2)
                                             (filter (\(Edge n1 _)->n1==label)
                                                     edges))
                                        dvEmptyH) idnodes)
 where
   idnodes = map node2idnode nodes

   createEdge tlabel =
     dvSimpleEdge e (fst (head (filter (\(_,l)->l==tlabel) idnodes))) dvEmptyH
              where e free

   -- transform node into node with a DvId:
   node2idnode (Node label) = (dvid,label)  where dvid free

