;;; curry-mode.el --- A Curry mode as a slight extension of the
;;;
;;; haskell-mode.el --- A Haskell mode with extensive font lock support

;; Copyright (C) 1992, 1997-1998 Simon Marlow, Graeme E Moss, and Tommy Thorn

;; Authors: 1997-1998 Graeme E Moss <gem@cs.york.ac.uk> and
;;                    Tommy Thorn <thorn@irisa.fr>,
;;          1992 Simon Marlow
;; Keywords: faces files Haskell
;; Version: 1.1
;; URL: http://www.cs.york.ac.uk/~gem/haskell-mode/

;;; This file is not part of GNU Emacs.

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.


;;; Commentary:

;; Purpose:
;;
;; To provide a pleasant mode to browse and edit Haskell files, with
;; support for fontification of standard Haskell keywords, symbols,
;; functions, etc.  Top-level declarations are also scanned and placed
;; in a menu.  This mode supports full Latin1 Haskell 1.4 as well as
;; Bird-style literate scripts.
;;
;; Installation:
;; 
;; Put in your ~/.emacs:
;;
;;    (setq auto-mode-alist
;;          (append auto-mode-alist
;;                  '(("\\.curry$"  . curry-mode)
;;                    ("\\.lcurry$" . literate-curry-mode))))
;;
;;    (autoload 'curry-mode "curry-mode"
;;       "Major mode for editing Curry programs." t)
;;    (autoload 'literate-curry-mode "curry-mode"
;;       "Major mode for editing literate Curry programs." t)
;;
;; with `curry-mode.el' accessible somewhere on the load-path.
;; To add a directory `~/lib/emacs' (for example) to the load-path,
;; add the following to .emacs:
;;
;;    (setq load-path (cons "~/lib/emacs" load-path))
;;
;; To turn font locking on for all Haskell buffers, add this to .emacs:
;;
;;    (setq haskell-use-font-lock t)
;;
;; To turn declaration scanning on for all Haskell buffers, add this
;; to .emacs:
;;
;;    (setq haskell-use-decl-scan t)
;;
;;
;; Customisation:
;;
;; The colours and level of font locking may be customised.  See the
;; documentation on `haskell-font-lock' for more details.
;;
;;
;; History:
;;
;; This mode is based on an editing mode by Simon Marlow 11/1/92
;; and heavily modified by Graeme E Moss and Tommy Thorn 19/12/97.
;; 
;; If you have any problems or suggestions, after consulting the list
;; below, email gem@cs.york.ac.uk and thorn@irisa.fr quoting the
;; version of the mode you are using, the version of emacs you are
;; using, and a small example of the problem or suggestion.  Note that
;; font locking and declaration scanning require a reasonably recent
;; version of Emacs.
;;
;; Version 1.1:
;;   Added support for declaration scanning under XEmacs via
;;   func-menu.  Moved operators to level two fontification.
;;
;; Version 1.0:
;;   Added a nice indention support from Heribert Schuetz
;;   <Heribert.Schuetz@informatik.uni-muenchen.de>:
;;
;;     I have just hacked an Emacs Lisp function which you might prefer
;;     to `indent-relative' in haskell-mode.el.  See below.  It is not
;;     really Haskell-specific because it does not take into account
;;     keywords like `do', `of', and `let' (where the layout rule
;;     applies), but I already find it useful.
;;
;;   Cleaned up the imenu support.  Added support for literate scripts.
;;
;; Version 0.103 [HWL]:
;;   From Hans Wolfgang Loidl <hwloidl@dcs.gla.ac.uk>:
;;
;;   I (HWL) added imenu support by copying the appropriate functions
;;   from hugs-mode.  A menu-bar item "Declarations" is now added in
;;   haskell mode.  The new code, however, needs some clean-up.
;;
;; Version 0.102:
;;
;;   Moved C-c C-c key binding to comment-region.  Leave M-g M-g to do
;;   the work.  comment-start-skip is changed to comply with comment-start.
;;
;; Version 0.101:
;;
;;   Altered indent-line-function to indent-relative.
;;
;; Version 0.100:
;; 
;;   First official release.

;; Present Limitations/Future Work (contributions are most welcome!):
;;
;; . The TAB key moves the point through the indent-points of
;;   previous lines but it would be nice to have it more
;;   Haskell-specific.
;;
;; . Nested comments are not highlighted properly, eg. {- {- -} -}
;;   The first closing dash-brace returns the face to default.  It is
;;   not known to us if this is actually possible with font lock.
;;
;; . Debatable whether `()' `[]' `(->)' `(,)' `(,,)' etc. should be
;;   highlighted as constructors or not.  Should the `->' in
;;   `id :: a -> a' be considered a constructor or a keyword?  If so,
;;   how do we distinguish this from `\x -> x'?  What about the `\'?
;;
;; . Unicode is still a mystery...  has anyone used it yet?  We still
;;   support Latin-ISO-8859-1 though (the character set of Haskell 1.3).
;;
;; . Font lock fails on:
;;   - `\' at the beginning of a line not starting a string continuation,
;;     eg. when defining a lambda expression;
;;   - `--' or `{-' used within strings;
;;   - `'''' (three apostrophes) is treated as a Haskell-quoted apostrophe.
;;
;; . Because strings are not defined at the syntax level, this causes
;;   confusion, eg. matching parantheses.
;;
;; . XEmacs can support both `--' comments and `{- -}' comments
;;   simultaneously.  If XEmacs is detected, this should be used.
;; 
;; . Declarations requiring information extending beyond starting line
;;   don't get scanned properly, eg.
;;   > class Eq a =>
;;   >       Test a
;;
;; . Infix operators bound using infix notation and paranthesised
;;   first argument are not scanned or navigated over properly, eg.
;;   > (+++) :: [a] -> [a] -> [a]
;;   > (x:xs) +++ ys = x : (xs +++ ys)
;;   but should cause minimal problems.
;;
;; . Comments placed in the midst of the first few lexemes of a
;;   declaration will cause havoc, eg.
;;   > infixWithComments :: Int -> Int -> Int
;;   > x {-nastyComment-} `infixWithComments` y = x + y
;;   but are not worth worrying about.
;;
;; . Would prefer "Variables" to "Functions" declarations group but
;;   have kept latter for compatibility with haskell-doc.
;;
;; . Would be nice to scan other top-level declarations such as
;;   methods of a class, datatype field labels...  any more?
;;
;; . haskell-doc hooks key to haskell-mode-map but this ignores
;;   literate-haskell-mode-map.
;;
;; . Would like RET in literate mode to add a ">" at the start of a
;;   line when previous line starts with ">".  Or would "> " be
;;   better?
;;
;; . Support for Green Card?
;;
;; . Support for \begin{code}..\end{code} style of literate scripts?
;;
;; . Re-running (literate-)haskell-imenu should not cause the problems
;;   that it does.  The ability to turn off scanning and font lock
;;   (individually) would also be useful.  (Note that re-running
;;   (literate-)haskell-mode seems to cause no problems.)
;;
;; . Inconsistency: we define the start of a declaration in `imenu' as
;;   the start of the line the declaration starts on, but in
;;   `func-menu' as the start of the name that the declaration is
;;   given (eg. "class Eq a => Ord a ..." starts at "class" in `imenu'
;;   but at "Ord" in `func-menu').  This avoids rescanning of the
;;   buffer by the goto functions of `func-menu' but allows `imenu' to
;;   have the better definition of the start of the declaration (IMO).
;;
;; . `func-menu' cannot cope well with spaces in declaration names.
;;   This is unavoidable in "instance Eq Int" (changing the spaces to
;;   underscores would cause rescans of the buffer).  Note though that
;;   `fume-prompt-function-goto' (usually bound to "C-c g") does cope
;;   with spaces okay.
;;
;; . Would like to extend the goto functions given by `func-menu'
;;   under XEmacs to FSF Emacs.  Would have to implement these
;;   ourselves as `imenu' does not provide them.
;;
;; . `func-menu' uses its own syntax table when grabbing a declaration
;;   name to lookup (why doesn't it use the syntax table of the
;;   buffer?) so some declaration names will not be grabbed correctly,
;;   eg. "fib'" will be grabbed as "fib" since "'" is not a word or
;;   symbol constituent under the syntax table `func-menu' uses.


;; Structure of code:
;;
;; 1) Basic mode set-up.
;; 2) Support for font locking.
;; 3) Support for declaration scanning.
;;

;; These three components are independent, ie. (1) calls (2) and (3),
;; and (2) and (3) do not depend on any of the others.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Basic mode set-up
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Version of mode.
(defconst haskell-version "1.1"
  "haskell-mode version number.")
(defun haskell-version ()
  "Echo the current version of haskell-mode in the minibuffer."
  (interactive)
  (message "Using haskell-mode version %s" haskell-version))

;; Customization variables.
(defvar haskell-use-font-lock nil
  "If non-nil, `haskell-mode' and `literate-haskell-mode' will use
font locking to colour scripts.")

(defvar haskell-use-decl-scan nil
  "If non-nil, `haskell-mode' and `literate-haskell-mode' will scan
for declarations and place these in a menu using the `imenu' package
(Emacs) or `func-menu' package (XEmacs).")

;; Are we running FSF Emacs or XEmacs?
(defvar haskell-running-xemacs
  (string-match "Lucid\\|XEmacs" emacs-version)
  "non-nil if we are running XEmacs, nil otherwise.")

;; Mode maps.
(defvar haskell-mode-map ()
  "Keymap used in Haskell mode.")

(if (not haskell-mode-map)
    (progn
      (setq haskell-mode-map (make-sparse-keymap))
      (define-key haskell-mode-map "\C-c\C-c" 'comment-region)
      (define-key haskell-mode-map "\M-e" 'haskell-forward-decl)
      (define-key haskell-mode-map "\M-a" 'haskell-backward-decl)))

(defvar literate-haskell-mode-map ()
  "Keymap used in literate Haskell mode.")

(if (not literate-haskell-mode-map)
    (progn
      (setq literate-haskell-mode-map (copy-keymap haskell-mode-map))
      (define-key literate-haskell-mode-map "\M-e"
	'literate-haskell-forward-decl)
      (define-key literate-haskell-mode-map "\M-a"
	'literate-haskell-backward-decl)))

;; Syntax table.
(defvar haskell-mode-syntax-table (make-syntax-table)
  "Syntax table used in (literate) Haskell mode.  `haskell-font-lock' will
alter this to suit font locking.")

;; Various mode variables.
(defun haskell-vars ()
  (kill-all-local-variables)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'comment-start)
  (setq comment-start "-- ")
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "[-{]- *")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'haskell-comment-indent)
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'indent-relative-ext)
  )

;; The main mode functions
(defun curry-mode ()
  "Major mode for editing Haskell programs.  Last adapted for Haskell
1.4.  Blank lines separate paragraphs, comments start with `--'.
\\<haskell-mode-map>\\[indent-for-comment] will place a comment at an appropriate place on the current line.
\\[indent-for-tab-command] positions the cursor at the next indent point of previous non-blank
lines.
Use \\[newline-and-indent] to do a newline and indent to the first indent point of previous
non-blank lines.
\\[font-lock-fontify-block] re-colours 16 (or ARG) lines surrounding point if font lock is
enabled (this is necessary to colour comments enclosed by {- and -} over
more than one line).
\\[haskell-forward-decl] and \\[haskell-backward-decl] move forward and backward to the start of a declaration.

Under XEmacs with declaration scanning turned on, the following keys
are also defined:

\\[fume-list-functions] lists the declarations of the current buffer,
\\[fume-prompt-function-goto] prompts for a declaration to move to, and
\\[fume-mouse-function-goto] moves to the declaration whose name is at point.

Use `haskell-version' to find out what version this is.

Uses `haskell-font-lock' for font locking if the value of
`haskell-use-font-lock' is non-nil.

Uses `haskell-decl-scan' for declaration scanning if the value of
`haskell-use-decl-scan' is non-nil.

Entry to this mode calls the value of haskell-mode `haskell-mode-hook'
if non-nil."

  (interactive)
  (haskell-vars)
  (setq major-mode 'curry-mode)
  (setq mode-name "Curry")
  (use-local-map haskell-mode-map)
  (set-syntax-table haskell-mode-syntax-table)
  (if haskell-use-font-lock (haskell-font-lock))
  (if haskell-use-decl-scan (haskell-decl-scan))
  (run-hooks 'haskell-mode-hook))

(defun literate-curry-mode ()
  "As `curry-mode' but for Bird-style literate scripts.

Uses `literate-haskell-font-lock' for font locking if the value of
`haskell-use-font-lock' is non-nil.

Uses `literate-haskell-decl-scan' for declaration scanning if the value of
`haskell-use-decl-scan' is non-nil.

Entry to this mode calls the value of `haskell-mode-hook' and then
`literate-haskell-mode-hook' if non-nil."

  (interactive)
  (haskell-vars)
  (setq major-mode 'literate-curry-mode)
  (setq mode-name "Literate-Curry")
  (use-local-map literate-haskell-mode-map)
  (set-syntax-table haskell-mode-syntax-table)
  (if haskell-use-font-lock (literate-haskell-font-lock))
  (if haskell-use-decl-scan (literate-haskell-decl-scan))
  (run-hooks 'haskell-mode-hook)
  (run-hooks 'literate-haskell-mode-hook))

;; Find the indentation level for a comment.
(defun haskell-comment-indent ()
  (skip-chars-backward " \t")
  ;; if the line is blank, put the comment at the beginning,
  ;; else at comment-column
  (if (bolp) 0 (max (1+ (current-column)) comment-column)))

;; Partly stolen from `indent-relative' in indent.el:
(defun indent-relative-ext ()
  "Space out to under next visible indent point.
Indent points are positions of non-whitespace following whitespace in
lines preceeding point. A position is visible if it is to the left of
the first non-whitespace of every nonblank line between the position and
the current line. If there is no visible indent point beyond the current
column, `tab-to-tab-stop' is done instead."
  (interactive)
  (let ((start-column (current-column))
	(invisible-from nil)		; `nil' means infinity here
	indent )
    (catch 'indent-relative-ext-break
      (save-excursion
	(while (progn (beginning-of-line)
		      (> (point) (point-min)))
	  (forward-line -1)
	  (if (not (looking-at "[ \t]*\n"))
	      (let ((this-indentation (current-indentation)))
		(if (or (not invisible-from)
			(< this-indentation invisible-from))
		    (if (> this-indentation start-column)
			(setq invisible-from this-indentation)
		      (let ((end (save-excursion (forward-line 1) (point))))
			(move-to-column start-column)
			;; Is start-column inside a tab on this line?
			(if (> (current-column) start-column)
			    (backward-char 1))
			(or (looking-at "[ \t]")
			    (skip-chars-forward "^ \t" end))
			(skip-chars-forward " \t" end)
			(let ((col (current-column)))
			  (setq indent
				(if (or (= (point) end)
					(and invisible-from
					     (> col invisible-from) ) )
				    invisible-from
				  col ) ) )
			(throw 'indent-relative-ext-break nil) ) ) ) ) ) ) ) )
    (if indent
	(let ((opoint (point-marker)))
	  (delete-region (point) (progn (skip-chars-backward " \t") (point)))
	  (indent-to indent 0)
	  (if (> opoint (point))
	      (goto-char opoint))
	  (move-marker opoint nil))
      (tab-to-tab-stop)) ) )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Support for font-lock-mode
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar haskell-font-lock-keywords ()
  "The default definitions used by font lock for fontification of
Haskell scripts.  This variable is set by `haskell-font-lock' and then
used by `font-lock-defaults'.")

(defvar haskell-font-lock-keywords-1 ()
  "Medium level font lock definitions for non-literate Haskell.")

(defvar haskell-font-lock-keywords-2 ()
  "High level font lock definitions for non-literate Haskell.")

(defvar literate-haskell-font-lock-keywords ()
  "The default definitions used by font lock for fontification of
Bird-style literate Haskell scripts.  This variable is set by
`literate-haskell-font-lock' and then used by `font-lock-defaults'.")

(defvar literate-haskell-font-lock-keywords-1 ()
  "Medium level font lock definitions for literate Haskell.")

(defvar literate-haskell-font-lock-keywords-2 ()
  "High level font lock definitions for literate Haskell.")

;; The font lock regular expressions.
(defun haskell-font-lock-keywords-create (literate level)
  "Creates appropriate LEVEL (1 or 2) of fontification definitions for
(LITERATE) Haskell scripts.  Returns keywords suitable for
`font-lock-keywords'."
  (let* (;; Bird-style literate scripts start a line of code with
	 ;; "^>", otherwise a line of code starts with "^".
	 (line-prefix (if literate "^>" "^"))

	 ;; Most names are borrowed from the lexical syntax of the Haskell
	 ;; report.
         ;; Some of these definitions have been superseded by using the
         ;; syntax table instead.

	 (ASCsymbol "-!#$%&*+./<=>?@\\\\^|~") 
         ;; Put the minus first to make it work in ranges.
         (ISOsymbol "\241-\277\327\367")
         (ISOlarge  "\300-\326\330-\337")
         (ISOsmall  "\340-\366\370-\377")
         (small
          (concat "a-z" ISOsmall))
         (large
          (concat "A-Z" ISOlarge))
         (symbol
          (concat ASCsymbol ISOsymbol))

         ;; We allow _ as the first char to fit GHC
         (varid
          (concat "\\b\\([" small large "0-9'_]+\\)\\b"))
         (conid
          (concat "\\b\\([" large "][" small large "0-9'_]*\\)\\b"))
	 (sym
	  (concat "[" symbol ":]+"))

         ;; Reserved operations
         ;(reservedsym
         ; '(".." "::" "=" "\\" "|" "<-" "->" "@" "~" "=>"))
         ;; Reserved identifiers
         ;(reservedid
         ; '("as" "case" "choice" "class" "data" "default" "deriving" "do" "else" "eval"
         ;   "flex" "free" "hiding" "if" "import" "in" "infix" "infixl" "infixr"
         ;   "instance" "let" "module" "newtype" "of" "qualified" "rigid" "then"
         ;   "type" "where"))
	 ;; make-regexp applied to reservedsym and reservedid creates
	 ;; the following regexps
	 ;; We add "\\`" and "\\'" to speed up sole later use of
	 ;; reserved sym.
	 (reservedsym
	  "\\`\\([=@\\|~]\\|->\\|\\.\\.\\|::\\|<-\\|=>\\)\\'")
	 (reservedid
	  "\\b\\(as\\|c\\(ase\\|hoice\\|lass\\)\\|d\\(ata\\|e\\(fault\\|riving\\)\\|o\\)\\|e\\(lse\\|val\\)\\|f\\(lex\\|ree\\)\\|hiding\\|i\\([fn]\\|mport\\|n\\(fix\\(\\|[lr]\\)\\|stance\\)\\)\\|let\\|module\\|newtype\\|of\\|qualified\\|rigid\\|t\\(hen\\|ype\\)\\|where\\)\\b")

         ;; This unreadable regexp matches strings and character
         ;; constants.  We need to do this with one regexp to handle
         ;; stuff like '"':"'".  The regexp is actually just the
         ;; composition of "([^"\\]|\\.)*" for strings and
         ;; '([^\\]|\\.[^']*)' for characters.   
         (string-and-char
          (concat "\\(\\(\"\\|" line-prefix "[ \t]*\\\\\\)\\([^\"\\\\]\\|\\\\.\\)*\\(\"\\|\\\\$\\)\\|'\\([^\\\\]\\|\\\\.[^']*\\)'\\)"))

	 ;; Top-level declarations
	 ;; These are not included as they don't work well (yet).
;	 (topdecl1
;	  (concat line-prefix "\\(" varid "\\)\\(\\s-\\|::\\|=\\||\\)"))
;	 (topdecl2
;	  (concat line-prefix varid "\\s-*\\(" sym "\\)"))
;	 (topdecl3
;	  (concat line-prefix "(\\(" sym "\\))"))

	 font-lock-keywords)

    (setq font-lock-keywords
          ;; There must be a better way to do this list construction...
	  (append
;;
;; NOTICE the ordering below is significant
;;
	   '(("--.*$" 0 'haskell-comment-face t))
	   ;; Expensive.
	   (list (list string-and-char 1 ''haskell-string-char-face))
	   '(("()" 0 'haskell-constructor-face))
	   '(("(,*)" 0 'haskell-constructor-face))
	   '(("\\[\\]" 0 'haskell-constructor-face))
	   '(("(->)" 0 'haskell-constructor-face))
	   ;; Expensive.
	   (list (list reservedid 1 ''haskell-keyword-face))
	   (if (eq level 2)
	     (list (list (concat "\`" varid "\`") 0 ''haskell-operator-face)))
	   ;; Expensive.
	   (list (list conid 1 ''haskell-constructor-face))
	   ;; Very expensive.
	   (list (list sym 0
		       (append '(cond)
			       (list (list (list 'string-match
						 reservedsym
						 '(match-string 0))
					   ''haskell-keyword-face))
			       '(((string-match "\\`:" (match-string 0))
				  'haskell-constructor-face))
			       (if (eq level 2)
				   '((t 'haskell-operator-face))))))
	   ;;	   (list topdecl1 1 ''haskell-definition-face t)
	   ;;	   (list topdecl2 1 ''haskell-definition-face t)
	   ;;	   (list topdecl3 1 ''haskell-definition-face t)
	   ))
      (if literate
	  (setq font-lock-keywords
		(append '(("^[^>\n].*$" 0 'haskell-comment-face t))
			font-lock-keywords
			'(("^>" 0 'haskell-default-face t)))))
      font-lock-keywords))

(defun haskell-font-lock-defaults-create (literate)
  "Makes local variable `font-lock-defaults' suitable for Haskell font locking.
If LITERATE is non-nil then the font locking is made suitable for literate
Haskell scripts."
  (setq literate-haskell-font-lock-keywords-1
	(haskell-font-lock-keywords-create t 1))
  (setq literate-haskell-font-lock-keywords-2
	(haskell-font-lock-keywords-create t 2))
  (setq literate-haskell-font-lock-keywords
	literate-haskell-font-lock-keywords-1)
  (setq haskell-font-lock-keywords-1
	(haskell-font-lock-keywords-create nil 1))
  (setq haskell-font-lock-keywords-2
	(haskell-font-lock-keywords-create nil 2))
  (setq haskell-font-lock-keywords
	haskell-font-lock-keywords-1)
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults
	(if literate
	    '((literate-haskell-font-lock-keywords
	       literate-haskell-font-lock-keywords-1
	       literate-haskell-font-lock-keywords-2) nil nil)
	  '((haskell-font-lock-keywords
	     haskell-font-lock-keywords-1
	     haskell-font-lock-keywords-2) nil nil))))

;; Faces required for font locking.
(defun haskell-faces ()
  "Defines faces required for Haskell font locking.

See `haskell-font-lock'."

  ;; XEmacs does not have a simple function for making the faces but
  ;; makes them when `require'd which was done by haskell-font-lock,
  ;; so we don't need to explicitly make them for XEmacs, and in fact
  ;; we shouldn't as an error will be produced.
  (if (fboundp 'font-lock-make-faces) (font-lock-make-faces))
  (copy-face 'font-lock-keyword-face 'haskell-keyword-face)
  (copy-face 'font-lock-type-face 'haskell-constructor-face)
  (copy-face 'font-lock-string-face 'haskell-string-char-face)
  (copy-face 'font-lock-function-name-face 'haskell-operator-face)
  (copy-face 'font-lock-comment-face 'haskell-comment-face)
  (copy-face 'default 'haskell-default-face)
;  (copy-face 'font-lock-reference-face 'haskell-definition-face)
  )

;; Haskell-like function for creating [from..to].
(defun enum-from-to (from to)
  (if (> from to)
      ()
    (cons from (enum-from-to (1+ from) to))))

;; Syntax required for font locking.
(defun haskell-syntax ()
  "Changes the current buffer's syntax table to suit Haskell font locking.

See `haskell-font-lock'."
  ;; It's easier for us to manually set the ISO Latin1 syntax as I'm
  ;; not sure what libraries are available and how they differ from
  ;; Haskell, eg. the iso-syntax library of Emacs 19.34 defines \241
  ;; as punctuation for good reasons but this conflicts with Haskell
  ;; so we would have to redefine it.  It's simpler for us to set the
  ;; syntax table according to the Haskell report for all of the 8-bit
  ;; characters.
  (setq haskell-syntax-table (make-syntax-table))
  (modify-syntax-entry ?\  " " haskell-syntax-table)
  (modify-syntax-entry ?\t " " haskell-syntax-table)
  (modify-syntax-entry ?\" " " haskell-syntax-table)
  (modify-syntax-entry ?\' "w" haskell-syntax-table)
  (modify-syntax-entry ?_  "w" haskell-syntax-table)
  (modify-syntax-entry ?\( "()" haskell-syntax-table)
  (modify-syntax-entry ?\) ")(" haskell-syntax-table)
  (modify-syntax-entry ?[  "(]" haskell-syntax-table)
  (modify-syntax-entry ?]  ")[" haskell-syntax-table)
  (modify-syntax-entry ?{  "(}1" haskell-syntax-table)
  (modify-syntax-entry ?}  "){4" haskell-syntax-table)
  (modify-syntax-entry ?-  "_ 23" haskell-syntax-table)
  (modify-syntax-entry ?\` "$`" haskell-syntax-table)
  (mapcar (lambda (x)
	    (modify-syntax-entry x "_" haskell-syntax-table))
	  (concat "!#$%&*+./:<=>?@\\\\^|~"
		  (enum-from-to ?\241 ?\277)
		  "\327\367"))
  (mapcar (lambda (x)
	    (modify-syntax-entry x "w" haskell-syntax-table))
	  (concat (enum-from-to ?\300 ?\326)
		  (enum-from-to ?\330 ?\337)
		  (enum-from-to ?\340 ?\366)
		  (enum-from-to ?\370 ?\377)))
  (set-syntax-table haskell-syntax-table))

;; The main functions.
(defun haskell-font-lock ()
  "Turns on font locking in current buffer for Haskell 1.4 scripts.

Changes the current buffer's `font-lock-defaults' and syntax table, and
adds the following faces:

   haskell-keyword-face      for reserved keywords and syntax,
   haskell-constructor-face  for data- and type-constructors, class names,
                             and module names,
   haskell-string-char-face  for strings and characters,
   haskell-operator-face     for symbolic and alphanumeric operators,
   haskell-comment-face      for comments, and
   haskell-default-face      for ordinary code.

The faces are initialised to the following font lock defaults:

   haskell-keyword-face      font-lock-keyword-face
   haskell-constructor-face  font-lock-type-face
   haskell-string-char-face  font-lock-string-face
   haskell-operator-face     font-lock-function-name-face
   haskell-comment-face      font-lock-comment-face
   haskell-default-face      <default face>

Two levels of fontification are defined: level one (the default)
and level two (more colour).  The former does not colour operators.
Use the variable `font-lock-maximum-decoration' to chose
non-default levels of fontification.  For example, adding this to
.emacs:

  (setq font-lock-maximum-decoration '((haskell-mode . 2) (t . 0)))

uses level two fontification for haskell-mode and default level for
all other modes.  See documentation on this variable for further
details.  Note that literate-haskell-mode is considered a separate
mode.

To alter an attribute of a face, add a hook, for example: to change
the foreground colour of comments to brown, add the following line
to .emacs:

  (add-hook 'haskell-mode-hook
      (lambda ()
          (set-face-foreground 'haskell-comment-face \"brown\")))

Note that the colours available vary from system to system.  To see
what colours are available on your system, call
`list-colors-display' from emacs.

To turn font locking on for all Haskell buffers, add this to .emacs:

  (add-hook 'haskell-mode-hook 'turn-on-font-lock)

To turn font locking on or off for the current buffer, call `font-lock-mode'.

Literate Haskell scripts are supported by `literate-haskell-font-lock'."

  (interactive)
  (require 'font-lock)
  (haskell-syntax)
  (haskell-faces)
  (haskell-font-lock-defaults-create nil)
  (turn-on-font-lock))

(defun literate-haskell-font-lock ()
  "Turns on font locking in current buffer for Bird-style literate
Haskell 1.4 scripts.

See `haskell-font-lock'."

  (interactive)
  (require 'font-lock)
  (haskell-syntax)
  (haskell-faces)
  (haskell-font-lock-defaults-create t)
  (turn-on-font-lock))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Support for declaration scanning.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; The imenu support is based on code taken from `hugs-mode',
;;; thanks go to Chris Van Humbeeck.

;;###autoload
;; As `cl' defines macros that `imenu' uses, we must require them at
;; compile time.
(eval-when-compile
  ;; `imenu' isn't used in XEmacs.
  (if (not (string-match "Lucid\\|XEmacs" emacs-version))
      (progn
	(require 'cl)
	(require 'imenu))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; General declaration scanning functions.

(defun haskell-match-string (num)
  "As `match-string' except that the string is stripped of properties."
  (format "%s" (match-string num)))

(defvar haskell-start-keywords-re
  (concat "\\(\\<"
	  "class\\|data\\|i\\(mport\\|n\\(fix\\(\\|[lr]\\)\\|stance\\)\\)\\|"
	  "module\\|primitive\\|type\\|newtype"
	  "\\)\\>")
  "Keywords that may start a declaration.") 

(defun haskell-get-variable (prefix)
  "Assuming point is looking at the regexp PREFIX followed by the
start of a declaration (perhaps in the middle of a series of
declarations concerning a single variable), if this declaration is a
value binding or type signature, return the variable involved, and
otherwise returns nil.  Point is not moved in either case."
  ;; PROBLEM: I assume here that the first argument to an infix
  ;; operator is a word (syntactically) which is not true in general,
  ;; eg. `(x:xs) +++ ys = ...'.  Very hard to separate the first
  ;; argument of an infix operator from the operator itself.
  (if (re-search-forward (concat "\\=" prefix) (point-max) t)
      (let (name)
	;; Keyword.
	(if (looking-at haskell-start-keywords-re)
	    (setq name nil)
	  (if (looking-at "(\\(\\s_+\\))")
	      ;; Paranthesised symbolic variable.
	      (setq name (haskell-match-string 1))
	    (if (looking-at "\\(\\sw+\\)\\s-*\\(\\s_+\\|`\\sw+`\\)")
		(let ((match1 (haskell-match-string 1))
		      (match2 (haskell-match-string 2)))
		  ;; Weed out `::', `=' and `|' from potential infix
		  ;; symbolic variable.
		  (if (member match2 '("::" "=" "|"))
		      ;; Variable identifier.
		      (setq name match1)
		    (if (string-match "\\``\\(\\sw+\\)`\\'" match2)
			;; Infix variable identifier.
			(setq name (match-string 1 match2))
		      ;; Infix symbolic variable.
		      (setq name match2))))
	      (if (looking-at "\\(\\sw+\\)\\>")
		  ;; Variable identifier.
		  (setq name (haskell-match-string 1))
		;; Declaration not a type signature or value binding.
		;; Return nil.
		(setq name nil)))))
	;; Return the result.
	(re-search-backward prefix)
	name)))

(defun haskell-move-to-start-regexp (inc regexp)
  "Move to beginning of line that succeeds/preceeds (INC = 1/-1)
current line that starts with REGEXP and is not in font-lock-comment-face."
  ;; Making this defsubst instead of defun appears to have little or
  ;; no effect on efficiency.  It is probably not called enough to do
  ;; so.
  (while (and (= (forward-line inc) 0)
	      (or (not (looking-at regexp))
		  (eq (get-text-property (point) 'face)
		      'font-lock-comment-face)))))

(defvar literate-haskell-decl-scan-line-prefix "> "
  "The regexp that matches the start of a line of literate code for
the purposes of imenu support.  Current value is \"> \" as we assume
top-level declarations start at column 3.  Must not contain the
special \"^\" regexp as we may not use the regexp at the start of a
regexp string.  Note this is only for imenu support.")

(defun haskell-move-to-decl (direction literate fix)
  "General function for moving to the start of a declaration, either
forwards or backwards from point, with normal or with literate
scripts.  If DIRECTION is t, then forward, else backward.  If LITERATE
is t, then treat as literate scripts, else normal scripts.  Returns
point if point is left at the start of a declaration, and nil
otherwise, ie. because point is at the beginning or end of the buffer
and no declaration starts there.  If FIX is t, then point does not
move if already at the start of a declaration."
  ;; As `haskell-get-variable' cannot separate an infix variable
  ;; identifier out of a value binding with non-alphanumeric first
  ;; argument, this function will treat such value bindings as
  ;; separate from the declarations surrounding it.
  (interactive)
  (let* (;; The variable typed or bound in the current series of
	 ;; declarations.
	 name
	 ;; The variable typed or bound in the new declaration.
	 newname
	 ;; Hack to solve hard problem for literate scripts that start
	 ;; with a declaration.  We are in the abyss if point is
	 ;; before start of this declaration.
	 abyss
	 (line-prefix (if literate literate-haskell-decl-scan-line-prefix ""))
	 ;; The regexp to match for the start of a declaration.
	 (start-decl-re (concat line-prefix "\\(\\sw\\|(\\)"))
	 (increment (if direction 1 -1))
	 (bound (if direction (point-max) (point-min))))
    (progn
      ;; Move to beginning of line that starts the "current
      ;; declaration" (dependent on DIRECTION and FIX), and then get
      ;; the variable typed or bound by this declaration, if any.
      (let (;; Where point was at call of function.
	    (here (point))
	    ;; Where the declaration on this line (if any) starts.
	    (start (progn
		     (beginning-of-line)
		     (if (looking-at start-decl-re)
			 (match-beginning 1)))))
	(if (and start
		 ;; This complicated boolean determines whether we
		 ;; should include the declaration that starts on the
		 ;; current line as the "current declaration" or not.
		 (or (and (or (and direction (not fix))
			      (and (not direction) fix))
			  (>= here start))
		     (and (or (and direction fix)
			      (and (not direction) (not fix)))
			  (> here start))))
	    ;; If so, we are already at start of the current line, so
	    ;; do nothing.
	    ()
	  ;; If point was before start of a declaration on the
	  ;; first line of the buffer (possible for literate
	  ;; scripts) then we are in the abyss.
	  (if (and start (bobp))
	      (setq abyss t)
	    ;; Otherwise we move to the start of the first declaration
	    ;; on a line preceeding the current one.
	    (haskell-move-to-start-regexp -1 start-decl-re))))
      ;; If we are in the abyss, position and return as appropriate.
      (if abyss
	  (if (not direction)
	      nil
	    (re-search-forward (concat "\\=" line-prefix) (point-max) t)
	    (point))
	;; Get the variable typed or bound by this declaration, if any. 
	(setq name (haskell-get-variable line-prefix))
	(if (not name)
	    ;; If no such variable, stop at the start of this
	    ;; declaration if moving backward, or move to the next
	    ;; declaration if moving forward.
	    (if direction
		(haskell-move-to-start-regexp 1 start-decl-re))
	  ;; If there is a variable, find the first
	  ;; succeeding/preceeding declaration that does not type or
	  ;; bind it.  Check for reaching start/end of buffer.
	  (haskell-move-to-start-regexp increment start-decl-re)
	  (while (and (/= (point) bound)
		      (and (setq newname (haskell-get-variable line-prefix))
			   (string= name newname)))
	    (setq name newname)
	    (haskell-move-to-start-regexp increment start-decl-re))
	  ;; If we are going backward, and have either reached a new
	  ;; declaration or the beginning of a buffer that does not
	  ;; start with a declaration, move forward to start of next
	  ;; declaration (which must exist).  Otherwise, we are done.
	  (if (and (not direction)
		   (or (and (looking-at start-decl-re)
			    (not (string= name
					  ;; Note we must not use
					  ;; newname here as this may
					  ;; not have been set if we
					  ;; have reached the beginning
					  ;; of the buffer.
					  (haskell-get-variable line-prefix))))
		       (and (not (looking-at start-decl-re))
			    (bobp))))
	      (haskell-move-to-start-regexp 1 start-decl-re)))
	;; Store whether we are at the start of a declaration or not.
	;; Used to calculate final result.
	(let ((at-start-decl (looking-at start-decl-re)))
	  ;; If we are at the beginning of a line, move over
	  ;; line-prefix, if present at point.
	  (if (bolp)
	      (re-search-forward (concat "\\=" line-prefix) (point-max) t))
	  ;; Return point if at the start of a declaration and nil
	  ;; otherwise.
	  (if at-start-decl (point) nil))))))

(defun haskell-backward-decl ()
  "Move point backward to the first character preceeding the current
point that starts a top-level declaration.  A series of declarations
concerning one variable is treated as one declaration by this
function.  So, if point is within a top-level declaration then move it
to the start of that declaration.  If point is already at the start of
a top-level declaration, then move it to the start of the preceeding
declaration.  Returns point if point is left at the start of a
declaration, and nil otherwise, ie. because point is at the beginning
of the buffer and no declaration starts there."
  (interactive)
  (haskell-move-to-decl nil nil nil))

(defun haskell-forward-decl ()
  "As `haskell-backward-decl' but forward."
  (interactive)
  (haskell-move-to-decl t nil nil))

(defun literate-haskell-backward-decl ()
  "As `haskell-backward-decl' but for literate scripts."
  (interactive)
  (haskell-move-to-decl nil t nil))

(defun literate-haskell-forward-decl ()
  "As `literate-haskell-backward-decl' but forward."
  (interactive)
  (haskell-move-to-decl t t nil))

(defun haskell-generic-find-next-decl (literate)
  "Find the name, position and type of the declaration at or after
point.  Returns `((name . (start-position . name-position)) . type)'
if one exists and nil otherwise.  The start-position is at the start
of the declaration, and the name-position is at the start of the name
of the declaration.  The name is a string, the positions are buffer
positions and the type is one of the symbols `variable', `datatype',
`class', `import' and `instance'."
  (let (;; The name, type and name-position of the declaration to
	;; return.
	name
	type
	name-pos
	;; Buffer positions marking the start and end of the space
	;; containing a declaration.
	start
	end)
    ;; Stop when we are at the end of the buffer or when a valid
    ;; declaration is grabbed.
    (while (not (or (eobp) name))
      ;; Move forward to next declaration at or after point.
      (haskell-move-to-decl t literate t)
      ;; Start and end of search space is currently just the starting
      ;; line of the declaration.
      (setq start (point)
	    end   (progn (end-of-line) (point)))
      (goto-char start)
      (cond
       ;; If the start of the top-level declaration does not begin
       ;; with a starting keyword, then (if legal) must be a type
       ;; signature or value binding, and the variable concerned is
       ;; grabbed.
       ((not (looking-at haskell-start-keywords-re))
	(setq name (haskell-get-variable ""))
	(if name
	    (progn
	      (setq type 'variable)
	      (looking-at (concat ".*\\(" (regexp-quote name) "\\)"))
	      (setq name-pos (match-beginning 1)))))
       ;; User-defined datatype declaration.
       ((re-search-forward "\\=\\(data\\|newtype\\|type\\)\\>" end t)
	(re-search-forward "=>" end t)
	(if (looking-at "[ \t]*\\(\\sw+\\)")
	    (progn
	      (setq name (haskell-match-string 1))
	      (setq name-pos (match-beginning 1))
	      (setq type 'datatype))))
       ;; Class declaration.
       ((re-search-forward "\\=class\\>" end t)
	(re-search-forward "=>" end t)
	(if (looking-at "[ \t]*\\(\\sw+\\)")
	    (progn
	      (setq name (haskell-match-string 1))
	      (setq name-pos (match-beginning 1))
	      (setq type 'class))))
       ;; Import declaration.
       ((looking-at "import[ \t]+\\(qualified[ \t]+\\)?\\(\\sw+\\)")
	(setq name (haskell-match-string 2))
	(setq name-pos (match-beginning 2))
	(setq type 'import))
       ;; Instance declaration.
       ((re-search-forward "\\=instance[ \t]+" end t)
	(re-search-forward "=>[ \t]+" end t)
	;; The instance "title" starts just after the `instance' (and
	;; any context) and finishes just before the _first_ `where'
	;; if one exists.  This solution is ugly, but I can't find a
	;; nicer one---a simple regexp will pick up the last `where',
	;; which may be rare but nevertheless...
	(setq name-pos (point))
	(setq name (format "%s"
			   (buffer-substring
			    (point)
			    (progn			       
			      ;; Look for a `where'.
			      (if (re-search-forward "\\<where\\>" end t)
				  ;; Move back to just before the `where'.
				  (progn
				    (re-search-backward "\\s-where")
				    (point))
				;; No `where' so move to last non-whitespace
				;; before `end'.
				(progn
				  (goto-char end)
				  (skip-chars-backward " \t")
				  (point)))))))
	;; If we did not manage to extract a name, cancel this
	;; declaration (eg. when line ends in "=> ").
	(if (string-match "^[ \t]*$" name) (setq name nil))
	(setq type 'instance)))
      ;; Move past start of current declaration.
      (goto-char end))
    ;; If we have a valid declaration then return it, otherwise return
    ;; nil.
    (if name
	(cons (cons name (cons start name-pos)) type)
      nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Declaration scanning via `imenu'.

(defun haskell-create-imenu-index ()
  "Non-literate Haskell version of `haskell-generic-create-menu-index'."
  (interactive)
  (haskell-generic-create-imenu-index nil))

(defun literate-haskell-create-imenu-index ()
  "Literate Haskell version of `haskell-generic-create-menu-index'."
  (interactive)
  (haskell-generic-create-imenu-index t))

(defun haskell-generic-create-imenu-index (literate)
  "Function for finding imenu declarations in (literate) Haskell mode.
Finds all declarations (classes, variables, imports, instances and
datatypes) in a Haskell file for the imenu package."
  ;; Each list has elements of the form `(INDEX-NAME . INDEX-POSITION)'.
  ;; These lists are nested using `(INDEX-TITLE . INDEX-ALIST)'.
  (let* ((index-alist '())
	 (index-class-alist '())   ;; Classes
	 (index-var-alist '())     ;; Variables
	 (index-imp-alist '())     ;; Imports
	 (index-inst-alist '())    ;; Instances
	 (index-type-alist '())    ;; Datatypes
	 ;; Variables for showing progress.
	 (bufname (buffer-name))
	 (divisor-of-progress (/ (point-max) 100))
	 ;; The result we wish to return.
	 result)
    (goto-char (point-min))
    ;; Loop forwards from the beginning of the buffer through the
    ;; starts of the top-level declarations.
    (while (< (point) (point-max))
      (message "Scanning declarations in %s... (%3d%%)" bufname
	       (/ (point) divisor-of-progress))
      ;; Grab the next declaration.
      (setq result (haskell-generic-find-next-decl literate))
      (if result
	  ;; If valid, extract the components of the result.
	  (let* ((name-posns (car result))
		 (name (car name-posns))
		 (posns (cdr name-posns))
		 (start-pos (car posns))
		 (type (cdr result))
		 ;; Place `(name . start-pos)' in the correct alist.
		 (alist (cond
			 ((eq type 'variable) 'index-var-alist)
			 ((eq type 'datatype) 'index-type-alist)
			 ((eq type 'class) 'index-class-alist)
			 ((eq type 'import) 'index-imp-alist)
			 ((eq type 'instance) 'index-inst-alist))))
	    (set alist (cons (cons name start-pos) (eval alist))))))
    ;; Now sort all the lists, label them, and place them in one list.
    (message "Sorting declarations in %s..." bufname)
    (and index-type-alist
	 (push (cons "Datatypes"
		     (sort index-type-alist 'haskell-imenu-label-cmp))
	       index-alist))
    (and index-inst-alist
	 (push (cons "Instances"
		     (sort index-inst-alist 'haskell-imenu-label-cmp))
	       index-alist))
    (and index-imp-alist
	 (push (cons "Imports"
		     (sort index-imp-alist 'haskell-imenu-label-cmp))
	       index-alist))
    (and index-var-alist
	 (push (cons "Functions"
		     (sort index-var-alist 'haskell-imenu-label-cmp))
	       index-alist))
    (and index-class-alist
	 (push (cons "Classes"
		     (sort index-class-alist 'haskell-imenu-label-cmp))
	       index-alist))
    (message "Sorting declarations in %s...done" bufname)
    ;; Return the alist.
    index-alist))

(defun haskell-imenu-label-cmp (el1 el2)
  "Predicate to compare labels in lists produced by
`haskell-create-imenu-index'."
  (string< (car el1) (car el2)))

(defun haskell-imenu ()
  "Non-literate version of `haskell-generic-imenu'."
  (haskell-generic-imenu nil))

(defun literate-haskell-imenu ()
  "Literate version of `haskell-generic-imenu'."
  (interactive)
  (haskell-generic-imenu t))

(defun haskell-generic-imenu (literate)
  "Install imenu for (literate if LITERATE non-nil) Haskell mode."
  ;; Would prefer to toggle imenu but can't see how to turn it off...
  (require 'imenu)
  ;; In emacs-20's imenu we have to bind some functions first -- HWL
  (if (and  (not haskell-running-xemacs)
	    (>= (string-to-number (substring emacs-version 0 2)) 20)
	    (not (fboundp 'imenu-extract-index-name-function)))
    (setq imenu-extract-index-name-function
	  (if literate (function literate-haskell-create-imenu-index)
	    (function haskell-create-imenu-index))))
  (setq imenu-create-index-function
	(if literate (function literate-haskell-create-imenu-index)
	  (function haskell-create-imenu-index)))
  (if (fboundp 'imenu-add-to-menubar)
      (imenu-add-to-menubar "Declarations")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Declaration scanning via `func-menu'.

(defun haskell-func-menu-next (buffer)
  "Non-literate Haskell version of `haskell-generic-func-menu-next'." 
  (haskell-generic-func-menu-next nil buffer)) 

(defun literate-haskell-func-menu-next (buffer)
  "Literate Haskell version of `haskell-generic-func-menu-next'."
  (haskell-generic-func-menu-next t buffer)) 

(defun haskell-generic-func-menu-next (literate buffer)
  "Returns `(name . pos)' of next declaration."
  (set-buffer buffer)
  (let ((result (haskell-generic-find-next-decl literate)))
    (if result
	(let* ((name-posns (car result))
	       (name (car name-posns))
	       (posns (cdr name-posns))
	       (name-pos (cdr posns))
	       ;(type (cdr result))
	       )
	  (cons ;(concat
		 ;; func-menu has problems with spaces, and adding a
		 ;; qualifying keyword will not allow the "goto fn"
		 ;; functions to work properly.  Sigh.
;		 (cond
;		  ((eq type 'variable) "")
;		  ((eq type 'datatype) "datatype ")
;		  ((eq type 'class) "class ")
;		  ((eq type 'import) "import ")
;		  ((eq type 'instance) "instance "))
		 name;)
		name-pos))
      nil)))

(defvar haskell-func-menu-regexp "^\\(\\sw\\|(\\)"
  "Regexp to match the start of a possible declaration.")

(defvar literate-haskell-func-menu-regexp
  (concat "^" literate-haskell-decl-scan-line-prefix "\\(\\sw\\|(\\)")
  "As `haskell-func-menu-regexp' but for literate scripts.")

(defun haskell-func-menu ()
  "Uses `func-menu' to establish declaration scanning."
  (require 'func-menu)
  (make-local-variable 'fume-menubar-menu-name)
  (setq fume-menubar-menu-name "Declarations")
  (make-local-variable 'fume-function-name-regexp-alist)
  (setq fume-function-name-regexp-alist
	'((haskell-mode . haskell-func-menu-regexp)))
  (make-local-variable 'fume-find-function-name-method-alist)
  (setq fume-find-function-name-method-alist
	'((haskell-mode . haskell-func-menu-next)))
  (fume-add-menubar-entry)
  (make-local-variable 'haskell-mode-map)
  (define-key haskell-mode-map "\C-cl" 'fume-list-functions)
  (define-key haskell-mode-map "\C-cg" 'fume-prompt-function-goto)
  (define-key haskell-mode-map [(meta button1)] 'fume-mouse-function-goto)
  (use-local-map haskell-mode-map))

(defun literate-haskell-func-menu ()
  "As `haskell-func-menu' for literate scripts."
  (require 'func-menu)
  (make-local-variable 'fume-menubar-menu-name)
  (setq fume-menubar-menu-name "Declarations")
  (make-local-variable 'fume-function-name-regexp-alist)
  (setq fume-function-name-regexp-alist
	'((literate-haskell-mode . literate-haskell-func-menu-regexp)))
  (make-local-variable 'fume-find-function-name-method-alist)
  (setq fume-find-function-name-method-alist
	'((literate-haskell-mode . literate-haskell-func-menu-next)))
  (fume-add-menubar-entry)
  (make-local-variable 'literate-haskell-mode-map)
  (define-key literate-haskell-mode-map "\C-cl" 'fume-list-functions)
  (define-key literate-haskell-mode-map "\C-cg" 'fume-prompt-function-goto)
  (define-key literate-haskell-mode-map [(meta button1)]
    'fume-mouse-function-goto)
  (use-local-map literate-haskell-mode-map))


;; The main functions to turn on declaration scanning.
(defun haskell-decl-scan ()
  "Turn on declaration scanning for Haskell mode.  Top-level
declarations are scanned and listed in the menu item \"Declarations\".
Selecting an item from this menu will take point to the start of the
declaration.

This may link with `haskell-doc-mode' (only for FSF Emacs currently).

Assumes the common convention that top-level declarations start at the
first column."
  (interactive)
  (if haskell-running-xemacs
      (haskell-func-menu)
    (haskell-imenu)))

(defun literate-haskell-decl-scan ()
  "Turn on declaration scanning for literate Haskell mode.  See
`haskell-decl-scan'.

Assumes the common convention that top-level declarations start at the
third column, ie. after \"> \"."
  (interactive)
  (if haskell-running-xemacs
      (literate-haskell-func-menu)
    (literate-haskell-imenu)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Provide ourselves:

(provide 'haskell-mode)

;;; haskell-mode.el ends here
