---------------------------------------------------------------------
-- A protocol GUI for the Curry test tool:
--
-- The GUI process messages sent by running tests with the test tool
-- and summarizes the results in a GUI.
--
-- Michael Hanus, October 2004
---------------------------------------------------------------------
-- A protocol GUI for the currytest tool

import Ports
import Tk
import Read
import Assertion
import System
import List(intersperse)
import FlatCurry
import Dynamic

---------------------------------------------------------------------
-- Check arguments and call main function:
main = do
  args <- getArgs
  case args of
   "-window":modnames -> startGUI modnames
   a:as -> mapIO_ (testModule putStr "") (map stripCurrySuffix (a:as))
   _ -> putStrLn $ "ERROR: Illegal arguments for currytest: " ++
                   concat (intersperse " " args) ++ "\n" ++
                   "Usage: currytest [-window] <module_names>"

startGUI modnames = do
  guiname <- newGuiName
  msgs <- openNamedPort guiname
  mapIO_ addTestModule (map stripCurrySuffix modnames)
  initmods <- getAllTestModules
  runControlledWidget "CurryTest Tool"
                      (protocol_gui guiname initmods)
                      msgs

stripCurrySuffix n = let rn = reverse n in
  if take 6 rn == "yrruc." then reverse (drop 6 rn) else
  if take 7 rn == "yrrucl." then reverse (drop 7 rn) else n

-- Generate unique name for GUI port:
newGuiName = do
  pid <- getPID
  return ("currytest"++show pid)

-- A text edit widget with vertical scrollbar.
TkTextEditVScroll confs =
   TkRow []
     [TkTextEdit ([TkRef txtref, TkFill]++confs),
      TkScrollV txtref [TkFillY]]     where txtref free

---------------------------------------------------------------------
-- The definition of the protocol GUI together with a handler
-- "ext_handler" that is responsible to handler the external messages
-- sent during running the test on the program:
protocol_gui guiname initmods =
 (TkRow []
   [TkCol [TkLeft]
     [TkRow [] [TkLabel [TkText "Testing Curry module:"],
                TkEntry [TkRef rmod, TkFillX, TkBackground "white"]],
      TkRow [] [TkLabel [TkText "Test cases:"],
                TkEntry [TkRef rtestnum, TkText "0", TkWidth 5],
                TkLabel [TkText "Failures:"],
                TkEntry [TkRef rfailnum, TkText "0", TkWidth 5,
                         TkBackground "green"]],
      TkLabel [TkText "Failed test cases:"],
      TkTextEditScroll [TkRef rfail, TkHeight 10, TkWidth 60,
                        TkBackground "green"],
      TkLabel [TkText "Test protocol:"],
      TkTextEditScroll [TkRef rprot, TkHeight 15, TkWidth 60,
                        TkBackground "white"]],
    TkCol [TkLeft]
     [TkRow [TkLeft] [TkButton openfile [TkText "Add test module"],
                      TkButton starttest [TkText "Test all modules"],
                      TkButton tkExit [TkText "Exit"]],
      TkLabel [TkText "Modules stored for testing:"],
      TkTextEditVScroll [TkRef rmods, TkHeight 10, TkWidth 60,
                         TkText (concat (intersperse "\n" initmods)),
                         TkBackground "white"],
      TkLabel [TkText "Compilation messages:"],
      TkTextEditScroll [TkRef rcmsgs, TkHeight 15, TkWidth 60,
                        TkBackground "white"]]
   ], ext_handler)
 where
   rmod,rtestnum,rfailnum,rfail,rprot,rcmsgs,rmods free

   openfile gp =
      tkGetOpenFileWithTypes curryFileTypes >>= \filename ->
      if null filename then done else
      addTestModule (stripCurrySuffix filename) >>
      updateModules gp

   starttest gp = startTestModule gp

   -- test a module, if present:
   startTestModule gp =
     getNextTestModule >>=  \nextmod ->
     updateModules gp >>
     maybe done (\m -> testModule (printCompMsg gp) guiname m) nextmod

   -- print a compilation message in corresponding widget:
   printCompMsg gp msg = tkAppendValue rcmsgs msg gp

   -- update list of modules in widget:
   updateModules gp =
     getAllTestModules >>= \mods ->
     tkSetValue rmods (concat (intersperse "\n" mods)) gp

   ext_handler (TestModule m) gp = do
      tkConfig rmod (TkBackground "green") gp
      tkSetValue rmod m gp
      tkAppendValue rprot
                    (take 60 (repeat '=')++"\nTesting module: "++m++"\n") gp

   ext_handler TestFinished gp = do
      tkConfig rmod (TkBackground "white") gp
      tkSetValue rmod "" gp
      startTestModule gp

   ext_handler (TestCase s b) gp = do
      tkUpdate incrText rtestnum gp
      if b then done
           else do tkConfig rfailnum (TkBackground "red") gp
                   tkConfig rfail (TkBackground "red") gp
                   tkUpdate incrText rfailnum gp
                   mod <- tkGetValue rmod gp
                   tkAppendValue rfail ("Module: "++mod++"\n") gp
                   tkAppendValue rfail (s++"\n") gp
      tkAppendValue rprot s gp

-- Curry file types:
curryFileTypes = [("Curry Files",".curry"),
                  ("Literate Curry files",".lcurry")]

-- increment number text string:
incrText s = show (readInt s + 1)


-- Predicate to hold all modules still to be tested:
testModuleName :: String -> Dynamic
testModuleName = dynamic

-- Add a new module for testing:
addTestModule :: String -> IO ()
addTestModule mod = assert (testModuleName mod)

-- Does a module exists for testing?
existsTestModule :: IO Bool
existsTestModule = do
  mbmod <- getDynamicSolution testModuleName
  return (mbmod/=Nothing)

-- Get list of all modules to be tested:
getAllTestModules :: IO [String]
getAllTestModules = getDynamicSolutions testModuleName

-- Get next module to be tested and retract it:
getNextTestModule :: IO (Maybe String)
getNextTestModule = do
  maybemod <- getDynamicSolution testModuleName
  maybe (return Nothing)
        (\mod -> retract (testModuleName mod) >> return (Just mod))
        maybemod

---------------------------------------------------------------------------
-- Main function to test a module:
-- Arg 1: function for printing compilation messages
-- Arg 2: name of the GUI port (or empty string if no gui defined)
-- Arg 3: module name
testModule prtmsg guiname modname = do
  prtmsg ("Loading module \""++modname++"\"...\n")
  prog_or_error <- tryReadFlatCurry modname
  testModuleIfPossible prtmsg guiname modname prog_or_error

testModuleIfPossible prtmsg guiname _ (Right errmsg) = do
  prtmsg ("ERROR: compilation not successful:\n\n"++errmsg++"\n")
  showTestEnd guiname
testModuleIfPossible prtmsg guiname modname (Left prog) =
  execTestFunctions prtmsg guiname modname (getTestFunctionNames prog)

execTestFunctions prtmsg guiname _ [] = do
  prtmsg "No test functions found.\n\n"
  showTestEnd guiname
execTestFunctions prtmsg guiname modname (f:fs) = do
  prtmsg ("Exported top-level test functions:\n"
            ++ concatMap (++" ") (f:fs) ++ "\n\n")
  if guiname/="" then showTestMod guiname modname else done
  pakcshome <- getEnviron "PAKCSHOME"
  let testgoal =
         "putStrLn (take 60 (repeat (chr 61))) >> " ++
         "putStrLn (\"Testing module \\\""++modname++"\\\"...\") >> " ++
         concat
           (intersperse " `seqStrActions` "
             (map (("checkAssertion " ++
                    if guiname/="" then "(showTestCase \""++guiname++"\") "
                                   else "return ") ++) (f:fs))) ++
         " >>= writeAssertResult \""++modname++".testlog\" " ++
         " >> putStrLn (take 60 (repeat (chr 61)))" ++
         (if guiname/="" then " >> showTestEnd \""++guiname++"\"" else "")
  --putStrLn testgoal
  system ("echo ':l "++modname++"\n"++testgoal++" ' | " ++
          pakcshome++"/bin/pakcs -quiet -Dshowfcyload=no 2>&1" ++
          if guiname=="" then "" else " &")
  done


-- Extract all test functions from a module:
getTestFunctionNames (Prog _ _ _ funs _) =
   map funcname . filter isExportedFunc . filter hasAssertType $ funs
 where
   isExportedFunc (Func _ _ vis _ _) = vis==Public

   funcname (Func (_,fname) _ _ _ _) = fname


hasAssertType (Func _ _ _ texp _) =
 case texp of
   TCons tc _ -> tc==("Assertion","Assertion")
   _          -> False


-- Tries to read a FlatCurry program.
-- Returns either (Left prog) (if reading was successful)
-- or (Right msg) where msg is the string of error messages from the parser
-- Requires environment variable PAKCSHOME (for loading the parser).
tryReadFlatCurry :: String -> IO (Either Prog String)
tryReadFlatCurry mname =
  do pakcshome <- getEnviron "PAKCSHOME"
     pofile <- getPOFilename
     system (pakcshome++"/bin/parsecurry -fcy "++mname++" > "++pofile++" 2>&1")
     ret <- system ("test -f "++mname++".fcy")
     returnReadFlatCurry (ret==0) pofile
 where
   -- check whether parsing was ok and return appropriate value:
   returnReadFlatCurry True pofile =
     do prog <- readFlatCurryFile (mname++".fcy")
        system ("rm "++pofile)
        return (Left prog)
   returnReadFlatCurry False pofile =
     do msgs <- readFile pofile
        system ("rm "++pofile)
        return (Right msgs)

   -- compute name for auxiliary file for parser outputs:
   getPOFilename =
     do pid <- getPID
        return ("/tmp/pakcsoutput_"++show pid)
